/*
 * Operating system kernel abstraction -- events
 *
 * Copyright (C) 2009 Cambridge Silicon Radio Ltd.
 *
 * Refer to LICENSE.txt included with this source code for details on
 * the license terms.
 */
#ifndef __OSKA_EVENT_H
#define __OSKA_EVENT_H

#include <oska/types.h>

#ifdef __OSKA_API_DOC
/**
 * @defgroup event Events
 */

/**
 * Implementation defined event object.
 */
typedef implementation_defined os_event_t;

/**
 * Initialize an event.
 *
 * Callable from: thread context.
 *
 * @param evt event to initialize.
 *
 * @ingroup event
 */
void os_event_init(os_event_t *evt);

/**
 * Destroy an event.
 *
 * Destroy an event previously initialized with os_event_init(),
 * freeing any associated resources.
 *
 * Callable from: thread context.
 *
 * @param evt event to destroy
 *
 * @ingroup event
 */
void os_event_destroy(os_event_t *evt);

/**
 * Wait for an event to be raised.
 *
 * Callable from: thread context.
 *
 * @param evt event to wait for.
 *
 * @return the set of events raised since the last call to
 * os_event_wait().
 *
 * @ingroup event
 */
uint16_t os_event_wait(os_event_t *evt);

/**
 * Wait for an event to be raised.
 *
 * The wait may be interrupted by a signal or other OS-specific
 * asynchronous event.
 *
 * Callable from: thread context.
 *
 * @param evt event to wait for.
 *
 * @return the set of events raised since the last call to
 * os_event_wait().
 *
 * @ingroup event
 */
uint16_t os_event_wait_interruptible(os_event_t *evt)

/**
 * Wait (with a timeout) for an event to be raised.
 *
 * Only one thread should wait on a particular event.
 *
 * Callable from: thread context.
 *
 * @param evt event to wait for.
 * @param timeout_ms time to wait for an event (in milliseconds).
 *
 * @return the set of events raised since the last call to
 * os_event_wait().
 * @return 0 if no event was raised before the timeout expired.
 *
 * @ingroup event
 */
uint16_t os_event_wait_timed(os_event_t *evt, unsigned timeout_ms);

/**
 * Raise an event, waking the thread waiting in os_event_wait().
 *
 * Callable from: thread context.
 *
 * @param evt event to wait for.
 * @param events the set of events to raise.
 *
 * @ingroup event
 */
void os_event_raise(os_event_t *evt, uint16_t events);

#endif /* __OSKA_API_DOC */

#ifdef linux
#  include <../linux/event.h>
#elif defined(ZSDE)
#  include <../zoran/event.h>
#elif defined(KMDF_MAJOR_VERSION)
#  include <../windows/kmdf/event.h>
#else
#  error <oska/event.h> not provided for this OS
#endif

#endif /* #ifndef __OSKA_EVENT_H */
