/*
 * Operating system kernel abstraction -- semaphores
 *
 * Copyright (C) 2007 Cambridge Silicon Radio Ltd.
 *
 * Refer to LICENSE.txt included with this source code for details on
 * the license terms.
 */
#ifndef __OSKA_SEMAPHORE_H
#define __OSKA_SEMAPHORE_H

#ifdef __OSKA_API_DOC
/**
 * @defgroup semaphore Semaphores
 */

/**
 * Implementation defined semaphore object.
 */
typedef implementation_defined os_semaphore_t;

/**
 * Initialize a semaphore to zero.
 *
 * Callable from: thread context.
 *
 * @param sem semaphore to initialize.
 *
 * @ingroup semaphore
 */
void os_semaphore_init(os_semaphore_t *sem);

/**
 * Destroy a semaphore.
 *
 * Destroy a semaphore previously initialized with
 * os_semaphore_init(), freeing any associated resources.
 *
 * Callable from: thread context.
 *
 * @param sem semaphore to destroy
 *
 * @ingroup semaphore
 */
void os_semaphore_destroy(os_semaphore_t *sem);

/**
 * Wait for an event to be posted to the semaphore.
 *
 * Callable from: thread context.
 *
 * @param sem semaphore to wait on.
 *
 * @ingroup semaphore
 */
void os_semaphore_wait(os_semaphore_t *sem);

/**
 * Wait for an event to be posted to the semaphore, or some time has
 * elapsed.
 *
 * Callable from: thread context.
 *
 * @param sem     semaphore to wait on
 * @param time_ms maximum time to wait (in ms).
 *
 * @return 0 the the semaphore was posted.
 * @return -ETIMEDOUT if the maximum time to wait was exceeded.
 *
 * @ingroup semaphore
 */
int os_semaphore_wait_timed(os_semaphore_t *sem, int time_ms);

/**
 * Post an event to a semaphore, waking at least one of the threads
 * waiting on it.
 *
 * Callable from: any context.
 *
 * @param sem semaphore to post to.
 *
 * @ingroup semaphore
 */
void os_semaphore_post(os_semaphore_t *sem);

#endif /* __OSKA_API_DOC */

#ifdef linux
#  include <../linux/semaphore.h>
#elif defined(ZSDE)
#  include <../zoran/semaphore.h>
#else
#  error <oska/semaphore.h> not provided for this OS
#endif

#endif /* #ifndef __OSKA_SEMAPHORE_H */
