/*
 * Operating system kernel abstraction -- timing
 *
 * Copyright (C) 2007 Cambridge Silicon Radio Ltd.
 *
 * Refer to LICENSE.txt included with this source code for details on
 * the license terms.
 */
#ifndef __OSKA_TIME_H
#define __OSKA_TIME_H

#ifdef __OSKA_API_DOC
/**
 * @defgroup time Timing and delays
 */

/**
 * Return the current time (since an arbitrary point) in milliseconds.
 *
 * The resolution depends on the OS but would typically be 10 ms or
 * better.
 *
 * Callable from: any context.
 *
 * @return current time in ms.
 *
 * @ingroup time
 */
unsigned long os_current_time_ms(void);

/**
 * Sleep for a minimum length of time.
 *
 * Callable from: thread context.
 *
 * @param ms minimum time to sleep (in ms).
 *
 * @ingroup time
 */
int os_sleep_ms(unsigned ms);

/**
 * Delay a minimum length of time.
 *
 * This may busy-wait and should not be used for long delays (use
 * os_sleep_ms() instead).
 *
 * Callable from: any context.
 *
 * @param us minimum time to delay (in us).
 *
 * @ingroup time
 */
int os_delay_us(unsigned us);

/**
 * Check if some time has elapsed.
 *
 * Callable from: any context.
 *
 * @param a a time returned by os_current_time_ms().
 * @param b a time returned by os_current_time_ms().
 *
 * @return true if time a is after time b.
 *
 * @ingroup time
 */
bool os_time_after(unsigned long a, unsigned long b);

#endif /* __OSKA_API_DOC */

#ifdef linux
#  include <../linux/time.h>
#elif defined(ZSDE)
#  include <../zoran/time.h>
#elif defined(KMDF_MAJOR_VERSION)
#  include <../windows/kmdf/time.h>
#else
#  error <oska/time.h> not provided for this OS
#endif

/*
 * A platform might provide a better implementation of os_time_after()
 * (e.g., with stronger type checking).
 */
#ifndef os_time_after
#  define os_time_after(a, b) ((long)(b) - (long)(a) < 0)
#endif

#endif /* #ifndef __OSKA_TIME_H */
