/*
 * Operating system kernel abstraction -- timers.
 *
 * Copyright (C) 2009 Cambridge Silicon Radio Ltd.
 *
 * Refer to LICENSE.txt included with this source code for details on
 * the license terms.
 */
#ifndef __OSKA_TIMER_H
#define __OSKA_TIMER_H

#ifdef __OSKA_API_DOC
/**
 * @defgroup timer Timers
 *
 * A timer is an object which calls a user-supplied function at some
 * future time (the expiry time).  Timers may be cancelled before they
 * expire.
 */

/**
 * Implementation defined timer object.
 *
 * @ingroup timer
 */
typedef implementation_defined os_timer_t;

/**
 * Timer expiry function.
 *
 * The function called when a timer expires must be of this type.
 *
 * @param arg the argument supplied when the timer was initialized.
 *
 * @ingroup timer
 */
typedef void (*os_timer_func_t)(void *arg);

/**
 * Initialize a timer object.
 *
 * Callable from: thread context.
 *
 * @param timer the timer to initialize.
 * @param func  the function to call when the timer expires.
 * @param arg   the argument to pass to \e func.
 *
 * @ingroup timer
 */
void os_timer_init(os_timer_t *timer, os_timer_func_t func, void *arg);

/**
 * Destroy a timer object.
 *
 * If this timer function is currently being executed, this function
 * waits for it to complete before returning.
 *
 * Callable from: thread context.
 *
 * @param timer the timer to destroy.
 */
void os_timer_destroy(os_timer_t *timer);

/**
 * Set a timer's expiry time.
 *
 * This sets the time at which the timer will expire.  If the timer is
 * already set (but not expired) it is cancelled and reset with the
 * new time.  If the timer has expired, the timer is reset.
 *
 * Callable from: any context.
 *
 * @param timer      the timer to start.
 * @param expires_ms elapsed time (in milliseconds) until the timer expires.
 *
 * @ingroup timer
 */
void os_timer_set(os_timer_t *timer, unsigned long expires_ms);

/**
 * Cancel a timer.
 *
 * This cancels a timer that was previously started.  If the timer has
 * already expired this has no effect.
 *
 * Callable from: any context.
 *
 * @note The timer function may still be running after this function
 * returns.
 *
 * @param timer the timer to stop.
 *
 * @ingroup timer
 */
void os_timer_cancel(os_timer_t *timer);

#endif /* __OSKA_API_DOC */

typedef void (*os_timer_func_t)(void *arg);

#ifdef linux
#  include <../linux/timer.h>
#elif defined (KMDF_MAJOR_VERSION)
#  include <../windows/kmdf/timer.h>
#else
#  error <oska/timer.h> not provided for this OS
#endif

#endif /* #ifndef __OSKA_TIMER_H */
