/*
 * Operating system kernel abstraction -- tracing messages.
 *
 * Copyright (C) 2009 Cambridge Silicon Radio Ltd.
 *
 * Refer to LICENSE.txt included with this source code for details on
 * the license terms.
 */
#ifndef __OSKA_TRACE_H
#define __OSKA_TRACE_H

#ifdef __OSKA_API_DOC
/**
 * @defgroup trace Trace messages
 *
 * Trace messages are printf-style messages used for run-time
 * diagnostics.  There are four levels of severity for the messages:
 * error, warning, informational, and debug.
 *
 * How these messages are read is platform specific.  Each module that
 * uses tracing may also have to provide additional platform specific
 * information.
 *
 * <h2>Linux</h2>
 *
 * Trace messages are sent to the standard kernel message log
 * (dmesg). If a module defines OS_TRACE_PREFIX then this string is
 * prefixed to all trace messages.
 *
 * For example,
 * \code
 * #define OS_TRACE_PREFIX "module: "
 * #include <oska/trace.h>
 * \endcode
 *
 * <h2>Windows KMDF</h2>
 *
 * WPP software tracing is used.  Each driver must:
 *
 * - Add <tt>RUN_WPP=\$(SOURCES) -km -scan:\$(OSKADIR)\\windows\\kmdf\\%trace.h</tt>
 *   to the Sources files.
 *
 * - Include the tracing message header in all files using tracing.
 * \code
 * #ifdef OS_TRACE_INCLUDE_TMH
 * #include "filename.tmh"
 * #endif
 * \endcode
 * - Call %WPP_INIT_TRACING() in the driver's %DriverEntry() routine
 *   and call %WPP_CLEANUP() as required.
 *
 * - (Optional) Provide a driver specific WPP_CONTROL_GUIDS
 *   definition.  For example,
 * \code
 * #define WPP_CONTROL_GUIDS                         \
 *     WPP_DEFINE_CONTROL_GUID(ModuleTraceGuid       \
 *         (210ba3cc,a3a9,42d9,add9,5cfc6d806bae),   \
 *         WPP_DEFINE_BIT(OS_TRACE_DFLT)             \
 *     )
 * #include <oska/trace.h>
 * \endcode
 *
 * <h2>X.org server log</h2>
 *
 * Trace messages from X.org drivers can be sent to the X.org server
 * log.  The driver must \#define OS_TRACE_USE_XORG before including
 * <oska/trace.h>
 */

/**
 * An error trace message.
 *
 * The printf-style format string must be a constant string literal.
 * The c, d, p, s, and x conversion specifiers and the field length
 * and pad character should be supported (e.g., \%08x).
 *
 * Callable from: any context.
 *
 * @param format printf-style format string.
 * @param ...    arguments for \a format.
 *
 * @ingroup trace
 */
void os_trace_err(const char *format, ...);

/**
 * A warning trace message.
 *
 * Callable from: any context.
 *
 * @param format printf-style format string.
 * @param ...    arguments for \a format.
 *
 * @see os_trace_err
 *
 * @ingroup trace
 */
void os_trace_warn(const char *format, ...);

/**
 * A informational trace message.
 *
 * Callable from: any context.
 *
 * @param format printf-style format string.
 * @param ...    arguments for \a format.
 *
 * @see os_trace_err
 *
 * @ingroup trace
 */
void os_trace_info(const char *format, ...);

/**
 * A debug trace message.
 *
 * Callable from: any context.
 *
 * @param format printf-style format string.
 * @param ...    arguments for \a format.
 *
 * @see os_trace_err
 *
 * @ingroup trace
 */
void os_trace_dbg(const char *format, ...);

#endif /* __OSKA_API_DOC */

#ifdef linux
#  include <../linux/trace.h>
#elif defined(KMDF_MAJOR_VERSION)
#  include <../windows/kmdf/trace.h>
#else
#  error <oska/trace.h> not provided for this OS
#endif

#endif /* #ifndef __OSKA_TRACE_H */
