/*
 * Operating system kernel abstraction -- misc. utility functions
 *
 * Copyright (C) 2007 Cambridge Silicon Radio Ltd.
 *
 * Refer to LICENSE.txt included with this source code for details on
 * the license terms.
 */
#ifndef __OSKA_UTIL_H
#define __OSKA_UTIL_H

#include <oska/types.h>

#ifdef __OSKA_API_DOC
/**
 * @defgroup util Miscellaneous utilties
 */

/**
 * Take some debug-related action if the expression is false.
 *
 * The action taken when the expression is false is platform
 * dependent.  For example, the platform could do nothing (e.g., for
 * non debug builds), terminatate the current thread, break into a
 * debugger, or halt the entire system.
 */
void OS_ASSERT(expression);

/**
 * Return a CPU (native) byte ordered 16-bit word in little endian
 * byte order.
 *
 * @param x CPU ordered word.
 *
 * @return the word in little-endian byte order.
 *
 * @ingroup util
 */
uint16_t os_cpu_to_le16(uint16_t x);

/**
 * Return a little endian byte ordered 16-bit word in CPU (native)
 * byte order.
 *
 * @param x little endian word.
 *
 * @return the word in native byte order.
 *
 * @ingroup util
 */
uint16_t os_le16_to_cpu(uint16_t x);

/**
 * Return a CPU (native) byte ordered 32-bit word in little endian
 * byte order.
 *
 * @param x CPU ordered word.
 *
 * @return the word in little-endian byte order.
 *
 * @ingroup util
 */
uint32_t os_cpu_to_le32(uint32_t x);

/**
 * Return a little endian byte ordered 32-bit word in CPU (native)
 * byte order.
 *
 * @param x little endian word.
 *
 * @return the word in native byte order.
 *
 * @ingroup util
 */
uint32_t os_le32_to_cpu(uint32_t x);

/**
 * Return a CPU (native) byte ordered 64-bit word in little endian
 * byte order.
 *
 * @param x CPU ordered word.
 *
 * @return the word in little-endian byte order.
 *
 * @ingroup util
 */
uint64_t os_cpu_to_le64(uint64_t x);

/**
 * Return a little endian byte ordered 64-bit word in CPU (native)
 * byte order.
 *
 * @param x little endian word.
 *
 * @return the word in native byte order.
 *
 * @ingroup util
 */
uint64_t os_le64_to_cpu(uint64_t x);

#endif /* __OSKA_API_DOC */

#ifdef linux
#  include <../linux/util.h>
#elif defined(ZSDE)
#  include <../zoran/util.h>
#elif defined(KMDF_MAJOR_VERSION)
#  include <../windows/kmdf/util.h>
#else
#  error <oska/util.h> not provided for this OS
#endif

#endif /* #ifndef __OSKA_UTIL_H */
