/*
 * OSKA Posix event functions.
 *
 * Copyright (C) 2010 Cambridge Silicon Radio Ltd.
 *
 * Refer to LICENSE.txt included with this source code for details on
 * the license terms.
 */

#include <oska/event.h>

#include <sys/time.h>
#include <errno.h>

void os_event_init(os_event_t *evt)
{
    pthread_mutex_init(&evt->mutex, NULL);
    pthread_cond_init(&evt->cond, NULL);
    evt->events = 0;
}

void os_event_destroy(os_event_t *evt)
{
    pthread_cond_destroy(&evt->cond);
    pthread_mutex_destroy(&evt->mutex);
}


uint16_t os_event_wait(os_event_t *evt)
{
    uint16_t e;

    pthread_mutex_lock(&evt->mutex);

    while (evt->events == 0) {
        pthread_cond_wait(&evt->cond, &evt->mutex);
    }
    e = evt->events;
    evt->events = 0;

    pthread_mutex_unlock(&evt->mutex);

    return e;
}

uint16_t os_event_wait_timed(os_event_t *evt, unsigned timeout_ms)
{
    uint16_t e;
    struct timeval cur_time;
    struct timespec end_time;
    int sec = timeout_ms / 1000;
    int msec = timeout_ms % 1000;
    int ret;

    gettimeofday(&cur_time, NULL);

    if (((cur_time.tv_usec / 1000) + msec) >= 1000)
    {
        msec = ((cur_time.tv_usec / 1000) + msec) % 1000;
        sec += ((cur_time.tv_usec / 1000) + msec) / 1000;
    }
    end_time.tv_sec = cur_time.tv_sec + sec;
    end_time.tv_nsec = cur_time.tv_usec * 1000 +  msec * 1000;

    pthread_mutex_lock(&evt->mutex);

    while (evt->events == 0) {
        ret = pthread_cond_timedwait(&evt->cond, &evt->mutex, &end_time);
        if (ret < 0 && errno == ETIMEDOUT) {
            break;
        }
    }
    e = evt->events;
    evt->events = 0;

    pthread_mutex_unlock(&evt->mutex);

    return e;
}


void os_event_raise(os_event_t *evt, uint16_t events)
{
    pthread_mutex_lock(&evt->mutex);
    evt->events |= events;
    pthread_cond_signal(&evt->cond);
    pthread_mutex_unlock(&evt->mutex);
}
