/*
 * OSKA Posix implementation -- inter-thread message queues.
 *
 * Copyright (C) 2010 Cambridge Silicon Radio Ltd.
 *
 * Refer to LICENSE.txt included with this source code for details on
 * the license terms.
 */
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>

#include <oska/mq.h>

/**
 * Initialize a message queue.
 *
 * @param mq          The message queue.
 * @param message_len Length of messages that will be enqueued.
 *
 * @return 0 on success; -ve error code otherwise.
 */
int os_mq_init(os_mq_t *mq, size_t message_len)
{
    int flags;
    int ret;

    mq->fds[OS_MQ_POSIX_READ_FD] = -1;
    mq->mesg_len = message_len;

    ret = pipe(mq->fds);
    if (ret < 0) {
        goto error;
    }

    flags = fcntl(mq->fds[OS_MQ_POSIX_READ_FD], F_GETFL, 0);
    if (flags < 0) {
        goto error;
    }
    flags |= O_NONBLOCK;
    ret = fcntl(mq->fds[OS_MQ_POSIX_READ_FD], F_SETFL, flags);
    if (ret < 0) {
        goto error;
    }

    return 0;

  error:
    os_mq_destroy(mq);
    return -errno;
}

/**
 * Destroy a message queue.
 *
 * @param mq The message queue.
 */
void os_mq_destroy(os_mq_t *mq)
{
    if (mq) {
        if (mq->fds[OS_MQ_POSIX_READ_FD] >= 0) {
            close(mq->fds[OS_MQ_POSIX_READ_FD]);
            close(mq->fds[OS_MQ_POSIX_WRITE_FD]);
        }
    }
}

/**
 * Dequeue a message from the queue.
 *
 * @param mq      The message queue.
 * @param message Buffer to read the message into.
 *
 * @return 0 if a message was retrieved.
 * @return -EAGAIN if no message was available.
 * @return -EIO if a low-level I/O error occurred.
 */
int os_mq_dequeue(os_mq_t *mq, void *message)
{
    uint8_t *buf = message;
    int len = 0;
    int ret;

    do {
        ret = read(mq->fds[OS_MQ_POSIX_READ_FD], buf + len, mq->mesg_len - len);
        if (ret < 0) {
            if (errno == EINTR) {
                continue;
            }
            return -errno;
        }
        len += ret;
    } while (len < mq->mesg_len);

    return 0;
}

/**
 * Enqueue a message to the queue.
 *
 * @param mq      The message queue.
 * @param message The message to enqueue.
 *
 * @return 0 on success; -ve error code otherwise.
 */
int os_mq_enqueue(os_mq_t *mq, void *message)
{
    uint8_t *buf = message;
    int len = 0;
    int ret;

    do {
        ret = write(mq->fds[OS_MQ_POSIX_WRITE_FD], buf + len, mq->mesg_len - len);
        if (ret < 0) {
            if (errno == EINTR) {
                continue;
            }
            return -errno;
        }
        len += ret;
    } while (len < mq->mesg_len);

    return 0;
}
