/*
 * OSKA Posix implementation -- tracing messages.
 *
 * Copyright (C) 2010 Cambridge Silicon Radio Ltd.
 *
 * Refer to LICENSE.txt included with this source code for details on
 * the license terms.
 */
#include <stdarg.h>
#include <stdio.h>
#include <sys/time.h>
#include <time.h>

#include <oska/trace.h>

#define OS_TRACE_BUF_LEN 256

static void os_trace_print(char *buf, const char *format, va_list args)
{
    size_t len = 0;
    struct timeval tv;
    struct tm tm;
    int ret;

    ret = gettimeofday(&tv, NULL);
    if (ret < 0) {
        goto no_time;
    }
    if (localtime_r(&tv.tv_sec, &tm) == NULL) {
        goto no_time;
    }
    len += strftime(buf + len, OS_TRACE_BUF_LEN - len, "%T", &tm);
    len += snprintf(buf + len, OS_TRACE_BUF_LEN - len, ".%03ld ", tv.tv_usec / 1000);

no_time:
    vsnprintf(buf + len, OS_TRACE_BUF_LEN - len, format, args);
}

void os_trace_stdio(const char *format, ...)
{
    va_list args;
    char buf[OS_TRACE_BUF_LEN];

    va_start(args, format);
    os_trace_print(buf, format, args);
    va_end(args);

    fputs(buf, stdout);
    fflush(stdout);
}

/* Flags for log messages (copied from <xorg/os.h> to avoid having to
   have the X.org headers installed). */
typedef enum {
    X_PROBED,                   /* Value was probed */
    X_CONFIG,                   /* Value was given in the config file */
    X_DEFAULT,                  /* Value is a default */
    X_CMDLINE,                  /* Value was given on the command line */
    X_NOTICE,                   /* Notice */
    X_ERROR,                    /* Error message */
    X_WARNING,                  /* Warning message */
    X_INFO,                     /* Informational message */
    X_NONE,                     /* No prefix */
    X_NOT_IMPLEMENTED,          /* Not implemented */
    X_UNKNOWN = -1              /* unknown -- this must always be last */
} MessageType;

/*
 * Provide a weak definition of xf86DrvMsgVerb() so applications that
 * don't #define OS_TRACE_USE_XORG can link against liboska without
 * having to also link against the X.org libraries.
 */
void xf86DrvMsgVerb(int scrnIndex, MessageType type, int verb,
                    const char *format, ...) __attribute__((weak));

/* xf86VDrvMsgVerb() checks if scrnIndex is >=0 before printing the
   driver prefix. */
#define NO_SCREEN -1

void os_trace_xorg_err(const char *format, ...)
{
    va_list args;
    char buf[OS_TRACE_BUF_LEN];

    va_start(args, format);
    os_trace_print(buf, format, args);
    va_end(args);

    xf86DrvMsgVerb(NO_SCREEN, X_ERROR, 1, buf);
}

void os_trace_xorg_warn(const char *format, ...)
{
    va_list args;
    char buf[OS_TRACE_BUF_LEN];

    va_start(args, format);
    os_trace_print(buf, format, args);
    va_end(args);
    
    xf86DrvMsgVerb(NO_SCREEN, X_WARNING, 1, buf);
}

void os_trace_xorg_info(const char *format, ...)
{
    va_list args;
    char buf[OS_TRACE_BUF_LEN];

    va_start(args, format);
    os_trace_print(buf, format, args);
    va_end(args);
    
    xf86DrvMsgVerb(NO_SCREEN, X_INFO, 1, buf);
}

void os_trace_xorg_dbg(const char *format, ...)
{
    va_list args;
    char buf[OS_TRACE_BUF_LEN];

    va_start(args, format);
    os_trace_print(buf, format, args);
    va_end(args);
    
    xf86DrvMsgVerb(NO_SCREEN, X_INFO, 2, buf);
}

void xf86DrvMsgVerb(int scrnIndex, MessageType type, int verb,
                    const char *format, ...)
{
}
