/*
 * Core BT Type-A operations.
 *
 * Copyright (C) 2006-2008 Cambridge Silicon Radio Ltd.
 *
 * Refer to LICENSE.txt included with this source code for details on
 * the license terms.
 */
#include <oska/all.h>
#include <oska/trace.h>

#include <sdioemb/sdio.h>
#include <sdioemb/sdio_cis.h>
#include <sdioemb/sdio_bt_a.h>
#include <csr_sdio.h>
#include <csr_sdio_lib.h>

#define BC_FIRMWARE_START_TIMEOUT_MS 100
#define SDIO_CSR_TO_HOST_SCRATCH0_INITED 0x1

/**
 * Are read acknowledges required?
 *
 * @param bt the BT device.
 *
 * @return true if read acknowledges are required; false otherwise.
 */
static bool needs_read_ack(struct sdioemb_bt_a_dev *bt)
{
    /* CSR chips don't need read acks. */
    if (bt->func->sdioId.manfId == SDIO_MANF_ID_CSR) {
        return false;
    }
    return true;
}

/**
 * Disable read acknowledges if supported by the hardware.
 *
 * @param bt the BT device.
 */
static void disable_read_acks(struct sdioemb_bt_a_dev *bt)
{
    CsrResult result;

    if (bt->needs_read_ack) {
        return;
    }

    result = CsrSdioWrite8(bt->func, SDIO_BT_A_RETRY_CTRL, RTC_SET);
    if (result) {
        os_trace_warn("%s: could not disable read acks (%d)",
                      bt->name, CsrSdioResultToStatus(result));
        bt->needs_read_ack = true;
    }
}


static void start_func(struct sdioemb_bt_a_dev *bt)
{
    int timeout_ms = BC_FIRMWARE_START_TIMEOUT_MS;

    sdioemb_bt_a_set_sleep_state(bt, SLEEP_STATE_AWAKE);
    CsrSdioInterruptEnable(bt->func);

    /*
     * For some hardware the I/O Ready bit is set before the device is
     * ready. In this case we wait for firmware to start by polling
     * bit 0 in TO_HOST_SCRATCH0.
     */
    if (bt->wait_for_firmware) {
        while (--timeout_ms) {
            uint8_t scratch0;
            CsrSdioF0Read8(bt->func, SDIO_CSR_TO_HOST_SCRATCH0, &scratch0);
            if (scratch0 == SDIO_CSR_TO_HOST_SCRATCH0_INITED) {
                break;
            }
            os_sleep_ms(1);
        }
        if (timeout_ms == 0) {
            os_trace_warn("%s: timed out waiting for firmware to start", bt->name);
        }
    }

    disable_read_acks(bt);
    CsrSdioWrite8(bt->func, SDIO_BT_A_INT_EN, EN_INTRD);
}

/**
 * Start a BT Type-A function.
 *
 * The function is enabled, read acknowledges are disabled (if
 * supported by the hardware) and finally interrupts are enabled.
 *
 * The sleep state is also set to AWAKE to ensure the software and
 * hardware deep sleep states are in sync.
 *
 * @param bt the BT device.
 */
void sdioemb_bt_a_start(struct sdioemb_bt_a_dev *bt)
{
    CsrSdioFunctionEnable(bt->func);
    start_func(bt);
}

/**
 * Stop a BT Type-A function.
 *
 * The function is disabled.
 *
 * @param bt the BT device.
 */
void sdioemb_bt_a_stop(struct sdioemb_bt_a_dev *bt)
{
    CsrSdioFunctionActive(bt->func);
    CsrSdioWrite8(bt->func, SDIO_BT_A_INT_EN, 0);
    CsrSdioInterruptDisable(bt->func);
    CsrSdioFunctionDisable(bt->func);
}

/**
 * Restart the function if a hardware reset occurred.
 *
 * If the hardware is reset by an internal watchdog (or similar), the
 * function must be reenabled prior to attempting any SDIO commands to
 * it.
 *
 * A watchdog reset is identified by the I/O Enable bit being set and
 * the Extended I/O Enable bit being clear.  In all uses of the
 * function we expect I/O Enable to be set so on the Extended I/O
 * Enable bit is checked.
 *
 * @param bt the BT device.
 *
 * @return 1 if a reset had occurred; 0 if a reset had not occurred;
 * -ve on error (-ENODEV, -EIO etc).
 */
int sdioemb_bt_a_check_for_reset(struct sdioemb_bt_a_dev *bt)
{
    uint8_t ext_io_en;
    CsrResult result;

    result = CsrSdioF0Read8(bt->func, SDIO_CSR_EXT_IO_EN, &ext_io_en);
    if (result) {
        return CsrSdioResultToStatus(result);
    }
    if (!(ext_io_en & (1 << bt->func->sdioId.sdioFunction))) {
        os_trace_warn("%s: watchdog reset detected", bt->name);
        CsrSdioFunctionReenable(bt->func);
        start_func(bt);
        return 1;
    }
    return 0;
}

/**
 * Write a packet to the hardware.
 *
 * The entire packet is written in a single SDIO command.
 *
 * I/O errors cause the packet write to be retried, fatal errors (too
 * many retries) will be reported to the caller.
 *
 * @param bt     the BT device.
 * @param packet packet to write including the Type-A header.
 * @param len    total length of the packet.
 *
 * @return 0 on success; -ve on error.
 */
int sdioemb_bt_a_send(struct sdioemb_bt_a_dev *bt, const uint8_t *packet, size_t len)
{
    CsrResult result;
    int retry = 0;

    do {
        result = CsrSdioWrite(bt->func, SDIO_BT_A_TD, packet, len);
        if (result == CSR_RESULT_SUCCESS)
            return 0;

        os_trace_warn("%s: write packet failed (%d), retrying",
                      bt->name, CsrSdioResultToStatus(result));
        CsrSdioWrite8(bt->func, SDIO_BT_A_TX_PKT_CTRL, PC_WRT);
    } while (++retry < bt->max_tx_retries);

    os_trace_err("%s: maximum number of write retries exceeded", bt->name);
    sdioemb_bt_a_check_for_reset(bt);

    return CsrSdioResultToStatus(result);
}

/**
 * Read a packet from the hardware.
 *
 * The 4 octet Type-A transport header and part of the payload data is
 * read.  The header is used to determine the total length of the
 * packet and the remaining data (if any) is read (using a read padded
 * to a whole number of blocks).
 *
 * If an I/O error occurs the packet read is retried.  Fatal errors
 * (no memory, too many retries) cause the packet to be silently
 * discarded.
 *
 * The read packet ready interrupt is cleared prior to reading (or
 * re-reading) a packet so the hardware will make the packet data
 * available.
 *
 * @param bt            the BT device.
 * @param buffer_handle returns the buffer handled supplied for this packet.
 *
 * @return 0 on success; -ve on error.
 */
static int rx_packet(struct sdioemb_bt_a_dev *bt, void **buffer_handle)
{
    CsrResult result;
    int ret = 0;
    int retry = 0;
    size_t packet_len = 0;
    size_t remaining_len = 0;
    uint8_t *buffer;
    size_t buffer_min_len;
    size_t to_copy;

    do {
        CsrSdioWrite8(bt->func, SDIO_BT_A_INTRD, CL_INTRD);

        /* Read header and part of the packet data. */
        result = CsrSdioRead(bt->func, SDIO_BT_A_RD, bt->header, SDIO_BT_A_MIN_READ);
        ret = CsrSdioResultToStatus(result);
        if (ret < 0) {
            goto retry_read;
        }

        /*
         * Calculate the size of buffer needed, request it and copy
         * the already read data into it.
         */
        if (packet_len == 0) {
            packet_len = sdioemb_bt_a_packet_len(bt->header);

            if (packet_len < SDIO_BT_A_HEADER_LEN || packet_len > SDIO_BT_A_PACKET_LEN_MAX) {
                os_trace_err("%s: bad rx packet length (%d)", bt->name, packet_len);
                ret = -EINVAL;
                goto done;
            }

            if (packet_len > SDIO_BT_A_MIN_READ) {
                /*
                 * Pad the read of the rest of the data to either a
                 * whole number of blocks (to reduce the number of
                 * CMD53s); or a whole number of words (in the case
                 * host controller can't efficiently DMA partial
                 * words).
                 */
                size_t block_size = 4;

                remaining_len = packet_len - SDIO_BT_A_MIN_READ;
                if (remaining_len > bt->func->blockSize) {
                    block_size = bt->func->blockSize;
                }
                remaining_len = ((remaining_len + (block_size - 1)) / block_size) * block_size;
                buffer_min_len = SDIO_BT_A_MIN_READ + remaining_len;
                to_copy = SDIO_BT_A_MIN_READ;
            } else {
                buffer_min_len = packet_len;
                to_copy = packet_len;
            }

            ret = bt->get_rx_buffer(bt, bt->header, buffer_min_len, &buffer, buffer_handle);
            if (ret < 0) {
                os_trace_err("%s: no buffer for Rx packet", bt->name);
                goto done;
            }
            memcpy(buffer, bt->header, to_copy);
        }

        /* Read any remaining data. */
        if (remaining_len) {
            result = CsrSdioRead(bt->func, SDIO_BT_A_RD,
                                 buffer + SDIO_BT_A_MIN_READ, remaining_len);
            ret = CsrSdioResultToStatus(result);
        }
        if (ret == 0)
            goto done;

      retry_read:
        CsrSdioWrite8(bt->func, SDIO_BT_A_RX_PKT_CTRL, PC_RRT);
    } while (++retry < bt->max_rx_retries);

  done:
    if (ret != -ENODEV) {
        /* Acknowledge read if the hardware requires it, or if an error
           occured (so the hardware discards the packet). */
        if (bt->needs_read_ack || ret < 0) {
            CsrSdioWrite8(bt->func, SDIO_BT_A_RX_PKT_CTRL, 0x00);
        }
        if (retry >= bt->max_rx_retries) {
            os_trace_err("%s: maximum number of read retries exceeded", bt->name);
            sdioemb_bt_a_check_for_reset(bt);
        }
    }
    return ret;
}

/**
 * Set the deep sleep state.
 *
 * The deep sleep state management is necessary as SDIO commands to
 * the BT Type-A function fail when the hardware is in deep sleep,
 * thus prior to reading or writing packets we must ensure the chip is
 * awake.
 *
 * The following state transitions are handled.
 *
 * AWAKE -> TORPID: In response to an idle timer expiring (or
 * similar), simply write TORPID to the state register to permit the
 * chip to enter deep sleep.
 *
 * TORPID -> DROWSY: In response to the host software wishing to write
 * a packet, set the state to DROWSY and wait for the state to be
 * AWAKE.
 *
 * DROWSY/TORPID -> AWAKE: An interrupt indicates that the chip is
 * awake, set the state to AWAKE and wake up any threads waiting for
 * AWAKE.
 *
 * @param bt    the BT device.
 * @param state state to transition to.
 */
void sdioemb_bt_a_set_sleep_state(struct sdioemb_bt_a_dev *bt, enum sdio_sleep_state state)
{
    CsrResult result;

    if (state == bt->sleep_state) {
        return;
    }

    if (bt->sleep_state == SLEEP_STATE_TORPID) {
        CsrSdioFunctionActive(bt->func);
    }
    if (state == SLEEP_STATE_TORPID) {
        CsrSdioFunctionIdle(bt->func);
    }

    /*
     * The awake interrupt is sourced from the 'core clock running'
     * signal, this is deasserted 1 slow clock (1 kHz) cycle after the
     * core clock is stopped.  If the DROWSY state is set within 1 ms
     * of the core clock being stopped then a spurious wake up
     * interrupt occurs.
     *
     * Setting AWAKE and then waiting 1 ms before setting DROWSY
     * ensures we don't set DROWSY during this unsafe window.
     */
    if (state == SLEEP_STATE_DROWSY) {
        CsrSdioF0Write8(bt->func, SDIO_CSR_SLEEP_STATE,
                        SDIO_CSR_SLEEP_STATE_FUNC(bt->func->sdioId.sdioFunction)
                        | SLEEP_STATE_AWAKE);
        os_sleep_ms(1);
    }

    result = CsrSdioF0Write8(bt->func, SDIO_CSR_SLEEP_STATE,
                             SDIO_CSR_SLEEP_STATE_FUNC(bt->func->sdioId.sdioFunction) | state);
    if (result) {
        return;
    }
    bt->sleep_state = state;

    bt->sleep_state_changed(bt);
}

/**
 * Handle an interrupt from the function.
 *
 * The chip will only raise an interrupts if it is awake and will not
 * enter deep sleep while an interrupt is asserted, so we first set
 * the sleep state to AWAKE.
 *
 * The sources of interrupt is then identified and handled.  The three
 * possible sources are:
 *
 *   1. Read packet ready.  The packet is read from the hardware and
 *      added to the received packet queue.
 *
 *   2. Awake.  This may only occur when the sleep state is DROWSY and
 *      no action other than setting the sleep state is required.
 *
 *   3. Generic interrupt.  The hardware provides a facility for the
 *      firmware to raise an interrupt (via the SDIO_HOST_INT
 *      register).  The current firmware makes no use of this so we
 *      just clear the interrupt.
 *
 * @param bt the BT device.
 */
void sdioemb_bt_a_handle_interrupt(struct sdioemb_bt_a_dev *bt)
{
    uint8_t rdint;
    CsrResult result;
    int err;
    int wake_up_expected;
    void *buffer_handle = NULL;

    wake_up_expected = bt->sleep_state == SLEEP_STATE_DROWSY;
    sdioemb_bt_a_set_sleep_state(bt, SLEEP_STATE_AWAKE);

    /* If a wake up isn't expected, the only interrupt source is
       'packet read ready' so skip reading the INTRD register. */
    if (wake_up_expected) {
        result = CsrSdioRead8(bt->func, SDIO_BT_A_INTRD, &rdint);
        if (result) {
            if (result != CSR_SDIO_RESULT_NO_DEVICE
                && sdioemb_bt_a_check_for_reset(bt) == 1) {
                result = CsrSdioRead8(bt->func, SDIO_BT_A_INTRD, &rdint);
            }
        }
        if (result || !(rdint & INTRD)) {
            CsrSdioInterruptAcknowledge(bt->func);
            return;
        }
    }

    err = rx_packet(bt, &buffer_handle);
    CsrSdioInterruptAcknowledge(bt->func);
    if (buffer_handle) {
        bt->receive(bt, buffer_handle, err);
    }
}


/**
 * Turn on the reception of packets.
 *
 * Packet reception is on by default so this only needs to be called
 * if sdioemb_bt_a_rx_off() was previously called.
 *
 * This enables the Type-A packet read ready interrupt.
 *
 * @param bt the BT device.
 */
void sdioemb_bt_a_rx_on(struct sdioemb_bt_a_dev *bt)
{
    if (bt->rx_off) {
        CsrSdioWrite8(bt->func, SDIO_BT_A_INT_EN, EN_INTRD);
        bt->rx_off = 0;
    }
}

/**
 * Turn off the reception of packets.
 *
 * Use this it flow control packets being read from the device if
 * there is insufficient space to receive new packets.
 *
 * Reenable packet reception with sdioemb_bt_a_rx_on().
 *
 * This disables the Type-A packet read ready interrupt.
 *
 * @param bt the BT device.
 */
void sdioemb_bt_a_rx_off(struct sdioemb_bt_a_dev *bt)
{
    if (!bt->rx_off) {
        CsrSdioWrite8(bt->func, SDIO_BT_A_INT_EN, 0x00);
        bt->rx_off = 1;
    }
}


/**
 * Initialize the bluetooth device structure.
 *
 * Call this once before calling sdioemb_bt_a_start().
 *
 * @param bt the BT device.
 * @param func the Bluetooth SDIO function.
 *
 * @return 0 on sucess.
 * @return -ENOMEM if the header receive buffer could not be allocated.
 */
int sdioemb_bt_a_setup(struct sdioemb_bt_a_dev *bt, CsrSdioFunction *func)
{
    bt->header = os_alloc(SDIO_BT_A_MIN_READ);
    if (bt->header == NULL) {
        return -ENOMEM;
    }

    bt->func              = func;
    bt->max_tx_retries    = 3;
    bt->max_rx_retries    = 3;
    bt->needs_read_ack    = needs_read_ack(bt);
    bt->wait_for_firmware = 0;
    bt->sleep_state       = SLEEP_STATE_TORPID; /* safe state to assume */

    return 0;
}

void sdioemb_bt_a_cleanup(struct sdioemb_bt_a_dev *bt)
{
    os_free(bt->header);
}
