/*
 * Sample SDIO function driver.
 *
 * Copyright (C) 2010 Cambridge Silicon Radio Ltd.
 *
 * Refer to LICENSE.txt included with this source code for details on
 * the license terms.
 *
 * This is a sample function driver for a CSR generic function.  It
 * just enables the function and interrupt and triggers (then
 * acknowledges) interrupts every 2 seconds.
 */
#if defined(__linux__)
#  define OS_TRACE_PREFIX "sample: "
#endif
#include <oska/trace.h>
#include <oska/all.h>

#include <csr_sdio.h>
#include <csr_sdio_lib.h>
#include <sdioemb/sdio.h>
#include <sdioemb/sdio_csr.h>

#ifdef OS_TRACE_INCLUDE_TMH
#include "sample.tmh"
#endif

struct sample_data {
    os_thread_t work_thread;
    os_event_t  interrupt_event;
};

static void work_func(void *arg)
{
    CsrSdioFunction *func = arg;
    struct sample_data *sd = func->driverData;
    uint16_t i = 0, a, b;
  
    while (!os_thread_should_stop(&sd->work_thread)) {
        /* Write some data to local RAM and read it back. */
        a = i++;
        CsrSdioWrite16(func, (0x8000 << 1), a);
        CsrSdioRead16(func, (0x8000 << 1), &b);
        if (a != b) {
            os_trace_err("a (%u) != b (%u)", a, b);
        }

        /* Trigger an interrupt by writing to the generic function's
           SDIO_HOST_INT register. */
        CsrSdioWrite8(func, (0xF92F << 1), 1);

        os_event_wait(&sd->interrupt_event);
        if (os_thread_should_stop(&sd->work_thread)) {
            break;
        }

        os_trace_info("acknowledge interrupt");

        /* Acknowledge the interrupt. */
        CsrSdioF0Write8(func, SDIO_CSR_HOST_INT, 1);
        CsrSdioInterruptAcknowledge(func);

        os_sleep_ms(2000);
    }
}

static int start_function(CsrSdioFunction *func)
{
    struct sample_data *sd = func->driverData;
    char name[32];
    CsrResult result;
    int ret;

    result = CsrSdioFunctionEnable(func);
    if (result != CSR_RESULT_SUCCESS) {
        os_trace_err("failed to enable function: %d", result);
        return CsrSdioResultToStatus(result);
    }

    result = CsrSdioInterruptEnable(func);
    if (result != CSR_RESULT_SUCCESS) {
        os_trace_err("failed to enable interrupt: %d", result);
        return CsrSdioResultToStatus(result);
    }

    snprintf(name, sizeof(name), "sample");
    ret = os_thread_create(&sd->work_thread, name, work_func, func);
    if (ret < 0) {
        os_trace_err("failed to start work thread: %d", result);
        return ret;
    }
    return 0;
}

static void stop_function(CsrSdioFunction *func)
{
    struct sample_data *sd = func->driverData;
    
    os_thread_stop(&sd->work_thread, &sd->interrupt_event);
    CsrSdioInterruptDisable(func);
    CsrSdioFunctionDisable(func);
}

static struct sample_data *alloc_drv_data(CsrSdioFunction *func)
{
    struct sample_data *sd;

    sd = os_alloc(sizeof(*sd));
    if (sd == NULL) {
        return NULL;
    }
    func->driverData = sd;

    os_event_init(&sd->interrupt_event);

    return sd;
}

static void free_drv_data(struct sample_data *sd)
{
    os_event_destroy(&sd->interrupt_event);
    os_free(sd);
}


static void sample_inserted(CsrSdioFunction *func)
{
    struct sample_data *sd;
    int ret;

    os_trace_info("probe %04x:%04x F%d",
                  func->sdioId.manfId, func->sdioId.cardId, func->sdioId.sdioFunction);

    sd = alloc_drv_data(func);
    if (sd == NULL) {
        ret = -ENOMEM;
        goto out;
    }

    ret = start_function(func);
    if (ret < 0) {
        goto out;
    }

  out:
    if (ret < 0) {
        CsrSdioInterruptDisable(func);
        CsrSdioFunctionDisable(func);
        free_drv_data(sd);
    }
    CsrSdioInsertedAcknowledge(func, CsrSdioStatusToResult(ret));
}

static void sample_removed(CsrSdioFunction *func)
{
    struct sample_data *sd = func->driverData;

    os_trace_info("remove");

    stop_function(func);
    free_drv_data(sd);

    CsrSdioRemovedAcknowledge(func);
}

static void sample_dsr(CsrSdioFunction *func)
{
    struct sample_data *sd = func->driverData;

    os_trace_info("interrupt");

    /*
     * The interrupt callback is called in a non-thread context so all
     * we can do is raise an event to wake a thread to handle the
     * interrupt.
     */
    os_event_raise(&sd->interrupt_event, 1);
}

static CsrSdioInterruptDsrCallback sample_isr(CsrSdioFunction *func)
{
    return sample_dsr;
}

static void sample_suspend(CsrSdioFunction *func)
{
    os_trace_info("suspend");

    stop_function(func);

    /* Don't need to retain power when suspended. */
    CsrSdioPowerOff(func);

    CsrSdioSuspendAcknowledge(func, CSR_RESULT_SUCCESS);
}

static void sample_resume(CsrSdioFunction *func)
{
    os_trace_info("resume");

    /* Suspend powered off the function, turn it back on. */
    CsrSdioPowerOn(func);

    start_function(func);

    CsrSdioResumeAcknowledge(func, CSR_RESULT_SUCCESS);
}

static CsrSdioFunctionId sample_id_table[] = {
    { SDIO_MANF_ID_CSR, SDIO_CARD_ID_CSR_DASH, 2, CSR_SDIO_ANY_SDIO_INTERFACE },
    { SDIO_MANF_ID_CSR, SDIO_CARD_ID_CSR_CINDERELLA, 1, CSR_SDIO_ANY_SDIO_INTERFACE },
};

static CsrSdioFunctionDriver sample_driver = {
    sample_inserted,
    sample_removed,
    sample_isr,
    sample_suspend,
    sample_resume,
    sample_id_table,
    sizeof(sample_id_table) / sizeof(sample_id_table[0]),
};

/**
 * Register the sample function driver.
 */
int sample_register(void)
{
    CsrResult result;
    
    result = CsrSdioFunctionDriverRegister(&sample_driver);
    return CsrSdioResultToStatus(result);
}

/**
 * Unregister the sample function driver.
 */
void sample_unregister(void)
{
    CsrSdioFunctionDriverUnregister(&sample_driver);
}
