/*
 * Synergy compatible API.
 *
 * Copyright (C) 2010 Cambridge Silicon Radio Ltd.
 *
 * Refer to LICENSE.txt included with this source code for details on
 * the license terms.
 */
#include "csr_sdio_internal.h"

struct glue_driver {
    struct sdioemb_func_driver driver; /* must be first */
    CsrSdioFunctionDriver *functionDriver;
};

static int glue_probe(struct sdioemb_dev *fdev)
{
    struct glue_driver *glue_driver = (struct glue_driver *)fdev->driver;
    struct glue_func *glue;
    CsrStatus ret;

    glue = os_alloc(sizeof(*glue));
    if (glue == NULL) {
        return -ENOMEM;
    }
    fdev->drv_data = glue;

    glue->function.sdioId.manfId = fdev->vendor_id;
    glue->function.sdioId.cardId = fdev->device_id;
    glue->function.sdioId.sdioFunction = fdev->function;
    glue->function.sdioId.sdioInterface = fdev->interface;
    glue->function.blockSize = fdev->blocksize;
    glue->function.features = CSR_SDIO_FEATURE_BYTE_MODE
        | CSR_SDIO_FEATURE_DMA_CAPABLE_MEM_REQUIRED;
    glue->function.cardHandle = (void *)fdev->slot_id;
    glue->function.osDevice = fdev->os_device;
    glue->function.priv = glue;
    glue->fdev = fdev;

    os_event_init(&glue->callback_event);

    glue_driver->functionDriver->inserted(&glue->function);
    os_event_wait(&glue->callback_event);

    ret = CsrSdioResultToStatus(glue->inserted_result);
    if (CSR_STATUS_FAILURE(ret)) {
        os_free(glue);
    }
    return ret;
}

static void glue_remove(struct sdioemb_dev *fdev)
{
    struct glue_driver *glue_driver = (struct glue_driver *)fdev->driver;
    struct glue_func *glue = fdev->drv_data;

    glue_driver->functionDriver->removed(&glue->function);
    os_event_wait(&glue->callback_event);
    os_free(glue);
}

static void glue_interrupt(struct sdioemb_dev *fdev)
{
    struct glue_driver *glue_driver = (struct glue_driver *)fdev->driver;
    struct glue_func *glue = fdev->drv_data;
    CsrSdioInterruptDsrCallback dsr;

    dsr = glue_driver->functionDriver->intr(&glue->function);
    if (dsr) {
        dsr(&glue->function);
    }
}

static void glue_suspend(struct sdioemb_dev *fdev)
{
    struct glue_driver *glue_driver = (struct glue_driver *)fdev->driver;
    struct glue_func *glue = fdev->drv_data;

    glue_driver->functionDriver->suspend(&glue->function);
    os_event_wait(&glue->callback_event);
}

static void glue_resume(struct sdioemb_dev *fdev)
{
    struct glue_driver *glue_driver = (struct glue_driver *)fdev->driver;
    struct glue_func *glue = fdev->drv_data;

    glue_driver->functionDriver->resume(&glue->function);
    os_event_wait(&glue->callback_event);
}

CsrResult CsrSdioFunctionDriverRegister(CsrSdioFunctionDriver *functionDriver)
{
    struct glue_driver *glue_driver;
    struct sdioemb_id_table *ids;
    unsigned i;
    int ret;

    glue_driver = os_alloc(sizeof(*glue_driver)
                           + (functionDriver->idsCount + 1) * sizeof(struct sdioemb_id_table));
    if (glue_driver == NULL) {
        return CSR_RESULT_FAILURE;
    }
    functionDriver->priv = glue_driver;

    ids = (struct sdioemb_id_table *)&glue_driver[1];
    for (i = 0; i < functionDriver->idsCount; i++) {
        ids[i].vendor_id = functionDriver->ids[i].manfId;
        ids[i].device_id = functionDriver->ids[i].cardId;
        ids[i].function  = functionDriver->ids[i].sdioFunction;
        ids[i].interface = functionDriver->ids[i].sdioInterface;
    }

    glue_driver->driver.name     = "glue";
    glue_driver->driver.id_table = ids;
    glue_driver->driver.probe    = glue_probe;
    glue_driver->driver.remove   = glue_remove;
    glue_driver->driver.card_int_handler = glue_interrupt;
    glue_driver->driver.suspend  = glue_suspend;
    glue_driver->driver.resume   = glue_resume;

    glue_driver->functionDriver = functionDriver;

    ret = sdioemb_driver_register(&glue_driver->driver);
    if (ret < 0) {
        os_free(glue_driver);
        return CsrSdioStatusToResult(ret);
    }
    return CSR_RESULT_SUCCESS;
}

void CsrSdioFunctionDriverUnregister(CsrSdioFunctionDriver *functionDriver)
{
    struct glue_driver *glue_driver = functionDriver->priv;

    sdioemb_driver_unregister(&glue_driver->driver);
    os_free(glue_driver);
}


void CsrSdioInsertedAcknowledge(CsrSdioFunction *function, CsrResult result)
{
    struct glue_func *glue = function->priv;

    glue->inserted_result = result;
    os_event_raise(&glue->callback_event, 1);
}

void CsrSdioRemovedAcknowledge(CsrSdioFunction *function)
{
    struct glue_func *glue = function->priv;

    os_event_raise(&glue->callback_event, 1);
}

void CsrSdioSuspendAcknowledge(CsrSdioFunction *function, CsrResult result)
{
    struct glue_func *glue = function->priv;

    os_event_raise(&glue->callback_event, 1);
}

void CsrSdioResumeAcknowledge(CsrSdioFunction *function, CsrResult result)
{
    struct glue_func *glue = function->priv;

    os_event_raise(&glue->callback_event, 1);
}

CsrResult CsrSdioFunctionEnable(CsrSdioFunction *function)
{
    struct glue_func *glue = function->priv;
    CsrStatus ret;

    ret = sdioemb_enable_function(glue->fdev);
    return CsrSdioStatusToResult(ret);
}

CsrResult CsrSdioFunctionDisable(CsrSdioFunction *function)
{
    struct glue_func *glue = function->priv;
    CsrStatus ret;

    ret = sdioemb_disable_function(glue->fdev);
    return CsrSdioStatusToResult(ret);
}

CsrResult CsrSdioInterruptEnable(CsrSdioFunction *function)
{
    struct glue_func *glue = function->priv;
    CsrStatus ret;

    ret = sdioemb_interrupt_enable(glue->fdev);
    return CsrSdioStatusToResult(ret);
}

CsrResult CsrSdioInterruptDisable(CsrSdioFunction *function)
{
    struct glue_func *glue = function->priv;
    CsrStatus ret;

    ret = sdioemb_interrupt_disable(glue->fdev);
    return CsrSdioStatusToResult(ret);
}

void CsrSdioInterruptAcknowledge(CsrSdioFunction *function)
{
    struct glue_func *glue = function->priv;

    sdioemb_interrupt_acknowledge(glue->fdev);
}

CsrResult CsrSdioBlockSizeSet(CsrSdioFunction *function, CsrUint16 blockSize)
{
    struct glue_func *glue = function->priv;
    CsrStatus ret;

    ret = sdioemb_set_block_size(glue->fdev, blockSize);
    if (ret >= 0) {
        function->blockSize = glue->fdev->blocksize;
    }
    return CsrSdioStatusToResult(ret);
}

CsrResult CsrSdioMaxBusClockFrequencySet(CsrSdioFunction *function, CsrUint32 maxFrequency)
{
    struct glue_func *glue = function->priv;

    sdioemb_set_max_bus_freq(glue->fdev, maxFrequency);
    return CSR_RESULT_SUCCESS;
}

CsrResult CsrSdioRead8(CsrSdioFunction *function, CsrUint32 address, CsrUint8 *data)
{
    struct glue_func *glue = function->priv;
    CsrStatus ret;

    ret = sdioemb_read8(glue->fdev, address, data);
    return CsrSdioStatusToResult(ret);
}

CsrResult CsrSdioRead16(CsrSdioFunction *function, CsrUint32 address, CsrUint16 *data)
{
    struct glue_func *glue = function->priv;
    CsrStatus ret;

    ret = sdioemb_read16(glue->fdev, address, data);
    return CsrSdioStatusToResult(ret);
}

CsrResult CsrSdioWrite8(CsrSdioFunction *function, CsrUint32 address, CsrUint8 data)
{
    struct glue_func *glue = function->priv;
    CsrStatus ret;

    ret = sdioemb_write8(glue->fdev, address, data);
    return CsrSdioStatusToResult(ret);
}

CsrResult CsrSdioWrite16(CsrSdioFunction *function, CsrUint32 address, CsrUint16 data)
{
    struct glue_func *glue = function->priv;
    CsrStatus ret;

    ret = sdioemb_write16(glue->fdev, address, data);
    return CsrSdioStatusToResult(ret);
}

CsrResult CsrSdioF0Read8(CsrSdioFunction *function, CsrUint32 address, CsrUint8 *data)
{
    struct glue_func *glue = function->priv;
    CsrStatus ret;

    ret = sdioemb_f0_read8(glue->fdev, address, data);
    return CsrSdioStatusToResult(ret);
}

CsrResult CsrSdioF0Write8(CsrSdioFunction *function, CsrUint32 address, CsrUint8 data)
{
    struct glue_func *glue = function->priv;
    CsrStatus ret;

    ret = sdioemb_f0_write8(glue->fdev, address, data);
    return CsrSdioStatusToResult(ret);
}

CsrResult CsrSdioRead(CsrSdioFunction *function, CsrUint32 address,
                      void *data, CsrUint32 length)
{
    struct glue_func *glue = function->priv;
    CsrStatus ret;

    ret = sdioemb_read(glue->fdev, address, data, length);
    return CsrSdioStatusToResult(ret);
}

CsrResult CsrSdioWrite(CsrSdioFunction *function, CsrUint32 address,
                       const void *data, CsrUint32 length)
{
    struct glue_func *glue = function->priv;
    CsrStatus ret;

    ret = sdioemb_write(glue->fdev, address, data, length);
    return CsrSdioStatusToResult(ret);
}

CsrResult CsrSdioPowerOn(CsrSdioFunction *function)
{
    struct glue_func *glue = function->priv;

    sdioemb_power_on(glue->fdev);
    return CSR_RESULT_SUCCESS;
}

void CsrSdioPowerOff(CsrSdioFunction *function)
{
    struct glue_func *glue = function->priv;

    sdioemb_power_off(glue->fdev);
}

CsrResult CsrSdioHardReset(CsrSdioFunction *function)
{
    struct glue_func *glue = function->priv;
    int ret;

    ret = sdioemb_hard_reset(glue->fdev);
    if (ret == 1) {
        return CSR_SDIO_RESULT_NOT_RESET;
    }
    return CsrSdioStatusToResult(ret);
}

void CsrSdioFunctionActive(CsrSdioFunction *function)
{
    /* Functions are activated automatically when a command is
       started. */
}

void CsrSdioFunctionIdle(CsrSdioFunction *function)
{
    struct glue_func *glue = function->priv;

    sdioemb_idle_function(glue->fdev);
}
