/*
 * Card detection and initialization.
 *
 * Copyright (C) 2007-2008 Cambridge Silicon Radio Ltd.
 *
 * Refer to LICENSE.txt included with this source code for details on
 * the license terms.
 */
#include <sdioemb/sdio_cis.h>

#include "sdio_internal.h"
#include "sdio_config.h"

#ifdef OS_TRACE_INCLUDE_TMH
#include "sdio_card.tmh"
#endif

static struct sdioemb_card_io_ops sdio_io_ops = {
    sdio_io_read8,
    sdio_io_read16,
    sdio_io_write8,
    sdio_io_write16,
    sdio_io_read,
    sdio_io_write,
};

static int sdio_card_init_sequence(struct sdioemb_slot *slot)
{
    struct sdioemb_slot_priv *slotp = slot->priv;
    int ret;
    union sdioemb_response response;
    uint32_t ocr, rca;
    int timeout;

    /* CMD0. Needed to start some cards. */
    ret = sdio_raw_cmd(slot, 0, 0, SDIOEMB_CMD_FLAG_RESP_NONE, &response);
    if (ret)
        goto cmd_error;

    /* CMD5. Read OCR. */
    ret = sdio_raw_cmd(slot, 5, 0, SDIOEMB_CMD_FLAG_RESP_R4, &response);
    if (ret)
        goto cmd_error;
    ocr = response.r4;
    if (!(ocr & SDIO_OCR_VOLTAGE_3V3)) {
        os_trace_err("%s: card does not support 3.3V", slot->name);
        goto fatal_card_error;
    }
    slotp->num_functions = (ocr & SDIO_OCR_NUM_FUNCS_MASK) >> SDIO_OCR_NUM_FUNCS_OFFSET;

    /* CMD5. Set operating voltage to 3.3V and wait for card to be
     * ready. */
    for (timeout = SDIOEMB_CARD_READY_TIMEOUT_MS; timeout > 0; timeout--) {
        ret = sdio_raw_cmd(slot, 5, SDIO_OCR_VOLTAGE_3V3, SDIOEMB_CMD_FLAG_RESP_R4, &response);
        if (ret) {
            goto cmd_error;
        }
        ocr = response.r4;
        if (ocr & SDIO_OCR_CARD_READY) {
            break;
        }
        os_sleep_ms(1);
    }
    if (timeout == 0) {
        os_trace_err("%s: card not ready", slot->name);
        goto fatal_card_error;
    }

    if (slot->type == SDIOEMB_SLOT_TYPE_SD) {
        /* CMD3. Set/read RCA. */
        ret = sdio_raw_cmd(slot, 3, 0, SDIOEMB_CMD_FLAG_RESP_R6, &response);
        if (ret)
            goto cmd_error;
        rca = response.r6 & 0xffff0000;

        /* CMD7. Select card. */
        ret = sdio_raw_cmd(slot, 7, rca, SDIOEMB_CMD_FLAG_RESP_R1B, &response);
        if (ret)
            goto cmd_error;
    }

    /* Successful card initialization. */
    return 0;

  cmd_error:
    return -EIO;
  fatal_card_error:
    return -ENODEV;
}

static int sdio_card_init(struct sdioemb_slot *slot)
{
    struct sdioemb_slot_priv *slotp = slot->priv;
    int t;
    int ret;

    for (t = 0; t < SDIOEMB_CARD_INIT_RETRY_MAX; t++) {
        ret = sdio_card_init_sequence(slot);
        if (ret == -ENODEV)
            goto fatal_card_error;
        if (ret == 0)
            break;
    }
    if (t == SDIOEMB_CARD_INIT_RETRY_MAX) {
        goto fatal_card_error;
    }

    slotp->io_ops = sdio_io_ops;

    return 0;

  fatal_card_error:
    /* Card isn't functional, present, or is not supported. */
    return -ENODEV;
}

int sdio_card_start(struct sdioemb_slot *slot)
{
    struct sdioemb_slot_priv *slotp = slot->priv;
    int r;

    sdio_event_log(slot, SDIOEMB_EVENT_POWER_ON);

    slot->card_power(slot, SDIOEMB_POWER_3V3);
    os_sleep_ms(SDIO_POWER_UP_TIME_MS);
    slotp->card_powered = 1;

    sdio_event_log(slot, SDIOEMB_EVENT_CLOCK_FREQ, SDIO_CLOCK_FREQ_MIN);
    slot->clock_freq = SDIO_CLOCK_FREQ_MIN;
    slot->bus_width = 1;

    if (slot->type == SDIOEMB_SLOT_TYPE_SPI_CSPI && cspi_is_enabled(slot)) {
        r = cspi_card_init(slot);
    } else {
        r = sdio_card_init(slot);
    }
    if (r) {
        os_trace_err("%s: card initialization failed", slot->name);
    }
    return r;
}

void sdio_card_stop(struct sdioemb_slot *slot)
{
    struct sdioemb_slot_priv *slotp = slot->priv;

    sdio_event_log(slot, SDIOEMB_EVENT_POWER_OFF);

    slot_set_bus_freq(slot, SDIOEMB_BUS_FREQ_OFF);
    slot->card_power(slot, SDIOEMB_POWER_OFF);
    slotp->card_powered = 0;
}

static int sdio_card_read_info_f0(struct sdioemb_dev *fdev)
{
    struct sdioemb_dev_priv *fdevp = fdev->priv;
    struct sdioemb_slot *slot = fdevp->slot;
    struct sdioemb_slot_priv *slotp = slot->priv;
    uint8_t card_caps, high_speed;
    uint16_t manfid[2];
    uint8_t funce[CISTPL_FUNCE_00_SIZE];
    int ret;

    sdioemb_f0_read8(fdev, SDIO_CCCR_CARD_CAPS, &card_caps);
    if (card_caps & SDIO_CCCR_CARD_CAPS_LSC) {
        slotp->max_bus_width = (card_caps & SDIO_CCCR_CARD_CAPS_4BLS) ? 4 : 1;
    } else {
        slotp->max_bus_width = 4;
    }

    sdioemb_f0_read8(fdev, SDIO_CCCR_HIGH_SPEED, &high_speed);
    slotp->supports_high_speed = (high_speed & SDIO_CCCR_HIGH_SPEED_SHS) ? 1 : 0;

    /* read common CIS ptr */
    ret = sdio_cis_read_ptr_reg(fdev, SDIO_CCCR_CIS_PTR, &fdevp->cis_ptr);
    if (ret) {
        return ret;
    }
    if (fdevp->cis_ptr < 0x1000 || fdevp->cis_ptr > 0x17000) {
        os_trace_err("%s: invalid common CIS ptr (0x%06x)", slot->name, fdevp->cis_ptr);
        return -EINVAL;
    }

    /* read manfid from CIS */
    ret = sdioemb_cis_get_tuple(fdev, CISTPL_MANFID, &manfid, CISTPL_MANFID_SIZE);
    if (ret) {
        return ret;
    }
    fdev->vendor_id = os_le16_to_cpu(manfid[0]);
    fdev->device_id = os_le16_to_cpu(manfid[1]);

    /* read maximum bus clock frequency from CIS */
    slotp->max_card_freq = SDIO_CLOCK_FREQ_MIN;
    if (sdioemb_cis_get_tuple(fdev, CISTPL_FUNCE, &funce, CISTPL_FUNCE_00_SIZE) == 0) {
        /* decode TPLFE_MAX_TRAN_SPEED */
        static const int val[16] = {
            0 /* reserved */, 10, 12, 13, 15, 20, 25, 30, 35, 40, 45, 50, 55, 60, 70, 80,
        };
        static const int unit_10hz[8] = {
            10000, 100000, 1000000, /* rest are reserved */
        };
        int max_tran_speed = val[(funce[3] & 0x78) >> 3] * unit_10hz[funce[3] & 0x7];
        if (max_tran_speed > 0) {
            if (max_tran_speed > slot->caps.max_bus_freq) {
                max_tran_speed = slot->caps.max_bus_freq;
            }
            if (sdioemb_max_bus_freq != 0 && max_tran_speed > sdioemb_max_bus_freq) {
                max_tran_speed = sdioemb_max_bus_freq;
            }
            slotp->max_card_freq = max_tran_speed;
        } else {
            os_trace_warn("%s: invalid TPLFE_MAX_TRAN_SPEED (%02x)", slot->name, funce[3]);
        }
    }

    os_trace_dbg("%s: initialized card %04x:%04x", slot->name, fdev->vendor_id, fdev->device_id);

    return 0;
}

static int sdio_card_read_info_fn(struct sdioemb_dev *fdev)
{
    struct sdioemb_dev_priv *fdevp = fdev->priv;
    struct sdioemb_slot *slot = fdevp->slot;
    int func = fdev->function;
    uint8_t funce_dat[CISTPL_FUNCE_01_SIZE];
    int ret;
    
    /* read func CIS ptr */
    ret = sdio_cis_read_ptr_reg(fdev, SDIO_FBR_CIS_PTR(func), &fdevp->cis_ptr);
    if (ret) {
        return ret;
    }
    if (fdevp->cis_ptr < 0x1000 || fdevp->cis_ptr > 0x17000) {
        os_trace_err("%s: invalid function %d CIS ptr (0x%06x)",
                     slot->name, func, fdevp->cis_ptr);
        return -EINVAL;
    }

    /* iface type */
    ret = sdioemb_f0_read8(fdev, SDIO_FBR_STD_IFACE(func), &fdev->interface);
    if (ret) {
        return ret;
    }

    /* max blocksize from CIS */
    ret = sdioemb_cis_get_tuple(fdev, CISTPL_FUNCE, funce_dat, CISTPL_FUNCE_01_SIZE);
    if (ret) {
        return ret;
    }
    fdev->max_blocksize = (funce_dat[0x0c] & 0xff) | ((funce_dat[0x0d] & 0xff) << 8);

    os_trace_dbg("%s: F%d: iface: %02x, blk size: %d",
                 slot->name, func, fdev->interface, fdev->max_blocksize);

    return 0;
}

static int sdio_card_read_info(struct sdioemb_dev *fdev)
{
    struct sdioemb_slot *slot = fdev->priv->slot;
    struct sdioemb_slot_priv *slotp = slot->priv;
    int ret;

    if (fdev->function == 0) {
        return sdio_card_read_info_f0(fdev);
    }

    ret = sdio_card_read_info_fn(fdev);
    if (ret) {
        return ret;
    }
    fdev->vendor_id = slotp->functions[0]->vendor_id;
    fdev->device_id = slotp->functions[0]->device_id;

    return 0;
}

static void sdio_card_configure_f0(struct sdioemb_dev *fdev)
{
    struct sdioemb_slot *slot = fdev->priv->slot;
    struct sdioemb_slot_priv *slotp = slot->priv;
    int bus_width = 1;

    /*
     * Ensure F0 register are in a sensible state:
     *  - disable all functions and interrupts.
     *  - disable card detect resistor.
     *  - enable continuous SPI interrupt.
     */
    sdioemb_f0_write8(fdev, SDIO_CCCR_IO_EN, 0x00);
    sdioemb_f0_write8(fdev, SDIO_CCCR_INT_EN, 0x00);
    sdioemb_f0_write8(fdev, SDIO_CCCR_BUS_IFACE_CNTL,
                      SDIO_CCCR_BUS_IFACE_CNTL_CD_R_DISABLE
                      | SDIO_CCCR_BUS_IFACE_CNTL_ECSI);

    if (slot->type == SDIOEMB_SLOT_TYPE_SD) {
        /* Configure 4 bit mode if supported by host and card. */
        if (slot->caps.max_bus_width == 4 && slotp->max_bus_width == 4) {
            bus_width = 4;
        }
        slot_set_bus_width(slot, bus_width);
    }
}

static void sdio_card_configure_func(struct sdioemb_dev *fdev)
{
    if (fdev->function == 0) {
        sdio_card_configure_f0(fdev);
    }

    sdioemb_idle_function(fdev);
}

static int sdio_card_init_func(struct sdioemb_slot *slot, int func)
{
    struct sdioemb_dev *fdev;
    struct sdioemb_dev_priv *fdevp;
    int ret = -EIO;

    fdev = sdio_dev_alloc(slot, func);
    if (!fdev) {
        return -ENOMEM;
    }
    fdevp = fdev->priv;

    ret = sdio_card_read_info(fdev);
    if (ret) {
        goto error;
    }

    return 0;

  error:
    sdioemb_put_function(fdev);
    return ret;
}

void sdio_card_configure(struct sdioemb_slot *slot)
{
    struct sdioemb_slot_priv *slotp = slot->priv;
    int f;

    /*
     * Try enabling CSPI mode if the slot supports it.
     */
    if (slot->type == SDIOEMB_SLOT_TYPE_SPI_CSPI) {
        cspi_enable(slot);
    }

    for (f = 0; f < SDIOEMB_MAX_FUNCTIONS; f++) {
        struct sdioemb_dev *fdev = slotp->functions[f];
        if (fdev != NULL) {
            sdio_card_configure_func(fdev);
            fdev->blocksize = 0;
            sdio_func_set_block_size(fdev, 0);
        }
    }
}

/**
 * Notify the SDIO layer that a card has been inserted.
 *
 * Callable from: thread context.
 *
 * @param slot the slot for the inserted card.
 *
 * @return 0 if the inserted call was successfull initialized and enumerated.
 * @return -ve if the initialization failed.
 *
 * @ingroup sdriver
 */
int sdioemb_card_inserted(struct sdioemb_slot *slot)
{
    struct sdioemb_slot_priv *slotp = slot->priv;
    int ret = 0;
    int f;

    os_mutex_lock(&sdio_core_mutex);

    if (slotp->card_present)
        goto out;

    os_trace_info("%s: card inserted", slot->name);

    slotp->card_present = 1;

    ret = sdio_card_start(slot);
    if (ret < 0) {
        goto out;
    }
    os_trace_dbg("%s: card has %d functions", slot->name, slotp->num_functions);

    ret = sdio_card_init_func(slot, 0);
    if (ret < 0) {
        goto out;
    }
    for (f = 1; f <= slotp->num_functions; f++) {
        ret = sdio_card_init_func(slot, f);
        if (ret < 0) {
            goto out;
        }
    }
    sdio_card_configure(slot);

    /*
     * Add successfully initialized functions to the core.
     */
    for (f = 0; f < SDIOEMB_MAX_FUNCTIONS; f++) {
        struct sdioemb_dev *fdev = slotp->functions[f];
        if (fdev) {
            ret = sdio_dev_add(fdev);
            if (ret < 0) {
                sdioemb_put_function(fdev);
                slotp->functions[f] = NULL;
                goto out;
            }
        }
    }

  out:
    os_mutex_unlock(&sdio_core_mutex);

    if (ret < 0) {
        sdioemb_card_removed(slot);
    }

    return ret;
}

/**
 * Notify the SDIO layer that a card has been removed
 *
 * Callable from: thread context.
 *
 * @param slot the slot for the removed card.
 *
 * @ingroup sdriver
 */
void sdioemb_card_removed(struct sdioemb_slot *slot)
{
    struct sdioemb_slot_priv *slotp = slot->priv;
    int f;

    os_mutex_lock(&sdio_core_mutex);

    if (!slotp->card_present) {
        goto out;
    }

    os_trace_info("%s: card removed", slot->name);

    slotp->card_present = 0;

    for (f = SDIOEMB_MAX_FUNCTIONS-1; f >= 0; f--) {
        struct sdioemb_dev *fdev = slotp->functions[f];
        if (fdev) {
            sdio_dev_del(fdev);
        }
    }
    for (f = 0; f < SDIOEMB_MAX_FUNCTIONS; f++) {
        struct sdioemb_dev *fdev = slotp->functions[f];
        sdioemb_put_function(fdev);
    }

    sdio_card_stop(slot);

  out:
    os_mutex_unlock(&sdio_core_mutex);
}

/**
 * Perform a hard reset of the card (if the slot is capable) and
 * reinitialize the card.
 *
 * @param fdev an SDIO device of the card.
 *
 * @return 0 if a hard reset was performed and the card successfully
 * reinitialized.
 * @return 1 if the slot is not capable of hard resets.
 * @return -ve if the reinitialization failed.
 *
 * @ingroup fdriver
 */
int sdioemb_hard_reset(struct sdioemb_dev *fdev)
{
    struct sdioemb_slot *slot = fdev->priv->slot;
    struct sdioemb_slot_priv *slotp = slot->priv;
    int reset;

    if (!slot->hard_reset) {
        return 1;
    }

    os_mutex_lock(&slotp->card_mutex);

    reset = slot->hard_reset(slot);
    if (reset == 0) {
        slotp->current_clock_freq = 0; /* slot driver may have changed the clock */
        sdio_event_log(slot, SDIOEMB_EVENT_RESET);
        reset = sdio_card_start(slot);
        if (reset == 0) {
            sdio_card_configure(slot);
            sdio_func_set_max_bus_freq(fdev, fdev->priv->max_freq);
        }
    }

    os_mutex_unlock(&slotp->card_mutex);

    return reset;
}

#if SDIOEMB_CARD_IS_REMOVABLE

static void sdio_card_detect_thread_func(void *data)
{
    struct sdioemb_slot *slot = data;
    struct sdioemb_slot_priv *slotp = slot->priv;

    slotp->card_present = 0;

    while (!os_thread_should_stop(&slotp->card_detect_thread)) {
        int card_present = slot->card_present(slot);
        if (card_present) {
            sdioemb_card_inserted(slot);
        }
        if (!card_present) {
            sdioemb_card_removed(slot);
        }
        os_try_suspend_thread(&slotp->card_detect_thread);
        os_sleep_ms(sdioemb_card_poll_interval_ms);
    }
}

/* Called by the driver init, starts the card manager thread. */
int sdio_card_detect_init(struct sdioemb_slot *slot)
{
    struct sdioemb_slot_priv *slotp = slot->priv;
    char name[12];

    snprintf(name, sizeof(name), "sdio_cd/%d", slotp->id);

    return os_thread_create(&slotp->card_detect_thread, name,
                            sdio_card_detect_thread_func, slot);
}

void sdio_card_detect_exit(struct sdioemb_slot *slot)
{
    struct sdioemb_slot_priv *slotp = slot->priv;

    os_thread_stop(&slotp->card_detect_thread, NULL);

    sdioemb_card_removed(slot);
}

#else /* !SDIOEMB_CARD_IS_REMOVABLE */

int sdio_card_detect_init(struct sdioemb_slot *slot)
{
    return sdioemb_card_inserted(slot);
}

void sdio_card_detect_exit(struct sdioemb_slot *slot)
{
    sdioemb_card_removed(slot);
}

#endif
