/*
 * Command queuing and dispatch.
 *
 * Copyright (C) 2007 Cambridge Silicon Radio Ltd.
 *
 * Refer to LICENSE.txt included with this source code for details on
 * the license terms.
 */
#include "sdio_internal.h"

#ifdef OS_TRACE_INCLUDE_TMH
#include "sdio_cmd.tmh"
#endif

static int start_or_queue_cmd(struct sdioemb_slot *slot, struct sdioemb_cmd *cmd)
{
    struct sdioemb_slot_priv *slotp = slot->priv;
    int new_tail;
    os_int_status_t istate;
    int ret = 0;

    os_spinlock_lock_intsave(&slotp->lock, &istate);

    if (slotp->busy) {
        new_tail = (slotp->cmd_q_tail + 1) % SDIOEMB_CMD_QUEUE_LEN;

        if (slotp->cmd_q_head != new_tail) {
            slotp->cmd_q[slotp->cmd_q_tail] = cmd;
            slotp->cmd_q_tail = new_tail;
        } else {
            ret = -EAGAIN;
        }
        cmd = NULL;
    } else {
        slotp->busy = 1;
        slotp->active |= 1 << cmd->owner->function;
    }

    os_spinlock_unlock_intrestore(&slotp->lock, &istate);

    if (cmd) {
        slot_start_cmd(slot, cmd);
    }

    return ret;
}


static void start_next_cmd(struct sdioemb_slot *slot)
{
    struct sdioemb_slot_priv *slotp = slot->priv;
    struct sdioemb_cmd *cmd;
    os_int_status_t istate;

    os_spinlock_lock_intsave(&slotp->lock, &istate);

    if (slotp->cmd_q_head != slotp->cmd_q_tail) {
        cmd = slotp->cmd_q[slotp->cmd_q_head];
        slotp->cmd_q_head = (slotp->cmd_q_head + 1) % SDIOEMB_CMD_QUEUE_LEN;
        slotp->active |= 1 << cmd->owner->function;
    } else {
        slotp->busy = 0;
        cmd = NULL;
    }

    os_spinlock_unlock_intrestore(&slotp->lock, &istate);

    if (cmd) {
        slot_start_cmd(slot, cmd);
    }
}

/**
 * Submit an SDIO command to a device.
 *
 * The command will be started if the slot is not busy with another
 * command, otherwise it is queued and will be executed later.
 *
 * The queue has enough space for 2 commands per SDIO function.
 *
 * @param fdev the SDIO device submitting the command.
 * @param cmd  the command to submit.
 *
 * @return \e 0 if the command was queued.
 * @return \e -EAGAIN if the command queue is full.
 * @return \e -EINVAL if the data transfer is invalid (unsupported
 * buffer alignment or the length is not a valid byte or block mode
 * transfer).
 *
 * @see sdio_dev::io
 *
 * @ingroup fdriver
 */
int sdioemb_start_cmd(struct sdioemb_dev *fdev, struct sdioemb_cmd *cmd)
{
    struct sdioemb_dev_priv *fdevp = fdev->priv;
    struct sdioemb_slot *slot = fdevp->slot;
    int ret;

    /*
     * Sanity check data transfers:
     *  - length is a valid block or byte mode transfer).
     *  - some slot drivers can't handle odd-byte aligned buffers.
     */
    if (cmd->data) {
        if ((cmd->len > fdev->blocksize && cmd->len % fdev->blocksize != 0)
            || cmd->len == 0)
        {
            os_trace_warn("%s: F%d: invalid length (cmd->len = %d)",
                          slot->name, fdev->function, cmd->len);
            return -EINVAL;
        }
        if ((unsigned long)cmd->data & 1) {
            os_trace_warn("%s: F%d: unsupported alignment (cmd->data = %p)",
                          slot->name, fdev->function, cmd->data);
            return -EINVAL;
        }
    }

    cmd->owner = fdev;

    ret = start_or_queue_cmd(slot, cmd);
    if (ret) {
        return ret;
    }
    return 0;
}


/**
 * Notify the SDIO layer that an SDIO commmand has completed.
 *
 * Slot drivers should set the \a cmd's status and response and call
 * this when a command has completed (either successfully or not).
 *
 * Callable from: interrupt context.
 *
 * @param slot the slot which completed the command.
 * @param cmd  the completed command.
 *
 * @ingroup sdriver
 */
void sdioemb_cmd_complete(struct sdioemb_slot *slot, struct sdioemb_cmd *cmd)
{
    sdio_event_log_set_response(slot, cmd);

    if (cmd->status && !slot->card_present(slot)) {
        cmd->status = SDIOEMB_CMD_ERR_NO_CARD;
    }

    if (cmd->flags & SDIOEMB_CMD_FLAG_RAW) {
        cmd->callback(cmd);
    } else {
        start_next_cmd(slot);
        cmd->callback(cmd);
    }
}
