/*
 * SDIO device management.
 *
 * Copyright (C) 2007-2009 Cambridge Silicon Radio Ltd.
 *
 * Refer to LICENSE.txt included with this source code for details on
 * the license terms.
 */
#include "sdio_config.h"
#include "sdio_internal.h"
#include <sdioemb/version.h>

#ifdef OS_TRACE_INCLUDE_TMH
#include "sdio_core.tmh"
#endif

/**
 * Mutex protecting card inserts/removes and driver registration.
 */
os_mutex_t sdio_core_mutex;

static struct sdioemb_func_driver *registered_fdrivers[SDIOEMB_FDRIVERS_MAX];

static struct sdioemb_dev *available_funcs[SDIOEMB_FDEVS_MAX];

static void sdio_dev_release(void *arg);

/**
 * Initialize the SDIO stack.
 *
 * This must be called from the OS specific entry point of the sdio
 * module.
 */
void sdio_core_init(void)
{
    os_trace_info("CSR SDIO driver (sdioemb) v%d%s", SDIOEMB_RELEASE, SDIOEMB_RELEASE_EXTRA);

    os_mutex_init(&sdio_core_mutex);
}

/**
 * Clean up the SDIO stack.
 *
 * This must be called from the OS specific exit function of the sdio
 * module.
 */
void sdio_core_clean_up(void)
{
    os_mutex_destroy(&sdio_core_mutex);
}

/**
 * Try binding a function to a driver.
 *
 * This can be used on platforms (e.g., Windows) that do their own
 * matching between drivers and devices.  This function is called
 * instead of sdioemb_add_function().
 *
 * @param fdriver the function driver.
 * @param fdev    the SDIO function.
 *
 * @return 0 if the function was bound to the driver.
 * @return -ENODEV if the driver doesn't support the function.
 * @return -ve error code if the driver's probe returned an error.
 */
int sdioemb_driver_probe(struct sdioemb_func_driver *fdriver, struct sdioemb_dev *fdev)
{
    struct sdioemb_id_table *id;
    int ret;

    for (id = fdriver->id_table; id->vendor_id != 0; id++) {
        if ((id->vendor_id == SDIOEMB_ANY_ID || id->vendor_id == fdev->vendor_id) &&
            (id->device_id == SDIOEMB_ANY_ID || id->device_id == fdev->device_id) &&
            ((id->function == SDIOEMB_ANY_FUNC && fdev->function != SDIOEMB_UIF_FUNC)
             || id->function == fdev->function) &&
            (id->interface == SDIOEMB_ANY_IFACE || id->interface == fdev->interface))
        {
            sdioemb_power_on(fdev);
            sdioemb_set_block_size(fdev, 0);
            sdioemb_set_max_bus_freq(fdev, SDIOEMB_BUS_FREQ_DEFAULT);

            fdev->driver = fdriver;
            ret = fdriver->probe(fdev);
            if (ret != 0) {
                sdioemb_power_off(fdev);
                fdev->driver = NULL;
            }
            return ret;
        }
    }
    return -ENODEV;
}

/**
 * Unbind a function from a driver.
 *
 * @param fdriver the function driver.
 * @param fdev    the SDIO function.
 */
void sdioemb_driver_remove(struct sdioemb_func_driver *fdriver, struct sdioemb_dev *fdev)
{
    fdriver->remove(fdev);
    fdev->driver = NULL;
    sdioemb_idle_function(fdev);
    sdioemb_power_off(fdev);
}

/**
 * Register a function driver with the core.
 *
 * If any inserted cards that have functions that match \a fdriver's
 * list of supported functions then \a fdriver's probe() function will
 * be called for those functions.
 *
 * @param fdriver the function driver to register.
 *
 * @returns 0 on success; -ve on error:
 *   -ENOMEM - maximum number of function drivers are already registered.
 *
 * @ingroup fdriver
 */
int sdioemb_driver_register(struct sdioemb_func_driver *fdriver)
{
    int i;

    os_mutex_lock(&sdio_core_mutex);

    for (i = 0; i < SDIOEMB_FDRIVERS_MAX; i++) {
        if (registered_fdrivers[i] == NULL) {
            registered_fdrivers[i] = fdriver;
            break;
        }
    }
    if (i == SDIOEMB_FDRIVERS_MAX) {
        os_mutex_unlock(&sdio_core_mutex);
        return -ENOMEM;
    }

    os_trace_dbg("registered %s driver", fdriver->name);

    for (i = 0; i < SDIOEMB_FDEVS_MAX; i++) {
        if (available_funcs[i]) {
            sdioemb_driver_probe(fdriver, available_funcs[i]);
        }
    }

    os_mutex_unlock(&sdio_core_mutex);

    return 0;
}

/**
 * Unregister a function driver from the core.
 *
 * \a fdriver's remove() will be called for any SDIO devices currently
 * attached to the driver.
 *
 * @param fdriver the function driver to unregister.
 *
 * @ingroup fdriver
 */
void sdioemb_driver_unregister(struct sdioemb_func_driver *fdriver)
{
    int i;

    os_trace_dbg("unregister %s driver", fdriver->name);

    os_mutex_lock(&sdio_core_mutex);

    /* Remove any functions attached to the driver. */
    for (i = 0; i < SDIOEMB_FDEVS_MAX; i++) {
        struct sdioemb_dev *fdev = available_funcs[i];
        if (fdev && fdev->driver == fdriver) {
            sdioemb_driver_remove(fdriver, fdev);
        }
    }

    /* Remove fdriver from the registered driver list. */
    for (i = 0; i < SDIOEMB_FDRIVERS_MAX; i++) {
        if (registered_fdrivers[i] == fdriver) {
            registered_fdrivers[i] = NULL;
        }
    }

    os_mutex_unlock(&sdio_core_mutex);
}

struct sdioemb_dev *sdio_dev_alloc(struct sdioemb_slot *slot, int func)
{
    struct sdioemb_slot_priv *slotp = slot->priv;
    struct sdioemb_dev *fdev;
    struct sdioemb_dev_priv *fdevp;

    fdev = os_alloc(sizeof(struct sdioemb_dev) + sizeof(struct sdioemb_dev_priv));
    if (!fdev) {
        return NULL;
    }
    slotp->functions[func] = fdev;

    fdev->priv     = (struct sdioemb_dev_priv *)&fdev[1];
    fdev->function = func;
    fdev->slot_id  = slotp->id;

    fdevp = fdev->priv;
    fdevp->slot = slot;

    os_refcount_init(&fdevp->refcount, sdio_dev_release, fdev);

    return fdev;
}

static void sdio_dev_free(struct sdioemb_dev *fdev)
{
    struct sdioemb_slot_priv *slotp;

    if (fdev == NULL)
        return;

    slotp = fdev->priv->slot->priv;

    slotp->functions[fdev->function] = NULL;

    os_refcount_destroy(&fdev->priv->refcount);
    os_free(fdev);
}

static void sdio_dev_release(void *arg)
{
    sdio_dev_free(arg);
}

/**
 * Increase the reference count for a function.
 *
 * @param fdev the function.
 *
 * @ingroup sdriver
 */
void sdioemb_get_function(struct sdioemb_dev *fdev)
{
    os_refcount_get(&fdev->priv->refcount);
}

/**
 * Decrease the reference count for a function.
 *
 * @param fdev the function.
 *
 * @ingroup sdriver
 */
void sdioemb_put_function(struct sdioemb_dev *fdev)
{
    if (fdev) {
        os_refcount_put(&fdev->priv->refcount);
    }
}


static void add_func(struct sdioemb_slot *slot, struct sdioemb_dev *fdev)
{
    int i;
    int ret;

    for (i = 0; i < SDIOEMB_FDEVS_MAX; i++) {
        if (available_funcs[i] == NULL) {
            available_funcs[i] = fdev;
            break;
        }
    }
    if (i == SDIOEMB_FDEVS_MAX) {
        os_trace_err("%s: no space for F%d", slot->name, fdev->function);
        return;
    }

    os_trace_dbg("%s: added F%d", slot->name, fdev->function);

    for (i = 0; i < SDIOEMB_FDRIVERS_MAX; i++) {
        if (registered_fdrivers[i]) {
            ret = sdioemb_driver_probe(registered_fdrivers[i], fdev);
            if (ret == 0) {
                break;
            }
        }
    }
}

static void del_func(struct sdioemb_slot *slot, struct sdioemb_dev *fdev)
{
    int i;

    if (fdev->driver) {
        sdioemb_driver_remove(fdev->driver, fdev);
    }

    for (i = 0; i < SDIOEMB_FDEVS_MAX; i++) {
        if (available_funcs[i] == fdev) {
            available_funcs[i] = NULL;
            os_trace_dbg("%s: deleted F%d", slot->name, fdev->function);
            break;
        }
    }
}

/**
 * Add a function to the list of functions available for use with
 * function drivers.
 *
 * A slot driver must call this if it returned
 * #SDIOEMB_ADD_FUNC_DEFERRED in its add_function() notifier.
 *
 * Alternatively, sdioemb_driver_probe() can be called instead.
 *
 * @param fdev the SDIO function to be added.
 *
 * @ingroup sdriver
 */
void sdioemb_add_function(struct sdioemb_dev *fdev)
{
    struct sdioemb_slot *slot = fdev->priv->slot;

    os_mutex_lock(&sdio_core_mutex);
    add_func(slot, fdev);
    os_mutex_unlock(&sdio_core_mutex);
}

/**
 * Remove a function from the list of functions available for use with
 * function drivers.
 *
 * This may be used if the slot driver called sdioemb_add_function().
 *
 * @param fdev the SDIO function to be removed.
 *
 * @ingroup sdriver
 */
void sdioemb_del_function(struct sdioemb_dev *fdev)
{
    struct sdioemb_slot *slot = fdev->priv->slot;

    os_mutex_lock(&sdio_core_mutex);
    del_func(slot, fdev);
    os_mutex_unlock(&sdio_core_mutex);
}

int sdio_dev_add(struct sdioemb_dev *fdev)
{
    struct sdioemb_slot *slot = fdev->priv->slot;
    int ret = SDIOEMB_ADD_FUNC_NOW;

    if (slot->notifs.add_function) {
        ret = slot->notifs.add_function(slot, fdev);
    }

    switch (ret) {
    case SDIOEMB_ADD_FUNC_NOW:
        add_func(slot, fdev);
        return 0;
    case SDIOEMB_ADD_FUNC_DEFERRED:
        return 0;
    default:
        return ret;
    }
}

void sdio_dev_del(struct sdioemb_dev *fdev)
{
    struct sdioemb_slot *slot = fdev->priv->slot;

    del_func(slot, fdev);

    if (slot->notifs.del_function) {
        slot->notifs.del_function(slot, fdev);
    }
}
