/*
 * Manage a log of the most recent SDIO events (commands and
 * interrupts etc.).
 *
 * Enable the log by #define'ing SDIOEMB_DEBUG_EVENT_LOG_LEN to the
 * maximum number of events to keep in the log.
 *
 * Copyright (C) 2007 Cambridge Silicon Radio Ltd.
 *
 * Refer to LICENSE.txt included with this source code for details on
 * the license terms.
 */

#ifndef _SDIO_EVENT_LOG_H
#define _SDIO_EVENT_LOG_H

#ifdef __linux
#  define SDIOEMB_DEBUG_HAVE_EVENT_LOG
#endif

#ifndef SDIOEMB_DEBUG_EVENT_LOG_LEN
#  define SDIOEMB_DEBUG_EVENT_LOG_LEN 0
#endif

#ifndef SDIOEMB_DEBUG_DATA_BUFFER_LEN
#  define SDIOEMB_DEBUG_DATA_BUFFER_LEN 0
#endif

#if SDIOEMB_DEBUG_EVENT_LOG_LEN > 0 && defined(SDIOEMB_DEBUG_HAVE_EVENT_LOG)

#define SDIOEMB_DEBUG_EVENT_LOG

enum sdio_event_type {
    SDIOEMB_EVENT_CMD, SDIOEMB_EVENT_CARD_INT, SDIOEMB_EVENT_INT_ACKED,
    SDIOEMB_EVENT_POWER_ON, SDIOEMB_EVENT_POWER_OFF, SDIOEMB_EVENT_RESET,
    SDIOEMB_EVENT_CLOCK_FREQ,
};

struct sdio_event_entry {
    unsigned seq_num;
    unsigned long time_ms;
    enum sdio_event_type type;
    union {
        struct {
            unsigned flags;
            union {
                struct sdioemb_cmd_resp sdio;
                struct cspi_cmd_resp cspi;
            };
            int status;
#if SDIOEMB_DEBUG_DATA_BUFFER_LEN > 0
            struct {
                unsigned total_length;
                unsigned length;
                unsigned start;
                uint8_t *buffer;
                struct sdio_event_entry *next;
            } data;
#endif
        } cmd;
        unsigned functions; /*< bit mask of relevant functions */
        int val;
    };
};

/* A ring buffer of SDIO events. */
struct sdio_event_log {
    struct sdio_event_entry *entries;
    unsigned head;
    unsigned tail;
    unsigned seq_num;
    os_spinlock_t lock;

#if SDIOEMB_DEBUG_DATA_BUFFER_LEN > 0
    /*
     * So that we don't display old data but can limit ourselves to
     * the last n octets of interesting data I play some clever tricks
     * here.  I store the sdio_cmd_event's that refer to data in a
     * linked list and also store the length of data blocks in this
     * list.  When I add a new block to the tail of this list that
     * increases the data size past the size of sdio_data_buffer I
     * chop things off the head of the list until the size is OK
     * again.
     */
    uint8_t *data_buffer;
    unsigned data_start;
    unsigned data_total;
    struct sdio_event_entry *data_head;
    struct sdio_event_entry *data_tail;
#endif
};

#if SDIOEMB_DEBUG_DATA_BUFFER_LEN > 0
extern unsigned sdioemb_debug_event_log_data_len;
#endif

void sdio_event_log_init(struct sdioemb_slot *slot);
void sdio_event_log_deinit(struct sdioemb_slot *slot);
void sdio_event_log(struct sdioemb_slot *slot, enum sdio_event_type type, ...);
void sdio_event_log_set_response(struct sdioemb_slot *slot, const struct sdioemb_cmd *cmd);

struct sdio_event_entry *sdio_event_log_get_entry(struct sdioemb_slot *slot, unsigned n);
uint8_t sdio_event_log_get_data(struct sdio_event_entry *e, unsigned i);

void sdio_os_event_log_init(struct sdioemb_slot *slot);
void sdio_os_event_log_deinit(struct sdioemb_slot *slot);

#else

#undef SDIOEMB_DEBUG_EVENT_LOG

#define sdio_event_log_init(s)
#define sdio_event_log_deinit(s)
#define sdio_event_log(s, t, ...)
#define sdio_event_log_set_response(s, c)

#endif

#endif /* #ifndef _SDIO_EVENT_LOG_H */
