/*
 * Card interrupt management.
 *
 * Copyright (C) 2007-2008 Cambridge Silicon Radio Ltd.
 *
 * Refer to LICENSE.txt included with this source code for details on
 * the license terms.
 */
#include "sdio_internal.h"

static void sdioemb_interrupt_unmask(struct sdioemb_dev *fdev)
{
    struct sdioemb_dev_priv *fdevp = fdev->priv;
    struct sdioemb_slot *slot = fdevp->slot;
    struct sdioemb_slot_priv *slotp = slot->priv;
    os_int_status_t istate;
    unsigned int_enabled;

    os_spinlock_lock_intsave(&slotp->lock, &istate);

    int_enabled = slotp->int_enabled | (1 << fdev->function);
    if (slotp->int_enabled == 0) {
        slot->enable_card_int(slot);
    }
    slotp->int_enabled = int_enabled;

    os_spinlock_unlock_intrestore(&slotp->lock, &istate);
}

static void sdioemb_interrupt_mask(struct sdioemb_dev *fdev)
{
    struct sdioemb_dev_priv *fdevp = fdev->priv;
    struct sdioemb_slot *slot = fdevp->slot;
    struct sdioemb_slot_priv *slotp = slot->priv;
    os_int_status_t istate;
    unsigned int_enabled;

    os_spinlock_lock_intsave(&slotp->lock, &istate);

    int_enabled = slotp->int_enabled & ~(1 << fdev->function);
    if (int_enabled == 0) {
        slot->disable_card_int(slot);
    }
    slotp->int_enabled = int_enabled;

    os_spinlock_unlock_intrestore(&slotp->lock, &istate);
}

static int sdioemb_interrupt_write_int_en(struct sdioemb_dev *fdev)
{
    struct sdioemb_dev_priv *fdevp = fdev->priv;
    struct sdioemb_slot *slot = fdevp->slot;
    struct sdioemb_slot_priv *slotp = slot->priv;

    if (fdev->function == SDIOEMB_UIF_FUNC) {
        return 0;
    }
     
    return sdioemb_f0_write8(fdev, SDIO_CCCR_INT_EN,
                             slotp->int_enabled | SDIO_CCCR_INT_EN_MIE);
}


/**
 * Enable SDIO interrupts for a function.
 *
 * Interrupts are enabled by writing the function's enable bit in the
 * Int Enable CCCR register.
 *
 * Callable from: thread context.
 *
 * @param fdev the SDIO function to enable interrupts for.
 *
 * @return 0 on success.
 * @return -ETIMEDOUT or -EIO if the Int Enable register write failed.
 * 
 * @ingroup fdriver
 */
int sdioemb_interrupt_enable(struct sdioemb_dev *fdev)
{
    struct sdioemb_dev_priv *fdevp = fdev->priv;
    struct sdioemb_slot *slot = fdevp->slot;
    struct sdioemb_slot_priv *slotp = slot->priv;
    unsigned bit;
    int ret = 0;

    bit = 1 << fdev->function;

    os_mutex_lock(&slotp->card_mutex);

    if (!(slotp->int_enabled & bit)) {
        sdioemb_interrupt_unmask(fdev);
        ret = sdioemb_interrupt_write_int_en(fdev);
        if (ret < 0) {
            sdioemb_interrupt_mask(fdev);
        }
    }

    os_mutex_unlock(&slotp->card_mutex);

    return ret;
}


/**
 * Disable SDIO interrupts for a function.
 *
 * Interrupts are disable by clearing the function's enable bit in the
 * Int Enable CCCR register.
 *
 * Callable from: thread context.
 *
 * @param fdev the SDIO function to disable interrupts for.
 *
 * @return 0 on success.
 * @return -ETIMEDOUT or -EIO if the Int Enable register write failed.
 *
 * @see sdioemb_interrupt_enable().
 *
 * @ingroup fdriver
 */
int sdioemb_interrupt_disable(struct sdioemb_dev *fdev)
{
    struct sdioemb_dev_priv *fdevp = fdev->priv;
    struct sdioemb_slot *slot = fdevp->slot;
    struct sdioemb_slot_priv *slotp = slot->priv;
    unsigned bit;
    int ret = 0;

    bit = 1 << fdev->function;

    os_mutex_lock(&slotp->card_mutex);

    if (slotp->int_enabled & bit) {
        sdioemb_interrupt_mask(fdev);
        ret = sdioemb_interrupt_write_int_en(fdev);
    }

    os_mutex_unlock(&slotp->card_mutex);

    return ret;
}

/**
 * Notify the SDIO layer that the card interrupt is asserted.
 *
 * Callable from: interrupt context.
 *
 * @param slot the slot with the asserted interrupt.
 *
 * @ingroup sdriver
 */
void sdioemb_interrupt(struct sdioemb_slot *slot)
{
    struct sdioemb_slot_priv *slotp = slot->priv;
    int f;
    os_int_status_t istate;

    slot->disable_card_int(slot);

    os_spinlock_lock_intsave(&slotp->lock, &istate);

    sdio_event_log(slot, SDIOEMB_EVENT_CARD_INT, slotp->int_enabled);

    slotp->int_not_acked = slotp->int_enabled;

    for (f = 0; f < SDIOEMB_MAX_FUNCTIONS; f++) {
        struct sdioemb_dev *fdev = slotp->functions[f];
        if (fdev && slotp->int_enabled & (1 << f)) {
            fdev->driver->card_int_handler(fdev);
        }
    }

    os_spinlock_unlock_intrestore(&slotp->lock, &istate);
}

/**
 * Acknowledge an SDIO interrupt.
 *
 * Function drivers should acknowledge interrupts when they have either:
 *
 * 1. Cleared the interrupt source.
 *
 * 2. Disabled interrupts with sdioemb_interrupt_disable().
 *
 * 3. Determined that the interrupt was for a different function.
 *
 * Callable from: thread context.
 *
 * @param fdev the function acknowledging the interrupt.
 *
 * @ingroup fdriver
 */
void sdioemb_interrupt_acknowledge(struct sdioemb_dev *fdev)
{
    struct sdioemb_dev_priv *fdevp = fdev->priv;
    struct sdioemb_slot *slot = fdevp->slot;
    struct sdioemb_slot_priv *slotp = slot->priv;
    os_int_status_t istate;
    unsigned bit;

    bit = 1 << fdev->function;

    os_spinlock_lock_intsave(&slotp->lock, &istate);

    if (slotp->int_not_acked & bit) {
        sdio_event_log(slot, SDIOEMB_EVENT_INT_ACKED, bit);

        slotp->int_not_acked &= ~bit;
        if (slotp->int_enabled && slotp->int_not_acked == 0) {
            slot->enable_card_int(slot);
        }
    }

    os_spinlock_unlock_intrestore(&slotp->lock, &istate);
}
