/*
 * Internal/private definitions.
 *
 * Copyright (C) 2007-2008 Cambridge Silicon Radio Ltd.
 *
 * Refer to LICENSE.txt included with this source code for details on
 * the license terms.
 */
#ifndef _SDIO_LAYER_H
#define _SDIO_LAYER_H

#include <oska/all.h>
#include <oska/refcount.h>

#include <sdioemb/sdio.h>
#include <sdioemb/sdio_api.h>
#include <sdioemb/slot_api.h>
#include <sdioemb/trace.h>

#include "sdio_event_log.h"
#include "sdio_config.h"

typedef int (*sdioemb_card_io_read8_t)(struct sdioemb_dev *fdev, int function,
                                       uint32_t addr, uint8_t *data);
typedef int (*sdioemb_card_io_read16_t)(struct sdioemb_dev *fdev, int function,
                                        uint32_t addr, uint16_t *data);
typedef int (*sdioemb_card_io_write8_t)(struct sdioemb_dev *fdev, int function,
                                        uint32_t addr, uint8_t data);
typedef int (*sdioemb_card_io_write16_t)(struct sdioemb_dev *fdev, int function,
                                         uint32_t addr, uint16_t data);
typedef int (*sdioemb_card_io_read_t)(struct sdioemb_dev *fdev, int function,
                                      uint32_t addr, uint8_t *data, size_t len);
typedef int (*sdioemb_card_io_write_t)(struct sdioemb_dev *fdev, int function,
                                       uint32_t addr, const uint8_t *data, size_t len);

struct sdioemb_card_io_ops {
    sdioemb_card_io_read8_t   read8;
    sdioemb_card_io_read16_t  read16;
    sdioemb_card_io_write8_t  write8;
    sdioemb_card_io_write16_t write16;
    sdioemb_card_io_read_t    read;
    sdioemb_card_io_write_t   write;
};


#define SDIOEMB_MAX_FUNCTIONS (SDIO_MAX_FUNCTIONS)

/*
 * The command queue needs to be large enough for each function driver
 * to start two commands -- one from a thread and one from its
 * interrupt handler.
 *
 * The length is rounded up to a power of two to avoid a divide when
 * managing the ring buffer.
 */
#define SDIOEMB_CMD_QUEUE_LEN 32

struct sdioemb_slot_priv {
    int id;

    unsigned card_present:1;
    unsigned card_powered:1;

    int num_functions;
    struct sdioemb_dev *functions[SDIOEMB_MAX_FUNCTIONS];
    int max_bus_width;
    int max_card_freq;
    int current_clock_freq;
    unsigned supports_high_speed:1;

    struct sdioemb_card_io_ops io_ops;

    os_spinlock_t lock;

    unsigned active;
    unsigned power;
    unsigned int_enabled;
    unsigned int_not_acked;

    int busy;
    struct sdioemb_cmd *cmd_q[SDIOEMB_CMD_QUEUE_LEN];
    int cmd_q_head;
    int cmd_q_tail;
#if SDIOEMB_CARD_IS_REMOVABLE
    os_thread_t card_detect_thread;
#endif
    os_mutex_t card_mutex;

#ifdef SDIOEMB_DEBUG_EVENT_LOG
    struct sdio_event_log event_log;
#endif
};

struct sdioemb_dev_priv {
    struct sdioemb_slot *slot;
    os_refcount_t refcount;
    uint32_t cis_ptr;
    int max_freq;
    struct sdioemb_cmd *queued_cmd;
};

void sdio_core_init(void);
void sdio_core_clean_up(void);
struct sdioemb_dev *sdio_dev_alloc(struct sdioemb_slot *slot, int func);
int sdio_dev_add(struct sdioemb_dev *fdev);
void sdio_dev_del(struct sdioemb_dev *fdev);

int sdio_card_detect_init(struct sdioemb_slot *slot);
void sdio_card_detect_exit(struct sdioemb_slot *slot);
int sdio_card_start(struct sdioemb_slot *slot);
void sdio_card_stop(struct sdioemb_slot *slot);
void sdio_card_configure(struct sdioemb_slot *slot);

void slot_start_cmd(struct sdioemb_slot *slot, struct sdioemb_cmd *cmd);
void slot_set_max_bus_freq(struct sdioemb_slot *slot, int freq);
void slot_set_bus_freq(struct sdioemb_slot *slot, int freq);
int slot_set_bus_width(struct sdioemb_slot *slot, int bus_width);

int sdio_func_set_block_size(struct sdioemb_dev *fdev, uint16_t blksz);
void sdio_func_set_max_bus_freq(struct sdioemb_dev *fdev, int max_freq);

int sdio_cis_read_ptr_reg(struct sdioemb_dev *fdev, uint32_t addr, uint32_t *ptr);

int sdio_raw_cmd(struct sdioemb_slot *slot, uint8_t cmd_id, uint32_t arg, unsigned flags,
                 union sdioemb_response *response);
int cspi_raw_word_cmd(struct sdioemb_slot *slot, uint8_t cmd_id, uint32_t addr, uint16_t *word);

int sdio_io_read8(struct sdioemb_dev *fdev, int func, uint32_t addr, uint8_t *data);
int sdio_io_read16(struct sdioemb_dev *fdev, int func, uint32_t addr, uint16_t *data);
int sdio_io_write8(struct sdioemb_dev *fdev, int func, uint32_t addr, uint8_t data);
int sdio_io_write16(struct sdioemb_dev *fdev, int func, uint32_t addr, uint16_t data);
int sdio_io_read(struct sdioemb_dev *fdev, int func, uint32_t addr, uint8_t *data, size_t len);
int sdio_io_write(struct sdioemb_dev *fdev, int func, uint32_t addr, const uint8_t *data, size_t len);

int cspi_io_read8(struct sdioemb_dev *fdev, int func, uint32_t addr, uint8_t *data);
int cspi_io_read16(struct sdioemb_dev *fdev, int func, uint32_t addr, uint16_t *data);
int cspi_io_write8(struct sdioemb_dev *fdev, int func, uint32_t addr, uint8_t data);
int cspi_io_write16(struct sdioemb_dev *fdev, int func, uint32_t addr, uint16_t data);
int cspi_io_read(struct sdioemb_dev *fdev, int func, uint32_t addr, uint8_t *data, size_t len);
int cspi_io_write(struct sdioemb_dev *fdev, int func, uint32_t addr, const uint8_t *data, size_t len);

int cspi_is_enabled(struct sdioemb_slot *slot);
int cspi_card_init(struct sdioemb_slot *slot);
int cspi_enable(struct sdioemb_slot *slot);

extern int sdioemb_card_poll_interval_ms;
extern int sdioemb_max_bus_freq;
extern uint16_t sdioemb_max_block_size;

extern os_mutex_t sdio_core_mutex;

/*
 * Hooks for managing OS-specific device structures.
 */
int sdio_os_dev_add(struct sdioemb_dev *fdev);
void sdio_os_dev_del(struct sdioemb_dev *fdev);

#endif /* #ifndef _SDIO_LAYER_H */
