/*
 * Power management.
 *
 * Copyright (C) 2007 Cambridge Silicon Radio Ltd.
 *
 * Refer to LICENSE.txt included with this source code for details on
 * the license terms.
 */
#include "sdio_internal.h"

/**
 * Request power to the card.
 *
 * If the card is not powered, apply power and re-initialize the card.
 * No per-function state of the card is restored and the driver should
 * re-enable the function and perform any other function specific
 * initialization.
 *
 * This is called by the core before calling a function driver's
 * probe() method.
 *
 * @param fdev the SDIO device requesting power.
 *
 * @ingroup fdriver
 */
void sdioemb_power_on(struct sdioemb_dev *fdev)
{
    struct sdioemb_slot *slot = fdev->priv->slot;
    struct sdioemb_slot_priv *slotp = slot->priv;

    os_mutex_lock(&slotp->card_mutex);

    slotp->power |= 1 << fdev->function;

    if (!slotp->card_powered) {
        sdio_card_start(slot);
        sdio_card_configure(slot);
        sdio_func_set_max_bus_freq(fdev, fdev->priv->max_freq);
    }

    os_mutex_unlock(&slotp->card_mutex);
}

/**
 * Allow the card to be powered off.
 *
 * If no other function is requesting power, the card is powered
 * off. The driver must call sdioemb_power_on() before performing any
 * more I/O to the card.
 *
 * Interrupts are disabled by calling sdioemb_interrupt_disable() as the
 * DAT1/INT line may no longer be pulled high when power is removed.
 * If there is an unhandled interrupt it is acknowledged with
 * sdioemb_interrupt_acknowledge() as it may have been for this function.
 *
 * This is called by the core after calling a function driver's
 * remove() method.
 *
 * @param fdev the SDIO device not requiring power.
 *
 * @ingroup fdriver
 */
void sdioemb_power_off(struct sdioemb_dev *fdev)
{
    struct sdioemb_slot *slot = fdev->priv->slot;
    struct sdioemb_slot_priv *slotp = slot->priv;

    sdioemb_interrupt_disable(fdev);
    sdioemb_interrupt_acknowledge(fdev);

    os_mutex_lock(&slotp->card_mutex);

    slotp->power &= ~(1 << fdev->function);

    if (!slotp->power && slotp->card_powered) {
        sdio_card_stop(slot);
    }

    os_mutex_unlock(&slotp->card_mutex);
}

static void suspend_func(struct sdioemb_dev *fdev)
{
    if (fdev->driver) {
        if (fdev->driver->suspend) {
            fdev->driver->suspend(fdev);
        } else {
            sdioemb_power_off(fdev);
        }
    }
}

static void resume_func(struct sdioemb_dev *fdev)
{
    if (fdev->driver) {
        if (fdev->driver->resume) {
            fdev->driver->resume(fdev);
        } else {
            sdioemb_power_on(fdev);
        }
    }
}

/**
 * Suspend an SDIO function.
 *
 * This puts an SDIO Function into a low power (D1/D2) state.  The
 * function driver's \e suspend() callback will be called.
 *
 * If the function driver doesn't provide a suspend callback, the
 * function is powered off with sdioemb_power_off().
 *
 * A suspended function may be resumed by calling
 * sdioemb_resume_function().
 *
 * @note This is useful on some platforms (e.g., Windows) where the
 * power management events are sent to child devices (the function)
 * before the parent device (the slot).  On platforms where power
 * management events are sent to parent devices first (e.g., Linux)
 * then calling this function is not required.
 *
 * @param fdev the function to suspend.
 *
 * @ingroup fdriver
 */
void sdioemb_suspend_function(struct sdioemb_dev *fdev)
{
    os_mutex_lock(&sdio_core_mutex);
    suspend_func(fdev);
    os_mutex_unlock(&sdio_core_mutex);
}

/**
 * Resume an SDIO function.
 *
 * This puts an SDIO Function into an operational (D0) power state.
 * The function driver's \e resume() callback will be called.
 *
 * If the function driver doesn't provide a resume callback, the
 * function is powered on with sdioemb_power_on().
 *
 * See sdioemb_suspend_function() for notes on when this is required.
 *
 * @param fdev the function to resume.
 *
 * @ingroup fdriver
 */
void sdioemb_resume_function(struct sdioemb_dev *fdev)
{
    os_mutex_lock(&sdio_core_mutex);
    resume_func(fdev);
    os_mutex_unlock(&sdio_core_mutex);
}

/**
 * Notify the SDIO driver that the suspend event has been received.
 *
 * @param slot the slot being suspended.
 * 
 * @ingroup sdriver
 */
void sdioemb_suspend(struct sdioemb_slot *slot)
{
    struct sdioemb_slot_priv *slotp = slot->priv;
    int f;

    os_mutex_lock(&sdio_core_mutex);

    /*
     * Pass the event to all device drivers that
     * have registered a suspend handler.
     */
    for (f = 0; f < SDIOEMB_MAX_FUNCTIONS; f++) {
        struct sdioemb_dev *fdev = slotp->functions[f];
        if (fdev) {
            suspend_func(fdev);
        }
    }
    slot_set_bus_freq(slot, SDIOEMB_BUS_FREQ_OFF);

    os_mutex_unlock(&sdio_core_mutex);
}


/**
 * Notify the SDIO driver that the resume event has been received.
 *
 * @param slot the slot being resumed.
 * 
 * @ingroup sdriver
 */
void sdioemb_resume(struct sdioemb_slot *slot)
{
    struct sdioemb_slot_priv *slotp = slot->priv;
    int f;

    os_mutex_lock(&sdio_core_mutex);

    /*
     * Pass the event to all device drivers that
     * have registered a resume handler.
     */
    for (f = 0; f < SDIOEMB_MAX_FUNCTIONS; f++) {
        struct sdioemb_dev *fdev = slotp->functions[f];
        if (fdev) {
            resume_func(fdev);
        }
    }

    os_mutex_unlock(&sdio_core_mutex);
}
