/*
 * Slot driver management.
 *
 * Copyright (C) 2007 Cambridge Silicon Radio Ltd.
 *
 * Refer to LICENSE.txt included with this source code for details on
 * the license terms.
 */
#include "sdio_config.h"
#include "sdio_internal.h"

#ifdef OS_TRACE_INCLUDE_TMH
#include "sdio_slot.tmh"
#endif

/*
 * A list of registered slots is kept so that: sdioemb_slot_unregister()
 * can be called with an unregistered slot; and the slot's position in
 * the slot list is the slot ID.
 */
struct sdioemb_slot *registered_slots[SDIOEMB_SLOTS_MAX];

/*
 * Add slot to registered slot list.
 */
static int slot_add(struct sdioemb_slot *slot)
{
    int i;

    os_mutex_lock(&sdio_core_mutex);
    for (i = 0; i < SDIOEMB_SLOTS_MAX; i++) {
        if (registered_slots[i] == NULL) {
            registered_slots[i] = slot;
            slot->priv->id = i;
            break;
        }
    }
    os_mutex_unlock(&sdio_core_mutex);

    if (i == SDIOEMB_SLOTS_MAX) {
        return -EINVAL;
    }
    return 0;
}

/*
 * Remove slot from registered slot list.
 */
static int slot_del(struct sdioemb_slot *slot)
{
    int i;

    os_mutex_lock(&sdio_core_mutex);
    for (i = 0; i < SDIOEMB_SLOTS_MAX; i++) {
        if (registered_slots[i] == slot) {
            registered_slots[i] = NULL;
            break;
        }
    }
    os_mutex_unlock(&sdio_core_mutex);

    if (i == SDIOEMB_SLOTS_MAX) {
        return -EINVAL;
    }
    return 0;
}

/**
 * Allocate and initialize and SDIO slot driver structure.
 *
 * Callable from: thread context.
 *
 * @param drv_data_size size of slot driver's drv_data.
 *
 * @return the allocated structure; or NULL if no memory could be
 * allocated.
 *
 * @ingroup sdriver
 */
struct sdioemb_slot *sdioemb_slot_alloc(size_t drv_data_size)
{
    struct sdioemb_slot *slot;
    struct sdioemb_slot_priv *slotp;

    slot = os_alloc(sizeof(struct sdioemb_slot)
                    + sizeof(struct sdioemb_slot_priv)
                    + drv_data_size);
    if (!slot)
        return NULL;

    slotp = slot->priv = (struct sdioemb_slot_priv *)&slot[1];
    slot->drv_data = (uint8_t *)slot->priv + sizeof(struct sdioemb_slot_priv);

    os_spinlock_init(&slotp->lock);
    os_mutex_init(&slotp->card_mutex);

    /* Sensible defaults for some capabilities. */
    slot->caps.max_bus_freq = SDIO_CLOCK_FREQ_NORMAL_SPD;

    return slot;
}

/**
 * Free a slot_driver allocated with sdioemb_slot_alloc().
 *
 * Callable from: thread context.
 *
 * @param slot the slot driver to free.
 *
 * @ingroup sdriver
 */
void sdioemb_slot_free(struct sdioemb_slot *slot)
{
    struct sdioemb_slot_priv *slotp = slot->priv;

    os_mutex_destroy(&slotp->card_mutex);
    os_spinlock_destroy(&slotp->lock);
    os_free(slot);
}


/**
 * Register a slot driver with the SDIO layer.
 *
 * Callable from: thread context.
 *
 * @param slot the slot driver to register.
 *
 * @return 0 on success; -ve on error: -ENOMEM - maximum number of
 * slots are already registered.
 *
 * @see SDIOEMB_SLOTS_MAX
 *
 * @ingroup sdriver
 */
int sdioemb_slot_register(struct sdioemb_slot *slot)
{
    int ret;

    ret = slot_add(slot);
    if (ret < 0)
        return ret;

    if (!slot->name[0]) {
        snprintf(slot->name, sizeof(slot->name), "slot%d", slot->priv->id);
    }

    sdio_event_log_init(slot);

    ret = sdio_card_detect_init(slot);
    if (ret) {
        slot_del(slot);
        sdio_event_log_deinit(slot);
        return ret;
    }

    os_trace_info("registered slot '%s'", slot->name);

    return 0;
}

/**
 * Unregister a slot driver from the SDIO layer.
 *
 * Callable from: thread context.
 *
 * @param slot the slot driver to unregister.
 *
 * @ingroup sdriver
 */
void sdioemb_slot_unregister(struct sdioemb_slot *slot)
{
    if (slot_del(slot) < 0) {
        return;
    }

    os_trace_info("unregister slot '%s'", slot->name);

    sdio_card_detect_exit(slot);
    sdio_event_log_deinit(slot);
}

/**
 * Start a command after setting the bus to the active frequency (if necessary).
 *
 * @param slot the slot to perform the command.
 * @param cmd  the command to start.
 */
void slot_start_cmd(struct sdioemb_slot *slot, struct sdioemb_cmd *cmd)
{
    struct sdioemb_slot_priv *slotp = slot->priv;
    int ret;

    cmd->status = SDIOEMB_CMD_IN_PROGRESS;
    sdio_event_log(slot, SDIOEMB_EVENT_CMD, cmd);

    slot_set_bus_freq(slot, slot->clock_freq);
    slot->clock_freq = slotp->current_clock_freq;

    ret = slot->start_cmd(slot, cmd);
    if (ret < 0) {
        cmd->status = SDIOEMB_CMD_ERR_OTHER;
        sdioemb_cmd_complete(slot, cmd);
    }
}

/**
 * Set the maximum bus frequency for this slot.
 *
 * High speed mode is enabled (or disabled) on the card if required.
 *
 * @param slot the slot.
 * @param max_freq maximum bus frequency in Hz.
 */
void slot_set_max_bus_freq(struct sdioemb_slot *slot, int freq)
{
    struct sdioemb_slot_priv *slotp = slot->priv;
    struct sdioemb_dev *f0 = slotp->functions[0];

    if (slotp->supports_high_speed) {
        if (freq > SDIO_CLOCK_FREQ_NORMAL_SPD
            && slot->clock_freq <= SDIO_CLOCK_FREQ_NORMAL_SPD) {
            sdioemb_f0_write8(f0, SDIO_CCCR_HIGH_SPEED, SDIO_CCCR_HIGH_SPEED_EHS);
        }
        if (freq <= SDIO_CLOCK_FREQ_NORMAL_SPD
            && slot->clock_freq > SDIO_CLOCK_FREQ_NORMAL_SPD) {
            sdioemb_f0_write8(f0, SDIO_CCCR_HIGH_SPEED, 0x00);
        }
        sdioemb_idle_function(f0);
    }

    sdio_event_log(slot, SDIOEMB_EVENT_CLOCK_FREQ, freq);

    /*
     * To reduce power consumption on the card we reduce the clock
     * immediately and defer increases until the next command.
     */
    if (freq < slot->clock_freq) {
        slot_set_bus_freq(slot, freq);
        slot->clock_freq = slotp->current_clock_freq;
    } else {
        slot->clock_freq = freq;
    }
}

/**
 * Set the bus frequency.
 *
 * @param slot the slot to perform the command.
 * @param freq frequence value.
 */
void slot_set_bus_freq(struct sdioemb_slot *slot, int freq)
{
    struct sdioemb_slot_priv *slotp = slot->priv;

    /* set clock rate and start it if it isn't already running at the
     * correct speed. */
    
    if (slotp->current_clock_freq != freq) {
        if (slot->set_bus_freq != NULL) {
            slotp->current_clock_freq = slot->set_bus_freq(slot, freq);
        }
    }
}


/**
 * Set a slot's bus width.
 *
 * The requested width is set on both the host and the card.
 *
 * @param slot the slot.
 * @param bus_width the new bus width.
 *
 * @return 0 on success
 * @return -ve on failure.
 */
int slot_set_bus_width(struct sdioemb_slot *slot, int bus_width)
{
    struct sdioemb_slot_priv *slotp = slot->priv;
    struct sdioemb_dev *f0 = slotp->functions[0];
    int ret;
    uint8_t bus_iface_cntl;

    ret = sdioemb_f0_read8(f0, SDIO_CCCR_BUS_IFACE_CNTL, &bus_iface_cntl);
    if (ret < 0) {
        goto out;
    }
    if (bus_width == 4) {
        bus_iface_cntl |= SDIO_CCCR_BUS_IFACE_CNTL_4BIT_BUS;
    } else {
        bus_iface_cntl &= ~SDIO_CCCR_BUS_IFACE_CNTL_4BIT_BUS;
    }
    ret = sdioemb_f0_write8(f0, SDIO_CCCR_BUS_IFACE_CNTL, bus_iface_cntl);
    if (ret < 0) {
        goto out;
    }

    if (slot->set_bus_width) {
        ret = slot->set_bus_width(slot, bus_width);
        if (ret < 0) {
            goto out;
        }
    }

  out:
    sdioemb_idle_function(f0);
    if (ret < 0) {
        /* Try to recover from errors by restoring the old bus width. */
        if (slot->set_bus_width) {
            slot->set_bus_width(slot, slot->bus_width);
        }
    } else {
        slot->bus_width = bus_width;
    }
    return ret;
}
