/*
 * Windows driver helpers.
 *
 * Copyright (C) 2009 Cambridge Silicon Radio Ltd.
 *
 * Refer to LICENSE.txt included with this source code for details on
 * the license terms.
 */
#ifndef SDIOEMB_WINDOWS_H
#define SDIOEMB_WINDOWS_H

#include <wdf.h>

/**
 * @defgroup windows Windows Drivers
 *
 * @brief API extensions for slot and function drivers on Windows.
 *
 * Sdioemb makes use of the Windows Driver Model to bind function
 * drivers to function device.  When a card is inserted into a slot
 * the following occurs.
 *
 * -# The slot driver's add_function() notifier callback is called for
 *    each enumerated function.
 * -# The slot driver creates a child PDO for the function.
 * -# The add_function callback returns #SDIOEMB_ADD_FUNC_DEFERRED to
 *    indicate that the function is not yet ready for use.
 * -# Windows loads an appropriate driver for the function.
 * -# The function driver calls sdioemb_driver_register() in the
 *    DriverEntry routine.
 * -# In the EvtDriverDeviceAdd callback, the function driver obtains
 *    sdioemb's own device structure for that function (struct
 *    sdioemb_dev).
 * -# When the function device first enters the D0 state, the function
 *    driver calls sdioemb_add_function().  This makes the function
 *    available to function drivers registered with sdioemb.
 * -# The driver's probe() callback is called.
 *
 * <h3>Slot Drivers</h3>
 *
 * Slot drivers must be kernel-mode WDF drivers.
 *
 * - Allocate the slots with a driver data structure whose first field
 *   is a struct sdioemb_wdf_slot_data.
 * - Before registering the slot, the driver must call
 *   sdioemb_wdf_create_child_list().  This may be done in the driver's
 *   \e EvtDriverDeviceAdd callback.
 * - Register the slot with sdioemb_wdf_slot_register().
 * - After unregistering the slot (with sdioemb_wdf_slot_unregister())
 *   call sdioemb_wdf_destroy_child_list(). The may be done in the
 *   FDO's \e EvtCleanup callback.
 * - Do not call sdioemb_suspend() or sdioemb_resume() when exiting
 *   or entering D0 state.  The function driver is responsible for
 *   its own suspend/resume events.
 *
 * <h3>Function Drivers</h3>
 *
 * Slot drivers must be kernel-mode WDF (or WDM) drivers.
 *
 * - Create an INF file for the function driver that matches the same
 *   devices as specified in the sdioemb_driver_register() call.
 * - Call sdioemb_driver_register() in the DriverEntry routine.
 * - In the \e EvtDriverDeviceAdd callback obtain sdioemb's function
 *   device via the GUID_DRVINTERFACE_SDIOEMB_FUNC driver interface
 *   (struct sdioemb_wdf_interface).
 * - When entering the D0 state for the first time call
 *   sdioemb_add_function().
 * - Call sdioemb_suspend_function() or sdioemb_resume_function() when
 *   exiting/entering D0 state.
 * - Call sdioemb_del_function() when entering the D3Final state.
 * - Dereference the driver interface in the device object's
 *   EvtCleanupCallback.
 * - Call sdioemb_driver_unregister() in the driver's EvtCleanupCallback.
 */

/**
 * Driver data required by the WDF slot library.
 *
 * Slot drivers must include this as the first field in their driver
 * specific slot data structure.
 *
 * @ingroup windows
 */
struct sdioemb_wdf_slot_data {
    /**
     * The device for the slot's controller.
     */
    WDFDEVICE device;

    /**
     * The child list for the inserted card's functions.
     */
    WDFCHILDLIST child_list;
};

NTSTATUS sdioemb_wdf_query_ulong(WDFDEVICE device,
                                 const UNICODE_STRING *value_name, ULONG *value);

NTSTATUS sdioemb_wdf_create_child_list(WDFDEVICE device, struct sdioemb_slot *slot);
void sdioemb_wdf_destroy_child_list(struct sdioemb_slot *slot);

int sdioemb_wdf_slot_register(struct sdioemb_slot *slot);
void sdioemb_wdf_slot_unregister(struct sdioemb_slot *slot);

void sdioemb_wdf_event_log_init(PWDFDEVICE_INIT device_init);
NTSTATUS sdioemb_wdf_event_log_create(WDFDEVICE device);

#endif /* #ifndef SDIOEMB_WINDOWS_H */
