/*
 * Character device interface.
 *
 * Copyright (C) 2007-2008 Cambridge Silicon Radio Ltd.
 *
 * Refer to LICENSE.txt included with this source code for details on
 * the license terms.
 */
#include <oska/trace.h>

#include "bt_a_dev.h"

#include <linux/kernel-compat.h>

/* Maximum number of Type-A devices to support. */
#define BT_A_DEVS_MAX 16

static int idle_timeout_ms = 100;

module_param(idle_timeout_ms, int, 0444);
MODULE_PARM_DESC(idle_timeout_ms, "Idle time (in ms) before entering deep-sleep");

static dev_t          bt_a_dev_first;
static struct class * bt_a_class;
static DEFINE_SPINLOCK(bt_a_minors_lock);
static DECLARE_BITMAP(bt_a_minors, BT_A_DEVS_MAX);

static int sdio_bt_a_get_minor(void)
{
    int minor;

    spin_lock(&bt_a_minors_lock);
    minor = find_first_zero_bit(bt_a_minors, BT_A_DEVS_MAX);
    set_bit(minor, bt_a_minors);
    spin_unlock(&bt_a_minors_lock);

    return minor;
}

static void sdio_bt_a_put_minor(int minor)
{
    spin_lock(&bt_a_minors_lock);
    clear_bit(minor, bt_a_minors);
    spin_unlock(&bt_a_minors_lock);
}

static void sdio_bt_a_get(struct bt_a_device *bt)
{
    atomic_inc(&bt->ref_count);
}

static void sdio_bt_a_put(struct bt_a_device *bt)
{
    if (atomic_dec_and_test(&bt->ref_count)) {
        kfree(bt);
    }
}

static void torpid_work(struct work_struct *work)
{
    struct bt_a_device *bt = container_of(work, struct bt_a_device, torpid_work);

    mutex_lock(&bt->mutex);

    if (bt->bdev.sleep_state == SLEEP_STATE_AWAKE) {
        sdioemb_bt_a_set_sleep_state(&bt->bdev, SLEEP_STATE_TORPID);
    }

    mutex_unlock(&bt->mutex);
}

static void idle_timer_func(unsigned long arg)
{
    struct bt_a_device *bt = (struct bt_a_device *)arg;

    queue_work(bt->workqueue, &bt->torpid_work);
}

static void reset_idle_timer(struct bt_a_device *bt)
{
    mod_timer(&bt->idle_timer, jiffies + msecs_to_jiffies(idle_timeout_ms));
}

static void sdio_bt_a_irq_work(struct work_struct *work)
{
    struct bt_a_device *bt = container_of(work, struct bt_a_device, irq_work);

    mutex_lock(&bt->mutex);
    sdioemb_bt_a_handle_interrupt(&bt->bdev);
    mutex_unlock(&bt->mutex);
}

static void bt_a_dsr(CsrSdioFunction *func)
{
    struct bt_a_device *bt = func->driverData;

    reset_idle_timer(bt);
    queue_work(bt->workqueue, &bt->irq_work);
}

static int bt_a_get_rx_buffer(struct sdioemb_bt_a_dev *bdev, const uint8_t *header,
                              size_t buffer_min_len, uint8_t **buffer, void **buffer_handle)
{
    uint8_t *packet;

    packet = kmalloc(buffer_min_len, GFP_KERNEL);
    if (packet == NULL) {
        return -ENOMEM;
    }
    *buffer = packet;
    *buffer_handle = packet;
    return 0;
}

static void bt_a_receive(struct sdioemb_bt_a_dev *bdev, void *buffer_handle, int status)
{
    struct bt_a_device *bt = bdev->drv_data;
    uint8_t *packet = buffer_handle;
    struct rx_queue_entry *rx_entry;

    /* Discard unsuccessful packets. */
    if (status < 0) {
        kfree(packet);
        return;
    }

    rx_entry = kzalloc(sizeof(struct rx_queue_entry), GFP_KERNEL);
    if (!rx_entry) {
        os_trace_err("%s: no memory for Rx packet", bt->bdev.name);
        goto error;
    }
    rx_entry->data = packet;
    rx_entry->len  = sdioemb_bt_a_packet_len(packet);
    rx_entry->read = 0;

    rx_queue_add(bt, rx_entry);
    return;

  error:
    kfree(packet);
}

static void bt_a_sleep_state_changed(struct sdioemb_bt_a_dev *bdev)
{
    struct bt_a_device *bt = bdev->drv_data;

    if (bdev->sleep_state == SLEEP_STATE_AWAKE) {
        wake_up(&bt->drowsy_waitq);
    } else if (bdev->sleep_state == SLEEP_STATE_DROWSY) {
        long r;
        mutex_unlock(&bt->mutex);
        r = wait_event_timeout(bt->drowsy_waitq, bdev->sleep_state == SLEEP_STATE_AWAKE,
                               msecs_to_jiffies(BC_DEEP_SLEEP_EXIT_TIMEOUT_MS));
        mutex_lock(&bt->mutex);
        if (r == 0) {
            os_trace_warn("%s: timed out waiting for chip to wake up", bt->bdev.name);
        }
    }
}

static int sdio_bt_a_read(struct file *file, char __user *buf, size_t len, loff_t *offset)
{
    struct bt_a_device *bt = file->private_data;
    struct rx_queue_entry *rx_entry = NULL;
    size_t read;

    if (!bt->card_present)
        return -ENODEV;
    if (!access_ok(VERIFY_WRITE, buf, len))
        return -EFAULT;

    mutex_lock(&bt->mutex);
    
    for (;;) {
        rx_entry = rx_queue_head(bt);
        if (rx_entry)
            break;

        mutex_unlock(&bt->mutex);
        if (file->f_flags & O_NONBLOCK)
            return -EAGAIN;
        if (wait_event_interruptible(bt->read_waitq, bt->rx_queue_len > 0))
            return -ERESTARTSYS;
        mutex_lock(&bt->mutex);
    }

    read = min(len, rx_entry->len - rx_entry->read);
    if (__copy_to_user(buf, rx_entry->data + rx_entry->read, read)) {
        mutex_unlock(&bt->mutex);
        return -EFAULT;
    }
    rx_entry->read += read;

    if (rx_entry->read == rx_entry->len && read < len) {
        rx_queue_remove(bt, rx_entry);
        kfree(rx_entry->data);
        kfree(rx_entry);
    }

    mutex_unlock(&bt->mutex);

    return read;
}

static int sdio_bt_a_write(struct file *file, const char __user *buf, size_t len, loff_t *offset)
{
    struct bt_a_device *bt = file->private_data;
    uint8_t *pkt;
    int ret = 0;

    pkt = kmalloc(len, GFP_KERNEL);
    if (pkt == NULL) {
        ret = -ENOMEM;
        goto out;
    }
    if (copy_from_user(pkt, buf, len)) {
        ret = -EFAULT;
        goto out;
    }

    /* sanity check packet */
    if (len < SDIO_BT_A_HEADER_LEN || len > SDIO_BT_A_PACKET_LEN_MAX
        || sdioemb_bt_a_packet_len(pkt) != len) {
        ret = -EINVAL;
        goto out;
    }

    mutex_lock(&bt->mutex);

    if (!bt->card_present) {
        mutex_unlock(&bt->mutex);
        ret = -ENODEV;
        goto out;
    }

    reset_idle_timer(bt);

    if (bt->bdev.sleep_state == SLEEP_STATE_TORPID) {
        sdioemb_bt_a_set_sleep_state(&bt->bdev, SLEEP_STATE_DROWSY);

        /* card may have been removed while waiting for it to awake. */
        if (!bt->card_present) {
            mutex_unlock(&bt->mutex);
            ret = -ENODEV;
            goto out;
        }
    }

    ret = sdioemb_bt_a_send(&bt->bdev, pkt, len);

    mutex_unlock(&bt->mutex);

  out:
    kfree(pkt);
    return ret < 0 ? ret : len;
}

static unsigned sdio_bt_a_poll(struct file *file, poll_table *wait)
{
    struct bt_a_device *bt = file->private_data;
    unsigned mask = 0;

    poll_wait(file, &bt->read_waitq, wait);
    mask |= POLLOUT | POLLWRNORM;
    if (bt->rx_queue_len)
        mask |= POLLIN | POLLRDNORM;
    if (!bt->card_present)
        mask |= POLLERR;

    return mask;
}

static int sdio_bt_a_open(struct inode *inode, struct file *file)
{
    struct bt_a_device *bt = container_of(inode->i_cdev, struct bt_a_device, cdev);
    int ret = 0;

    mutex_lock(&bt->mutex);

    if (!bt->card_present) {
        ret = -ENODEV;
    } else {
        sdio_bt_a_get(bt);
        file->private_data = bt;
    }

    mutex_unlock(&bt->mutex);

    return ret;
}

static int sdio_bt_a_release(struct inode *inode, struct file *file)
{
    struct bt_a_device *bt = container_of(inode->i_cdev, struct bt_a_device, cdev);

    sdio_bt_a_put(bt);
    return 0;
}

static struct file_operations bt_a_fops = {
    .owner      = THIS_MODULE,
    .open       = sdio_bt_a_open,
    .release    = sdio_bt_a_release,
    .read       = sdio_bt_a_read,
    .write      = sdio_bt_a_write,
    .poll       = sdio_bt_a_poll,
};

static void bt_a_inserted(CsrSdioFunction *func)
{
    struct bt_a_device *bt;
    char *dev_name;
    int ret = -ENOMEM;

    bt = kzalloc(sizeof(struct bt_a_device), GFP_KERNEL);
    if (!bt)
        goto error;
    func->driverData = bt;

    bt->func = func;

    cdev_init(&bt->cdev, &bt_a_fops);

    bt->minor = sdio_bt_a_get_minor();
    bt->card_present = 1;

    mutex_init(&bt->mutex);
    /* Single thread workqueue since accesses to the SDIO must be
     * serialized anyway. */
    bt->workqueue = create_singlethread_workqueue("sdio_bt_a");
    if (!bt->workqueue)
        goto error;
    INIT_WORK(&bt->irq_work, sdio_bt_a_irq_work);
    spin_lock_init(&bt->rx_queue_lock);
    INIT_LIST_HEAD(&bt->rx_queue);
    init_waitqueue_head(&bt->read_waitq);
    init_waitqueue_head(&bt->drowsy_waitq);

    bt->idle_timer.data = (unsigned long)bt;
    bt->idle_timer.function = idle_timer_func;
    init_timer(&bt->idle_timer);
    INIT_WORK(&bt->torpid_work, torpid_work);

    atomic_set(&bt->ref_count, 1);

    ret = cdev_add(&bt->cdev, MKDEV(MAJOR(bt_a_dev_first),
                                    MINOR(bt_a_dev_first) + bt->minor), 1);
    if (ret)
        goto error_wq;

    switch (func->sdioId.sdioInterface) {
    case SDIO_STD_IFACE_BT_TYPE_A_AMP:
        dev_name = "sdio_bt_amp%d";
        break;
    default:
        dev_name = "sdio_bt_a%d";
        break;
    }

    if (!device_create(bt_a_class, NULL, bt->cdev.dev, NULL, dev_name, bt->minor))
        goto error_device_create;

    ret = sdioemb_bt_a_setup(&bt->bdev, func);
    if (ret < 0) {
        goto error_bt_a_init;
    }
    sprintf(bt->bdev.name, dev_name, bt->minor);
    bt->bdev.get_rx_buffer       = bt_a_get_rx_buffer;
    bt->bdev.receive             = bt_a_receive;
    bt->bdev.sleep_state_changed = bt_a_sleep_state_changed;
    bt->bdev.drv_data            = bt;

    fix_quirks(bt);

    mutex_lock(&bt->mutex);
    sdioemb_bt_a_start(&bt->bdev);
    mutex_unlock(&bt->mutex);

    goto out;

  error_bt_a_init:
    device_destroy(bt_a_class, bt->cdev.dev);    
  error_device_create:
    cdev_del(&bt->cdev);
  error_wq:
    destroy_workqueue(bt->workqueue);
    sdio_bt_a_put_minor(bt->minor);
  error:
    kfree(bt);
  out:
    CsrSdioInsertedAcknowledge(func, ret);
}

static void bt_a_removed(CsrSdioFunction *func)
{
    struct bt_a_device *bt = func->driverData;

    mutex_lock(&bt->mutex);
    bt->card_present = 0;
    sdioemb_bt_a_stop(&bt->bdev);
    mutex_unlock(&bt->mutex);

    del_timer_sync(&bt->idle_timer);
    flush_workqueue(bt->workqueue);
    destroy_workqueue(bt->workqueue);
    sdioemb_bt_a_cleanup(&bt->bdev);
    device_destroy(bt_a_class, bt->cdev.dev);
    cdev_del(&bt->cdev);

    sdio_bt_a_put_minor(bt->minor);
    sdio_bt_a_put(bt);

    CsrSdioRemovedAcknowledge(func);
}

static CsrSdioInterruptDsrCallback bt_a_isr(CsrSdioFunction *func)
{
    return bt_a_dsr;
}

static CsrSdioFunctionId bt_a_ids[] = {
    { SDIO_MANF_ID_CSR, CSR_SDIO_ANY_CARD_ID, CSR_SDIO_ANY_SDIO_FUNCTION,
      SDIO_STD_IFACE_BT_TYPE_A, },
    { SDIO_MANF_ID_CSR, CSR_SDIO_ANY_CARD_ID, CSR_SDIO_ANY_SDIO_FUNCTION,
      SDIO_STD_IFACE_BT_TYPE_A_AMP, },
};

static CsrSdioFunctionDriver bt_a_driver = {
    bt_a_inserted,
    bt_a_removed,
    bt_a_isr,
    NULL,
    NULL,
    bt_a_ids,
    2,
};

static int __init bt_a_init(void)
{
    CsrResult result;
    int ret;

    ret = alloc_chrdev_region(&bt_a_dev_first, 0, BT_A_DEVS_MAX, "sdio_bt_a");
    if (ret)
        goto error;

    bt_a_class = class_create(THIS_MODULE, "bt_a");
    if (IS_ERR(bt_a_class)) {
        ret = PTR_ERR(bt_a_class);
        goto error;
    }

    result = CsrSdioFunctionDriverRegister(&bt_a_driver);
    if (result) {
        ret = CsrSdioResultToStatus(result);
        goto error;
    }

    return 0;

  error:
    class_destroy(bt_a_class);
    if (bt_a_dev_first)
        unregister_chrdev_region(bt_a_dev_first, BT_A_DEVS_MAX);
    return ret;
}

static void __exit bt_a_exit(void)
{
    CsrSdioFunctionDriverUnregister(&bt_a_driver);

    class_destroy(bt_a_class);
    unregister_chrdev_region(bt_a_dev_first, BT_A_DEVS_MAX);
}

module_init(bt_a_init);
module_exit(bt_a_exit);

MODULE_AUTHOR("David Vrabel <david.vrabel@csr.com>");
MODULE_DESCRIPTION("CSR's SDIO Bluetooth Type-A driver");
MODULE_LICENSE("GPL and additional rights");
