/*
 * Handling of quirky hardware/firmware.
 *
 * Copyright (C) 2007 Cambridge Silicon Radio Ltd.
 *
 * Refer to LICENSE.txt included with this source code for details on
 * the license terms.
 */
#include "bt_a_dev.h"

static int ack_reads = -1;

module_param(ack_reads,int,0444);
MODULE_PARM_DESC(ack_reads, "Acknowledge packet reads (0=no acks; 1=ack)");

/**
 * Fix quirky hardware.
 *
 * Depending on the manufacturer and card ID, feature flags could be
 * set or cleared, or the hardware could be configured etc.
 *
 * @param bt the BT device.
 */
void fix_quirks(struct bt_a_device *bt)
{
    uint16_t manf_id;
    uint16_t card_id;

    manf_id = bt->func->sdioId.manfId;
    card_id = bt->func->sdioId.cardId;

    /*
     * BlueCore6 quirks:
     *
     *  - I/O Aborts reset the Type-A block and result in read/write
     *    retries reading/writing bad data.
     *  - write retries produce bad data (even when no I/O Abort is done).
     *  - read acks are required.
     *  - must poll bit0 of TO_HOST_SCRATCH0 after enabling function.
     */
    if (manf_id == SDIO_MANF_ID_CSR && card_id == SDIO_CARD_ID_CSR_BC6) {
        bt->bdev.max_rx_retries = 1;
        bt->bdev.max_tx_retries = 1;
        bt->bdev.needs_read_ack = 1;
        bt->bdev.wait_for_firmware = 1;
    }

    /*
     * Permit the user to override the read acknowledge mode -- useful
     * for testing purposes.
     */
    if (ack_reads != -1) {
        bt->bdev.needs_read_ack = ack_reads;
    }
}
