/*
 * Synergy SDIO API implementation for Linux MMC stack.
 *
 * Copyright (C) 2010 Cambridge Silicon Radio Ltd.
 *
 * Refer to LICENSE.txt included with this source code for details on
 * the license terms.
 */
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/completion.h>
#include <linux/wait.h>
#include <linux/sched.h>
#include <linux/pm.h>
#include <linux/delay.h>
#include <linux/mmc/core.h>
#include <linux/mmc/card.h>
#include <linux/mmc/host.h>
#include <linux/mmc/sdio_func.h>

#include <csr_sdio.h>
#include <csr_sdio_lib.h>
#include <sdioemb/sdio.h>
#include <sdioemb/sdio_csr.h>

/*
 * Some parts of the Synergy SDIO API do not map well onto the Linux
 * SDIO API.
 *
 * - Host claim/release.
 *
 * Linux requires that the host is claimed (with sdio_claim_host())
 * while performing any SDIO operation.  This implementation claims
 * and releases around each individual SDIO operation.
 *
 * - Interrupts.
 *
 * On Linux, interrupts are threaded so the interrupt handler must
 * call the registered ISR (and DSR if necessary) and then wait until
 * CsrSdioInterruptAcknowledge() is called.  The host shall be
 * released before calling the ISR/DSR and and then reclaimed before
 * returning.
 *
 * - Driver registration.
 *
 * This registers itself with the Linux stack as supporting /all/ CSR
 * chips.  A seperate list of functions and function drivers is kept
 * and additional matching is done.
 */

struct glue_driver {
    struct list_head node;
    CsrSdioFunctionDriver *functionDriver;
};

enum {
    GLUE_BIT_INT_PENDING = 0,
    GLUE_BIT_INT_ACKED   = 1,
};

struct glue_func {
    CsrSdioFunction function; /* must be first */
    struct list_head node;
    struct sdio_func *func;
    struct glue_driver *glue_driver;
    struct mutex irq_mutex;
    struct completion complete;
    unsigned long bits;
    wait_queue_head_t wq;
    CsrResult acknowledge_result;
};

static DEFINE_MUTEX(glue_mutex);
static LIST_HEAD(glue_drivers);
static LIST_HEAD(glue_funcs);

#define IO_EN_TIMEOUT_MS 5000
#define RESET_TIMEOUT_MS 200

CsrStatus CsrSdioResultToStatus(CsrResult result)
{
    switch (result) {
    case CSR_RESULT_SUCCESS:
        return 0;
    case CSR_RESULT_FAILURE:
        return -ENOMEM;
    case CSR_SDIO_RESULT_INVALID_VALUE:
        return -EINVAL;
    case CSR_SDIO_RESULT_NO_DEVICE:
        return -ENODEV;
    case CSR_SDIO_RESULT_CRC_ERROR:
        return -EIO;
    case CSR_SDIO_RESULT_TIMEOUT:
        return -ETIMEDOUT;
    case CSR_SDIO_RESULT_NOT_RESET:
        return 0;
    default:
        return -EIO; /* arbitrary */
    }
}
EXPORT_SYMBOL(CsrSdioResultToStatus);

CsrResult CsrSdioStatusToResult(CsrStatus status)
{
    if (status >= 0) {
        return CSR_RESULT_SUCCESS;
    }

    switch (status) {
    case -EIO:
        return CSR_SDIO_RESULT_CRC_ERROR;
    case -ETIMEDOUT:
        return CSR_SDIO_RESULT_TIMEOUT;
    case -ENODEV:
    case -ENOMEDIUM:
        return CSR_SDIO_RESULT_NO_DEVICE;
    case -ENOMEM:
        return CSR_RESULT_FAILURE; /* There is no CSR_RESULT_NO_MEMORY. */
    case -EINVAL:
        return CSR_SDIO_RESULT_INVALID_VALUE;
    default:
        return CSR_RESULT_FAILURE;
    }
}
EXPORT_SYMBOL(CsrSdioStatusToResult);

/*
 * Does an SDIO function driver work with this SDIO function?
 *
 * FIXME: move this generic code elsewhere?
 */
CsrBool CsrSdioFunctionDriverMatch(CsrSdioFunctionDriver *functionDriver, CsrSdioFunction *function)
{
    int i;
    CsrSdioFunctionId *a, *b;

    b = &function->sdioId;

    for (i = 0; i < functionDriver->idsCount; i++) {
        a = &functionDriver->ids[i];

        if ((a->manfId == CSR_SDIO_ANY_MANF_ID || a->manfId == b->manfId) &&
            (a->cardId == CSR_SDIO_ANY_CARD_ID || a->cardId == b->cardId) &&
            (a->sdioFunction == CSR_SDIO_ANY_SDIO_FUNCTION || a->sdioFunction == b->sdioFunction) &&
            (a->sdioInterface == CSR_SDIO_ANY_SDIO_INTERFACE || a->sdioInterface == b->sdioInterface))
        {
            return TRUE;
        }
    }
    return FALSE;
}


static CsrResult wait_for_acknowledge(struct glue_func *glue_func)
{
    wait_for_completion(&glue_func->complete);
    INIT_COMPLETION(glue_func->complete);
    return glue_func->acknowledge_result;
}

static CsrResult try_insert(struct glue_driver *glue_driver, struct glue_func *glue_func)
{
    CsrResult result = CSR_SDIO_RESULT_NO_DEVICE;

    if (CsrSdioFunctionDriverMatch(glue_driver->functionDriver, &glue_func->function)) {
        glue_func->glue_driver = glue_driver;
        glue_driver->functionDriver->inserted(&glue_func->function);
        result = wait_for_acknowledge(glue_func);
        if (result) {
            glue_func->glue_driver = NULL;
        }
    }
    return result;
}

static void try_remove(struct glue_driver *glue_driver, struct glue_func *glue_func)
{
    if (glue_driver && glue_func->glue_driver == glue_driver) {
        glue_driver->functionDriver->removed(&glue_func->function);
        glue_func->glue_driver = NULL;
        wait_for_acknowledge(glue_func);
    }
}

static void add_func(struct glue_func *glue_func)
{
    struct glue_driver *glue_driver;
    CsrResult result;

    mutex_lock(&glue_mutex);

    list_add_tail(&glue_func->node, &glue_funcs);

    list_for_each_entry(glue_driver, &glue_drivers, node) {
        result = try_insert(glue_driver, glue_func);
        if (result == CSR_RESULT_SUCCESS) {
            break;
        }
    }

    mutex_unlock(&glue_mutex);
}

static void del_func(struct glue_func *glue_func)
{
    mutex_lock(&glue_mutex);

    list_del(&glue_func->node);
    try_remove(glue_func->glue_driver, glue_func);

    mutex_unlock(&glue_mutex);
}

static void add_driver(struct glue_driver *glue_driver)
{
    struct glue_func *glue_func;

    mutex_lock(&glue_mutex);

    list_add_tail(&glue_driver->node, &glue_drivers);

    list_for_each_entry(glue_func, &glue_funcs, node) {
        try_insert(glue_driver, glue_func);
    }

    mutex_unlock(&glue_mutex);
}

static void del_driver(struct glue_driver *glue_driver)
{
    struct glue_func *glue_func;

    mutex_lock(&glue_mutex);

    list_del(&glue_driver->node);

    list_for_each_entry(glue_func, &glue_funcs, node) {
        try_remove(glue_driver, glue_func);
    }

    mutex_unlock(&glue_mutex);
}


static int glue_probe(struct sdio_func *func, const struct sdio_device_id *id)
{
    struct glue_func *glue_func;

    glue_func = kzalloc(sizeof(*glue_func), GFP_KERNEL);
    if (glue_func == NULL) {
        return -ENOMEM;
    }
    sdio_set_drvdata(func, glue_func);

    glue_func->function.sdioId.manfId        = func->vendor;
    glue_func->function.sdioId.cardId        = func->device;
    glue_func->function.sdioId.sdioFunction  = func->num;
    glue_func->function.sdioId.sdioInterface = func->class;
    glue_func->function.blockSize            = func->cur_blksize;
    glue_func->function.features             = CSR_SDIO_FEATURE_BYTE_MODE
        | CSR_SDIO_FEATURE_DMA_CAPABLE_MEM_REQUIRED;
    glue_func->function.cardHandle           = func->card;
    glue_func->function.osDevice             = &func->dev;
    glue_func->function.priv                 = glue_func;

    glue_func->func = func;
    mutex_init(&glue_func->irq_mutex);
    init_completion(&glue_func->complete);
    init_waitqueue_head(&glue_func->wq);

    add_func(glue_func);
    return 0;
}

static void glue_remove(struct sdio_func *func)
{
    struct glue_func *glue_func = sdio_get_drvdata(func);

    del_func(glue_func);
    kfree(glue_func);
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,32)
#ifdef CONFIG_PM

static int glue_suspend(struct device *dev)
{
    struct sdio_func *func = dev_to_sdio_func(dev);
    struct glue_func *glue_func = sdio_get_drvdata(func);
    CsrResult result;

    if (glue_func->glue_driver->functionDriver->suspend) {
        glue_func->glue_driver->functionDriver->suspend(&glue_func->function);
        result = wait_for_acknowledge(glue_func);
    }

    return CsrSdioResultToStatus(result);
}

static int glue_resume(struct device *dev)
{
    struct sdio_func *func = dev_to_sdio_func(dev);
    struct glue_func *glue_func = sdio_get_drvdata(func);
    CsrResult result;

    if (glue_func->glue_driver->functionDriver->resume) {
        glue_func->glue_driver->functionDriver->resume(&glue_func->function);
        result = wait_for_acknowledge(glue_func);
    }

    return CsrSdioResultToStatus(result);
}

#endif /* CONFIG_PM */
#endif /* kernel >= 2.6.32 */


CsrResult CsrSdioFunctionDriverRegister(CsrSdioFunctionDriver *functionDriver)
{
    struct glue_driver *glue_driver;

    glue_driver = kzalloc(sizeof(*glue_driver), GFP_KERNEL);
    if (glue_driver == NULL) {
        return CSR_RESULT_FAILURE;
    }
    functionDriver->priv = glue_driver;

    glue_driver->functionDriver = functionDriver;

    add_driver(glue_driver);

    return CSR_RESULT_SUCCESS;
}
EXPORT_SYMBOL(CsrSdioFunctionDriverRegister);

void CsrSdioFunctionDriverUnregister(CsrSdioFunctionDriver *functionDriver)
{
    struct glue_driver *glue_driver = functionDriver->priv;

    del_driver(glue_driver);
    kfree(glue_driver);
}
EXPORT_SYMBOL(CsrSdioFunctionDriverUnregister);

void CsrSdioInsertedAcknowledge(CsrSdioFunction *function, CsrResult result)
{
    struct glue_func *glue_func = function->priv;

    glue_func->acknowledge_result = result;
    complete(&glue_func->complete);
}
EXPORT_SYMBOL(CsrSdioInsertedAcknowledge);

void CsrSdioRemovedAcknowledge(CsrSdioFunction *function)
{
    struct glue_func *glue_func = function->priv;

    complete(&glue_func->complete);
}
EXPORT_SYMBOL(CsrSdioRemovedAcknowledge);

void CsrSdioSuspendAcknowledge(CsrSdioFunction *function, CsrResult result)
{
    struct glue_func *glue_func = function->priv;

    glue_func->acknowledge_result = result;
    complete(&glue_func->complete);
}
EXPORT_SYMBOL(CsrSdioSuspendAcknowledge);

void CsrSdioResumeAcknowledge(CsrSdioFunction *function, CsrResult result)
{
    struct glue_func *glue_func = function->priv;

    glue_func->acknowledge_result = result;
    complete(&glue_func->complete);
}
EXPORT_SYMBOL(CsrSdioResumeAcknowledge);


void glue_irq_handler(struct sdio_func *func)
{
    struct glue_func *glue_func = sdio_get_drvdata(func);
    CsrSdioInterruptDsrCallback dsr;

    set_bit(GLUE_BIT_INT_PENDING, &glue_func->bits);
    sdio_release_host(func);

    if (glue_func->glue_driver->functionDriver->intr) {
        dsr = glue_func->glue_driver->functionDriver->intr(&glue_func->function);
        if (dsr) {
            dsr(&glue_func->function);
        }
    }

    wait_event(glue_func->wq, test_and_clear_bit(GLUE_BIT_INT_ACKED, &glue_func->bits));

    sdio_claim_host(func);
    clear_bit(GLUE_BIT_INT_PENDING, &glue_func->bits);
    wake_up(&glue_func->wq);
}

CsrResult CsrSdioInterruptEnable(CsrSdioFunction *function)
{
    struct glue_func *glue_func = function->priv;
    struct sdio_func *func = glue_func->func;
    int ret;

    mutex_lock(&glue_func->irq_mutex);

    sdio_claim_host(func);
    ret = sdio_claim_irq(func, glue_irq_handler);
    sdio_release_host(func);

    mutex_unlock(&glue_func->irq_mutex);

    return CsrSdioStatusToResult(ret);
}
EXPORT_SYMBOL(CsrSdioInterruptEnable);

CsrResult CsrSdioInterruptDisable(CsrSdioFunction *function)
{
    struct glue_func *glue_func = function->priv;
    struct sdio_func *func = glue_func->func;
    struct mmc_host *host = func->card->host;
    sdio_irq_handler_t *irq_handler;
    int ret;

    /*
     * Since we release and reclaim the host in this function use an
     * additional mutex to prevent races between interrupt enable and
     * disable.
     */
    mutex_lock(&glue_func->irq_mutex);

    /*
     * Ug, this is messy.
     *
     * If this is the last function with an interrupt handler then
     * sdio_release_irq() will deadlock if we're currently inside the
     * interrupt handler.  To avoid this we unclaim the host,
     * acknowledge the interrupt and wait for the interrupt handler to
     * reclaim the host and then try to reclaim it.
     *
     * Since the interrupt is not actually cleared, a livelock may
     * occur with the interrupt handler constantly being
     * re-entered. The irq thread sleeps for a bit (allowing us to
     * claim the host) if process_sdio_pending_irqs() returns an
     * error.  func->irq_handler is set to NULL to cause this.
     *
     * func->irq_handler must be restored before calling
     * sdio_release_irq() or it won't correctly stop the irq thread.
     */
    sdio_claim_host(func);
    irq_handler = func->irq_handler;
    func->irq_handler = NULL;
    while (host->sdio_irqs == 1 && test_bit(GLUE_BIT_INT_PENDING, &glue_func->bits)) {
        sdio_release_host(func);
        CsrSdioInterruptAcknowledge(function);
        wait_event(glue_func->wq, test_bit(GLUE_BIT_INT_PENDING, &glue_func->bits) == 0);
        sdio_claim_host(func);
    }
    func->irq_handler = irq_handler;
    ret = sdio_release_irq(func);
    sdio_release_host(func);

    mutex_unlock(&glue_func->irq_mutex);

    return CsrSdioStatusToResult(ret);
}
EXPORT_SYMBOL(CsrSdioInterruptDisable);

void CsrSdioInterruptAcknowledge(CsrSdioFunction *function)
{
    struct glue_func *glue_func = function->priv;

    set_bit(GLUE_BIT_INT_ACKED, &glue_func->bits);
    wake_up(&glue_func->wq);
}
EXPORT_SYMBOL(CsrSdioInterruptAcknowledge);


CsrResult CsrSdioRead8(CsrSdioFunction *function, CsrUint32 address, CsrUint8 *data)
{
    struct glue_func *glue_func = function->priv;
    struct sdio_func *func = glue_func->func;
    int err = 0;

    sdio_claim_host(func);
    *data = sdio_readb(func, address, &err);
    sdio_release_host(func);

    return CsrSdioStatusToResult(err);
}
EXPORT_SYMBOL(CsrSdioRead8);

CsrResult CsrSdioRead16(CsrSdioFunction *function, CsrUint32 address, CsrUint16 *data)
{
    struct glue_func *glue_func = function->priv;
    struct sdio_func *func = glue_func->func;
    uint8_t h, l;
    int err = 0;

    sdio_claim_host(func);
    l = sdio_readb(func, address, &err);
    if (err < 0) {
        goto out;
    }
    h = sdio_readb(func, address+1, &err);
    if (err < 0) {
        goto out;
    }
    *data = (h << 8) | l;
  out:
    sdio_release_host(func);

    return CsrSdioStatusToResult(err);
}
EXPORT_SYMBOL(CsrSdioRead16);

CsrResult CsrSdioWrite8(CsrSdioFunction *function, CsrUint32 address, CsrUint8 data)
{
    struct glue_func *glue_func = function->priv;
    struct sdio_func *func = glue_func->func;
    int err = 0;

    sdio_claim_host(func);
    sdio_writeb(func, data, address, &err);
    sdio_release_host(func);

    return CsrSdioStatusToResult(err);
}
EXPORT_SYMBOL(CsrSdioWrite8);

CsrResult CsrSdioWrite16(CsrSdioFunction *function, CsrUint32 address, CsrUint16 data)
{
    struct glue_func *glue_func = function->priv;
    struct sdio_func *func = glue_func->func;
    uint8_t h, l;
    int err = 0;

    h = (data >> 8) & 0xff;
    l = (data >> 0) & 0xff;

    sdio_claim_host(func);
    sdio_writeb(func, h, address+1, &err);
    if (err < 0) {
        goto out;
    }
    sdio_writeb(func, l, address, &err);
  out:
    sdio_release_host(func);

    return CsrSdioStatusToResult(err);
}
EXPORT_SYMBOL(CsrSdioWrite16);

CsrResult CsrSdioF0Read8(CsrSdioFunction *function, CsrUint32 address, CsrUint8 *data)
{
    struct glue_func *glue_func = function->priv;
    struct sdio_func *func = glue_func->func;
    int err = 0;

    sdio_claim_host(func);
    *data = sdio_f0_readb(func, address, &err);
    sdio_release_host(func);

    return CsrSdioStatusToResult(err);
}
EXPORT_SYMBOL(CsrSdioF0Read8);

CsrResult CsrSdioF0Write8(CsrSdioFunction *function, CsrUint32 address, CsrUint8 data)
{
    struct glue_func *glue_func = function->priv;
    struct sdio_func *func = glue_func->func;
    int err = 0;

    sdio_claim_host(func);
    sdio_f0_writeb(func, data, address, &err);
    sdio_release_host(func);

    return CsrSdioStatusToResult(err);
}
EXPORT_SYMBOL(CsrSdioF0Write8);

CsrResult CsrSdioRead(CsrSdioFunction *function, CsrUint32 address, void *data, CsrUint32 length)
{
    struct glue_func *glue_func = function->priv;
    struct sdio_func *func = glue_func->func;
    int err;

    sdio_claim_host(func);
    err = sdio_readsb(func, data, address, length);
    sdio_release_host(func);

    return CsrSdioStatusToResult(err);
}
EXPORT_SYMBOL(CsrSdioRead);

CsrResult CsrSdioWrite(CsrSdioFunction *function, CsrUint32 address,
                       const void *data, CsrUint32 length)
{
    struct glue_func *glue_func = function->priv;
    struct sdio_func *func = glue_func->func;
    int err;

    sdio_claim_host(func);
    err = sdio_writesb(func, address, (void*)data, length);
    sdio_release_host(func);

    return CsrSdioStatusToResult(err);
}
EXPORT_SYMBOL(CsrSdioWrite);


CsrResult CsrSdioFunctionEnable(CsrSdioFunction *function)
{
    struct glue_func *glue_func = function->priv;
    struct sdio_func *func = glue_func->func;
    int err;

    sdio_claim_host(func);
    err = sdio_enable_func(func);
    sdio_release_host(func);

    return CsrSdioStatusToResult(err);
}
EXPORT_SYMBOL(CsrSdioFunctionEnable);

CsrResult CsrSdioFunctionDisable(CsrSdioFunction *function)
{
    struct glue_func *glue_func = function->priv;
    struct sdio_func *func = glue_func->func;
    int err;

    sdio_claim_host(func);
    err = sdio_disable_func(func);
    sdio_release_host(func);

    return CsrSdioStatusToResult(err);
}
EXPORT_SYMBOL(CsrSdioFunctionDisable);


static int wait_for_io_ready(struct sdio_func *func)
{
    uint8_t io_rdy;
    unsigned long timeout;
    int ret;

    timeout = jiffies + msecs_to_jiffies(IO_EN_TIMEOUT_MS);

    for (;;) {
        io_rdy = sdio_f0_readb(func, SDIO_CCCR_IO_READY, &ret);
        if (ret) {
            return CsrSdioStatusToResult(ret);
        }
        if (io_rdy & (1 << func->num)) {
            return CSR_RESULT_SUCCESS;
        }

        if (time_after(jiffies, timeout)) {
            break;
        }
        msleep(10);
    }
    return CSR_SDIO_RESULT_TIMEOUT;
}

CsrResult CsrSdioFunctionReenable(CsrSdioFunction *function)
{
    struct glue_func *glue_func = function->priv;
    struct sdio_func *func = glue_func->func;
    uint8_t io_en;
    unsigned long timeout;
    int ret;

    sdio_claim_host(func);

    /*
     * Write the function's enable bit to Extended I/O Enable.
     *
     * The register must be read and the bit checked as the enable bit
     * will remain cleared if the function is still held in reset.
     */
    timeout = jiffies + msecs_to_jiffies(RESET_TIMEOUT_MS);
    for (;;) {
        io_en = 1 << func->num;
        sdio_f0_writeb(func, io_en, SDIO_CSR_EXT_IO_EN, &ret);
        if (ret) {
            goto out;
        }

        io_en = sdio_f0_readb(func, SDIO_CSR_EXT_IO_EN, &ret);
        if (ret) {
            goto out;
        }
        if (io_en & (1 << func->num)) {
            break;
        }

        if (time_after(jiffies, timeout)) {
            ret = -ETIMEDOUT;
            goto out;
        }
        msleep(10);
    }
    ret = wait_for_io_ready(func);

  out:
    sdio_release_host(func);
    return CsrSdioStatusToResult(ret);
}
EXPORT_SYMBOL(CsrSdioFunctionReenable);


/*
 * With the Linux MMC stack, it's up to the host controller driver to
 * manage the clocks to reduce power consumption so
 * CsrSdioFunctionIdle() and CsrSdioFunctionActive() do nothing.
 */
void CsrSdioFunctionIdle(CsrSdioFunction *function)
{
}
EXPORT_SYMBOL(CsrSdioFunctionIdle);

void CsrSdioFunctionActive(CsrSdioFunction *function)
{
}
EXPORT_SYMBOL(CsrSdioFunctionActive);

void CsrSdioPowerOff(CsrSdioFunction *function)
{
}
EXPORT_SYMBOL(CsrSdioPowerOff);

CsrResult CsrSdioPowerOn(CsrSdioFunction *function)
{
    return CSR_SDIO_RESULT_NOT_RESET;
}
EXPORT_SYMBOL(CsrSdioPowerOn);

CsrResult CsrSdioHardReset(CsrSdioFunction *function)
{
    return CSR_SDIO_RESULT_NOT_RESET;
}
EXPORT_SYMBOL(CsrSdioHardReset);

CsrResult CsrSdioBlockSizeSet(CsrSdioFunction *function, CsrUint16 block_size)
{
    struct glue_func *glue_func = function->priv;
    struct sdio_func *func = glue_func->func;
    int ret;

    sdio_claim_host(func);
    ret = sdio_set_block_size(func, block_size);
    if (ret >= 0) {
        function->blockSize = func->cur_blksize;
    }
    sdio_release_host(func);

    return CsrSdioStatusToResult(ret);
}
EXPORT_SYMBOL(CsrSdioBlockSizeSet);

CsrResult CsrSdioMaxBusClockFrequencySet(CsrSdioFunction *function, CsrUint32 max_hz)
{
    struct glue_func *glue_func = function->priv;
    struct sdio_func *func = glue_func->func;
    struct mmc_host *host = func->card->host;
    struct mmc_ios *ios = &host->ios;

    sdio_claim_host(func);

    if (max_hz > host->f_max) {
        max_hz = host->f_max;
    }
    ios->clock = max_hz;
    host->ops->set_ios(host, ios);

    sdio_release_host(func);

    return CSR_RESULT_SUCCESS;
}
EXPORT_SYMBOL(CsrSdioMaxBusClockFrequencySet);


/*
 * Any CSR chip supported by Synergy.
 */
static const struct sdio_device_id csr_ids[] = {
    { SDIO_DEVICE(SDIO_MANF_ID_CSR, SDIO_CARD_ID_CSR_UNIFI_3) },
    { SDIO_DEVICE(SDIO_MANF_ID_CSR, SDIO_CARD_ID_CSR_UNIFI_4) },
    { SDIO_DEVICE(SDIO_MANF_ID_CSR, SDIO_CARD_ID_CSR_BC6) },
    { SDIO_DEVICE(SDIO_MANF_ID_CSR, SDIO_CARD_ID_CSR_BC7) },
    { SDIO_DEVICE(SDIO_MANF_ID_CSR, SDIO_CARD_ID_CSR_DASH) },
    { 0 },
};
MODULE_DEVICE_TABLE(sdio, csr_ids);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,32)
#ifdef CONFIG_PM

static struct dev_pm_ops unifi_pm_ops = {
    .suspend = glue_suspend,
    .resume  = glue_resume,
};

#define UNIFI_PM_OPS  (&unifi_pm_ops)

#else

#define UNIFI_PM_OPS  NULL

#endif /* CONFIG_PM */
#endif /* kernel >= 2.6.32 */

static struct sdio_driver synergy_driver = {
    .probe      = glue_probe,
    .remove     = glue_remove,
    .name       = "synergy",
    .id_table   = csr_ids,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,32)
    .drv.pm     = UNIFI_PM_OPS,
#endif /* kernel >= 2.6.32 */
};

static int __init synergy_mmc_init(void)
{
    return sdio_register_driver(&synergy_driver);
}
module_init(synergy_mmc_init);

static void __exit synergy_mmc_exit(void)
{
    sdio_unregister_driver(&synergy_driver);
}
module_exit(synergy_mmc_exit);

MODULE_DESCRIPTION("CSR Synergy SDIO for the Linux MMC stack");
MODULE_AUTHOR("Cambridge Silicon Radio Ltd.");
MODULE_LICENSE("GPL and additional rights");
