/*
 * Linux kernel module support.
 *
 * Copyright (C) 2007-2009 Cambridge Silicon Radio Ltd.
 *
 * Refer to LICENSE.txt included with this source code for details on
 * the license terms.
 */
#include <linux/module.h>
#include <linux/device.h>

#include "../core/sdio_internal.h"
#include <csr_sdio.h>
#include <csr_sdio_lib.h>

int sdioemb_card_poll_interval_ms = 500;
module_param_named(card_poll_interval, sdioemb_card_poll_interval_ms, int, 0644);
MODULE_PARM_DESC(card_poll_interval, "interval (in ms) between card detect polls");

int sdioemb_max_bus_freq = 0;
module_param_named(max_bus_freq, sdioemb_max_bus_freq, int, 0444);
MODULE_PARM_DESC(max_bus_freq, "maximum bus frequency (in Hz), or 0 for card's maximum");

uint16_t sdioemb_max_block_size = 0;
module_param_named(max_block_size, sdioemb_max_block_size, ushort, 0444);
MODULE_PARM_DESC(max_block_size, "maximum block size, or 0 for the card's maximum");

#if SDIOEMB_DEBUG_DATA_BUFFER_LEN > 0
module_param_named(debug_event_log_data_len, sdioemb_debug_event_log_data_len, uint, 0644);
MODULE_PARM_DESC(debug_event_log_data_len, "length of data to log from each command with data");
#endif

EXPORT_SYMBOL(sdioemb_driver_register);
EXPORT_SYMBOL(sdioemb_driver_unregister);
EXPORT_SYMBOL(sdioemb_set_block_size);
EXPORT_SYMBOL(sdioemb_set_max_bus_freq);
EXPORT_SYMBOL(sdioemb_set_bus_width);
EXPORT_SYMBOL(sdioemb_enable_function);
EXPORT_SYMBOL(sdioemb_disable_function);
EXPORT_SYMBOL(sdioemb_reenable_csr_function);
EXPORT_SYMBOL(sdioemb_idle_function);
EXPORT_SYMBOL(sdioemb_interrupt_enable);
EXPORT_SYMBOL(sdioemb_interrupt_disable);
EXPORT_SYMBOL(sdioemb_interrupt_acknowledge);
EXPORT_SYMBOL(sdioemb_cis_get_tuple);

EXPORT_SYMBOL(sdioemb_read8);
EXPORT_SYMBOL(sdioemb_read16);
EXPORT_SYMBOL(sdioemb_write8);
EXPORT_SYMBOL(sdioemb_write16);
EXPORT_SYMBOL(sdioemb_f0_read8);
EXPORT_SYMBOL(sdioemb_f0_write8);
EXPORT_SYMBOL(sdioemb_read);
EXPORT_SYMBOL(sdioemb_write);

EXPORT_SYMBOL(sdioemb_slot_alloc);
EXPORT_SYMBOL(sdioemb_slot_free);
EXPORT_SYMBOL(sdioemb_slot_register);
EXPORT_SYMBOL(sdioemb_slot_unregister);
EXPORT_SYMBOL(sdioemb_card_inserted);
EXPORT_SYMBOL(sdioemb_card_removed);
EXPORT_SYMBOL(sdioemb_hard_reset);
EXPORT_SYMBOL(sdioemb_interrupt);
EXPORT_SYMBOL(sdioemb_cmd_complete);

EXPORT_SYMBOL(sdioemb_power_on);
EXPORT_SYMBOL(sdioemb_power_off);
EXPORT_SYMBOL(sdioemb_suspend);
EXPORT_SYMBOL(sdioemb_resume);

/* Only for sdio_uif. */
EXPORT_SYMBOL(slot_set_bus_freq);

/* Synergy compatible API. */
EXPORT_SYMBOL(CsrSdioFunctionDriverRegister);
EXPORT_SYMBOL(CsrSdioFunctionDriverUnregister);
EXPORT_SYMBOL(CsrSdioInsertedAcknowledge);
EXPORT_SYMBOL(CsrSdioRemovedAcknowledge);
EXPORT_SYMBOL(CsrSdioSuspendAcknowledge);
EXPORT_SYMBOL(CsrSdioResumeAcknowledge);
EXPORT_SYMBOL(CsrSdioFunctionEnable);
EXPORT_SYMBOL(CsrSdioFunctionDisable);
EXPORT_SYMBOL(CsrSdioInterruptEnable);
EXPORT_SYMBOL(CsrSdioInterruptDisable);
EXPORT_SYMBOL(CsrSdioInterruptAcknowledge);
EXPORT_SYMBOL(CsrSdioBlockSizeSet);
EXPORT_SYMBOL(CsrSdioMaxBusClockFrequencySet);
EXPORT_SYMBOL(CsrSdioRead8);
EXPORT_SYMBOL(CsrSdioRead16);
EXPORT_SYMBOL(CsrSdioWrite8);
EXPORT_SYMBOL(CsrSdioWrite16);
EXPORT_SYMBOL(CsrSdioF0Read8);
EXPORT_SYMBOL(CsrSdioF0Write8);
EXPORT_SYMBOL(CsrSdioRead);
EXPORT_SYMBOL(CsrSdioWrite);
EXPORT_SYMBOL(CsrSdioPowerOn);
EXPORT_SYMBOL(CsrSdioPowerOff);
EXPORT_SYMBOL(CsrSdioHardReset);
EXPORT_SYMBOL(CsrSdioFunctionActive);
EXPORT_SYMBOL(CsrSdioFunctionIdle);

EXPORT_SYMBOL(CsrSdioFunctionReenable);

EXPORT_SYMBOL(CsrSdioResultToStatus);
EXPORT_SYMBOL(CsrSdioStatusToResult);

struct sdioe_dev {
    struct device dev;
};

struct bus_type sdioe_bus_type = {
    .name   = "sdioemb",
    /* Seems we can we get away with these being NULL since no drivers
       are ever added to the bus. */
    .match  = NULL,
    .probe  = NULL,
    .remove = NULL,
};

static void sdioe_device_release(struct device *dev)
{
    struct sdioe_dev *sdev = container_of(dev, struct sdioe_dev, dev);
    kfree(sdev);
}

static struct sdioe_dev *sdioe_device_create(struct sdioemb_dev *fdev)
{
    struct sdioemb_slot *slot = fdev->priv->slot;
    struct sdioe_dev *sdev;

    sdev = kzalloc(sizeof(struct sdioe_dev), GFP_KERNEL);
    if (sdev) {
        dev_set_name(&sdev->dev, "%s.%d", slot->name, fdev->function);
        sdev->dev.parent  = NULL;
        sdev->dev.bus     = &sdioe_bus_type;
        sdev->dev.release = sdioe_device_release;
    }
    return sdev;
}

static int sdioemb_linux_add_func(struct sdioemb_slot *slot, struct sdioemb_dev *fdev)
{
    struct sdioe_dev *sdev;
    int ret = -ENOMEM;

    sdev = sdioe_device_create(fdev);
    if (sdev) {
        ret = device_register(&sdev->dev);
        if (ret < 0) {
            kfree(sdev);
        } else {
            fdev->os_device = &sdev->dev;
        }
    }
    return ret < 0 ? ret : SDIOEMB_ADD_FUNC_NOW;
}

static void sdioemb_linux_del_func(struct sdioemb_slot *slot, struct sdioemb_dev *fdev)
{
    if (fdev->os_device) {
        device_unregister(fdev->os_device);
        fdev->os_device = NULL;
    }
}

int sdioemb_linux_slot_register(struct sdioemb_slot *slot)
{
    slot->notifs.add_function = sdioemb_linux_add_func;
    slot->notifs.del_function = sdioemb_linux_del_func;

    return sdioemb_slot_register(slot);
}
EXPORT_SYMBOL(sdioemb_linux_slot_register);

static int __init sdio_init(void)
{
    int err;

    err = bus_register(&sdioe_bus_type);
    if (err) {
        return err;
    }

    sdio_core_init();
    return 0;
}
module_init(sdio_init);

static void __exit sdio_exit(void)
{
    sdio_core_clean_up();
    bus_unregister(&sdioe_bus_type);
}
module_exit(sdio_exit);

MODULE_DESCRIPTION("Embedded SDIO device driver");
MODULE_AUTHOR("Cambridge Silicon Radio Ltd.");
MODULE_LICENSE("GPL and additional rights");
