/*
 * Test SDIO function driver.
 *
 * Copyright (C) 2010 Cambridge Silicon Radio Ltd.
 *
 * Refer to LICENSE.txt included with this source code for details on
 * the license terms.
 *
 * This is a test function driver for a CSR generic function.  There
 * are a number of tests for measuring performance of SDIO commands.
 * A debugfs file (sdio_test_ctrl) can be written to to trigger tests.
 *
 * Available commands:
 *
 *  "0 N" - N CMD52 reads.
 *  "1 N L" - N CMD53 reads of length L (buffers aligned on word boundaries).
 *  "2 N L" - N CMD53 reads of length L (buffers aligned on half-word boundaries).
 *  "3 N L" - CMD53 write/read of length L (checked, word aligned).
 *  "4 N L" - CMD53 write/read of length L (checked, half-word aligned).
 */
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/uaccess.h>
#include <linux/device.h>
#include <linux/debugfs.h>
#include <linux/slab.h>

#include <csr_sdio.h>
#include <sdioemb/sdio.h>
#include <sdioemb/sdio_csr.h>

struct test_data {
    struct device *dev;
    struct dentry *ctrl_file;
};

typedef int (*test_f)(CsrSdioFunction *func, int p1, int p2);

struct test_entry {
    const char *name;
    test_f run;
};

static int test_cmd52_perf(CsrSdioFunction *func, int p1, int p2)
{
    int i;
    uint8_t v;
    CsrResult result;

    for (i = 0; i < p1; i++) {
        result = CsrSdioF0Read8(func, SDIO_CSR_TO_HOST_SCRATCH0, &v);
        if (result) {
            return result;
        }
    }

    return 0;
}

static int test_cmd53_read_perf(CsrSdioFunction *func, int n, void *buf, size_t len)
{
    int i;
    CsrResult result;

    for (i = 0; i < n; i++) {
        result = CsrSdioRead(func, 0x13, buf, len);
        if (result) {
            return result;
        }
    }

    return 0;
}

static int test_cmd53_read_perf_word_aligned(CsrSdioFunction *func, int p1, int p2)
{
    void *buf;

    if (p2 <= 0 || p2 >= 4096) {
        return -EINVAL;
    }

    buf = kmalloc(p2, GFP_KERNEL);
    if (buf == NULL) {
        return -ENOMEM;
    }

    return test_cmd53_read_perf(func, p1, buf, p2);
}

static int test_cmd53_read_perf_hword_aligned(CsrSdioFunction *func, int p1, int p2)
{
    void *buf;
    CsrResult result;

    if (p2 <= 0 || p2 >= (4096-2)) {
        return -EINVAL;
    }

    buf = kmalloc(p2 + 2, GFP_KERNEL);
    if (buf == NULL) {
        return -ENOMEM;
    }

    result = test_cmd53_read_perf(func, p1, buf + 2, p2);

    kfree(buf);
    return result;
}


static int test_cmd53_rw_checked(CsrSdioFunction *func, int n, u8 *out, u8 *in, size_t l)
{
    struct test_data *td = func->driverData;
    int i, j;
    u8 p = 0;
    CsrResult result;

    for (i = 0; i < n; i++) {
        for (j = 0; j < l; j++) {
            out[j] = p++;
        }
        result = CsrSdioWrite(func, 0x12, out, l);
        if (result) {
            return result;
        }
        result = CsrSdioRead(func, 0x13, in, l);
        if (result) {
            return result;
        }
        if (memcmp(out, in, l) != 0) {
            dev_err(td->dev, "read buffer != written\n");
            return -EIO;
        }
    }
    return 0;
}

static int test_cmd53_rw_checked_word_aligned(CsrSdioFunction *func, int p1, int p2)
{
    void *buf;
    CsrResult result;

    if (p2 <= 0 || p2 >= (4096-2)) {
        return -EINVAL;
    }

    buf = kmalloc(8192, GFP_KERNEL);
    if (buf == NULL) {
        return -ENOMEM;
    }

    result = test_cmd53_rw_checked(func, p1, buf, buf + 4096, p2);

    kfree(buf);
    return result;
}

static int test_cmd53_rw_checked_hword_aligned(CsrSdioFunction *func, int p1, int p2)
{
    void *buf;
    CsrResult result;

    if (p2 <= 0 || p2 >= (4096-2)) {
        return -EINVAL;
    }

    buf = kmalloc(8192, GFP_KERNEL);
    if (buf == NULL) {
        return -ENOMEM;
    }

    result = test_cmd53_rw_checked(func, p1, buf + 2, buf + 4096 + 2, p2);

    kfree(buf);
    return result;
}

static struct test_entry tests[] = {
    {
        .name = "CMD52",
        .run  = test_cmd52_perf,
    },
    {
        .name = "CMD53 read (word aligned)",
        .run  = test_cmd53_read_perf_word_aligned,
    },
    {
        .name = "CMD53 read (hword aligned)",
        .run  = test_cmd53_read_perf_hword_aligned,
    },
    {
        .name = "CMD53 write/read (checked, word aligned)",
        .run = test_cmd53_rw_checked_word_aligned,
    },
    {
        .name = "CMD53 write/read (checked, hword aligned)",
        .run = test_cmd53_rw_checked_hword_aligned,
    },
};

static int ctrl_open(struct inode *inode, struct file *file)
{
    file->private_data = inode->i_private;

    return 0;
}

static ssize_t ctrl_write(struct file *file, const char __user *buf,
                          size_t len, loff_t *off)
{
    CsrSdioFunction *func = file->private_data;
    struct test_data *td = func->driverData;
    char *cmd;
    int ret;
    int idx, p1;
    int p2 = 0;
    struct timespec a, b;

    cmd = kmalloc(len, GFP_KERNEL);
    if (cmd == NULL) {
        ret = -ENOMEM;
        goto out;
    }
    if (copy_from_user(cmd, buf, len) != 0) {
        ret = -EFAULT;
        goto out;
    }

    ret = sscanf(cmd, "%d %d %d", &idx, &p1, &p2);
    if (ret < 2 || idx < 0 || idx >= ARRAY_SIZE(tests)) {
        dev_err(td->dev, "invalid command\n");
        ret = -EINVAL;
        goto out;
    }

    getnstimeofday(&a);
    ret = tests[idx].run(func, p1, p2);
    getnstimeofday(&b);
    if (ret >= 0) {
        unsigned long us;
        us = (b.tv_sec - a.tv_sec) * 1000000 + (b.tv_nsec - a.tv_nsec) / 1000;
        dev_info(td->dev, "test %s(%d, %d): %lu.%06lu s\n", tests[idx].name, p1, p2,
                 us / 1000000, us % 1000000);
    } else {
        dev_info(td->dev, "test %s(%d, %d): failed: %d\n", tests[idx].name, p1, p2, ret);
    }
  out:
    kfree(cmd);
    return ret < 0 ? ret : len;
}

static const struct file_operations ctrl_fops = {
        .open   = ctrl_open,
        .write  = ctrl_write,
        .read   = NULL,
        .llseek = no_llseek,
        .owner  = THIS_MODULE,
};

static int start_function(CsrSdioFunction *func)
{
    struct test_data *td = func->driverData;
    int ret;

    ret = CsrSdioFunctionEnable(func);
    if (ret < 0) {
        dev_err(td->dev, "failed to enable function: %d\n", ret);
        return ret;
    }

    ret = CsrSdioWrite8(func, (0xf935 << 1), 0x00);
    if (ret < 0) {
        dev_err(td->dev, "failed to set SDIO_MODE: %d\n", ret);
        return ret;
    }

    return 0;
}

static void stop_function(CsrSdioFunction *func)
{
    CsrSdioWrite8(func, (0xf935 << 1), 0x01);
    CsrSdioFunctionDisable(func);
}

static struct test_data *alloc_drv_data(CsrSdioFunction *func)
{
    struct test_data *td;

    td = kzalloc(sizeof(*td), GFP_KERNEL);
    if (td == NULL) {
        return NULL;
    }
    func->driverData = td;

    return td;
}

static void free_drv_data(struct test_data *td)
{
    kfree(td);
}


static void test_inserted(CsrSdioFunction *func)
{
    struct test_data *td;
    CsrResult result;

    td = alloc_drv_data(func);
    if (td == NULL) {
        result = CSR_RESULT_FAILURE;
        goto out;
    }

    td->dev = func->osDevice;
    td->ctrl_file = debugfs_create_file("sdio_test_ctrl", 0200, NULL, func, &ctrl_fops);

    result = start_function(func);
    if (result) {
        goto out;
    }

  out:
    if (result) {
        if (td) {
            stop_function(func);
            debugfs_remove(td->ctrl_file);
            free_drv_data(td);
        }
    }
    CsrSdioInsertedAcknowledge(func, result);
}

static void test_removed(CsrSdioFunction *func)
{
    struct test_data *td = func->driverData;

    stop_function(func);
    debugfs_remove(td->ctrl_file);
    free_drv_data(td);

    CsrSdioRemovedAcknowledge(func);
}

static CsrSdioFunctionId test_ids[] = {
    { SDIO_MANF_ID_CSR, SDIO_CARD_ID_CSR_DASH, 2, CSR_SDIO_ANY_SDIO_INTERFACE, },
};

static CsrSdioFunctionDriver test_driver = {
    test_inserted,
    test_removed,
    NULL, /* intr */
    NULL, /* suspend */
    NULL, /* resume */
    test_ids,
    sizeof(test_ids)/sizeof(test_ids[0]),
};

static int __init test_init(void)
{
    return CsrSdioFunctionDriverRegister(&test_driver);
}
module_init(test_init);

static void __exit test_exit(void)
{
    CsrSdioFunctionDriverUnregister(&test_driver);
}
module_exit(test_exit);

MODULE_DESCRIPTION("SDIO test driver");
MODULE_AUTHOR("Cambridge Silicon Radio Ltd.");
MODULE_LICENSE("GPL and additional rights");
