/*
 * SDIO Userspace Interface driver.
 *
 * Copyright (C) 2007-2008 Cambridge Silicon Radio Ltd.
 *
 * Refer to LICENSE.txt included with this source code for details on
 * the license terms.
 */
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/device.h>
#include <linux/mutex.h>
#include <linux/sched.h>
#include <linux/slab.h>

#include <asm/uaccess.h>

#include <linux/kernel-compat.h>

#include <csr_sdio.h>
#include <csr_sdio_lib.h>
#include <linux/sdioemb/uif.h>
#include <sdioemb/sdio.h>

#define UIF_MAX_CARDS 16

struct sdio_uif_dev {
    CsrSdioFunction    *fdev;
    int                 card_present;
    struct mutex        sdio_mutex;
    wait_queue_head_t   waitq;
    struct cdev         cdev;
    atomic_t            ref_count;

    int                 interrupt;
};

static dev_t          uif_first_dev;
static struct class * uif_class;

static void uif_get(struct sdio_uif_dev *ud)
{
    atomic_inc(&ud->ref_count);
}

static void uif_put(struct sdio_uif_dev *ud)
{
    if (atomic_dec_and_test(&ud->ref_count)) {
        kfree(ud);
    }
}


static inline int uif_num_functions(struct sdio_uif_dev *ud)
{
    return -EINVAL;
}

static int uif_do_cmd(struct sdio_uif_dev *ud, struct sdioemb_uif_cmd *cmd, uint8_t *data)
{
    CsrResult result;

    switch (cmd->type) {
    case SDIOEMB_UIF_CMD52_READ:
        if (cmd->function == 0) {
            result = CsrSdioF0Read8(ud->fdev, cmd->address, data);
        } else {
            result = CsrSdioRead8(ud->fdev, cmd->address, data);
        }
        break;
    case SDIOEMB_UIF_CMD52_WRITE:
        if (cmd->function == 0) {
            result = CsrSdioF0Write8(ud->fdev, cmd->address, *data);
        } else {
            result = CsrSdioWrite8(ud->fdev, cmd->address, *data);
        }
        break;
    case SDIOEMB_UIF_CMD53_READ:
        result = CsrSdioRead(ud->fdev, cmd->address, data, cmd->len);
        break;
    case SDIOEMB_UIF_CMD53_WRITE:
        result = CsrSdioWrite(ud->fdev, cmd->address, data, cmd->len);
        break;
    default:
        return -EINVAL;
    }
    return CsrSdioResultToStatus(result);
}

static int uif_cmd(struct sdio_uif_dev *ud, struct sdioemb_uif_cmd __user *cmd)
{
    struct sdioemb_uif_cmd scmd;
    uint8_t *data;
    int ret;

    if (copy_from_user(&scmd, cmd, sizeof(struct sdioemb_uif_cmd))) {
        return -EFAULT;
    }
    data = kmalloc(scmd.len, GFP_KERNEL);
    if (!data) {
        return -ENOMEM;
    }
    if (copy_from_user(data, scmd.data, scmd.len) == 0) {
        ret = uif_do_cmd(ud, &scmd, data);
        if (!ret && copy_to_user(scmd.data, data, scmd.len)) {
            ret = -EFAULT;
        }
    } else {
        ret = -EFAULT;
    }
    kfree(data);
    return ret;
}

static int uif_wait_for_interrupt(struct sdio_uif_dev *ud)
{
    int ret = 0;

    CsrSdioInterruptAcknowledge(ud->fdev);

    ret = wait_event_interruptible(ud->waitq, ud->interrupt);
    if (ret) {
        ret = -ERESTARTSYS;
    } else {
        ud->interrupt = 0;
    }

    return ret;
}

static int uif_set_bus_width(struct sdio_uif_dev *ud, int bus_width)
{
    return -ENOTSUPP;
}

static int uif_reinsert(struct sdio_uif_dev *ud)
{
    return -ENOTSUPP;
}

static int uif_set_bus_freq(struct sdio_uif_dev *ud, int bus_freq)
{
    return -ENOTSUPP;
}

static int uif_manf_id(struct sdio_uif_dev *ud)
{
    return ud->fdev->sdioId.manfId;
}

static int uif_card_id(struct sdio_uif_dev *ud)
{
    return ud->fdev->sdioId.cardId;
}

static int uif_get_std_if(struct sdio_uif_dev *ud, unsigned long func)
{
    return -ENOTSUPP;
}

static int uif_get_max_block_size(struct sdio_uif_dev *ud, unsigned long func)
{
    return -ENOTSUPP;
}

static int uif_get_block_size(struct sdio_uif_dev *ud, unsigned long func)
{
    return -ENOTSUPP;
}

static int uif_set_block_size(struct sdio_uif_dev *ud, unsigned long blksz, unsigned long func)
{
    return -ENOTSUPP;
}

static long uif_ioctl(struct file *file, unsigned cmd, unsigned long arg)
{
    struct sdio_uif_dev *ud = file->private_data;
    int ret;

    mutex_lock(&ud->sdio_mutex);

    if (!ud->card_present) {
        ret = -ENODEV;
        goto out;
    }

    switch (cmd) {
    case SDIOEMB_UIF_IOCQNUMFUNCS:
        ret = uif_num_functions(ud);
        break;
    case SDIOEMB_UIF_IOCCMD:
        ret = uif_cmd(ud, (struct sdioemb_uif_cmd __user *)arg);
        break;
    case SDIOEMB_UIF_IOCWAITFORINT:
        ret = uif_wait_for_interrupt(ud);
        break;
    case SDIOEMB_UIF_IOCTBUSWIDTH:
        ret = uif_set_bus_width(ud, arg);
        break;
    case SDIOEMB_UIF_IOCREINSERT:
        ret = uif_reinsert(ud);
        break;
    case SDIOEMB_UIF_IOCTBUSFREQ:
        ret = uif_set_bus_freq(ud, (int)arg);
        break;
    case SDIOEMB_UIF_IOCQMANFID:
        ret = uif_manf_id(ud);
        break;
    case SDIOEMB_UIF_IOCQCARDID:
        ret = uif_card_id(ud);
        break;
    case SDIOEMB_UIF_IOCQSTDIF:
        ret = uif_get_std_if(ud, arg);
        break;
    case SDIOEMB_UIF_IOCQMAXBLKSZ:
        ret = uif_get_max_block_size(ud, arg);
        break;
    case SDIOEMB_UIF_IOCQBLKSZ:
        ret = uif_get_block_size(ud, arg);
        break;
    case SDIOEMB_UIF_IOCTBLKSZ:
        ret = uif_set_block_size(ud, arg >> 8, arg & 0xff); /* low octet is function number */
        break;
    default:
        ret = -ENOTTY;
        break;
    }

  out:
    mutex_unlock(&ud->sdio_mutex);
    return ret;
}

static int uif_open(struct inode *inode, struct file *file)
{
    struct sdio_uif_dev *ud = container_of(inode->i_cdev, struct sdio_uif_dev, cdev);
    int ret = 0;

    mutex_lock(&ud->sdio_mutex);

    if (!ud->card_present)
        ret = -ENODEV;
    else {
        uif_get(ud);
        file->private_data = ud;
    }

    mutex_unlock(&ud->sdio_mutex);

    CsrSdioInterruptEnable(ud->fdev);

    return ret;
}

static int uif_release(struct inode *inode, struct file *file)
{
    struct sdio_uif_dev *ud = container_of(inode->i_cdev, struct sdio_uif_dev, cdev);

    CsrSdioInterruptDisable(ud->fdev);
    CsrSdioInterruptAcknowledge(ud->fdev);

    uif_put(ud);

    return 0;
}


static struct file_operations uif_fops = {
    .owner          = THIS_MODULE,
    .open           = uif_open,
    .release        = uif_release,
    .unlocked_ioctl = uif_ioctl,
};

static void uif_inserted(CsrSdioFunction *fdev)
{
    struct sdio_uif_dev *ud;
    dev_t devno;
    int ret = -ENOMEM;

    ud = kzalloc(sizeof(struct sdio_uif_dev), GFP_KERNEL);
    if (!ud)
        goto error_kzalloc;
    fdev->driverData = ud;

    ud->fdev = fdev;

    cdev_init(&ud->cdev, &uif_fops);

    ud->card_present = 1;

    mutex_init(&ud->sdio_mutex);
    init_waitqueue_head(&ud->waitq);

    atomic_set(&ud->ref_count, 1);

    devno = MKDEV(MAJOR(uif_first_dev), MINOR(uif_first_dev) + 0 /* FIXME */);
    ret = cdev_add(&ud->cdev, devno, 1);
    if (ret)
        goto error_cdev_add;

    if (!device_create(uif_class, NULL, ud->cdev.dev, NULL, "sdio_uif%d", 0 /* FIXME */))
        goto error_device_create;

    goto done;

  error_device_create:
    cdev_del(&ud->cdev);
  error_cdev_add:
    kfree(ud);
  error_kzalloc:
  done:
    CsrSdioInsertedAcknowledge(fdev, CsrSdioStatusToResult(ret));
}

static void uif_removed(CsrSdioFunction *fdev)
{
    struct sdio_uif_dev *ud = fdev->driverData;

    mutex_lock(&ud->sdio_mutex);
    ud->card_present = 0;
    mutex_unlock(&ud->sdio_mutex);

    device_destroy(uif_class, ud->cdev.dev);
    cdev_del(&ud->cdev);
    uif_put(ud);

    CsrSdioRemovedAcknowledge(fdev);
}

static void uif_dsr(CsrSdioFunction *fdev)
{
    struct sdio_uif_dev *ud = fdev->driverData;

    ud->interrupt = 1;
    wake_up_interruptible(&ud->waitq);
}

static CsrSdioInterruptDsrCallback uif_isr(CsrSdioFunction *fdev)
{
    return uif_dsr;
}

static CsrSdioFunctionId uif_ids[] = {
    { SDIO_MANF_ID_CSR, SDIO_CARD_ID_CSR_DASH, 2, CSR_SDIO_ANY_SDIO_INTERFACE, },
};

static CsrSdioFunctionDriver uif_driver = {
    uif_inserted,
    uif_removed,
    uif_isr,
    NULL,
    NULL,
    uif_ids,
    2,
};

static int __init sdio_uif_init(void)
{
    int ret;
    CsrResult result;

    ret = alloc_chrdev_region(&uif_first_dev, 0, UIF_MAX_CARDS, "sdio_uif");
    if (ret)
        goto error;

    uif_class = class_create(THIS_MODULE, "sdio_uif");
    if (!uif_class)
        goto error;

    result = CsrSdioFunctionDriverRegister(&uif_driver);
    if (result != CSR_RESULT_SUCCESS) {
        ret = CsrSdioResultToStatus(result);
        goto error;
    }

    return 0;

  error:
    if (uif_class)
        class_destroy(uif_class);
    if (uif_first_dev)
        unregister_chrdev_region(uif_first_dev, UIF_MAX_CARDS);
    return ret;
}

static void __exit sdio_uif_exit(void)
{
    CsrSdioFunctionDriverUnregister(&uif_driver);

    class_destroy(uif_class);
    unregister_chrdev_region(uif_first_dev, UIF_MAX_CARDS);
}

module_init(sdio_uif_init);
module_exit(sdio_uif_exit);

MODULE_DESCRIPTION("SDIO Userspace Interface driver (using Synergy API)");
MODULE_AUTHOR("Cambridge Silicon Radio Ltd.");
MODULE_LICENSE("GPL and additional rights");
