/*
 * Linux PCI SHC slot driver.
 *
 * Copyright (C) 2007 Cambridge Silicon Radio Ltd.
 *
 * Refer to LICENSE.txt included with this source code for details on
 * the license terms.
 *
 * References:
 *   [SDHC] SD Host Controller Standard Specification v1.00.
 */
#include <linux/kernel.h>
#include <linux/interrupt.h>
#include <linux/pci.h>
#include <linux/kernel-compat.h>

#include <asm/io.h>

#include <oska/all.h>
#include <oska/io.h>

#include <sdioemb/sdio.h>
#include <sdioemb/sdio_api.h>
#include <sdioemb/slot_api.h>
#include <sdioemb/slot_shc.h>
#include <sdioemb/linux.h>
#include <sdioemb/trace.h>

struct shc_data {
    struct sdioemb_shc shc;
    struct pci_dev *pdev;
    bool use_dma;
    dma_addr_t dma_addr;
};

/* Software timeout for commands. For working around buggy controllers
   that occasionally fail to complete a command. */
#define SLOT_SHC_LOCKUP_TIMEOUT_MS 2000

static void shc_enable_int(struct sdioemb_slot *slot, uint32_t ints)
{
    struct shc_data *shc = slot->drv_data;
    os_int_status_t istate;
    uint32_t stat_en;

    os_spinlock_lock_intsave(&shc->shc.lock, &istate);

    stat_en = os_io_read32(shc->shc.addr, SHC_INT_STATUS_EN);
    os_io_write32(shc->shc.addr, SHC_INT_STATUS_EN, stat_en | ints);

    (void)os_io_read32(shc->shc.addr, SHC_INT_STATUS_EN); /* flush posted writes */
    os_spinlock_unlock_intrestore(&shc->shc.lock, &istate);
}
 
static void shc_disable_int(struct sdioemb_slot *slot, uint32_t ints)
{
    struct shc_data *shc = slot->drv_data;
    os_int_status_t istate;
    uint32_t stat_en;

    os_spinlock_lock_intsave(&shc->shc.lock, &istate);

    stat_en = os_io_read32(shc->shc.addr, SHC_INT_STATUS_EN);
    os_io_write32(shc->shc.addr, SHC_INT_STATUS_EN, stat_en & ~ints);

    (void)os_io_read32(shc->shc.addr, SHC_INT_STATUS_EN); /* flush posted writes */
    os_spinlock_unlock_intrestore(&shc->shc.lock, &istate);
}

static void shc_cmd_complete(struct sdioemb_slot *slot, struct sdioemb_cmd *cmd)
{
    struct shc_data *shc = slot->drv_data;

    if (shc->use_dma) {
        pci_unmap_single(shc->pdev, shc->dma_addr, cmd->len,
                         (cmd->flags & SDIOEMB_CMD_FLAG_READ) ? DMA_FROM_DEVICE : DMA_TO_DEVICE);
    }

    sdioemb_cmd_complete(slot, cmd);
}

static int shc_set_bus_freq(struct sdioemb_slot *slot, int clk)
{
    struct shc_data *shc = slot->drv_data;

    return sdioemb_shc_set_bus_freq(&shc->shc, clk);
}

static int shc_set_bus_width(struct sdioemb_slot *slot, int bus_width)
{
    struct shc_data *shc = slot->drv_data;

    return sdioemb_shc_set_bus_width(&shc->shc, bus_width);
}

static int shc_start_cmd(struct sdioemb_slot *slot, struct sdioemb_cmd *cmd)
{
    struct shc_data *shc = slot->drv_data;

    if (cmd->data) {
        shc->use_dma = true;
        if ((shc->shc.quirks & SHC_QUIRK_DMA_NEEDS_DWORD_ALIGNED_ADDR)
            && ((unsigned long)cmd->data & 0x3)) {
            shc->use_dma = false;
        }
        if ((shc->shc.quirks & SHC_QUIRK_DMA_NEEDS_DWORD_ALIGNED_LEN)
            && ((unsigned long)cmd->len & 0x3)) {
            shc->use_dma = false;
        }
    } else {
        shc->use_dma = false;
    }
    if (shc->use_dma) {
        shc->dma_addr = pci_map_single(shc->pdev, cmd->data, cmd->len,
                                       (cmd->flags & SDIOEMB_CMD_FLAG_READ) ? DMA_FROM_DEVICE
                                       : DMA_TO_DEVICE);
    }

    sdioemb_shc_start_cmd(&shc->shc, cmd, shc->use_dma, shc->dma_addr);
    return 0;
}

static int shc_card_present(struct sdioemb_slot *slot)
{
    struct shc_data *shc = slot->drv_data;

    return sdioemb_shc_card_present(&shc->shc);
}

static void shc_enable_card_int(struct sdioemb_slot *slot)
{
    struct shc_data *shc = slot->drv_data;

    return sdioemb_shc_enable_card_int(&shc->shc);
}

static void shc_disable_card_int(struct sdioemb_slot *slot)
{
    struct shc_data *shc = slot->drv_data;

    return sdioemb_shc_disable_card_int(&shc->shc);
}

static int shc_hard_reset(struct sdioemb_slot *slot)
{
    struct shc_data *shc = slot->drv_data;

    return sdioemb_shc_hard_reset(&shc->shc);
}

static int shc_card_power(struct sdioemb_slot *slot, enum sdioemb_power power)
{
    struct shc_data *shc = slot->drv_data;
    int ret;

    ret = sdioemb_shc_card_power(&shc->shc, power);
    if (ret < 0) {
        return ret;
    }

    /* JMicron 385 chip needs config space write to enable the power switch */
    if (shc->pdev->vendor == PCI_VENDOR_ID_JMICRON
        && shc->pdev->device == PCI_DEVICE_ID_JMICRON_JMB38X_SD) {
        uint8_t config_byte;

        pci_read_config_byte(shc->pdev, 0xAE, &config_byte);

        if (power != SDIOEMB_POWER_OFF) {
            config_byte |= 0x47;
        } else {
            config_byte &= ~(0x47);
        }
        pci_write_config_byte(shc->pdev, 0xAE, config_byte);
    }
    return 0;
}

static irqreturn_t shc_int_handler(int irq, void *dev_id)
{
    struct sdioemb_slot *slot = dev_id;
    struct shc_data *shc = slot->drv_data;
    uint32_t int_stat;
    bool pending;
    os_int_status_t istate;

    os_spinlock_lock_intsave(&shc->shc.lock, &istate);
    pending = sdioemb_shc_isr(&shc->shc, &int_stat);
    os_spinlock_unlock_intrestore(&shc->shc.lock, &istate);
    if (!pending) {
        return IRQ_NONE;
    }
    sdioemb_shc_dsr(&shc->shc, int_stat);
    return IRQ_HANDLED;
}

static struct sdioemb_slot *shc_pci_slot_init(struct pci_dev *pci, int slot_num, unsigned long quirks)
{
    struct sdioemb_slot *slot;
    struct shc_data *shc;
    int ret;

    slot = sdioemb_slot_alloc(sizeof(struct shc_data));
    if (!slot) {
        goto err_alloc;
    }

    sprintf(slot->name, "%s-%d", dev_name(&pci->dev), slot_num);
    slot->set_bus_freq     = shc_set_bus_freq;
    slot->set_bus_width    = shc_set_bus_width;
    slot->start_cmd        = shc_start_cmd;
    slot->card_present     = shc_card_present;
    slot->card_power       = shc_card_power;
    slot->enable_card_int  = shc_enable_card_int;
    slot->disable_card_int = shc_disable_card_int;
    slot->hard_reset       = shc_hard_reset;

    shc = slot->drv_data;
    shc->pdev = pci;

    sdioemb_shc_init(&shc->shc);
    shc->shc.slot         = slot;
    shc->shc.enable_int   = shc_enable_int;
    shc->shc.disable_int  = shc_disable_int;
    shc->shc.cmd_complete = shc_cmd_complete;
    shc->shc.quirks       = quirks;

    shc->shc.addr = pci_iomap(pci, slot_num, pci_resource_len(pci, slot_num));
    if (!shc->shc.addr) {
        goto err_iomap;
    }

    ret = sdioemb_shc_start(&shc->shc);
    if (ret < 0) {
        goto err_start;
    }

    ret = request_irq(pci->irq, shc_int_handler, IRQF_SHARED, "slot_shc", slot);
    if (ret) {
        goto err_irq;
    }

    ret = sdioemb_linux_slot_register(slot);
    if (ret) {
        goto err_register;
    }

    return slot;

  err_register:
    free_irq(pci->irq, slot);
  err_irq:
    sdioemb_shc_stop(&shc->shc);
  err_start:
    pci_iounmap(pci, shc->shc.addr);
  err_iomap:
    sdioemb_shc_clean_up(&shc->shc);
    sdioemb_slot_free(slot);
  err_alloc:
    return NULL;
}

static void shc_pci_slot_cleanup(struct sdioemb_slot *slot)
{
    struct shc_data *shc = slot->drv_data;
    struct pci_dev *pci = shc->pdev;

    sdioemb_slot_unregister(slot);
    free_irq(pci->irq, slot);
    sdioemb_shc_stop(&shc->shc);
    pci_iounmap(pci, shc->shc.addr);
    sdioemb_shc_clean_up(&shc->shc);
    sdioemb_slot_free(slot);
}

static int shc_pci_probe(struct pci_dev *pci, const struct pci_device_id *id)
{
    int ret;
    uint8_t num_slots, s;
    struct sdioemb_slot **slots;
    unsigned long quirks;

    ret = pci_enable_device(pci);
    if (ret)
        goto err_enable;
    
    ret = pci_set_dma_mask(pci, DMA_32BIT_MASK);
    if (ret) {
        goto err_set_dma_mask;
    }
    pci_set_master(pci);

    if (pci->vendor == PCI_VENDOR_ID_JMICRON && pci->device == PCI_DEVICE_ID_JMICRON_JMB38X_SD) {
        /* Write to various config registers */
        /* The device will only allow a 12.5MHz SDIO clock unless bit 0 at 0xAD is clear */
        pci_read_config_byte(pci, 0xAD, &s);
        s &= ~(0x01);
        pci_write_config_byte(pci, 0xAD, s);

        /* The device does not meet high speed timing unless the delay register is set to 0x03
         * This should be the default according to JMicron, but it is 0x00 on this system
         */
        pci_read_config_byte(pci, 0xB1, &s);
        s &= 0xf0;
        s |= 0x03;
        pci_write_config_byte(pci, 0xB1, s);
    }

    pci_read_config_byte(pci, PCI_SHC_SLOT_INFO, &num_slots);
    num_slots = ((num_slots >> 4) & 0x07) + 1;

    slots = kzalloc(sizeof(struct sdio_driver_slot *) * (num_slots + 1), GFP_KERNEL);
    if (!slots) {
        ret = -ENOMEM;
        goto err_no_slots_mem;
    }
    
    quirks = id->driver_data;
    
    pci_set_drvdata(pci, slots);

    for (s = 0; s < num_slots; s++) {
        struct sdioemb_slot *slot;

        slot = shc_pci_slot_init(pci, s, quirks);
        if (!slot) {
            goto err_init;
        }
        slots[s] = slot;
    }

    return 0;

  err_init:
    for (s = 0; s < num_slots; s++) {
        if (slots[s]) {
            shc_pci_slot_cleanup(slots[s]);
        }
    }
    kfree(slots);
    pci_set_drvdata(pci, NULL);
  err_set_dma_mask:
  err_no_slots_mem:
    pci_disable_device(pci);
  err_enable:

    return ret;
}

static void shc_pci_remove(struct pci_dev *pci)
{
    struct sdioemb_slot **slots = pci_get_drvdata(pci);
    struct sdioemb_slot **s;

    if (slots) {
        s = slots;
        while (*s) {
            shc_pci_slot_cleanup(*s);
            s++;
        }

        kfree(slots);
    }
}

static int shc_pci_suspend(struct pci_dev *pci, pm_message_t state)
{
    struct sdioemb_slot **slots = pci_get_drvdata(pci);
    struct sdioemb_slot **s;

    if (slots) {
        s = slots;
        while (*s) {
            /* Pass the event to the SDIO driver. */
            sdioemb_suspend(*s);
            s++;
        }
    }

    return 0;
}

static int shc_pci_resume(struct pci_dev *pci)
{
    int ret;
    struct sdioemb_slot **slots = pci_get_drvdata(pci);
    struct sdioemb_slot **s;

    ret = 0;
    if (slots) {
        s = slots;
        while (*s) {
            struct shc_data *shc = (*s)->drv_data;
            /* Re-initialise h/w. */
            ret = sdioemb_shc_start(&shc->shc);
            if (ret != 0) {
                break;
            }
            /* Pass the event to the SDIO driver. */
            sdioemb_resume(*s);
            s++;
        }
    }

    return ret;
}


#ifndef PCI_CLASS_SYSTEM_SDHCI
#define PCI_CLASS_SYSTEM_SDHCI 0x0805
#endif

static struct pci_device_id shc_pci_id_table[] = {
    {
        .vendor      = PCI_VENDOR_ID_RICOH,
        .device      = PCI_DEVICE_ID_RICOH_R5C822,
        .subvendor   = PCI_ANY_ID,
        .subdevice   = PCI_ANY_ID,
        .driver_data = SHC_QUIRK_DMA_NEEDS_DWORD_ALIGNED_ADDR
                       | SHC_QUIRK_DMA_NEEDS_DWORD_ALIGNED_LEN
                       | SLOT_SHC_QUIRK_R5B_BROKEN,
    },
    /* The Arasan SHC card does not meet the rise time requirement for
     * high speed mode so high speed mode is very unreliable. */
    {
        PCI_DEVICE(0x1095, 0x0670),
        .driver_data = SHC_QUIRK_DMA_NEEDS_DWORD_ALIGNED_ADDR
                       | SLOT_SHC_QUIRK_HIGH_SPD_BROKEN
                       | SLOT_SHC_QUIRK_DATA_TIMEOUT_USES_SDCLK,
    },
    {
        .vendor      = PCI_VENDOR_ID_JMICRON,
        .device      = PCI_DEVICE_ID_JMICRON_JMB38X_SD,
        .subvendor   = PCI_ANY_ID,
        .subdevice   = PCI_ANY_ID,
        .driver_data = SHC_QUIRK_DMA_NEEDS_DWORD_ALIGNED_ADDR
                       | SHC_QUIRK_DMA_NEEDS_DWORD_ALIGNED_LEN
                       | SLOT_SHC_QUIRK_R5B_BROKEN
                       | SHC_QUIRK_RESET_EVERY_CMD_COMPLETE,
    },
    /* Improved JMicron controller, JMB387 */
    {
        .vendor      = PCI_VENDOR_ID_JMICRON,
        .device      = 0x2386,
        .subvendor   = PCI_ANY_ID,
        .subdevice   = PCI_ANY_ID,
        .driver_data = SLOT_SHC_QUIRK_R5B_BROKEN | SHC_QUIRK_RESET_EVERY_CMD_COMPLETE,
    },
    { PCI_DEVICE_CLASS((PCI_CLASS_SYSTEM_SDHCI << 8), 0xFFFF00) },
    { 0 },
};
MODULE_DEVICE_TABLE(pci, shc_pci_id_table);

static struct pci_driver shc_pci_driver = {
    .name     = "slot_shc",
    .id_table = shc_pci_id_table,
    .probe    = shc_pci_probe,
    .remove   = __devexit_p(shc_pci_remove),
    .suspend  = shc_pci_suspend,
    .resume   = shc_pci_resume,
};

static int __init shc_pci_init(void)
{
    return pci_register_driver(&shc_pci_driver);
}

static void __exit shc_pci_exit(void)
{
    pci_unregister_driver(&shc_pci_driver);
}

module_init(shc_pci_init);
module_exit(shc_pci_exit);

MODULE_DESCRIPTION("PCI SHC slot driver");
MODULE_AUTHOR("Cambridge Silicon Radio Ltd.");
MODULE_LICENSE("GPL and additional rights");
