/*
 * Linux USB SD Host Controller (USHC) slot driver.
 *
 * Copyright (C) 2008 Cambridge Silicon Radio Ltd.
 *
 * Refer to LICENSE.txt included with this source code for details on
 * the license terms.
 *
 * References:
 *   [USHC] USB SD Host Controller specification (CS-118793-SP)
 */
#include <linux/module.h>
#include <linux/usb.h>
#include <linux/kernel.h>
#include <linux/usb.h>
#include <linux/slab.h>

#include <linux/kernel-compat.h>

#include <sdioemb/sdio.h>
#include <sdioemb/slot_api.h>
#include <sdioemb/linux.h>

enum ushc_request {
    USHC_GET_CAPS  = 0x00,
    USHC_HOST_CTRL = 0x01,
    USHC_PWR_CTRL  = 0x02,
    USHC_CLK_FREQ  = 0x03,
    USHC_EXEC_CMD  = 0x04,
    USHC_READ_RESP = 0x05,
    USHC_RESET     = 0x06,
};

enum ushc_request_type {
    USHC_GET_CAPS_TYPE  = USB_DIR_IN  | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
    USHC_HOST_CTRL_TYPE = USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
    USHC_PWR_CTRL_TYPE  = USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
    USHC_CLK_FREQ_TYPE  = USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
    USHC_EXEC_CMD_TYPE  = USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
    USHC_READ_RESP_TYPE = USB_DIR_IN  | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
    USHC_RESET_TYPE     = USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
};

#define USHC_GET_CAPS_VERSION_MASK 0xff
#define USHC_GET_CAPS_3V3      (1 << 8)
#define USHC_GET_CAPS_3V0      (1 << 9)
#define USHC_GET_CAPS_1V8      (1 << 10)
#define USHC_GET_CAPS_HIGH_SPD (1 << 16)

#define USHC_HOST_CTRL_4BIT     (1 << 1)
#define USHC_HOST_CTRL_HIGH_SPD (1 << 0)

#define USHC_PWR_CTRL_OFF 0x00
#define USHC_PWR_CTRL_3V3 0x01
#define USHC_PWR_CTRL_3V0 0x02
#define USHC_PWR_CTRL_1V8 0x03

#define USHC_READ_RESP_BUSY        (1 << 4)
#define USHC_READ_RESP_ERR_TIMEOUT (1 << 3)
#define USHC_READ_RESP_ERR_CRC     (1 << 2)
#define USHC_READ_RESP_ERR_DAT     (1 << 1)
#define USHC_READ_RESP_ERR_CMD     (1 << 0)
#define USHC_READ_RESP_ERR_MASK    0x0f

struct ushc_cbw {
    __u8 signature;
    __u8 cmd_idx;
    __le16 block_size;
    __le32 arg;
} __attribute__((packed));

#define USHC_CBW_SIGNATURE 'C'

struct ushc_csw {
    __u8 signature;
    __u8 status;
    __le32 response;
} __attribute__((packed));

#define USHC_CSW_SIGNATURE 'S'

struct ushc_int_data {
    u8 status;
    u8 reserved[3];
};

#define USHC_INT_STATUS_SDIO_INT     (1 << 1)
#define USHC_INT_STATUS_CARD_PRESENT (1 << 0)


struct ushc_data {
    struct usb_device *usb_dev;
    struct sdioemb_slot *slot;

    struct urb *int_urb;
    struct ushc_int_data *int_data;

    struct urb *cbw_urb;
    struct ushc_cbw *cbw;

    struct urb *data_urb;

    struct urb *csw_urb;
    struct ushc_csw *csw;

    spinlock_t lock;
    struct sdioemb_cmd *current_cmd;
    u32 caps;
    u16 host_ctrl;
    unsigned long flags;

    bool set_bus_freq_pending;
    int pending_bus_freq;
    struct sdioemb_cmd *pending_cmd;
    struct work_struct set_bus_freq_work;
};

#define DISCONNECTED    0
#define INT_EN          1
#define IGNORE_NEXT_INT 2

static void data_callback(struct urb *urb);

static int ushc_hw_reset(struct ushc_data *ushc)
{
    return usb_control_msg(ushc->usb_dev, usb_sndctrlpipe(ushc->usb_dev, 0),
                           USHC_RESET, USHC_RESET_TYPE,
                           0, 0, NULL, 0, 100);
}

static int ushc_hw_get_caps(struct ushc_data *ushc)
{
    int ret;
    int version;

    ret = usb_control_msg(ushc->usb_dev, usb_rcvctrlpipe(ushc->usb_dev, 0),
                          USHC_GET_CAPS, USHC_GET_CAPS_TYPE,
                          0, 0, &ushc->caps, sizeof(ushc->caps), 100);
    if (ret < 0) {
        return ret;
    }

    ushc->caps = le32_to_cpu(ushc->caps);

    version = ushc->caps & USHC_GET_CAPS_VERSION_MASK;
    if (version != 0x02) {
        dev_err(&ushc->usb_dev->dev, "controller version %d is not supported\n", version);
        return -EINVAL;
    }

    return 0;
}
    

static int ushc_hw_set_host_ctrl(struct ushc_data *ushc, u16 mask, u16 val)
{
    u16 host_ctrl;
    int ret;

    host_ctrl = (ushc->host_ctrl & ~mask) | val;
    ret = usb_control_msg(ushc->usb_dev, usb_sndctrlpipe(ushc->usb_dev, 0),
                          USHC_HOST_CTRL, USHC_HOST_CTRL_TYPE,
                          host_ctrl, 0, NULL, 0, 100);
    if (ret >= 0) {
        ushc->host_ctrl = host_ctrl;
    }
    return ret;
}

static void int_callback(struct urb *urb)
{
    struct sdioemb_slot *slot = urb->context;
    struct ushc_data *ushc = slot->drv_data;

    /*
     * Ignore the card interrupt status on interrupt transfers that
     * were submitted while card interrupts where disabled.
     *
     * This avoid occasional spurious interrupts when enabling
     * interrupts immediately after clearing the source on the card.
     */

    if (!test_and_clear_bit(IGNORE_NEXT_INT, &ushc->flags)
        && test_bit(INT_EN, &ushc->flags)
        && ushc->int_data->status & USHC_INT_STATUS_SDIO_INT) {
        sdioemb_interrupt(slot);
    }

    if (ushc->int_urb->status == 0) {
        if (!test_bit(INT_EN, &ushc->flags)) {
            set_bit(IGNORE_NEXT_INT, &ushc->flags);
        }
        usb_submit_urb(ushc->int_urb, GFP_ATOMIC);
    }
}

static void cbw_callback(struct urb *urb)
{
    struct sdioemb_slot *slot = urb->context;
    struct ushc_data *ushc = slot->drv_data;

    if (urb->status != 0) {
        usb_unlink_urb(ushc->data_urb);
        usb_unlink_urb(ushc->csw_urb);
    }
}

static void data_callback(struct urb *urb)
{
    struct sdioemb_slot *slot = urb->context;
    struct ushc_data *ushc = slot->drv_data;

    if (urb->status != 0) {
        usb_unlink_urb(ushc->csw_urb);
    }
}

static void csw_callback(struct urb *urb)
{
    struct sdioemb_slot *slot = urb->context;
    struct ushc_data *ushc = slot->drv_data;
    struct sdioemb_cmd *cmd = ushc->current_cmd;
    int status;

    status = ushc->csw->status;

    if (urb->status != 0) {
        cmd->status = SDIOEMB_CMD_ERR_CMD_OTHER;
    } else if (status & USHC_READ_RESP_ERR_MASK) {
        if (status & USHC_READ_RESP_ERR_DAT) {
            cmd->status = SDIOEMB_CMD_ERR_DAT;
        } else {
            cmd->status = SDIOEMB_CMD_ERR_CMD;
        }
        if (status & USHC_READ_RESP_ERR_CRC) {
            cmd->status |= SDIOEMB_CMD_ERR_CRC;
        }
        if (status & USHC_READ_RESP_ERR_TIMEOUT) {
            cmd->status |= SDIOEMB_CMD_ERR_TIMEOUT;
        }
    }

    if (cmd->status == SDIOEMB_CMD_IN_PROGRESS) {
        cmd->status = SDIOEMB_CMD_OK;
    }

    cmd->sdio.response.r1 = le32_to_cpu(ushc->csw->response);

    sdioemb_cmd_complete(slot, cmd);
}

static int ushc_start_cmd(struct sdioemb_slot *slot, struct sdioemb_cmd *cmd)
{
    struct ushc_data *ushc = slot->drv_data;
    int ret;
    unsigned long flags;

    spin_lock_irqsave(&ushc->lock, flags);

    if (test_bit(DISCONNECTED, &ushc->flags)) {
        ret = -ENODEV;
        goto out;
    }

    if (ushc->set_bus_freq_pending) {
        ushc->pending_cmd = cmd;
        ret = 0;
        goto out;
    }

    /*
     * The Astoria's data FIFOs don't work with clock speeds < 5MHz so
     * limit commands with data to 6MHz or more.
     */
    if (cmd->data && slot->clock_freq < 6000000) {
        ret = -EINVAL;
        goto out;
    }

    ushc->current_cmd = cmd;

    /* start cmd with CBW */
    ushc->cbw->cmd_idx = cmd->sdio.cmd;
    if (cmd->data) {
        ushc->cbw->block_size = cpu_to_le16(cmd->owner->blocksize);
    } else {
        ushc->cbw->block_size = 0;
    }
    ushc->cbw->arg = cpu_to_le32(cmd->sdio.arg);

    ret = usb_submit_urb(ushc->cbw_urb, GFP_ATOMIC);
    if (ret < 0) {
        goto out;
    }

    /* submit data (if any) */
    if (cmd->data) {
        int pipe;

        if (cmd->flags & SDIOEMB_CMD_FLAG_READ)
            pipe = usb_rcvbulkpipe(ushc->usb_dev, 6);
        else
            pipe = usb_sndbulkpipe(ushc->usb_dev, 2);
            
        usb_fill_bulk_urb(ushc->data_urb, ushc->usb_dev, pipe,
                          cmd->data, cmd->len, data_callback, slot);
        ret = usb_submit_urb(ushc->data_urb, GFP_ATOMIC);
        if (ret < 0) {
            usb_unlink_urb(ushc->cbw_urb);
            goto out;
        }
    }

    /* submit CSW */
    ret = usb_submit_urb(ushc->csw_urb, GFP_ATOMIC);
    if (ret < 0) {
        usb_unlink_urb(ushc->cbw_urb);
        usb_unlink_urb(ushc->data_urb);
        goto out;
    }

out:
    spin_unlock_irqrestore(&ushc->lock, flags);
    return ret;
}

static void ushc_set_bus_freq_work(struct work_struct *work)
{
    struct ushc_data *ushc = container_of(work, struct ushc_data, set_bus_freq_work);
    struct sdioemb_slot *slot = ushc->slot;
    int clk = ushc->pending_bus_freq;
    struct sdioemb_cmd *cmd;
    int ret;

    ret = ushc_hw_set_host_ctrl(ushc, USHC_HOST_CTRL_HIGH_SPD,
                                clk > SDIO_CLOCK_FREQ_NORMAL_SPD ? USHC_HOST_CTRL_HIGH_SPD : 0);
    if (ret >= 0) {
        ret = usb_control_msg(ushc->usb_dev, usb_sndctrlpipe(ushc->usb_dev, 0),
                              USHC_CLK_FREQ, USHC_CLK_FREQ_TYPE,
                              clk & 0xffff, (clk >> 16) & 0xffff, NULL, 0, 100);
    }

    spin_lock_irq(&ushc->lock);
    ushc->set_bus_freq_pending = false;
    cmd = ushc->pending_cmd;
    ushc->pending_cmd = NULL;
    spin_unlock_irq(&ushc->lock);

    if (cmd) {
        if (ret >= 0) {
            ret = ushc_start_cmd(slot, cmd);
        }
        if (ret < 0) {
            cmd->status = SDIOEMB_CMD_ERR_CMD_OTHER;
            sdioemb_cmd_complete(slot, cmd);
        }
    }
}

static int ushc_set_bus_freq(struct sdioemb_slot *slot, int clk)
{
    struct ushc_data *ushc = slot->drv_data;
    unsigned long flags;

    /* Hardware can't detect interrupts while the clock is off. */
    if (clk == SDIOEMB_BUS_FREQ_IDLE) {
        clk = 400000;
    }

    spin_lock_irqsave(&ushc->lock, flags);
    ushc->set_bus_freq_pending = true;
    ushc->pending_bus_freq = clk;
    spin_unlock_irqrestore(&ushc->lock, flags);

    schedule_work(&ushc->set_bus_freq_work);

    return clk;
}

static int ushc_set_bus_width(struct sdioemb_slot *slot, int bus_width)
{
    struct ushc_data *ushc = slot->drv_data;

    return ushc_hw_set_host_ctrl(ushc, USHC_HOST_CTRL_4BIT,
                                 bus_width == 4 ? USHC_HOST_CTRL_4BIT : 0);
}

static int ushc_card_power(struct sdioemb_slot *slot, enum sdioemb_power power)
{
    struct ushc_data *ushc = slot->drv_data;
    u16 voltage;

    switch (power) {
    case SDIOEMB_POWER_OFF:
        voltage = USHC_PWR_CTRL_OFF;
        break;
    case SDIOEMB_POWER_3V3:
        if (!(ushc->caps & USHC_GET_CAPS_3V3)) {
            return -ENOTSUPP;
        }
        voltage = USHC_PWR_CTRL_3V3;
        break;
    default:
        return -EINVAL;
    }

    return usb_control_msg(ushc->usb_dev, usb_sndctrlpipe(ushc->usb_dev, 0),
                           USHC_PWR_CTRL, USHC_PWR_CTRL_TYPE,
                           voltage, 0, NULL, 0, 100);
}

static int ushc_card_present(struct sdioemb_slot *slot)
{
    struct ushc_data *ushc = slot->drv_data;

    return ushc->int_data->status & USHC_INT_STATUS_CARD_PRESENT;
}

static void ushc_enable_card_int(struct sdioemb_slot *slot)
{
    struct ushc_data *ushc = slot->drv_data;

    set_bit(INT_EN, &ushc->flags);
}

static void ushc_disable_card_int(struct sdioemb_slot *slot)
{
    struct ushc_data *ushc = slot->drv_data;

    clear_bit(INT_EN, &ushc->flags);
}

static void ushc_clean_up(struct sdioemb_slot *slot)
{
    struct ushc_data *ushc = slot->drv_data;

    usb_free_urb(ushc->int_urb);
    usb_free_urb(ushc->csw_urb);
    usb_free_urb(ushc->data_urb);
    usb_free_urb(ushc->cbw_urb);

    kfree(ushc->int_data);
    kfree(ushc->cbw);
    kfree(ushc->csw);
}

static int ushc_probe(struct usb_interface *intf, const struct usb_device_id *id)
{
    struct usb_device *usb_dev = interface_to_usbdev(intf);
    struct sdioemb_slot *slot;
    struct ushc_data *ushc;
    int ret = -ENOMEM;

    slot = sdioemb_slot_alloc(sizeof(struct ushc_data));
    if (slot == NULL) {
        return -ENOMEM;
    }
    ushc = slot->drv_data;
    usb_set_intfdata(intf, slot);

    ushc->usb_dev = usb_dev;
    ushc->slot = slot;

    strlcpy(slot->name, dev_name(&usb_dev->dev), sizeof(slot->name));
    slot->set_bus_freq     = ushc_set_bus_freq;
    slot->set_bus_width    = ushc_set_bus_width;
    slot->start_cmd        = ushc_start_cmd;
    slot->card_present     = ushc_card_present;
    slot->card_power       = ushc_card_power;
    slot->enable_card_int  = ushc_enable_card_int;
    slot->disable_card_int = ushc_disable_card_int;

    spin_lock_init(&ushc->lock);

    ret = ushc_hw_reset(ushc);
    if (ret < 0) {
        goto err;
    }

    /* Read capabilities. */
    ret = ushc_hw_get_caps(ushc);
    if (ret < 0) {
        goto err;
    }
    slot->caps.max_bus_freq = (ushc->caps & USHC_GET_CAPS_HIGH_SPD)
        ? SDIO_CLOCK_FREQ_HIGH_SPD : SDIO_CLOCK_FREQ_NORMAL_SPD;
    slot->caps.max_bus_width = 4;

    ushc->int_urb = usb_alloc_urb(0, GFP_KERNEL);
    if (ushc->int_urb == NULL) {
        goto err;
    }
    ushc->int_data = kzalloc(sizeof(struct ushc_int_data), GFP_KERNEL);
    if (ushc->int_data == NULL) {
        goto err;
    }
    usb_fill_int_urb(ushc->int_urb, ushc->usb_dev,
                     usb_rcvintpipe(usb_dev,
                                    intf->cur_altsetting->endpoint[0].desc.bEndpointAddress),
                     ushc->int_data, sizeof(struct ushc_int_data),
                     int_callback, slot,
                     intf->cur_altsetting->endpoint[0].desc.bInterval);

    ushc->cbw_urb = usb_alloc_urb(0, GFP_KERNEL);
    if (ushc->cbw_urb == NULL) {
        goto err;
    }
    ushc->cbw = kzalloc(sizeof(struct ushc_cbw), GFP_KERNEL);
    if (ushc->cbw == NULL) {
        goto err;
    }
    ushc->cbw->signature = USHC_CBW_SIGNATURE;

    usb_fill_bulk_urb(ushc->cbw_urb, ushc->usb_dev, usb_sndbulkpipe(usb_dev, 2),
                      ushc->cbw, sizeof(struct ushc_cbw),
                      cbw_callback, slot);

    ushc->data_urb = usb_alloc_urb(0, GFP_KERNEL);
    if (ushc->data_urb == NULL) {
        goto err;
    }

    ushc->csw_urb = usb_alloc_urb(0, GFP_KERNEL);
    if (ushc->csw_urb == NULL) {
        goto err;
    }
    ushc->csw = kzalloc(sizeof(struct ushc_cbw), GFP_KERNEL);
    if (ushc->csw == NULL) {
        goto err;
    }
    usb_fill_bulk_urb(ushc->csw_urb, ushc->usb_dev, usb_rcvbulkpipe(usb_dev, 6),
                      ushc->csw, sizeof(struct ushc_csw),
                      csw_callback, slot);

    INIT_WORK(&ushc->set_bus_freq_work, ushc_set_bus_freq_work);

    ret = sdioemb_linux_slot_register(slot);
    if (ret) {
        goto err;
    }

    ret = usb_submit_urb(ushc->int_urb, GFP_KERNEL);
    if (ret < 0) {
        sdioemb_slot_unregister(slot);
        goto err;
    }

    return 0;

  err:
    ushc_clean_up(slot);
    sdioemb_slot_free(slot);
    return ret;
}

static void ushc_disconnect(struct usb_interface *intf)
{
    struct sdioemb_slot *slot = usb_get_intfdata(intf);
    struct ushc_data *ushc = slot->drv_data;

    spin_lock_irq(&ushc->lock);
    set_bit(DISCONNECTED, &ushc->flags);
    spin_unlock_irq(&ushc->lock);

    usb_kill_urb(ushc->int_urb);
    usb_kill_urb(ushc->cbw_urb);
    usb_kill_urb(ushc->data_urb);
    usb_kill_urb(ushc->csw_urb);

    sdioemb_slot_unregister(slot);

    ushc_clean_up(slot);
}

static struct usb_device_id ushc_id_table[] = {
    /* CSR USB SD Host Controller */
    { USB_DEVICE(0x0a12, 0x5d10) }, /* FIXME: real device ID TBD */
    { },
};

static struct usb_driver ushc_driver = {
    .name       = "slot_ushc",
    .id_table   = ushc_id_table,
    .probe      = ushc_probe,
    .disconnect = ushc_disconnect,
};

static int __init ushc_init(void)
{
    return usb_register(&ushc_driver);
}
module_init(ushc_init);

static void __exit ushc_exit(void)
{
    usb_deregister(&ushc_driver);
}
module_exit(ushc_exit);

MODULE_DESCRIPTION("USB SD Host Controller slot driver");
MODULE_AUTHOR("Cambridge Silicon Radio Ltd.");
MODULE_LICENSE("GPL and additional rights");
