/*
 * Standard SD Host Controller library.
 *
 * Copyright (C) 2007-2009 Cambridge Silicon Radio Ltd.
 *
 * Refer to LICENSE.txt included with this source code for details on
 * the license terms.
 *
 * References:
 *   [SDHC] SD Host Controller Standard Specification v1.00.
 */
#include <oska/all.h>
#include <oska/io.h>

#include <sdioemb/sdio.h>
#include <sdioemb/sdio_api.h>
#include <sdioemb/slot_api.h>
#include <sdioemb/slot_shc.h>
#include <sdioemb/linux.h>
#include <sdioemb/trace.h>

#ifdef OS_TRACE_INCLUDE_TMH
#include "lib_shc.tmh"
#endif

/* Software timeout for commands. For working around buggy controllers
   that occasionally fail to complete a command. */
#define SLOT_SHC_LOCKUP_TIMEOUT_MS 2000

static int shc_hw_reset(struct sdioemb_shc *shc, uint8_t rst_bits)
{
    int timeout = SHC_RESET_TIMEOUT_MS;

    os_io_write8(shc->addr, SHC_SOFTWARE_RST, rst_bits);
    while (--timeout) {
        if (!(os_io_read8(shc->addr, SHC_SOFTWARE_RST) & rst_bits)) {
            break;
        }
        os_delay_us(1000);
    }
    if (!timeout) {
        os_trace_err("%s: timed out waiting for reset", shc->slot->name);
        return -ETIMEDOUT;
    }

    return 0;
}

static int shc_hw_init(struct sdioemb_shc *shc)
{
    uint32_t caps;
    int ret;
    int timeout;

    ret = shc_hw_reset(shc, SHC_SOFTWARE_RST_ALL);
    if (ret) {
        return ret;
    }

    /* read and check various capabilties */
    caps = os_io_read32(shc->addr, SHC_CAPS);

    shc->base_clk = 1000000 * SHC_CAPS_TO_BASE_CLK_FREQ(caps);
    if (shc->base_clk == 0) {
        os_trace_err("%s: cannot determine base SD clock frequency", shc->slot->name);
        return -EIO;
    }

    /* mask all interrupts */
    os_io_write32(shc->addr, SHC_INT_STATUS_EN, 0);

    /* allow any unmasked interrupt to raise an interrupt */
    os_io_write32(shc->addr, SHC_INT_SIGNAL_EN, SHC_INT_ALL);

    /* enable internal clock, disable SD clock, and wait for internal
     * clock to stabilize */
    os_io_write16(shc->addr, SHC_CLOCK_CTRL, SHC_CLOCK_CTRL_INT_CLK_EN);
    timeout = SHC_INT_CLK_STABLE_TIMEOUT_MS;
    while (--timeout) {
        if (os_io_read16(shc->addr, SHC_CLOCK_CTRL) & SHC_CLOCK_CTRL_INT_CLK_STABLE) {
            break;
        }
        os_sleep_ms(1);
    }
    if (!timeout) {
        os_trace_err("%s: timed out waiting for internal clock to stabilize", shc->slot->name);
        return -ETIMEDOUT;
    }

    return 0;
}

static void shc_hw_stop(struct sdioemb_shc *shc)
{
    os_io_write16(shc->addr, SHC_CLOCK_CTRL, 0);
}

static void shc_hw_set_high_speed(struct sdioemb_shc *shc, int hs)
{
    os_int_status_t istate;
    uint8_t host_ctrl;

    os_spinlock_lock_intsave(&shc->lock, &istate);

    host_ctrl = os_io_read8(shc->addr, SHC_HOST_CTRL) & ~SHC_HOST_CTRL_HIGH_SPD_EN;
    if (hs) {
        host_ctrl |= SHC_HOST_CTRL_HIGH_SPD_EN;
    }
    os_io_write8(shc->addr, SHC_HOST_CTRL, host_ctrl);

    os_spinlock_unlock_intrestore(&shc->lock, &istate);
}

static int shc_hw_set_clock_rate(struct sdioemb_shc *shc, uint32_t clk)
{
    uint16_t clk_reg;

    /* Stop the clock before changing the frequency. */
    clk_reg = os_io_read16(shc->addr, SHC_CLOCK_CTRL);
    clk_reg &= ~SHC_CLOCK_CTRL_SD_CLK_EN;
    os_io_write16(shc->addr, SHC_CLOCK_CTRL, clk_reg);

    shc_hw_set_high_speed(shc, clk > SDIO_CLOCK_FREQ_NORMAL_SPD);

    if (clk) {
        unsigned divisor;
        uint32_t timeout_clks;
        uint8_t timeout_ctrl;

        /* Find maximum possible bus freq <= clk (or the minimum the
           controller is capable of)... */
        for (divisor = 1; divisor < 256; divisor *= 2) {
            if (shc->base_clk / divisor <= clk) {
                break;
            }
        }
        clk = shc->base_clk / divisor;

        /* ...and start it. */
        clk_reg = SHC_CLOCK_CTRL_DIV(divisor) | SHC_CLOCK_CTRL_INT_CLK_EN;
        os_io_write16(shc->addr, SHC_CLOCK_CTRL, clk_reg);
        clk_reg |= SHC_CLOCK_CTRL_SD_CLK_EN;
        os_io_write16(shc->addr, SHC_CLOCK_CTRL, clk_reg);

        /*
         * Set the data timeout to 1 s.
         */
        if (shc->quirks & SLOT_SHC_QUIRK_DATA_TIMEOUT_USES_SDCLK) {
            timeout_clks = clk;
        } else {
            timeout_clks = shc->base_clk;
        }
        for (timeout_ctrl = 0; timeout_ctrl < SHC_TIMEOUT_CTRL_MAX; timeout_ctrl++) {
            if ((1u << (timeout_ctrl + 13)) >= timeout_clks) {
                break;
            }
        }
        os_io_write8(shc->addr, SHC_TIMEOUT_CTRL, timeout_ctrl);
    }

    return clk;
}

static void shc_hw_set_bus_width(struct sdioemb_shc *shc, int bus_width)
{
    os_int_status_t istate;
    uint8_t host_ctrl;

    os_spinlock_lock_intsave(&shc->lock, &istate);

    host_ctrl = os_io_read8(shc->addr, SHC_HOST_CTRL);
    if (bus_width == 4) {
        host_ctrl |= SHC_HOST_CTRL_4BIT;
    } else {
        host_ctrl &= ~SHC_HOST_CTRL_4BIT;
    }
    os_io_write8(shc->addr, SHC_HOST_CTRL, host_ctrl);

    os_spinlock_unlock_intrestore(&shc->lock, &istate);
}

/*
 * Interrupts are masked/unmasked using INT_STATUS_EN (not
 * INT_SIGNAL_EN) as this is what the card interrupt requires.
 */
static void shc_hw_enable_int_signal(struct sdioemb_shc *shc, uint32_t sigs)
{
    shc->enable_int(shc->slot, sigs);
}

static void shc_hw_disable_int_signal(struct sdioemb_shc *shc, uint32_t sigs)
{
    shc->disable_int(shc->slot, sigs);
}

static void shc_hw_read_block_pio(struct sdioemb_shc *shc)
{
    size_t len;
    uint32_t dword;
    int bytes = 0;
    size_t i;

    len = min(shc->block_size, shc->remaining);

    for (i = 0; i < len; i++) {
        if (bytes == 0) {
            dword = os_io_read32(shc->addr, SHC_BUFFER_DATA_PORT);
            bytes = 4;
        }

        *(shc->data++) = dword & 0xff;
        dword >>= 8;
        shc->remaining--;
        bytes--;
    }
}

static void shc_hw_write_block_pio(struct sdioemb_shc *shc)
{
    size_t len;
    uint32_t dword = 0;
    int bytes = 0;
    size_t i;

    len = min(shc->block_size, shc->remaining);

    for (i = 0; i < len; i++) {
        dword |= ((uint32_t)*(shc->data++)) << (bytes * 8);
        shc->remaining--;
        bytes++;

        if (bytes == 4 || shc->remaining == 0) {
            os_io_write32(shc->addr, SHC_BUFFER_DATA_PORT, dword);
            dword = 0;
            bytes = 0;
        }
    }
}

static void shc_hw_setup_data_pio(struct sdioemb_shc *shc)
{
    struct sdioemb_cmd *cmd = shc->current_cmd;

    shc->data = cmd->data;
    shc->remaining = cmd->len;
    shc->block_size = cmd->owner->blocksize;

    if (cmd->flags & SDIOEMB_CMD_FLAG_READ) {
        shc_hw_enable_int_signal(shc, SHC_INT_RD_BUF_RDY);
    } else {
        shc_hw_enable_int_signal(shc, SHC_INT_WR_BUF_RDY);
    }
}

static int int_stat_to_status(uint32_t int_stat)
{
    int status;

    if (int_stat & (SHC_INT_ERR_CMD_ALL | SHC_INT_ERR_DAT_ALL)) {
        if (int_stat & SHC_INT_ERR_CMD_ALL) {
            status = SDIOEMB_CMD_ERR_CMD;
        } else {
            status = SDIOEMB_CMD_ERR_DAT;
        }
        if (int_stat & (SHC_INT_ERR_CMD_TIMEOUT | SHC_INT_ERR_DAT_TIMEOUT)) {
            status |= SDIOEMB_CMD_ERR_TIMEOUT;
        } else {
            status |= SDIOEMB_CMD_ERR_CRC;
        }
    } else {
        status = SDIOEMB_CMD_OK;
    }
    return status;
}

static void shc_cmd_complete(struct sdioemb_shc *shc, int status)
{
    struct sdioemb_slot *slot = shc->slot;
    struct sdioemb_cmd *cmd = shc->current_cmd;
    os_int_status_t istate;

    os_spinlock_lock_intsave(&shc->lock, &istate);
    cmd = shc->current_cmd;
    shc->current_cmd = NULL;
    os_spinlock_unlock_intrestore(&shc->lock, &istate);

    if (cmd == NULL) {
        os_trace_err("%s: spurious command complete interrupt", slot->name);
        return;
    }

    cmd->status = status;

    /* disable transfer related interrupts */
    shc_hw_disable_int_signal(shc, SHC_INT_ERR_ALL | SHC_INT_CMD_COMPLETE
                              | SHC_INT_TRANSFER_COMPLETE
                              | SHC_INT_RD_BUF_RDY | SHC_INT_WR_BUF_RDY);

    os_timer_cancel(&shc->lockup_timer);

    /* Read response if it's valid. */
    if (!(cmd->status & SDIOEMB_CMD_ERR_CMD)) {
        switch (cmd->flags & SDIOEMB_CMD_FLAG_RESP_MASK) {
        case SDIOEMB_CMD_FLAG_RESP_NONE:
            break;
        case SDIOEMB_CMD_FLAG_RESP_R1:
        case SDIOEMB_CMD_FLAG_RESP_R1B:
        case SDIOEMB_CMD_FLAG_RESP_R4:
        case SDIOEMB_CMD_FLAG_RESP_R5:
        case SDIOEMB_CMD_FLAG_RESP_R5B:
        case SDIOEMB_CMD_FLAG_RESP_R6:
            cmd->sdio.response.r1 = os_io_read32(shc->addr, SHC_RESPONSE_0_31);
            break;
        default:
            os_trace_err("%s: response format not supported", slot->name);
        }
    }

    /*
     * After an error or abort, reset the DAT and CMD line state
     * machines.  See [SDHC] section 3.7.
     */
    if ( (shc->quirks & SHC_QUIRK_RESET_EVERY_CMD_COMPLETE)
         || cmd->status != SDIOEMB_CMD_OK
         || (cmd->flags & SDIOEMB_CMD_FLAG_ABORT)) {
        /* Some controllers do not like having both blocks reset at the same time */
        shc_hw_reset(shc, SHC_SOFTWARE_RST_CMD);
        shc_hw_reset(shc, SHC_SOFTWARE_RST_DAT);
    }

    shc->cmd_complete(slot, cmd);
}

static void shc_lockup_timer_fn(void *arg)
{
    struct sdioemb_shc *shc = arg;
    uint32_t int_stat;
    int status;

    int_stat = os_io_read32(shc->addr, SHC_INT_STATUS);
    status = int_stat_to_status(int_stat);
    if (status == SDIOEMB_CMD_OK) {
        if (int_stat & SHC_INT_CMD_COMPLETE) {
            status = SDIOEMB_CMD_ERR_DAT_OTHER;
        } else {
            status = SDIOEMB_CMD_ERR_CMD_OTHER;
        }
    }
    shc_cmd_complete(shc, status);
}

void sdioemb_shc_init(struct sdioemb_shc *shc)
{
    os_spinlock_init(&shc->lock);
    os_timer_init(&shc->lockup_timer, shc_lockup_timer_fn, shc);
}

void sdioemb_shc_clean_up(struct sdioemb_shc *shc)
{
    os_spinlock_destroy(&shc->lock);
}

int sdioemb_shc_start(struct sdioemb_shc *shc)
{
    struct sdioemb_slot *slot = shc->slot;
    int max_freq;
    int ret;

    ret = shc_hw_init(shc);
    if (ret < 0) {
        return ret;
    }

    max_freq = shc->base_clk;
    if (shc->quirks & SLOT_SHC_QUIRK_HIGH_SPD_BROKEN && max_freq > SDIO_CLOCK_FREQ_NORMAL_SPD) {
        max_freq = SDIO_CLOCK_FREQ_NORMAL_SPD;
    }

    slot->type               = SDIOEMB_SLOT_TYPE_SD;
    slot->caps.max_bus_freq  = max_freq;
    slot->caps.max_bus_width = 4;

    return 0;
}

void sdioemb_shc_stop(struct sdioemb_shc *shc)
{
    shc_hw_stop(shc);
}

bool sdioemb_shc_isr(struct sdioemb_shc *shc, uint32_t *int_stat)
{
    *int_stat = os_io_read32(shc->addr, SHC_INT_STATUS);
    if (*int_stat == 0) {
        return false;
    }
    /* Clear interrupt status. */
    os_io_write32(shc->addr, SHC_INT_STATUS, *int_stat);
    if (*int_stat & SHC_INT_CARD_INT) {
        os_io_write32(shc->addr, SHC_INT_STATUS_EN,
                      os_io_read32(shc->addr, SHC_INT_STATUS_EN) & ~SHC_INT_CARD_INT);
    }
    return true;
}

void sdioemb_shc_dsr(struct sdioemb_shc *shc, uint32_t int_stat)
{
    struct sdioemb_slot *slot = shc->slot;

    if (int_stat & SHC_INT_RD_BUF_RDY) {
        shc_hw_read_block_pio(shc);
    }
    if (int_stat & SHC_INT_WR_BUF_RDY) {
        shc_hw_write_block_pio(shc);
    }

    if (int_stat & SHC_INT_DMA) {
        /* Restart the DMA transfer at the same address. */
        os_io_write32(shc->addr, SHC_SYSTEM_ADDRESS, os_io_read32(shc->addr, SHC_SYSTEM_ADDRESS));
    }

    if (int_stat & (SHC_INT_ERR_ALL | SHC_INT_CMD_COMPLETE | SHC_INT_TRANSFER_COMPLETE)) {
        shc_cmd_complete(shc, int_stat_to_status(int_stat));
    }

    if (int_stat & SHC_INT_CARD_INT) {
        sdioemb_interrupt(slot);
    }
}

int sdioemb_shc_set_bus_freq(struct sdioemb_shc *shc, int clk)
{
    if (clk == SDIOEMB_BUS_FREQ_IDLE) {
        clk = 0;
    }
    return shc_hw_set_clock_rate(shc, clk);
}

int sdioemb_shc_set_bus_width(struct sdioemb_shc *shc, int bus_width)
{
    shc_hw_set_bus_width(shc, bus_width);
    return 0;
}

static const uint16_t resp_map[] = {
    0,                                                              /* None */
    SHC_CMD_RESP_48  | SHC_CMD_RESP_IDX_CHK | SHC_CMD_RESP_CRC_CHK, /* R1 */
    SHC_CMD_RESP_48B | SHC_CMD_RESP_IDX_CHK | SHC_CMD_RESP_CRC_CHK, /* R1B */
    SHC_CMD_RESP_136 | SHC_CMD_RESP_CRC_CHK,                        /* R2 */
    SHC_CMD_RESP_48,                                                /* R3 */
    SHC_CMD_RESP_48,                                                /* R4 */
    SHC_CMD_RESP_48  | SHC_CMD_RESP_IDX_CHK | SHC_CMD_RESP_CRC_CHK, /* R5 */
    SHC_CMD_RESP_48B | SHC_CMD_RESP_IDX_CHK | SHC_CMD_RESP_CRC_CHK, /* R5B */
    SHC_CMD_RESP_48  | SHC_CMD_RESP_IDX_CHK | SHC_CMD_RESP_CRC_CHK, /* R6 */
};

int sdioemb_shc_start_cmd(struct sdioemb_shc *shc, struct sdioemb_cmd *cmd,
                          bool use_dma, uint64_t dma_addr)
{
    struct sdioemb_slot *slot = shc->slot;
    uint32_t system_addr;
    uint16_t block_size, block_count;
    uint16_t transfer_mode, command;

    shc->current_cmd = cmd;

    /* The first command (which will always be a CMD0) must be
     * preceeded by at least 64 clocks. */
    if (cmd->sdio.cmd == 0) {
        os_delay_us(64 * 1000000 / slot->clock_freq + 1);
    }

    /* enable command complete, error, data transfer
     * etc. interrupts */
    shc_hw_enable_int_signal(shc, SHC_INT_ERR_CMD_ALL | SHC_INT_ERR_DAT_ALL | SHC_INT_DMA
                             | (cmd->data ? SHC_INT_TRANSFER_COMPLETE : SHC_INT_CMD_COMPLETE));

    /* Set up data for DMA, poke registers to start cmd */
    transfer_mode = 0;
    if (cmd->data) {
        if (!use_dma) {
            system_addr = 0;
            shc_hw_setup_data_pio(shc);
        } else {
            system_addr = (uint32_t)dma_addr;
            transfer_mode |= SHC_TRANSFER_MODE_DMA_EN;
        }
        if (cmd->len < cmd->owner->blocksize) {
            block_size = (uint32_t)cmd->len;
            block_count = 1;
        } else {
            block_size = cmd->owner->blocksize;
            block_count = cmd->len / block_size;
        }

        /*
         * All buffers are physically contiguous so setting a DMA
         * boundary shouldn't be necessary.  There is no way to do
         * this so just use the largest possible DMA boundary.
         */
        block_size |= SHC_BLOCK_SIZE_DMA_BOUNDARY_512K;
    } else {
        block_size = block_count = system_addr = 0;
    }

    if (cmd->flags & SDIOEMB_CMD_FLAG_READ) {
        transfer_mode |= SHC_TRANSFER_MODE_DATA_READ;
    }
    if (block_count > 0) {
        transfer_mode |= SHC_TRANSFER_MODE_MULTI_BLK | SHC_TRANSFER_MODE_BLK_CNT_EN;
    }

    command = SHC_CMD_IDX(cmd->sdio.cmd) | resp_map[cmd->flags & SDIOEMB_CMD_FLAG_RESP_MASK];
    if (cmd->data) {
        command |= SHC_CMD_DATA_PRESENT;
        if (shc->quirks & SLOT_SHC_QUIRK_R5B_BROKEN) {
            /* Swap R5B for R5 */
            if ((command & SHC_CMD_RESP_48B) == SHC_CMD_RESP_48B) {
                command &= ~(SHC_CMD_RESP_48B);
                command |= SHC_CMD_RESP_48;
            }
        }
    }
    if (cmd->flags & SDIOEMB_CMD_FLAG_ABORT) {
        command |= SHC_CMD_TYPE_ABORT;
    }

    if (system_addr)
        os_io_write32(shc->addr, SHC_SYSTEM_ADDRESS, system_addr);
    if (block_count > 0)
        os_io_write16(shc->addr, SHC_BLOCK_SIZE, block_size);
    if (block_count > 0)
        os_io_write16(shc->addr, SHC_BLOCK_COUNT, block_count);
    os_io_write32(shc->addr, SHC_ARG, cmd->sdio.arg);
    os_io_write16(shc->addr, SHC_TRANSFER_MODE, transfer_mode);

    os_timer_set(&shc->lockup_timer, SLOT_SHC_LOCKUP_TIMEOUT_MS);

    (void)os_io_read32(shc->addr, 0); /* flush posted writes */
    os_io_write16(shc->addr, SHC_CMD, command);

    return 0;
}

int sdioemb_shc_card_present(struct sdioemb_shc *shc)
{
    uint32_t state;

    state = os_io_read32(shc->addr, SHC_PRESENT_STATE);
    return state & SHC_PRESENT_STATE_CARD_PRESENT;
}

int sdioemb_shc_card_power(struct sdioemb_shc *shc, enum sdioemb_power power)
{
    uint32_t caps;
    uint8_t pwr_ctrl;

    caps = os_io_read32(shc->addr, SHC_CAPS);

    if( shc->quirks & SLOT_SHC_QUIRK_NO_VOLTAGE_CAPS )
    {
        /* Add in 3.3V to capabilites - the hardware supports this, but doesn't say! */
        caps |= SHC_CAPS_PWR_3V3;
    }
    
    switch (power) {
    case SDIOEMB_POWER_OFF:
        pwr_ctrl = 0;
        break;
    case SDIOEMB_POWER_3V3:
        if (caps & SHC_CAPS_PWR_3V3) {
            pwr_ctrl = SHC_PWR_CTRL_3V3;
            os_io_write8(shc->addr, SHC_PWR_CTRL, pwr_ctrl);
            pwr_ctrl |= SHC_PWR_CTRL_ON;
        } else {
            return -ENOTSUPP;
        }
        break;
    default:
        return -EINVAL;
    }

    os_io_write8(shc->addr, SHC_PWR_CTRL, pwr_ctrl);
    return 0;
}

void sdioemb_shc_enable_card_int(struct sdioemb_shc *shc)
{
    shc_hw_enable_int_signal(shc, SHC_INT_CARD_INT);
}

void sdioemb_shc_disable_card_int(struct sdioemb_shc *shc)
{
    shc_hw_disable_int_signal(shc, SHC_INT_CARD_INT);
}

int sdioemb_shc_hard_reset(struct sdioemb_shc *shc)
{
    uint8_t pwr_ctrl;

    shc_hw_set_clock_rate(shc, SDIOEMB_BUS_FREQ_OFF);

    /* Power cycle the card. */
    pwr_ctrl = os_io_read8(shc->addr, SHC_PWR_CTRL);
    os_io_write8(shc->addr, SHC_PWR_CTRL, 0);
    os_sleep_ms(100); /* short delay to allow voltage to drop */
    os_io_write8(shc->addr, SHC_PWR_CTRL, pwr_ctrl);

    return 0;
}

void sdioemb_shc_show_quirks(struct sdioemb_shc *shc)
{
    static const char *quirk_name[] = {
        "NO_VOLTAGE_CAPS", "R5B_BROKEN", "HIGH_SPD_BROKEN", "DATA_TIMEOUT_USES_SDCLK",
        "DMA_NEEDS_DWORD_ALIGNED_ADDR", "RESET_AFTER_EVERY_CMD_COMPLETE", "JMICRON_JMB38X",
        "DMA_NEEDS_DWORD_ALIGNED_LEN",
        NULL,
    };
    unsigned i;

    for (i = 0; quirk_name[i] != NULL; i++) {
        if (shc->quirks & (1 << i)) {
            os_trace_info("%s: has %s quirk", shc->slot->name, quirk_name[i]);
        }
    }
}
