/*
 * USB Standard SD Host Controller library.
 *
 * Copyright (C) 2007-2010 Cambridge Silicon Radio Ltd.
 *
 * Refer to LICENSE.txt included with this source code for details on
 * the license terms.
 *
 */
#include <oska/all.h>
#include <oska/io.h>

#include <sdioemb/sdio.h>
#include <sdioemb/sdio_api.h>
#include <sdioemb/slot_api.h>
#include <sdioemb/slot_ushc.h>
#include <sdioemb/linux.h>
#include <sdioemb/trace.h>

#ifdef OS_TRACE_INCLUDE_TMH
#include "lib_ushc.tmh"
#endif

/*
 * ushc_hw_reset
 *
 * Arguments
 *  ushc - pointer to a sdioemb_ushc structure
 *
 * Returns
 *  int
 */
static int ushc_hw_reset(struct sdioemb_ushc* ushc)
{
    return ushc->submit_vendor_request(ushc->slot, USHC_RESET, USHC_HOST_TO_DEVICE, USHC_RECIPIENT_DEVICE, 0, 0, NULL, 0);
}

/*
 * ushc_hw_init
 *
 * Arguments
 *  ushc - pointer to a sdioemb_ushc structure
 *
 * Returns
 *  int
 */
static int ushc_hw_init(struct sdioemb_ushc* ushc)
{
    return ushc_hw_reset(ushc);
}

/*
 * ushc_hw_set_bus_width
 *
 * Arguments
 *  ushc - pointer to a sdioemb_ushc structure
 *
 * Returns
 *  int
 */
static int ushc_hw_set_bus_width(struct sdioemb_ushc* ushc, int bus_width)
{
    return ushc_hw_set_host_ctrl(ushc, USHC_HOST_CTRL_4BIT, bus_width == 4 ? USHC_HOST_CTRL_4BIT : 0);
}

/*
 * ushc_cmd_complete
 *
 * Arguments
 *  ushc - pointer to a sdioemb_ushc structure
 *
 * Returns
 *  void
 */
static void ushc_cmd_complete(struct sdioemb_ushc* ushc, int status)
{
    struct sdioemb_slot *slot = ushc->slot;
    struct sdioemb_cmd *cmd = ushc->current_cmd;
    os_int_status_t istate;

    os_spinlock_lock_intsave(&ushc->lock, &istate);
    cmd = ushc->current_cmd;
    ushc->current_cmd = NULL;
    os_spinlock_unlock_intrestore(&ushc->lock, &istate);

    ushc->cmd_complete(slot, cmd);
}

/*
 * sdioemb_ushc_init
 *
 * Arguments
 *  ushc - pointer to a sdioemb_ushc structure
 *
 * Returns
 *  void
 */
void sdioemb_ushc_init(struct sdioemb_ushc* ushc)
{
    os_spinlock_init(&ushc->lock);
}

/*
 * sdioemb_ushc_clean_up
 *
 * Arguments
 *  ushc - pointer to a sdioemb_ushc structure
 *
 * Returns
 *  void
 */
void sdioemb_ushc_clean_up(struct sdioemb_ushc* ushc)
{
    os_spinlock_destroy(&ushc->lock);
}

/*
 * sdioemb_ushc_start
 *
 * Arguments
 *  ushc - pointer to a sdioemb_ushc structure
 *
 * Returns
 *  int
 */
int sdioemb_ushc_start(struct sdioemb_ushc* ushc)
{
    struct sdioemb_slot *slot = ushc->slot;
    int max_freq = 0;
    int status;

    status = ushc_hw_init(ushc);
    if(status >= 0) 
    {
        status = ushc_hw_get_caps(ushc);
        if(status >= 0) 
        {
            slot->caps.max_bus_width = 4;
            slot->type = SDIOEMB_SLOT_TYPE_SD;
            slot->caps.max_bus_freq = (ushc->controler_capabilities & USHC_GET_CAPS_HIGH_SPD) ? 
                                       SDIO_CLOCK_FREQ_HIGH_SPD : 
                                       SDIO_CLOCK_FREQ_NORMAL_SPD;
        }
    }

    return status;
}

/*
 * sdioemb_ushc_stop
 *
 * Arguments
 *  ushc - pointer to a sdioemb_ushc structure
 *
 * Returns
 *  void
 */
void sdioemb_ushc_stop(struct sdioemb_ushc* ushc)
{
}

/*
 * sdioemb_ushc_set_bus_freq
 *
 * Arguments
 *  ushc - pointer to a sdioemb_ushc structure
 *
 * Returns
 *  int
 */
int sdioemb_ushc_set_bus_freq(struct sdioemb_ushc* ushc, int clock)
{
    int status;

    status = ushc_hw_set_host_ctrl(ushc, 
                                   USHC_HOST_CTRL_HIGH_SPD, 
                                   clock > SDIO_CLOCK_FREQ_NORMAL_SPD ?  USHC_HOST_CTRL_HIGH_SPD : 0);
    if(status >= 0)
    {
        status = ushc_hw_submit_vendor_request(ushc, 
                                               USHC_CLK_FREQ,
                                               USHC_HOST_TO_DEVICE,
                                               USHC_RECIPIENT_DEVICE,
                                               clock & 0xffff,
                                               (clock >> 16) & 0xffff,
                                               NULL,
                                               0);

        if(status >= 0)
        {
            status = clock;
        }
    }

    return status;
}

/*
 * ushc_hw_set_host_ctrl
 *
 * Arguments
 *  ushc - pointer to a sdioemb_ushc structure
 *
 * Returns
 *  int
 */
static int ushc_hw_set_host_ctrl(struct sdioemb_ushc* ushc, uint16_t mask, uint16_t val)
{
    uint16_t controler_state;
    int status;

    controler_state = (ushc->controler_state & ~mask) | val;

    status = ushc->set_host_ctrl(ushc->slot, controler_state);

    if(status >= 0) 
    {
        ushc->controler_state = controler_state;
    }
    return status;
}


/*
 * ushc_hw_submit_vendor_request
 *
 * Arguments 
 *  ushc             - pointer to a sdioemb_ushc structure
 *  request          - Vendor-defined request code
 *  recipient        - request recipeint one of: USHC_RECIPIENT_DEVICE
 *                                               USHC_RECIPIENT_INTERFACE
 *                                               USHC_RECIPIENT_ENDPOINT
 *                                               USHC_RECIPIENT_OTHER
 *  direction        - request direction one of: USHC_HOST_TO_DEVICE
 *                                               USHC_DEVICE_TO_HOST
 *  value            - Value, specific to Request (can be 0)
 *  index            - Index, specific to Request (can be 0)
 *  io_buffer        - Pointer to a buffer for the request (can be NULL)
 *  io_buffer_length - Length, in bytes, of the buffer specified in io_buffer
 *
 * Returns
 *  int
 */
static int ushc_hw_submit_vendor_request(struct sdioemb_ushc* ushc,
                                         enum ushc_request request, 
                                         enum ushc_request_recipient recipient, 
                                         enum ushc_request_direction direction, 
                                         uint16_t value, 
                                         uint16_t index, 
                                         void* io_buffer, 
                                         uint32_t io_buffer_length)
{
    int status;

    status = ushc->submit_vendor_request(ushc->slot, 
                                         request, 
                                         recipient, 
                                         direction, 
                                         value, 
                                         index, 
                                         io_buffer, 
                                         io_buffer_length);

    return status;
}

/*
 * sdioemb_ushc_set_bus_width
 *
 * Arguments
 *  ushc - pointer to a sdioemb_ushc structure
 *
 * Returns
 *  int
 */
int sdioemb_ushc_set_bus_width(struct sdioemb_ushc* ushc, int bus_width)
{
    ushc_hw_set_bus_width(ushc, bus_width);
    return 0;
}

#define SLOT_USHC_LOCKUP_TIMEOUT_MS 2000
/*
 * sdioemb_ushc_start_cmd
 *
 * Arguments
 *  ushc - pointer to a sdioemb_ushc structure
 *  cmd  - pointer to a sdioemb_cmd structure
 *
 * Returns
 *  int
 */
int sdioemb_ushc_start_cmd(struct sdioemb_ushc* ushc, struct sdioemb_cmd *cmd)
{
    os_int_status_t istate;
    enum ushc_request_direction direction;
    uint16_t blocksize;
    int status;

    os_spinlock_lock_intsave(&ushc->lock, &istate);
    if(ushc->flags & DISCONNECTED)
    {
        status = -ENODEV;
        os_spinlock_unlock_intrestore(&ushc->lock, &istate);
        goto Leave;
    }
    ushc->current_cmd = cmd;
    os_spinlock_unlock_intrestore(&ushc->lock, &istate);

    /*
     * The Astoria's data FIFOs don't work with clock speeds < 5MHz so
     * limit commands with data to 6MHz or more.
     */
    if(cmd->data && ushc->slot->clock_freq < 6000000) 
    {
        status = -EINVAL;
        goto Leave;
    }

    if(cmd->data) 
    {
        blocksize = (uint16_t)cmd->owner->blocksize;
    } 
    else
    {
        blocksize = 0;
    }

    /*
     * Submit CBW
     */
    status = ushc->submit_cbw_request(ushc->slot, cmd->sdio.cmd, blocksize, cmd->sdio.arg);
    if(status <0)
    {
        goto Leave;
    }

    /*
     *  Submit any data
     */
    if(cmd->data)
    {
        if(cmd->flags & SDIOEMB_CMD_FLAG_READ)
        {
            direction = USHC_DEVICE_TO_HOST;
        }
        else
        {
            direction = USHC_HOST_TO_DEVICE;
        }

        status = ushc->submit_data_request(ushc->slot, direction, cmd->data, cmd->len);
        if(status <0)
        {
            goto Leave;
        }
    }

    /*
     * Submit CSW
     */
    status = ushc->submit_csw_request(ushc->slot);
    if(status <0)
    {
        goto Leave;
    }

Leave:
    return status;
}

/*
 * sdioemb_ushc_command_complete
 *
 * Arguments
 *  ushc     - pointer to a sdioemb_ushc structure
 *  status   - command status
 *  response - sdio command response
 *
 * Returns
 *  void
 */
void sdioemb_ushc_command_complete(struct sdioemb_ushc* ushc, uint8_t status, uint32_t respones)
{
    struct sdioemb_cmd* cmd = ushc->current_cmd;

    if(status & USHC_READ_RESP_ERR_MASK) 
    {
        if(status & USHC_READ_RESP_ERR_DAT) 
        {
            cmd->status = SDIOEMB_CMD_ERR_DAT;
        } 
        else 
        {
            cmd->status = SDIOEMB_CMD_ERR_CMD;
        }

        if(status & USHC_READ_RESP_ERR_CRC) 
        {
            cmd->status |= SDIOEMB_CMD_ERR_CRC;
        }
    
        if(status & USHC_READ_RESP_ERR_TIMEOUT) 
        {
            cmd->status |= SDIOEMB_CMD_ERR_TIMEOUT;
        }
    }

    cmd->sdio.response.r1 = respones;

    if(SDIOEMB_CMD_IN_PROGRESS == cmd->status) 
    {
        cmd->status = SDIOEMB_CMD_OK;
    }

    sdioemb_cmd_complete(ushc->slot, cmd);
}

/*
 * sdioemb_ushc_card_present
 *
 * Arguments
 *  ushc - pointer to a sdioemb_ushc structure
 *
 * Returns
 *  int
 */
int sdioemb_ushc_card_present(struct sdioemb_ushc* ushc)
{
    return ushc->interrupt_status & USHC_INT_STATUS_CARD_PRESENT;
}

/*
 * sdioemb_ushc_card_power
 *
 * Arguments
 *  ushc - pointer to a sdioemb_ushc structure
 *  power -
 *
 * Returns
 *  int
 */
int sdioemb_ushc_card_power(struct sdioemb_ushc* ushc, enum sdioemb_power power)
{
    uint16_t voltage;
    int status = 0;

    switch(power)
    {
        case SDIOEMB_POWER_OFF:
            voltage = USHC_PWR_CTRL_OFF;
        break;

        case SDIOEMB_POWER_3V3:
            if(!(ushc->controler_capabilities & USHC_GET_CAPS_3V3)) 
            {
                status = -ENOTSUPP;
            }
            voltage = USHC_PWR_CTRL_3V3;
        break;

        default:
            status = -EINVAL;
        break;
    }

    if(0 == status) 
    {
        status =  ushc->submit_vendor_request(ushc->slot, 
                                              USHC_PWR_CTRL, 
                                              USHC_HOST_TO_DEVICE, 
                                              USHC_RECIPIENT_DEVICE, 
                                              voltage, 
                                              0, 
                                              NULL, 
                                              0);
    }

    return status;
}

/*
 * ushc_hw_get_caps
 *
 * Arguments
 *  ushc - pointer to a sdioemb_ushc structure
 *
 * Returns
 *  int
 */
int ushc_hw_get_caps(struct sdioemb_ushc* ushc)
{
    int status;
    int version;

    status = ushc_hw_submit_vendor_request(ushc, 
                                           USHC_GET_CAPS,
                                           USHC_DEVICE_TO_HOST,
                                           USHC_RECIPIENT_DEVICE,
                                           0,
                                           0,
                                           &ushc->controler_capabilities,
                                           sizeof(ushc->controler_capabilities));
    if(status >= 0) 
    {
        version = ushc->controler_capabilities & USHC_GET_CAPS_VERSION_MASK;
        if (0x02 != version) 
        {
            status = -EINVAL;
        }
    }

    return status;
}

/*
 * sdioemb_ushc_isr
 *
 * Arguments
 *  ushc             - pointer to a sdioemb_ushc structure
 *  interrupt_status -
 *
 * Returns
 *  bool
 */
bool sdioemb_ushc_isr(struct sdioemb_ushc* ushc, uint8_t interrupt_status)
{
    bool status;
    os_int_status_t istate;

    os_spinlock_lock_intsave(&ushc->lock, &istate);
    ushc->interrupt_status = interrupt_status;
    /*
     * Ignore the card interrupt status on interrupt transfers that
     * were submitted while card interrupts where disabled.
     *
     * This avoids occasional spurious interrupts when enabling
     * interrupts immediately after clearing the source on the card.
     */
    if(ushc->flags & IGNORE_NEXT_INT)
    {
        ushc->flags &= ~IGNORE_NEXT_INT;
        os_spinlock_unlock_intrestore(&ushc->lock, &istate);
    }
    else
    {
        if((ushc->flags & INT_EN) && (ushc->interrupt_status & USHC_INT_STATUS_SDIO_INT)) 
        {
            os_spinlock_unlock_intrestore(&ushc->lock, &istate);
            sdioemb_interrupt(ushc->slot);
        }
        else
        {
            os_spinlock_unlock_intrestore(&ushc->lock, &istate);
        }
    }

    os_spinlock_lock_intsave(&ushc->lock, &istate);
    if(!(ushc->flags & INT_EN)) 
    {
        ushc->flags |= IGNORE_NEXT_INT;
    }
    os_spinlock_unlock_intrestore(&ushc->lock, &istate);

    return true;
}

/*
 * sdioemb_ushc_enable_card_int
 *
 * Arguments
 *  ushc - pointer to a sdioemb_ushc structure
 *
 * Returns
 *  void
 */
void sdioemb_ushc_enable_card_int(struct sdioemb_ushc* ushc)
{
    os_int_status_t istate;

    os_spinlock_lock_intsave(&ushc->lock, &istate);
    ushc->flags |= INT_EN;
    os_spinlock_unlock_intrestore(&ushc->lock, &istate);
}

/*
 * sdioemb_ushc_disable_card_int
 *
 * Arguments
 *  ushc - pointer to a sdioemb_ushc structure
 *
 * Returns
 *  void
 */
void sdioemb_ushc_disable_card_int(struct sdioemb_ushc* ushc)
{
    os_int_status_t istate;

    os_spinlock_lock_intsave(&ushc->lock, &istate);
    ushc->flags &= ~INT_EN;
    os_spinlock_unlock_intrestore(&ushc->lock, &istate);
}
