/*
 * 007_speed - measure CMD52 and CMD53 performance.
 *
 * Copyright (C) 2009 Cambridge Silicon Radio Ltd.
 *
 * Refer to LICENSE.txt included with this source code for details on
 * the license terms.
 *
 * Measure the performance of CMD52s and CMD53s.  Uses the similar
 * tests to those in 002_cmd52 and 006_cmd53.
 */
#include <string.h>
#include <sys/time.h>
#include <time.h>

#include "sdd_test.h"

#define COMMAND_RATE_MIN 1000 /* command/s */
#define READ_RATE_MIN 10 /* Mbit/s */
#define WRITE_RATE_MIN 10 /* Mbit/s */

struct test_card {
    int      generic_func;
    uint32_t sdio_mode;
    int      buffer_hdl_write;
    int      buffer_hdl_read;
};

#define DASH_SDIO_MODE (0xf935 << 1)

static struct test_card dash_data = {
    .generic_func     = 2,
    .sdio_mode        = DASH_SDIO_MODE,
    .buffer_hdl_write = 0x12,
    .buffer_hdl_read  = 0x13,
};

static struct test_card cinderella_data = {
    .generic_func     = 1,
    .sdio_mode        = 0,
    .buffer_hdl_write = 0x1,
    .buffer_hdl_read  = 0x1,
};

static const struct sdioemb_test_card supported_cards[] = {
    { SDIO_MANF_ID_CSR, SDIO_CARD_ID_CSR_DASH_D00, "dash (d00)", &dash_data },
    { SDIO_MANF_ID_CSR, SDIO_CARD_ID_CSR_DASH, "dash", &dash_data },
    { SDIO_MANF_ID_CSR, SDIO_CARD_ID_CSR_CINDERELLA, "cinderella", &cinderella_data },
    { },
};


#define BUFFER_LEN (16*1024)

static uint64_t time_us(void)
{
    struct timeval tv;

    gettimeofday(&tv, NULL);
    return tv.tv_sec * 1000000ull + tv.tv_usec;
}


static void test_007_speed(sdio_uif_t uif, void *arg)
{
    struct test_card *card;
    size_t block_size;
    int i, j, l;
    int ret;
    uint8_t *buf_out, *buf_in;
    uint64_t start, stop;
    unsigned command_rate = 0;
    double read_rate = 0, write_rate = 0;

    buf_out = malloc(2*BUFFER_LEN);
    if (!buf_out) {
        sdioemb_test_abort("transfer buffer");
    }
    buf_in = buf_out + BUFFER_LEN;

    card = sdioemb_test_check_cards(supported_cards);

    block_size = sdio_block_size(uif, card->generic_func);

    sdio_write8(uif, 0, SDIO_CCCR_IO_EN, 1 << card->generic_func);
    if (card->sdio_mode) {
        sdio_write8(uif, card->generic_func, card->sdio_mode, 0x00);
    }

    l = 4096;

    for (i = 0; i < l; i++) {
        buf_out[i] = rand()/(int)(((unsigned)RAND_MAX + 1) / 256);
    }

    for (i = 0; i < sdioemb_test_max_iters(); i++) {
        uint8_t r;
        unsigned cr;

        start = time_us();
        for (j = 0; j < 1000; j++) {
            ret = sdio_read8(uif, 0, SDIO_CSR_FROM_HOST_SCRATCH0, &r);
            if (ret < 0) {
                sdioemb_test_abort("sdioemb_read8");
            }
        }
        stop = time_us();

        cr = j * 1000000 / (stop - start);
        command_rate += cr;
        sdioemb_test_info("%u command/s", cr);
    }
    command_rate /= i;

    for (i = 0; i < sdioemb_test_max_iters(); i++) {
        double wr;

        start = time_us();
        for (j = 0; j < 1000; j++) {
            ret = sdio_write(uif, card->generic_func, card->buffer_hdl_write,
                             buf_out, l, block_size);
            if (ret < 0) {
                sdioemb_test_abort("sdio_write");
            }
        }
        stop = time_us();

        wr = l * j * 8 / (double)(stop - start);
        write_rate += wr;
        sdioemb_test_info("write %.3f Mbit/s", wr);
    }
    write_rate /= i;

    for (i = 0; i < sdioemb_test_max_iters(); i++) {
        double rr;

        start = time_us();
        for (j = 0; j < 1000; j++) {
            ret = sdio_read(uif, card->generic_func, card->buffer_hdl_read,
                            buf_in, l, block_size);
            if (ret < 0) {
                sdioemb_test_abort("sdio_read");
            }
        }
        stop = time_us();

        rr = l * j * 8 / (double)(stop - start);
        read_rate += rr;
        sdioemb_test_info("read %.3f Mbit/s", rr);
    }
    read_rate /= i;

    if (command_rate < COMMAND_RATE_MIN) {
        sdioemb_test_fail("%d < %d command/s", command_rate, COMMAND_RATE_MIN);
    }
    if (write_rate < WRITE_RATE_MIN) {
        sdioemb_test_fail("write %d < %d Mbit/s", write_rate, WRITE_RATE_MIN);
    }
    if (read_rate< READ_RATE_MIN) {
        sdioemb_test_fail("write %d < %d Mbit/s", write_rate, READ_RATE_MIN);
    }

    if (card->sdio_mode) {
        sdio_write8(uif, card->generic_func, card->sdio_mode, 0x01);
    }
    sdioemb_test_pass();
}

int main(int argc, char *argv[])
{
    return sdioemb_test_run(SDIOEMB_TEST_NAME, argc, argv,
                        test_007_speed, NULL, NULL, NULL);
}
