/*
 * 008_timeout - test detection of response and data timeouts.
 *
 * Copyright (C) 20089 Cambridge Silicon Radio Ltd.
 *
 * Refer to LICENSE.txt included with this source code for details on
 * the license terms.
 *
 * Tests CMD52s time out when expected by doing register reads/writes
 * to a disabled function.
 *
 * Tests CMD53s time out when expected by reading/writing from a
 * buffer handle that does not exist (the generic function is enabled
 * for this).
 */
#include <string.h>

#include "sdd_test.h"

struct test_card {
    int      generic_func;
    uint32_t sdio_mode;
    int      buffer_hdl_write;
    int      buffer_hdl_read;
};

#define DASH_SDIO_MODE (0xf935 << 1)

static struct test_card dash_data = {
    .generic_func     = 2,
    .sdio_mode        = DASH_SDIO_MODE,
    .buffer_hdl_write = 0x12,
    .buffer_hdl_read  = 0x13,
};

static struct test_card cinderella_data = {
    .generic_func     = 1,
    .sdio_mode        = 0,
    .buffer_hdl_write = 0x1,
    .buffer_hdl_read  = 0x1,
};

static const struct sdioemb_test_card supported_cards[] = {
    { SDIO_MANF_ID_CSR, SDIO_CARD_ID_CSR_DASH_D00, "dash (d00)", &dash_data },
    { SDIO_MANF_ID_CSR, SDIO_CARD_ID_CSR_DASH, "dash", &dash_data },
    { SDIO_MANF_ID_CSR, SDIO_CARD_ID_CSR_CINDERELLA, "cinderella", &cinderella_data },
    { },
};


#define BUFFER_LEN (16*1024)


static void test_006_cmd53(sdio_uif_t uif, void *arg)
{
    struct test_card *card;
    size_t block_size;
    int i;
    int ret;
    uint8_t *buf_out, *buf_in;

    buf_out = malloc(2*BUFFER_LEN);
    if (!buf_out) {
        sdioemb_test_abort("transfer buffer");
    }
    buf_in = buf_out + BUFFER_LEN;

    card = sdioemb_test_check_cards(supported_cards);

    block_size = sdio_block_size(uif, card->generic_func);

    sdio_write8(uif, 0, SDIO_CCCR_IO_EN, 1 << card->generic_func);
    if (card->sdio_mode) {
        sdio_write8(uif, card->generic_func, card->sdio_mode, 0x00);
    }

    for (i = 0; i < sdioemb_test_max_iters(); i++) {
        uint8_t r;

        /* Disable all functions. */
        ret = sdio_write8(uif, 0, SDIO_CCCR_IO_EN, 0);
        if (ret < 0) {
            sdioemb_test_abort("all functions not disabled");
        }

        ret = sdio_write8(uif, card->generic_func, 0xf93e << 1, 0x45);
        if (ret == 0 || (ret < 0 && ret != SDIO_ETIMEDOUT)) {
            sdioemb_test_fail("sdio_write8 didn't time out");
        }

        ret = sdio_read8(uif, card->generic_func, 0xf93e << 1, &r);
        if (ret == 0 || (ret < 0 && ret != SDIO_ETIMEDOUT)) {
            sdioemb_test_fail("sdio_read8 didn't time out");
        }

        ret = sdio_write(uif, card->generic_func, 0, buf_out, 64, block_size);
        if (ret == 0 || (ret < 0 && ret != SDIO_ETIMEDOUT)) {
            sdioemb_test_fail("sdio_write response didn't time out");
        }

        ret = sdio_read(uif, card->generic_func, 0, buf_in, 64, block_size);
        if (ret == 0 || (ret < 0 && ret != SDIO_ETIMEDOUT)) {
            sdioemb_test_fail("sdio_read response didn't time out");
        }

        /* Enable the generic function. */
        ret = sdio_write8(uif, 0, SDIO_CCCR_IO_EN, 1 << card->generic_func);
        if (ret < 0) {
            sdioemb_test_abort("generic function not enabled");
        }

        ret = sdio_write(uif, card->generic_func, 0, buf_out, 1024, block_size);
        if (ret == 0 || (ret < 0 && ret != SDIO_ETIMEDOUT)) {
            sdioemb_test_fail("sdio_write data didn't time out (%d)", ret);
        }

        ret = sdio_read(uif, card->generic_func, 0, buf_in, 1024, block_size);
        if (ret == 0 || (ret < 0 && ret != SDIO_ETIMEDOUT)) {
            sdioemb_test_fail("sdio_read data didn't time out (%d)", ret);
        }
    }

    if (card->sdio_mode) {
        sdio_write8(uif, card->generic_func, card->sdio_mode, 0x01);
    }
    sdioemb_test_pass();
}

int main(int argc, char *argv[])
{
    return sdioemb_test_run(SDIOEMB_TEST_NAME, argc, argv,
                        test_006_cmd53, NULL, NULL, NULL);
}
