/*
 * 009_bus_freq - test CMD52 and CMD53 reads and writes at different bus frequencies
 *
 * Copyright (C) 2008 Cambridge Silicon Radio Ltd.
 *
 * Refer to LICENSE.txt included with this source code for details on
 * the license terms.
 *
 * Tests CMD53 reads and writes using some buffer handles available
 * via the generic function.
 */
#include <string.h>

#include "sdd_test.h"

struct test_card {
    int      generic_func;
    uint32_t sdio_mode;
    int      buffer_hdl_write;
    int      buffer_hdl_read;
};

#define DASH_SDIO_MODE (0xf935 << 1)

static struct test_card dash_data = {
    .generic_func     = 2,
    .sdio_mode        = DASH_SDIO_MODE,
    .buffer_hdl_write = 0x12,
    .buffer_hdl_read  = 0x13,
};

static struct test_card cinderella_data = {
    .generic_func     = 1,
    .sdio_mode        = 0,
    .buffer_hdl_write = 0x1,
    .buffer_hdl_read  = 0x1,
};

static const struct sdioemb_test_card supported_cards[] = {
    { SDIO_MANF_ID_CSR, SDIO_CARD_ID_CSR_DASH_D00, "dash (d00)", &dash_data },
    { SDIO_MANF_ID_CSR, SDIO_CARD_ID_CSR_DASH, "dash", &dash_data },
    { SDIO_MANF_ID_CSR, SDIO_CARD_ID_CSR_CINDERELLA, "cinderella", &cinderella_data },
    { },
};

#define BUFFER_LEN (16*1024)


static void test_009_bus_freq(sdio_uif_t uif, void *arg)
{
    struct test_card *card;
    size_t block_size;
    int i, l;
    int ret;
    uint8_t *buf_out, *buf_in;
    int freq;

    buf_out = malloc(2*BUFFER_LEN);
    if (!buf_out) {
        sdioemb_test_abort("transfer buffer");
    }
    buf_in = buf_out + BUFFER_LEN;

    card = sdioemb_test_check_cards(supported_cards);

    block_size = sdio_block_size(uif, card->generic_func);

    sdio_write8(uif, 0, SDIO_CCCR_IO_EN, 1 << card->generic_func);
    if (card->sdio_mode) {
        sdio_write8(uif, card->generic_func, card->sdio_mode, 0x00);
    }

    if (sdio_card_id(uif) == SDIO_CARD_ID_CSR_DASH_D00) {
        /* map all buffer pages by writing 16 KiB. */
        l = 512;
        for (i = 0; i < 16*1024; i += l) {
            ret = sdio_write(uif, card->generic_func, card->buffer_hdl_write,
                             buf_out, l, block_size);
            if (ret < 0) {
                sdioemb_test_abort("write to map buffer");
            }
        }
    }

    for (freq = 50000000; freq > 400000/2; freq /= 2) {
        sdio_set_max_bus_freq(uif, freq);
        for (i = 0; i < sdioemb_test_max_iters(); i++) {
            uint8_t r;
            int b;

            /* Try some CMD52s */
            ret = sdio_write8(uif, 0, SDIO_CSR_FROM_HOST_SCRATCH0, i & 0xff);
            if (ret < 0) {
                sdioemb_test_fail("sdioemb_write8");
            }

            ret = sdio_read8(uif, 0, SDIO_CSR_FROM_HOST_SCRATCH0, &r);
            if (ret < 0) {
                sdioemb_test_fail("sdioemb_read8");
            }
            if (r != (i & 0xff)) {
                sdioemb_test_fail("data read (%d) != expected value (%d)", r, i & 0xff);
            }

            /* And some CMD53s, but only at higher data rates. */
            if (freq > 6000000) {
                l = 2048;
                for (b = 0; b < l; b++) {
                    buf_out[b] = rand()/(int)(((unsigned)RAND_MAX + 1) / 256);
                    buf_in[b] = 0xaa;
                }

                ret = sdio_write(uif, card->generic_func, card->buffer_hdl_write,
                                 buf_out, l, block_size);
                if (ret < 0) {
                    sdioemb_test_fail("sdio_write");
                }

                ret = sdio_read(uif, card->generic_func, card->buffer_hdl_read,
                                buf_in, l, block_size);
                if (ret < 0) {
                    sdioemb_test_fail("sdio_read");
                }
                if (sdio_card_id(uif) != SDIO_CARD_ID_CSR_CINDERELLA) {
                    if (memcmp(buf_in, buf_out, l) != 0) {
                        sdioemb_test_fail("read buffer != written buffer");
                    }
                }
            }
        }
    }

    sdio_set_max_bus_freq(uif, 50000000);

    if (card->sdio_mode) {
        sdio_write8(uif, card->generic_func, card->sdio_mode, 0x01);
    }
    if (sdio_card_id(uif) == SDIO_CARD_ID_CSR_CINDERELLA) {
        sdioemb_test_abort("commands successful but data not verified");
    }
    sdioemb_test_pass();
}

int main(int argc, char *argv[])
{
    return sdioemb_test_run(SDIOEMB_TEST_NAME, argc, argv,
                        test_009_bus_freq, NULL, NULL, NULL);
}
