/*
 * 010_block_size - test CMD53s with different block sizes.
 *
 * Copyright (C) 2007-2009 Cambridge Silicon Radio Ltd.
 *
 * Refer to LICENSE.txt included with this source code for details on
 * the license terms.
 *
 * Tests CMD53s at different block sizes from 4 to the maximum
 * supported by the card.
 */
#include <string.h>

#include "sdd_test.h"

struct test_card {
    int      generic_func;
    uint32_t sdio_mode;
    int      buffer_hdl_write;
    int      buffer_hdl_read;
    size_t   min_block_size;
};

#define DASH_SDIO_MODE (0xf935 << 1)

static struct test_card dash_data = {
    .generic_func     = 2,
    .sdio_mode        = DASH_SDIO_MODE,
    .buffer_hdl_write = 0x12,
    .buffer_hdl_read  = 0x13,
    .min_block_size   = 4,
};

static struct test_card cinderella_data = {
    .generic_func     = 1,
    .sdio_mode        = 0,
    .buffer_hdl_write = 0x1,
    .buffer_hdl_read  = 0x1,
    .min_block_size   = 64,
};

static const struct sdioemb_test_card supported_cards[] = {
    { SDIO_MANF_ID_CSR, SDIO_CARD_ID_CSR_DASH_D00, "dash (d00)", &dash_data },
    { SDIO_MANF_ID_CSR, SDIO_CARD_ID_CSR_DASH, "dash", &dash_data },
    { SDIO_MANF_ID_CSR, SDIO_CARD_ID_CSR_CINDERELLA, "cinderella", &cinderella_data },
    { },
};


#define BUFFER_LEN (16*1024)

static void test_010_block_size(sdio_uif_t uif, void *arg)
{
    struct test_card *card;
    size_t max_block_size, block_size;
    int i, l;
    int ret;
    uint8_t *buf_out, *buf_in;

    buf_out = malloc(2*BUFFER_LEN);
    if (!buf_out) {
        sdioemb_test_abort("transfer buffer");
    }
    buf_in = buf_out + BUFFER_LEN;

    card = sdioemb_test_check_cards(supported_cards);

    max_block_size = sdio_max_block_size(uif, card->generic_func);
    if (max_block_size == 0) {
        sdioemb_test_abort("maximum block size == 0\n");
    }

    sdio_write8(uif, 0, SDIO_CCCR_IO_EN, 1 << card->generic_func);
    if (card->sdio_mode) {
        sdio_write8(uif, card->generic_func, card->sdio_mode, 0x00);
    }

    if (sdio_card_id(uif) == SDIO_CARD_ID_CSR_DASH_D00) {
        /* map all buffer pages by writing 16 KiB. */
        l = 512;
        for (i = 0; i < 16*1024; i += l) {
            ret = sdio_write(uif, card->generic_func, card->buffer_hdl_write,
                             buf_out, l, max_block_size);
            if (ret < 0) {
                sdioemb_test_abort("write to map buffer");
            }
        }
    }
    
    /*
     * Test block sizes up to the maximum.
     */
    for (block_size = card->min_block_size; block_size <= max_block_size; block_size *= 2) {
        ret = sdio_set_block_size(uif, card->generic_func, block_size);
        if (ret < 0) {
            sdioemb_test_abort("F%d: set block size to %d", card->generic_func, block_size);
        }

        l = 1024; /* At least two blocks but less than 512. */

        for (i = 0; i < sdioemb_test_max_iters(); i++) {
            int b;

            for (b = 0; b < l; b++) {
                buf_out[b] = rand()/(int)(((unsigned)RAND_MAX + 1) / 256);
                buf_in[b] = 0xaa;
            }

            ret = sdio_write(uif, card->generic_func, card->buffer_hdl_write,
                             buf_out, l, block_size);
            if (ret < 0) {
                sdioemb_test_fail("sdio_write");
            }

            ret = sdio_read(uif, card->generic_func, card->buffer_hdl_read,
                            buf_in, l, block_size);
            if (ret < 0) {
                sdioemb_test_fail("sdio_read");
            }
            if (sdio_card_id(uif) != SDIO_CARD_ID_CSR_CINDERELLA) {
                if (memcmp(buf_in, buf_out, l) != 0) {
                    sdioemb_test_fail("read buffer != written buffer");
                }
            }
        }
    }

    ret = sdio_set_block_size(uif, card->generic_func, 0);
    if (ret < 0) {
        sdioemb_test_abort("F%d: set block size to default", card->generic_func);
    }
    if (card->sdio_mode) {
        sdio_write8(uif, card->generic_func, card->sdio_mode, 0x01);
    }
    if (sdio_card_id(uif) == SDIO_CARD_ID_CSR_CINDERELLA) {
        sdioemb_test_abort("commands successful but data not verified");
    }
    sdioemb_test_pass();
}

int main(int argc, char *argv[])
{
    return sdioemb_test_run(SDIOEMB_TEST_NAME, argc, argv,
                        test_010_block_size, NULL, NULL, NULL);
}
