/*
 * Synergy compatible API.
 *
 * Copyright (C) 2010 Cambridge Silicon Radio Ltd.
 *
 * Refer to LICENSE.txt included with this source code for details on
 * the license terms.
 */
#include <ntddk.h>
#include <initguid.h>

#include <oska/types.h>
#include <sdioemb/windows/sdio_api.h>
#include <sdioemb/windows/guid.h>
#include <csr_sdio.h>
#include <csr_sdio_lib.h>
#include <csr_sdio_wdf.h>

#define CSR_SDIO_POOL_TAG 'srsc'

struct glue_driver {
    struct sdioemb_func_driver driver; /* must be first */
    CsrSdioFunctionDriver *functionDriver;
    LIST_ENTRY node;
};

typedef struct glue_func {
    CsrSdioFunction function;
    struct sdioemb_wdf_interface iface;
    KEVENT callback_event;
    CsrResult inserted_result;
} GLUE_FUNC, *PGLUE_FUNC;
WDF_DECLARE_CONTEXT_TYPE_WITH_NAME(GLUE_FUNC, get_glue_func);

static LIST_ENTRY drivers_list;
static BOOLEAN drivers_initialized;

CsrStatus CsrSdioResultToStatus(CsrResult result)
{
    switch (result) {
    case CSR_RESULT_SUCCESS:
        return 0;
    case CSR_RESULT_FAILURE:
        return -ENOMEM;
    case CSR_SDIO_RESULT_INVALID_VALUE:
        return -EINVAL;
    case CSR_SDIO_RESULT_NO_DEVICE:
        return -ENODEV;
    case CSR_SDIO_RESULT_CRC_ERROR:
        return -EIO;
    case CSR_SDIO_RESULT_TIMEOUT:
        return -ETIMEDOUT;
    case CSR_SDIO_RESULT_NOT_RESET:
        return 0;
    default:
        return -EIO; /* arbitrary */
    }
}

CsrResult CsrSdioStatusToResult(CsrStatus status)
{
    if (!CSR_STATUS_FAILURE(status)) {
        return CSR_RESULT_SUCCESS;
    }

    switch (status) {
    case -EIO:
        return CSR_SDIO_RESULT_CRC_ERROR;
    case -ETIMEDOUT:
        return CSR_SDIO_RESULT_TIMEOUT;
    case -ENODEV:
        return CSR_SDIO_RESULT_NO_DEVICE;
    case -ENOMEM:
        return CSR_RESULT_FAILURE; /* There is no CSR_RESULT_NO_MEMORY. */
    case -EINVAL:
        return CSR_SDIO_RESULT_INVALID_VALUE;
    default:
        return CSR_RESULT_FAILURE;
    }
}

static int glue_probe(struct sdioemb_dev *fdev)
{
    struct glue_driver *glue_driver = (struct glue_driver *)fdev->driver;
    struct glue_func *glue = fdev->drv_data;

    glue_driver->functionDriver->inserted(&glue->function);
    KeWaitForSingleObject(&glue->callback_event, Executive, KernelMode, FALSE, NULL);
    return CsrSdioResultToStatus(glue->inserted_result);
}

static void glue_remove(struct sdioemb_dev *fdev)
{
    struct glue_driver *glue_driver = (struct glue_driver *)fdev->driver;
    struct glue_func *glue = fdev->drv_data;

    glue_driver->functionDriver->removed(&glue->function);
    KeWaitForSingleObject(&glue->callback_event, Executive, KernelMode, FALSE, NULL);
}

static void glue_interrupt(struct sdioemb_dev *fdev)
{
    struct glue_driver *glue_driver = (struct glue_driver *)fdev->driver;
    struct glue_func *glue = fdev->drv_data;
    CsrSdioInterruptDsrCallback dsr;

    dsr = glue_driver->functionDriver->intr(&glue->function);
    if (dsr) {
        dsr(&glue->function);
    }
}

static void glue_suspend(struct sdioemb_dev *fdev)
{
    struct glue_driver *glue_driver = (struct glue_driver *)fdev->driver;
    struct glue_func *glue = fdev->drv_data;

    glue_driver->functionDriver->suspend(&glue->function);
    KeWaitForSingleObject(&glue->callback_event, Executive, KernelMode, FALSE, NULL);
}

static void glue_resume(struct sdioemb_dev *fdev)
{
    struct glue_driver *glue_driver = (struct glue_driver *)fdev->driver;
    struct glue_func *glue = fdev->drv_data;

    glue_driver->functionDriver->resume(&glue->function);
    KeWaitForSingleObject(&glue->callback_event, Executive, KernelMode, FALSE, NULL);
}

CsrResult CsrSdioFunctionDriverRegister(CsrSdioFunctionDriver *functionDriver)
{
    struct glue_driver *glue_driver;
    size_t glue_driver_size;
    struct sdioemb_id_table *ids;
    unsigned i;
    int ret;

    if (!drivers_initialized) {
        InitializeListHead(&drivers_list);
        drivers_initialized = TRUE;
    }

    glue_driver_size = sizeof(*glue_driver)
        + (functionDriver->idsCount + 1) * sizeof(struct sdioemb_id_table);

    glue_driver = ExAllocatePoolWithTag(NonPagedPool, glue_driver_size, CSR_SDIO_POOL_TAG);
    if (glue_driver == NULL) {
        return CSR_RESULT_FAILURE;
    }
    functionDriver->priv = glue_driver;

    ids = (struct sdioemb_id_table*)&glue_driver[1];
    for (i = 0; i < functionDriver->idsCount; i++) {
        ids[i].vendor_id = functionDriver->ids[i].manfId;
        ids[i].device_id = functionDriver->ids[i].cardId;
        ids[i].function  = functionDriver->ids[i].sdioFunction;
        ids[i].interface = functionDriver->ids[i].sdioInterface;
    }

    glue_driver->driver.name     = "glue";
    glue_driver->driver.id_table = ids;
    glue_driver->driver.probe    = glue_probe;
    glue_driver->driver.remove   = glue_remove;
    glue_driver->driver.card_int_handler = glue_interrupt;
    glue_driver->driver.suspend  = glue_suspend;
    glue_driver->driver.resume   = glue_resume;

    glue_driver->functionDriver = functionDriver;

    InsertTailList(&drivers_list, &glue_driver->node);

    return CSR_RESULT_SUCCESS;
}

void CsrSdioFunctionDriverUnregister(CsrSdioFunctionDriver *functionDriver)
{
    struct glue_driver *glue_driver = functionDriver->priv;

    RemoveEntryList(&glue_driver->node);

    ExFreePoolWithTag(glue_driver, CSR_SDIO_POOL_TAG);
}

NTSTATUS CsrSdioWdfDeviceInit(WDFDEVICE device)
{
    struct glue_func *glue;
    WDF_OBJECT_ATTRIBUTES attribs;
    struct sdioemb_dev *fdev;
    NTSTATUS status;

    WDF_OBJECT_ATTRIBUTES_INIT_CONTEXT_TYPE(&attribs, GLUE_FUNC);

    status = WdfObjectAllocateContext(device, &attribs, &glue);
    if (!NT_SUCCESS(status)) {
        return status;
    }

    /* Get driver interface. */
    status = WdfFdoQueryForInterface(device, &GUID_DRVINTERFACE_SDIOEMB_FUNC,
                                     &glue->iface.header, sizeof(glue->iface), 1, NULL);
    if (!NT_SUCCESS(status)) {
        return status;
    }

    fdev = glue->iface.fdev;

    fdev->os_device = device;
    fdev->drv_data = glue;

    glue->function.sdioId.manfId = fdev->vendor_id;
    glue->function.sdioId.cardId = fdev->device_id;
    glue->function.sdioId.sdioFunction = (CsrUint8)fdev->function;
    glue->function.sdioId.sdioInterface = fdev->interface;
    glue->function.blockSize = fdev->blocksize;
    glue->function.features = CSR_SDIO_FEATURE_BYTE_MODE
        | CSR_SDIO_FEATURE_DMA_CAPABLE_MEM_REQUIRED;
    glue->function.cardHandle = (void *)fdev->slot_id;
    glue->function.osDevice = fdev->os_device;
    glue->function.priv = glue;

    KeInitializeEvent(&glue->callback_event, SynchronizationEvent, FALSE);

    return STATUS_SUCCESS;
}

void CsrSdioWdfDeviceCleanup(WDFDEVICE device)
{
    struct glue_func *glue = get_glue_func(device);

    if (glue && glue->iface.header.InterfaceDereference) {
        glue->iface.header.InterfaceDereference(glue->iface.header.Context);
    }
}

NTSTATUS CsrSdioWdfDeviceAdd(WDFDEVICE device)
{
    struct glue_func *glue = get_glue_func(device);
    LIST_ENTRY *node;
    int ret;

    for (node = drivers_list.Flink; node != &drivers_list; node = node->Flink) {
        struct glue_driver *driver = CONTAINING_RECORD(node, struct glue_driver, node);

        ret = glue->iface.driver_probe(&driver->driver, glue->iface.fdev);
        if (ret == 0) {
            break;
        }
    }

    return (NTSTATUS)ret;
}

void CsrSdioWdfDeviceDel(WDFDEVICE device)
{
    struct glue_func *glue = get_glue_func(device);

    if (glue->iface.fdev->driver) {
        glue->iface.driver_remove(glue->iface.fdev->driver, glue->iface.fdev);
    }
}

NTSTATUS CsrSdioWdfDeviceSuspend(WDFDEVICE device)
{
    struct glue_func *glue = get_glue_func(device);

    glue->iface.suspend_function(glue->iface.fdev);
    return STATUS_SUCCESS;
}

NTSTATUS CsrSdioWdfDeviceResume(WDFDEVICE device)
{
    struct glue_func *glue = get_glue_func(device);

    glue->iface.resume_function(glue->iface.fdev);
    return STATUS_SUCCESS;
}

void CsrSdioInsertedAcknowledge(CsrSdioFunction *function, CsrResult result)
{
    struct glue_func *glue = function->priv;

    glue->inserted_result = result;
    KeSetEvent(&glue->callback_event, 0, FALSE);
}

void CsrSdioRemovedAcknowledge(CsrSdioFunction *function)
{
    struct glue_func *glue = function->priv;

    KeSetEvent(&glue->callback_event, 0, FALSE);
}

void CsrSdioSuspendAcknowledge(CsrSdioFunction *function, CsrResult result)
{
    struct glue_func *glue = function->priv;

    KeSetEvent(&glue->callback_event, 0, FALSE);
}

void CsrSdioResumeAcknowledge(CsrSdioFunction *function, CsrResult result)
{
    struct glue_func *glue = function->priv;

    KeSetEvent(&glue->callback_event, 0, FALSE);
}

CsrResult CsrSdioFunctionEnable(CsrSdioFunction *function)
{
    struct glue_func *glue = function->priv;
    CsrStatus ret;

    ret = glue->iface.enable_function(glue->iface.fdev);
    return CsrSdioStatusToResult(ret);
}

CsrResult CsrSdioFunctionDisable(CsrSdioFunction *function)
{
    struct glue_func *glue = function->priv;
    CsrStatus ret;

    ret = glue->iface.disable_function(glue->iface.fdev);
    return CsrSdioStatusToResult(ret);
}

CsrResult CsrSdioInterruptEnable(CsrSdioFunction *function)
{
    struct glue_func *glue = function->priv;
    CsrStatus ret;

    ret = glue->iface.interrupt_enable(glue->iface.fdev);
    return CsrSdioStatusToResult(ret);
}

CsrResult CsrSdioInterruptDisable(CsrSdioFunction *function)
{
    struct glue_func *glue = function->priv;
    CsrStatus ret;

    ret = glue->iface.interrupt_disable(glue->iface.fdev);
    return CsrSdioStatusToResult(ret);
}

void CsrSdioInterruptAcknowledge(CsrSdioFunction *function)
{
    struct glue_func *glue = function->priv;

    glue->iface.interrupt_acknowledge(glue->iface.fdev);
}

CsrResult CsrSdioBlockSizeSet(CsrSdioFunction *function, CsrUint16 blockSize)
{
    struct glue_func *glue = function->priv;
    CsrStatus ret;

    ret = glue->iface.set_block_size(glue->iface.fdev, blockSize);
    if (ret >= 0) {
        function->blockSize = glue->iface.fdev->blocksize;
    }
    return CsrSdioStatusToResult(ret);
}

CsrResult CsrSdioMaxBusClockFrequencySet(CsrSdioFunction *function, CsrUint32 maxFrequency)
{
    struct glue_func *glue = function->priv;

    glue->iface.set_max_bus_freq(glue->iface.fdev, maxFrequency);
    return CSR_RESULT_SUCCESS;
}

CsrResult CsrSdioRead8(CsrSdioFunction *function, CsrUint32 address, CsrUint8 *data)
{
    struct glue_func *glue = function->priv;
    CsrStatus ret;

    ret = glue->iface.read8(glue->iface.fdev, address, data);
    return CsrSdioStatusToResult(ret);
}

CsrResult CsrSdioRead16(CsrSdioFunction *function, CsrUint32 address, CsrUint16 *data)
{
    struct glue_func *glue = function->priv;
    CsrStatus ret;

    ret = glue->iface.read16(glue->iface.fdev, address, data);
    return CsrSdioStatusToResult(ret);
}

CsrResult CsrSdioWrite8(CsrSdioFunction *function, CsrUint32 address, CsrUint8 data)
{
    struct glue_func *glue = function->priv;
    CsrStatus ret;

    ret = glue->iface.write8(glue->iface.fdev, address, data);
    return CsrSdioStatusToResult(ret);
}

CsrResult CsrSdioWrite16(CsrSdioFunction *function, CsrUint32 address, CsrUint16 data)
{
    struct glue_func *glue = function->priv;
    CsrStatus ret;

    ret = glue->iface.write16(glue->iface.fdev, address, data);
    return CsrSdioStatusToResult(ret);
}

CsrResult CsrSdioF0Read8(CsrSdioFunction *function, CsrUint32 address, CsrUint8 *data)
{
    struct glue_func *glue = function->priv;
    CsrStatus ret;

    ret = glue->iface.f0_read8(glue->iface.fdev, address, data);
    return CsrSdioStatusToResult(ret);
}

CsrResult CsrSdioF0Write8(CsrSdioFunction *function, CsrUint32 address, CsrUint8 data)
{
    struct glue_func *glue = function->priv;
    CsrStatus ret;

    ret = glue->iface.f0_write8(glue->iface.fdev, address, data);
    return CsrSdioStatusToResult(ret);
}

CsrResult CsrSdioRead(CsrSdioFunction *function, CsrUint32 address,
                      void *data, CsrUint32 length)
{
    struct glue_func *glue = function->priv;
    CsrStatus ret;

    ret = glue->iface.read(glue->iface.fdev, address, data, length);
    return CsrSdioStatusToResult(ret);
}

CsrResult CsrSdioWrite(CsrSdioFunction *function, CsrUint32 address,
                       const void *data, CsrUint32 length)
{
    struct glue_func *glue = function->priv;
    CsrStatus ret;

    ret = glue->iface.write(glue->iface.fdev, address, data, length);
    return CsrSdioStatusToResult(ret);
}

CsrResult CsrSdioPowerOn(CsrSdioFunction *function)
{
    struct glue_func *glue = function->priv;

    glue->iface.power_on(glue->iface.fdev);
    return CSR_RESULT_SUCCESS;
}

void CsrSdioPowerOff(CsrSdioFunction *function)
{
    struct glue_func *glue = function->priv;

    glue->iface.power_off(glue->iface.fdev);
}

CsrResult CsrSdioHardReset(CsrSdioFunction *function)
{
    struct glue_func *glue = function->priv;
    CsrStatus ret;

    ret = glue->iface.hard_reset(glue->iface.fdev);
    if (ret == 1) {
        return CSR_SDIO_RESULT_NOT_RESET;
    }
    return CsrSdioStatusToResult(ret);
}

void CsrSdioFunctionActive(CsrSdioFunction *function)
{
    /* Functions are activated automatically when a command is
       started. */
}

void CsrSdioFunctionIdle(CsrSdioFunction *function)
{
    struct glue_func *glue = function->priv;

    glue->iface.idle_function(glue->iface.fdev);
}

CsrResult CsrSdioFunctionReenable(CsrSdioFunction *function)
{
    struct glue_func *glue = function->priv;
    CsrStatus ret;

    ret = glue->iface.reenable_csr_function(glue->iface.fdev);
    return CsrSdioStatusToResult(ret);
}
