/*
 * Sample Windows KMDF function driver
 *
 * Copyright (C) 2010 Cambridge Silicon Radio Ltd.
 *
 * Refer to LICENSE.txt included with this source code for details on
 * the license terms.
 *
 * This is the Windows KMDF wrapper for the sample function driver in
 * ../../../cards/sample/sample.c.
 */
#define WPP_CONTROL_GUIDS                                               \
    WPP_DEFINE_CONTROL_GUID(SampleTraceGuid,                            \
                            (27ae9e5d,899a,40a4,9048,6578fff21083),     \
                            WPP_DEFINE_BIT(OS_TRACE_DFLT)               \
        )
#include <oska/trace.h>
#include <oska/all.h>
#include <csr_sdio_wdf.h>

#include "../../../cards/sample/sample.h"

#include "sample.tmh"

typedef struct sample_device_context {
    bool added;
} SAMPLE_DEVICE_CONTEXT, *PSAMPLE_DEVICE_CONTEXT;
WDF_DECLARE_CONTEXT_TYPE_WITH_NAME(SAMPLE_DEVICE_CONTEXT, sample_get_device_context);

DRIVER_INITIALIZE DriverEntry;

static EVT_WDF_OBJECT_CONTEXT_CLEANUP sample_driver_cleanup;
static EVT_WDF_DRIVER_DEVICE_ADD sample_evt_driver_device_add;
static EVT_WDF_OBJECT_CONTEXT_CLEANUP sample_device_evt_cleanup;
static EVT_WDF_DEVICE_D0_ENTRY sample_evt_device_d0_entry;
static EVT_WDF_DEVICE_D0_EXIT sample_evt_device_d0_exit;

NTSTATUS DriverEntry(
    IN OUT PDRIVER_OBJECT driver_object,
    IN PUNICODE_STRING registry_path)
{
    WDF_DRIVER_CONFIG config;
    WDF_OBJECT_ATTRIBUTES attribs;
    NTSTATUS status;
    int ret;

    WPP_INIT_TRACING(driver_object, registry_path);

    WDF_DRIVER_CONFIG_INIT(&config, sample_evt_driver_device_add);
    WDF_OBJECT_ATTRIBUTES_INIT(&attribs);
    attribs.EvtCleanupCallback = sample_driver_cleanup;

    status = WdfDriverCreate(driver_object, registry_path, &attribs, &config, WDF_NO_HANDLE);
    if (!NT_SUCCESS(status)) {
        WPP_CLEANUP(driver_object);
    }

    /*
     * Register the function driver.
     *
     * The driver won't be probed until we add a matching function in
     * sample_evt_d0_entry.
     */
    ret = sample_register();
    if (ret < 0) {
        return os_err_to_ntstatus(ret);
    }

    return STATUS_SUCCESS;
}

static void sample_driver_cleanup(
    IN WDFDRIVER driver)
{
    /*
     * Unregister the function driver.
     *
     * All functions will have previously been removed when the
     * entered the D3Final power state (see sample_evt_d0_exit).
     */
    sample_unregister();

    WPP_CLEANUP(WdfDriverWdmGetDriverObject(driver));
}

NTSTATUS sample_evt_driver_device_add(
    IN WDFDRIVER driver,
    IN PWDFDEVICE_INIT device_init)
{
    WDF_PNPPOWER_EVENT_CALLBACKS pnp_power_callbacks;
    WDF_OBJECT_ATTRIBUTES attribs;
    WDFDEVICE device;
    struct sample_device_context *dev_ctx;
    NTSTATUS status;

    WdfDeviceInitSetIoType(device_init, WdfDeviceIoDirect);

    WDF_PNPPOWER_EVENT_CALLBACKS_INIT(&pnp_power_callbacks);
    pnp_power_callbacks.EvtDeviceD0Entry = sample_evt_device_d0_entry;
    pnp_power_callbacks.EvtDeviceD0Exit = sample_evt_device_d0_exit;
    WdfDeviceInitSetPnpPowerEventCallbacks(device_init, &pnp_power_callbacks);

    WDF_OBJECT_ATTRIBUTES_INIT_CONTEXT_TYPE(&attribs, SAMPLE_DEVICE_CONTEXT);
    attribs.EvtCleanupCallback = sample_device_evt_cleanup;

    status = WdfDeviceCreate(&device_init, &attribs, &device);
    if (!NT_SUCCESS(status)) {
        return status;
    }

    dev_ctx = sample_get_device_context(device);

    /*
     * Initialize the Synergy SDIO interface for this device.
     */
    status = CsrSdioWdfDeviceInit(device);
    if (!NT_SUCCESS(status)) {
        os_trace_err("function interface initialization failed: %!STATUS!", status);
        return status;
    }

    return STATUS_SUCCESS;
}

VOID sample_device_evt_cleanup(
    IN WDFOBJECT object)
{
    WDFDEVICE device = object;
    struct sample_device_context *dev_ctx = sample_get_device_context(device);

    /*
     * Cleanup the Synergy SDIO interface for this device.
     */
    CsrSdioWdfDeviceCleanup(device);
}

NTSTATUS sample_evt_device_d0_entry(
    IN WDFDEVICE device,
    IN WDF_POWER_DEVICE_STATE previous_state)
{
    struct sample_device_context *dev_ctx = sample_get_device_context(device);

    os_trace_dbg("D0 entry");

    /*
     * Add the function the first time the D0 state is entered.
     *
     * On subsequent calls the function will have been suspended so
     * resume it.
     */
    if (!dev_ctx->added) {
        CsrSdioWdfDeviceAdd(device);
        dev_ctx->added = true;
    } else {
        CsrSdioWdfDeviceResume(device);
    }

    return STATUS_SUCCESS;
}

NTSTATUS sample_evt_device_d0_exit(
    IN WDFDEVICE device,
    IN WDF_POWER_DEVICE_STATE target_state)
{
    struct sample_device_context *dev_ctx = sample_get_device_context(device);

    os_trace_dbg("D0 exit");

    /*
     * Suspend the function unless the D3Final state is being entered.
     */
    switch (target_state) {
    case WdfPowerDeviceD3Final:
        CsrSdioWdfDeviceDel(device);
        dev_ctx->added = false; 
        break;
    default:
        CsrSdioWdfDeviceSuspend(device);
        break;
    }

    return STATUS_SUCCESS;
}
