/*
 * Windows KMDF SDIO function child device handling.
 *
 * Copyright (C) 2009-2010 Cambridge Silicon Radio Ltd.
 *
 * Refer to LICENSE.txt included with this source code for details on
 * the license terms.
 */
#include <oska/all.h>
#include <ntstrsafe.h>

#include <sdioemb/slot_api.h>
#include <sdioemb/windows/slot_api.h>
#include <sdioemb/windows/sdio_api.h>
#include <sdioemb/windows/guid.h>
#include "../../../core/sdio_internal.h"
#include "wdf_slot_internal.h"

#include "children.tmh"

struct sdioemb_wdf_child_id {
    WDF_CHILD_IDENTIFICATION_DESCRIPTION_HEADER header;
    int slot_id;
    uint16_t manf_id;
    uint16_t card_id;
    int function;
};

static DECLARE_CONST_UNICODE_STRING(default_function_name, L"SDIO Function");

static EVT_WDF_CHILD_LIST_CREATE_DEVICE sdioemb_wdf_evt_child_list_create_device;

static bool sdioemb_wdf_enable_uif(struct sdioemb_slot *slot)
{
    static DECLARE_CONST_UNICODE_STRING(enable_uif_value_name, L"EnableUif");
    struct sdioemb_wdf_slot_data *wdf_data = slot->drv_data;
    WDFDEVICE device = wdf_data->device;
    ULONG enable_uif;
    NTSTATUS status;

    status = sdioemb_wdf_query_ulong(device, &enable_uif_value_name, &enable_uif);
    if (!NT_SUCCESS(status)) {
        return false;
    }
    return !!enable_uif;
}

static void sdioemb_wdf_child_id_init(struct sdioemb_wdf_child_id *id, struct sdioemb_dev *fdev)
{
    WDF_CHILD_IDENTIFICATION_DESCRIPTION_HEADER_INIT(&id->header, sizeof(*id));

    id->slot_id   = fdev->slot_id;
    id->manf_id   = fdev->vendor_id;
    id->card_id   = fdev->device_id;
    id->function  = fdev->function;
}

static int sdioemb_wdf_add_function(struct sdioemb_slot *slot, struct sdioemb_dev *fdev)
{
    struct sdioemb_wdf_slot_data *wdf_data = slot->drv_data;
    struct sdioemb_wdf_child_id id;
    NTSTATUS status;

    if (fdev->function == SDIOEMB_UIF_FUNC && !sdioemb_wdf_enable_uif(slot)) {
        return SDIOEMB_ADD_FUNC_DEFERRED;
    }

    sdioemb_wdf_child_id_init(&id, fdev);

    status = WdfChildListAddOrUpdateChildDescriptionAsPresent(wdf_data->child_list,
                                                              &id.header, NULL);
    if (!NT_SUCCESS(status)) {
        return os_ntstatus_to_err(status);
    }

    return SDIOEMB_ADD_FUNC_DEFERRED;
}

static void sdioemb_wdf_del_function(struct sdioemb_slot *slot, struct sdioemb_dev *fdev)
{
    struct sdioemb_wdf_slot_data *wdf_data = slot->drv_data;
    struct sdioemb_wdf_child_id id;
    NTSTATUS status;

    sdioemb_wdf_child_id_init(&id, fdev);

    WdfChildListUpdateChildDescriptionAsMissing(wdf_data->child_list, &id.header);
}

static void sdioemb_wdf_interface_ref(void *context)
{
    struct sdioemb_dev *fdev = context;

    sdioemb_get_function(fdev);
}

static void sdioemb_wdf_interface_deref(void *context)
{
    struct sdioemb_dev *fdev = context;

    sdioemb_put_function(fdev);
}

static NTSTATUS sdioemb_wdf_evt_child_list_create_device(
    WDFCHILDLIST child_list,
    PWDF_CHILD_IDENTIFICATION_DESCRIPTION_HEADER id_header,
    PWDFDEVICE_INIT child_init)
{
    WDFDEVICE device = WdfChildListGetDevice(child_list);
    struct sdioemb_wdf_child_list_context *cl_ctx = sdioemb_wdf_get_child_list_context(child_list);
    struct sdioemb_slot *slot = cl_ctx->slot;
    struct sdioemb_wdf_child_id *id = (struct sdioemb_wdf_child_id *)id_header;
    struct sdioemb_dev *fdev;
    WDF_PDO_EVENT_CALLBACKS  child_callbacks;
    DECLARE_UNICODE_STRING_SIZE(str, 40);
    WDFDEVICE child;
    WDF_DEVICE_PNP_CAPABILITIES pnp_caps;
    WDF_DEVICE_POWER_CAPABILITIES power_caps;
    WDF_QUERY_INTERFACE_CONFIG interface_config;
    struct sdioemb_wdf_interface interface;
    NTSTATUS status;

    fdev = slot->priv->functions[id->function];

    WdfDeviceInitSetDeviceType(child_init, FILE_DEVICE_BUS_EXTENDER);

    WDF_PDO_EVENT_CALLBACKS_INIT(&child_callbacks);
    /* FIXME: need implementations for these. */
    //child_callbacks.EvtDeviceEnableWakeAtBus = ???;
    //child_callbacks.EvtDeviceDisableWakeAtBus = ???;
    WdfPdoInitSetEventCallbacks(child_init, &child_callbacks);

    RtlUnicodeStringPrintf(&str, L"SDIOEMB\\MANF_%04x&CARD_%04x&FUNC_%02x",
                           fdev->vendor_id, fdev->device_id, fdev->function);
    status = WdfPdoInitAssignDeviceID(child_init, &str);
    if (!NT_SUCCESS(status)) {
        return status;
    }
    status = WdfPdoInitAddHardwareID(child_init, &str);
    if (!NT_SUCCESS(status)) {
        return status;
    }
    RtlUnicodeStringPrintf(&str, L"%02x", fdev->function);
    status = WdfPdoInitAssignInstanceID(child_init, &str);
    if (!NT_SUCCESS(status)) {
        return status;
    }
    /* Compatible IDs for the standard interface (if it has one). */
    if (fdev->interface) {
        RtlUnicodeStringPrintf(&str, L"SDIOEMB\\MANF_%04x&CARD_%04x&IFACE_%02x",
                               fdev->vendor_id, fdev->device_id, fdev->interface);
        status = WdfPdoInitAddCompatibleID(child_init, &str);
        if (!NT_SUCCESS(status)) {
            return status;
        }
        RtlUnicodeStringPrintf(&str, L"SDIOEMB\\MANF_%04x&IFACE_%02x",
                               fdev->vendor_id, fdev->interface);
        status = WdfPdoInitAddCompatibleID(child_init, &str);
        if (!NT_SUCCESS(status)) {
            return status;
        }
        RtlUnicodeStringPrintf(&str, L"SDIOEMB\\IFACE_%02x",
                               fdev->interface);
        status = WdfPdoInitAddCompatibleID(child_init, &str);
        if (!NT_SUCCESS(status)) {
            return status;
        }
    }
    /* Compatible IDs for the UIF function. */
    if (fdev->function == SDIOEMB_UIF_FUNC) {
        RtlUnicodeStringPrintf(&str, L"SDIOEMB\\UIF");
        status = WdfPdoInitAddCompatibleID(child_init, &str);
        if (!NT_SUCCESS(status)) {
            return status;
        }
    }        
  
    RtlUnicodeStringPrintf(&str, L"Sdioemb slot %d, function %d", fdev->slot_id, fdev->function);
    status = WdfPdoInitAddDeviceText(child_init, &default_function_name, &str, 0x409);
    if (!NT_SUCCESS(status)) {
        return status;
    }
    WdfPdoInitSetDefaultLocale(child_init, 0x409);

    status = WdfDeviceCreate(&child_init, WDF_NO_OBJECT_ATTRIBUTES, &child);
    if (!NT_SUCCESS(status)) {
        return status;
    }

    WDF_DEVICE_PNP_CAPABILITIES_INIT(&pnp_caps);
    pnp_caps.SurpriseRemovalOK = WdfTrue;
    WdfDeviceSetPnpCapabilities(child, &pnp_caps);

    WDF_DEVICE_POWER_CAPABILITIES_INIT(&power_caps);
    /* FIXME: revist these capabiltities. */
    power_caps.DeviceD1 = WdfFalse;
    power_caps.DeviceD2 = WdfFalse;
    WdfDeviceSetPowerCapabilities(child, &power_caps);

    RtlZeroMemory(&interface, sizeof(interface));
    interface.header.Size    = sizeof(interface);
    interface.header.Version = 1;
    interface.header.Context = NULL;
    interface.header.InterfaceReference   = sdioemb_wdf_interface_ref;
    interface.header.InterfaceDereference = sdioemb_wdf_interface_deref;
    interface.header.Context = fdev;
    interface.fdev = fdev;

    interface.driver_probe          = sdioemb_driver_probe;
    interface.driver_remove         = sdioemb_driver_remove;
    interface.enable_function       = sdioemb_enable_function;
    interface.disable_function      = sdioemb_disable_function;
    interface.reenable_csr_function = sdioemb_reenable_csr_function;
    interface.idle_function         = sdioemb_idle_function;
    interface.read8                 = sdioemb_read8;
    interface.read16                = sdioemb_read16;
    interface.write8                = sdioemb_write8;
    interface.write16               = sdioemb_write16;
    interface.f0_read8              = sdioemb_f0_read8;
    interface.f0_write8             = sdioemb_f0_write8;
    interface.read                  = sdioemb_read;
    interface.write                 = sdioemb_write;
    interface.hard_reset            = sdioemb_hard_reset;
    interface.power_on              = sdioemb_power_on;
    interface.power_off             = sdioemb_power_off;
    interface.interrupt_enable      = sdioemb_interrupt_enable;
    interface.interrupt_disable     = sdioemb_interrupt_disable;
    interface.interrupt_acknowledge = sdioemb_interrupt_acknowledge;
    interface.suspend_function      = sdioemb_suspend_function;
    interface.resume_function       = sdioemb_resume_function;

    WDF_QUERY_INTERFACE_CONFIG_INIT(&interface_config,
                                    &interface.header,
                                    &GUID_DRVINTERFACE_SDIOEMB_FUNC,
                                    NULL);

    WdfDeviceAddQueryInterface(child, &interface_config);

    return STATUS_SUCCESS;
}

/**
 * Create and initialize a slot's child list.
 *
 * This must be called before registering the slot with
 * sdioemb_wdf_slot_register().
 *
 * @param device the FDO for the slot.
 * @param slot   the slot.
 *
 * @return An NTSTATUS value for the status of the child list creation
 * (success or failure).
 *
 * @ingroup windows
 */
NTSTATUS sdioemb_wdf_create_child_list(WDFDEVICE device, struct sdioemb_slot *slot)
{
    struct sdioemb_wdf_slot_data *wdf_data = slot->drv_data;
    WDF_CHILD_LIST_CONFIG config;
    WDF_OBJECT_ATTRIBUTES attribs;
    struct sdioemb_wdf_child_list_context *cl_ctx;
    NTSTATUS status;

    wdf_data->device = device;

    WDF_CHILD_LIST_CONFIG_INIT(&config, sizeof(struct sdioemb_wdf_child_id),
                               sdioemb_wdf_evt_child_list_create_device);
    WDF_OBJECT_ATTRIBUTES_INIT_CONTEXT_TYPE(&attribs, SDIOEMB_WDF_CHILD_LIST_CONTEXT);

    status = WdfChildListCreate(device, &config, &attribs, &wdf_data->child_list);
    if (!NT_SUCCESS(status)) {
        return status;
    }

    cl_ctx = sdioemb_wdf_get_child_list_context(wdf_data->child_list);
    cl_ctx->slot = slot;

    return STATUS_SUCCESS;
}

/**
 * Destroy a slot's child list.
 *
 * The slot must not be registered.
 *
 * @param slot the slot.
 *
 * @ingroup windows
 */
void sdioemb_wdf_destroy_child_list(struct sdioemb_slot *slot)
{
    struct sdioemb_wdf_slot_data *wdf_data = slot->drv_data;

    if (wdf_data->child_list) {
        WdfObjectDelete(wdf_data->child_list);
    }
}

/**
 * Register a WDF slot driver.
 *
 * This sets the add and delete function notifiers to create the
 * correct WDF child devices.
 *
 * @param slot the slot to be registered.
 *
 * @return 0 on success.
 * @return -ve error code on failure.
 *
 * @ingroup windows
 */
int sdioemb_wdf_slot_register(struct sdioemb_slot *slot)
{
    int ret;

    slot->notifs.add_function = sdioemb_wdf_add_function;
    slot->notifs.del_function = sdioemb_wdf_del_function;

    ret = sdioemb_slot_register(slot);
    if (ret < 0) {
        return ret;
    }
    sdioemb_wdf_event_log_add_slot(slot);
    return 0;
}

/**
 * Unregister a WDF slot driver.
 *
 * This unregisters the slot driver previously registered with
 * sdioemb_wdf_slot_register().
 *
 * @param slot the slot to be unregistered.
 *
 * @ingroup windows
 */
void sdioemb_wdf_slot_unregister(struct sdioemb_slot *slot)
{
    sdioemb_wdf_event_log_del_slot(slot);
    sdioemb_slot_unregister(slot);
}
