/*
 * Windows KMDF SDIO event logging.
 *
 * Copyright (C) 2010 Cambridge Silicon Radio Ltd.
 *
 * Refer to LICENSE.txt included with this source code for details on
 * the license terms.
 */
#include <oska/all.h>
#include <ntstrsafe.h>

#include <sdioemb/slot_api.h>
#include <sdioemb/cspi.h>
#include <sdioemb/windows/slot_api.h>
#include <sdioemb/windows/guid.h>
#include "../../../core/sdio_internal.h"
#include "../../../core/sdio_event_log.h"
#include "wdf_slot_internal.h"

#include "event_log.tmh"

typedef struct sdioemb_wdf_file_context {
    struct sdioemb_slot *slot;
    unsigned index;
} SDIOEMB_WDF_FILE_CONTEXT, *PSDIOEMB_WDF_FILE_CONTEXT;
WDF_DECLARE_CONTEXT_TYPE_WITH_NAME(SDIOEMB_WDF_FILE_CONTEXT, sdioemb_wdf_get_file_context);

static EVT_WDF_DEVICE_FILE_CREATE sdioemb_wdf_evt_file_create;
static EVT_WDF_IO_QUEUE_IO_READ sdioemb_wdf_evt_io_read;

static struct sdioemb_slot *sdioemb_wdf_event_log_find_slot(WDFDEVICE device,
                                                            UNICODE_STRING *file_name);

static void safe_printf(char **str, size_t *len, const char *format, ...)
{
    va_list args;
    size_t remaining_space, printed;

    va_start(args, format);
    RtlStringCbVPrintfExA(*str, *len, NULL, &remaining_space, 0, format, args);
    va_end(format);

    printed = *len - remaining_space;
    *str += printed;
    *len -= printed;
}

static size_t fill_buffer_from_entry(char *s, size_t len, struct sdio_event_entry *e)
{
    size_t l = len;

    safe_printf(&s, &l, "%6u[%7lu.%03lu]: ",
               e->seq_num, e->time_ms / 1000, e->time_ms % 1000);

    switch (e->type) {
    case SDIOEMB_EVENT_CMD:
        if (e->cmd.flags & SDIOEMB_CMD_FLAG_CSPI) {
            safe_printf(&s, &l, "CSPI cmd:%02x addr: %06x %s: %04x status: %02x resp: %02x\n",
                       e->cmd.cspi.cmd, e->cmd.cspi.addr,
                       e->cmd.cspi.cmd & CSPI_BURST ? "len" : "val", e->cmd.cspi.val,
                       e->cmd.status, e->cmd.cspi.response);
        } else {
            safe_printf(&s, &l, "CMD%02u arg: %08x stat: %02x resp: %08x\n",
                       e->cmd.sdio.cmd, e->cmd.sdio.arg, e->cmd.status, e->cmd.sdio.response.r1);
        }
#if SDIOEMB_DEBUG_DATA_BUFFER_LEN > 0
        if (e->cmd.data.total_length > 0) {
            unsigned int i;

            for (i = 0; i < e->cmd.data.length; i++) {
                if ((i % 16) == 0) {
                    safe_printf(&s, &l, "                     0x%04X:", i);
                }

                safe_printf(&s, &l, " %02X", sdio_event_log_get_data(e, i));

                if ((i % 16) == 15) {
                    safe_printf(&s, &l, "\n");
                }
            }
            if ((i % 16) != 0) {
                safe_printf(&s, &l, "\n");
            }
            if (e->cmd.data.total_length > e->cmd.data.length) {
                safe_printf(&s, &l, "                     ...\n");
            }
        }
#endif
        break;
    case SDIOEMB_EVENT_CARD_INT:
        safe_printf(&s, &l, "card interrupt: %02x\n", e->functions);
        break;
    case SDIOEMB_EVENT_INT_ACKED:
        safe_printf(&s, &l, "card interrupt acked: %02x\n", e->functions);
        break;
    case SDIOEMB_EVENT_POWER_ON:
        safe_printf(&s, &l, "power on\n");
        break;
    case SDIOEMB_EVENT_POWER_OFF:
        safe_printf(&s, &l, "power off\n");
        break;
    case SDIOEMB_EVENT_RESET:
        safe_printf(&s, &l, "hard reset\n");
        break;
    case SDIOEMB_EVENT_CLOCK_FREQ:
        safe_printf(&s, &l, "clock frequency request: %d kHz\n", e->val / 1000);
        break;
    }

    return len - l + 1;
}

static VOID sdioemb_wdf_evt_file_create(
    IN WDFDEVICE device,
    IN WDFREQUEST request,
    IN WDFFILEOBJECT file)
{
    struct sdioemb_wdf_file_context *fc = sdioemb_wdf_get_file_context(file);
    UNICODE_STRING *file_name;
    struct sdioemb_slot *slot;
    NTSTATUS status;

    file_name = WdfFileObjectGetFileName(file);

    slot = sdioemb_wdf_event_log_find_slot(device, file_name);
    if (slot == NULL) {
        status = STATUS_NOT_FOUND;
    } else {
        status = STATUS_SUCCESS;
        fc->slot = slot;
        fc->index = 0;
    }
    WdfRequestComplete(request, status);
}

static VOID sdioemb_wdf_evt_io_read(
    IN WDFQUEUE queue,
    IN WDFREQUEST request,
    size_t length)
{
    struct sdioemb_wdf_file_context *fc;
    struct sdio_event_entry *e;
    char *s;
    size_t l, written = 0;
    NTSTATUS status;

    fc = sdioemb_wdf_get_file_context(WdfRequestGetFileObject(request));

    e = sdio_event_log_get_entry(fc->slot, fc->index);
    if (e == NULL) {
        status = STATUS_END_OF_FILE;
    } else {
        status = WdfRequestRetrieveOutputBuffer(request, 0, &s, &l);
        if (NT_SUCCESS(status)) {
            written = fill_buffer_from_entry(s, l, e);
            fc->index++;
        }
    }
    WdfRequestCompleteWithInformation(request, status, written);
}

/**
 * Initialize a WDF device for event logging.
 *
 * @param device_init the device initialization information.
 *
 * @ingroup windows
 */
void sdioemb_wdf_event_log_init(PWDFDEVICE_INIT device_init)
{
    WDF_FILEOBJECT_CONFIG file_config;
    WDF_OBJECT_ATTRIBUTES attribs;

    WDF_FILEOBJECT_CONFIG_INIT(&file_config, sdioemb_wdf_evt_file_create, NULL, NULL);
    WDF_OBJECT_ATTRIBUTES_INIT_CONTEXT_TYPE(&attribs, SDIOEMB_WDF_FILE_CONTEXT);
    WdfDeviceInitSetFileObjectConfig(device_init, &file_config, &attribs);
}

/**
 * Create the event logging infrastructure for a WDF device.
 *
 * This creates the I/O queues etc. for the event logging for a WDF
 * device with one or more slots.
 *
 * @param device the WDF device.
 *
 * @return STATUS_SUCCESS on successfully creating the I/O queues.
 * @return A error status otherwise.
 *
 * @ingroup windows
 */
NTSTATUS sdioemb_wdf_event_log_create(WDFDEVICE device)
{
    struct sdioemb_wdf_event_log_info *eli;
    WDF_OBJECT_ATTRIBUTES attribs;
    WDF_IO_QUEUE_CONFIG io_queue_config;
    WDFQUEUE queue;
    NTSTATUS status;

    WDF_OBJECT_ATTRIBUTES_INIT_CONTEXT_TYPE(&attribs, SDIOEMB_WDF_EVENT_LOG_INFO);
    status = WdfObjectAllocateContext(device, &attribs, &eli);
    if (!NT_SUCCESS(status)) {
        return status;
    }

    WDF_IO_QUEUE_CONFIG_INIT_DEFAULT_QUEUE(&io_queue_config, WdfIoQueueDispatchSequential);
    io_queue_config.EvtIoRead = sdioemb_wdf_evt_io_read;
    status = WdfIoQueueCreate(device, &io_queue_config, WDF_NO_OBJECT_ATTRIBUTES, &queue);
    if (!NT_SUCCESS(status)) {
        return status;
    }

    return STATUS_SUCCESS;
}

void sdioemb_wdf_event_log_add_slot(struct sdioemb_slot *slot)
{
    struct sdioemb_wdf_slot_data *wdf_data = slot->drv_data;
    WDFDEVICE device = wdf_data->device;
    struct sdioemb_wdf_event_log_info *eli = sdioemb_wdf_get_event_log_info(device);
    DECLARE_UNICODE_STRING_SIZE(ref_str, 16);
    NTSTATUS status;

    RtlUnicodeStringPrintf(&ref_str, L"%d", slot->priv->id);
    status = WdfDeviceCreateDeviceInterface(device, &GUID_DEVINTERFACE_SDIOEMB_EVENT_LOG, &ref_str);
    if (!NT_SUCCESS(status)) {
        return;
    }
    WdfDeviceSetDeviceInterfaceState(device, &GUID_DEVINTERFACE_SDIOEMB_EVENT_LOG,
                                     &ref_str, TRUE);
    eli->slots[slot->priv->id] = slot;
}

void sdioemb_wdf_event_log_del_slot(struct sdioemb_slot *slot)
{
    struct sdioemb_wdf_slot_data *wdf_data = slot->drv_data;
    WDFDEVICE device = wdf_data->device;
    struct sdioemb_wdf_event_log_info *eli = sdioemb_wdf_get_event_log_info(device);
    DECLARE_UNICODE_STRING_SIZE(ref_str, 16);

    RtlUnicodeStringPrintf(&ref_str, L"%d", slot->priv->id);
    WdfDeviceSetDeviceInterfaceState(device, &GUID_DEVINTERFACE_SDIOEMB_EVENT_LOG,
                                     &ref_str, FALSE);
    eli->slots[slot->priv->id] = NULL;
}

static struct sdioemb_slot *sdioemb_wdf_event_log_find_slot(WDFDEVICE device,
                                                            UNICODE_STRING *file_name)
{
    struct sdioemb_wdf_event_log_info *eli = sdioemb_wdf_get_event_log_info(device);
    UNICODE_STRING name;
    ULONG id;
    NTSTATUS status;

    /* Remove leading '\'. */
    name = *file_name;
    if (name.Length > 0) {
        name.Buffer++;
        name.Length -= sizeof(*name.Buffer);
        name.MaximumLength -= sizeof(*name.Buffer);
    }

    status = RtlUnicodeStringToInteger(&name, 0, &id);
    if (!NT_SUCCESS(status)) {
        return NULL;
    }
    if (id >= SDIOEMB_SLOTS_MAX) {
        return NULL;
    }

    return eli->slots[id];
}
