/*
 * Sdioemb Windows KMDF SHC driver.
 *
 * Copyright (C) 2009 Cambridge Silicon Radio Ltd.
 *
 * Refer to LICENSE.txt included with this source code for details on
 * the license terms.
 */
#define WPP_CONTROL_GUIDS                                               \
    WPP_DEFINE_CONTROL_GUID(ShcTraceGuid,                               \
                            (567ef66f,7a84,4e10,ba48,2d5e3edd37ee),     \
                            WPP_DEFINE_BIT(OS_TRACE_DFLT)               \
        )
#include <initguid.h>
#include <wdmguid.h>
#include <sdioemb/windows/guid.h>

#include "shc_internal.h"

#include "driver.tmh"

DRIVER_INITIALIZE DriverEntry;

static EVT_WDF_OBJECT_CONTEXT_CLEANUP shc_driver_cleanup;
static EVT_WDF_OBJECT_CONTEXT_CLEANUP shc_device_evt_cleanup;
static EVT_WDF_DRIVER_DEVICE_ADD shc_evt_driver_device_add;
static EVT_WDF_DEVICE_PREPARE_HARDWARE shc_evt_device_prepare_hardware;
static EVT_WDF_DEVICE_RELEASE_HARDWARE shc_evt_device_release_hardware;
static EVT_WDF_DEVICE_D0_ENTRY shc_evt_device_d0_entry;
static EVT_WDF_DEVICE_D0_EXIT shc_evt_device_d0_exit;

NTSTATUS DriverEntry(
    IN OUT PDRIVER_OBJECT driver_object,
    IN PUNICODE_STRING registry_path)
{
    WDF_DRIVER_CONFIG config;
    WDF_OBJECT_ATTRIBUTES attribs;
    NTSTATUS status;

    WPP_INIT_TRACING(driver_object, registry_path);

    WDF_DRIVER_CONFIG_INIT(&config, shc_evt_driver_device_add);
    WDF_OBJECT_ATTRIBUTES_INIT(&attribs);
    attribs.EvtCleanupCallback = shc_driver_cleanup;

    status = WdfDriverCreate(driver_object, registry_path, &attribs, &config, WDF_NO_HANDLE);
    if (!NT_SUCCESS(status)) {
        WPP_CLEANUP(driver_object);
    }

    return status;
}

static void shc_driver_cleanup(
    IN WDFDRIVER driver)
{
    WPP_CLEANUP(WdfDriverWdmGetDriverObject(driver));
}

static NTSTATUS shc_evt_driver_device_add(
    IN WDFDRIVER driver,
    IN PWDFDEVICE_INIT device_init)
{
    WDF_PNPPOWER_EVENT_CALLBACKS pnp_power_callbacks;
    WDF_OBJECT_ATTRIBUTES attribs;
    WDFDEVICE device;
    struct shc_device_context *ctx;
    uint8_t num_slots;
    int i;
    NTSTATUS status;

    WdfDeviceInitSetIoType(device_init, WdfDeviceIoBuffered);

    WDF_PNPPOWER_EVENT_CALLBACKS_INIT(&pnp_power_callbacks);
    pnp_power_callbacks.EvtDevicePrepareHardware = shc_evt_device_prepare_hardware;
    pnp_power_callbacks.EvtDeviceReleaseHardware = shc_evt_device_release_hardware;
    pnp_power_callbacks.EvtDeviceD0Entry = shc_evt_device_d0_entry;
    pnp_power_callbacks.EvtDeviceD0Exit = shc_evt_device_d0_exit;
    WdfDeviceInitSetPnpPowerEventCallbacks(device_init, &pnp_power_callbacks);

    WDF_OBJECT_ATTRIBUTES_INIT(&attribs);
    WDF_OBJECT_ATTRIBUTES_INIT_CONTEXT_TYPE(&attribs, SHC_DEVICE_CONTEXT);
    attribs.EvtCleanupCallback = shc_device_evt_cleanup;

    sdioemb_wdf_event_log_init(device_init);

    status = WdfDeviceCreate(&device_init, &attribs, &device);
    if (!NT_SUCCESS(status)) {
        return status;
    }

    status = sdioemb_wdf_event_log_create(device);
    if (!NT_SUCCESS(status)) {
        return status;
    }

    ctx = shc_get_device_context(device);

    status = shc_registry_init(device);
    if (!NT_SUCCESS(status)) {
        return status;
    }

    status = shc_pci_config_init(device);
    if (!NT_SUCCESS(status)) {
        return status;
    }

    status = shc_interrupt_init(device);
    if (!NT_SUCCESS(status)) {
        return status;
    }

    status = shc_dma_init(device);
    if (!NT_SUCCESS(status)) {
        return status;
    }

    shc_pci_config_read8(device, PCI_SHC_SLOT_INFO, &num_slots);
    num_slots = ((num_slots >> 4) & 0x07) + 1;

    ctx->slots = os_alloc(num_slots * sizeof(*ctx->slots));
    if (ctx->slots == NULL) {
        return STATUS_INSUFFICIENT_RESOURCES;
    }
    ctx->num_slots = num_slots;

    for (i = 0; i < ctx->num_slots; i++) {
        status = shc_create_slot(device, &ctx->slots[i], i);
        if (ctx->slots[i] == NULL) {
            return status;
        }
    }

    return STATUS_SUCCESS;
}

static VOID shc_device_evt_cleanup(
    IN WDFOBJECT object)
{
    WDFDEVICE device = object;
    struct shc_device_context *ctx = shc_get_device_context(device);
    int i;

    for (i = 0; i < ctx->num_slots; i++) {
        shc_destroy_slot(object, ctx->slots[i]);
    }
    os_free(ctx->slots);

    shc_pci_config_cleanup(device);
    shc_registry_cleanup(device);
}

static NTSTATUS map_io_space(struct sdioemb_slot *slot, PCM_PARTIAL_RESOURCE_DESCRIPTOR bar)
{
    struct shc_data *shc = slot->drv_data;

    shc->shc.addr = MmMapIoSpace(bar->u.Memory.Start, bar->u.Memory.Length, MmNonCached);
    if (shc->shc.addr == NULL) {
        return STATUS_INSUFFICIENT_RESOURCES;
    }
    shc->addr_len = bar->u.Memory.Length;
    return STATUS_SUCCESS;
}

NTSTATUS shc_evt_device_prepare_hardware(
    WDFDEVICE device,
    WDFCMRESLIST resources,
    WDFCMRESLIST resources_translated)
{
    struct shc_device_context *ctx = shc_get_device_context(device);
    PCM_PARTIAL_RESOURCE_DESCRIPTOR descriptor;
    int s;
    unsigned i;
    NTSTATUS status;

    for (i = 0, s = 0; i < WdfCmResourceListGetCount(resources_translated)
             && s < ctx->num_slots; i++) {
        descriptor = WdfCmResourceListGetDescriptor(resources_translated, i);
        switch (descriptor->Type) {
        case CmResourceTypeMemory:
            status = map_io_space(ctx->slots[s++], descriptor); 
            if (!NT_SUCCESS(status)) {
                return status;
            }
            break;
        }
    }

    if (s != ctx->num_slots) {
        return STATUS_DEVICE_CONFIGURATION_ERROR;
    }

    return STATUS_SUCCESS;
}

NTSTATUS shc_evt_device_release_hardware(
    WDFDEVICE device,
    WDFCMRESLIST resources_translated)
{
    struct shc_device_context *ctx = shc_get_device_context(device);
    int i;

    for (i = 0; i < ctx->num_slots; i++) {
        struct sdioemb_slot *slot;
        struct shc_data *shc;

        slot = ctx->slots[i];
        shc = slot->drv_data;

        if (shc->shc.addr) {
            MmUnmapIoSpace(shc->shc.addr, shc->addr_len);
        }
    }

    return STATUS_SUCCESS;
}

NTSTATUS shc_evt_device_d0_entry(
    IN WDFDEVICE device,
    IN WDF_POWER_DEVICE_STATE previous_state)
{
    struct shc_device_context *ctx = shc_get_device_context(device);
    int i;
    int ret;

    /* JMicron 385 requires various PCI config space writes to work
     * properly. */
    if (ctx->quirks & SHC_QUIRK_JMICRON_JMB38X) {
        uint8_t s;

        /* The device will only allow a 12.5MHz SDIO clock unless bit 0
           at 0xAD is clear */
        shc_pci_config_read8(device, 0xAD, &s);
        s &= ~(0x01);
        shc_pci_config_write8(device, 0xAD, s);

        /* The device does not meet high speed timing unless the delay
           register is set to 0x03.  This should be the default
           according to JMicron, but it is 0x00 on this system. */
        shc_pci_config_read8(device, 0xB1, &s);
        s &= 0xf0;
        s |= 0x03;
        shc_pci_config_write8(device, 0xB1, s);
    }

    for (i = 0; i < ctx->num_slots; i++) {
        struct sdioemb_slot *slot;
        struct shc_data *shc;

        slot = ctx->slots[i];
        shc = slot->drv_data;

        sdioemb_shc_start(&shc->shc);
        if (!ctx->registered) {
            sdioemb_wdf_slot_register(slot);
        }
    }
    ctx->registered = true;

    return STATUS_SUCCESS;
}

NTSTATUS shc_evt_device_d0_exit(
    IN WDFDEVICE device,
    IN WDF_POWER_DEVICE_STATE target_state)
{
    struct shc_device_context *ctx = shc_get_device_context(device);
    int i;

    for (i = 0; i < ctx->num_slots; i++) {
        struct sdioemb_slot *slot;
        struct shc_data *shc;

        slot = ctx->slots[i];
        shc = slot->drv_data;

        switch (target_state) {
        case WdfPowerDeviceD3Final:
            sdioemb_wdf_slot_unregister(slot);
            /* fall through */
        default:
            sdioemb_shc_stop(&shc->shc);
            break;
        }
    }

    return STATUS_SUCCESS;
}
