/*
 * Windows KMDF generic PCI transport driver -- registry access.
 *
 * Copyright (C) 2009 Cambridge Silicon Radio Ltd.
 *
 * Refer to LICENSE.txt included with this source code for details on
 * the license terms.
 */
#include <intsafe.h>
#include "shc_internal.h"

#include "registry.tmh"

static const char *unknown_name = "(unknown)";

static const char *shc_get_name(WDFDEVICE device);
static void shc_free_name(WDFDEVICE device);
static NTSTATUS shc_get_quirks(WDFDEVICE device, ULONG *quirks);

NTSTATUS shc_registry_init(WDFDEVICE device)
{
    struct shc_device_context *dev_ctx = shc_get_device_context(device);
    NTSTATUS status;

    dev_ctx->name = shc_get_name(device);

    status = shc_get_quirks(device, &dev_ctx->quirks);
    if (!NT_SUCCESS(status)) {
        os_trace_warn("%s: assuming hardware has no quirks", dev_ctx->name);
        dev_ctx->quirks = 0;
    }

    return STATUS_SUCCESS;
}

void shc_registry_cleanup(WDFDEVICE device)
{
    shc_free_name(device);
}

static const char *shc_get_name(WDFDEVICE device)
{
    struct shc_device_context *dev_ctx = shc_get_device_context(device);
    WCHAR *buf;
    size_t len;
    UNICODE_STRING str;
    UCHAR *ansi_buf;
    ULONG ansi_len;
    ANSI_STRING ansi_str;
    NTSTATUS status;

    status = WdfDeviceQueryProperty(device, DevicePropertyLocationInformation,
                                    0, NULL, &len);
    if (status != STATUS_BUFFER_TOO_SMALL) {
        return unknown_name;
    }

    buf = os_alloc(len);
    if (buf == NULL) {
        return unknown_name;
    }

    status = WdfDeviceQueryProperty(device, DevicePropertyLocationInformation,
                                    len, buf, &len);
    if (!NT_SUCCESS(status)) {
        return unknown_name;
    }

    RtlInitUnicodeString(&str, buf);
    ansi_len = RtlUnicodeStringToAnsiSize(&str);
    if (ansi_len == 0 || ansi_len > USHORT_MAX) {
        return unknown_name;
    }
    ansi_buf = os_alloc(ansi_len);
    if (ansi_buf == NULL) {
        return unknown_name;
    }
    ansi_str.MaximumLength = (USHORT)ansi_len;
    ansi_str.Buffer = ansi_buf;
    RtlUnicodeStringToAnsiString(&ansi_str, &str, FALSE);

    os_free(buf);

    return ansi_str.Buffer;
}

static void shc_free_name(WDFDEVICE device)
{
    struct shc_device_context *dev_ctx = shc_get_device_context(device);

    if (dev_ctx->name != unknown_name) {
        os_free((char *)dev_ctx->name);
    }
}

static NTSTATUS shc_get_quirks(WDFDEVICE device, ULONG *quirks)
{
    static DECLARE_CONST_UNICODE_STRING(sdioemb_key_name, L"Sdioemb");
    static DECLARE_CONST_UNICODE_STRING(quirks_value_name, L"Quirks");
    struct shc_device_context *dev_ctx = shc_get_device_context(device);
    WDFKEY hardware_key = NULL, sdioemb_key = NULL;
    NTSTATUS status;

    status = WdfDeviceOpenRegistryKey(device, PLUGPLAY_REGKEY_DEVICE, KEY_READ,
                                      WDF_NO_OBJECT_ATTRIBUTES, &hardware_key);
    if (!NT_SUCCESS(status)) {
        os_trace_err("%s: unable to open hardware key: %!STATUS!", dev_ctx->name, status);
        goto out;
    }
    status = WdfRegistryOpenKey(hardware_key, &sdioemb_key_name, KEY_READ,
                                WDF_NO_OBJECT_ATTRIBUTES, &sdioemb_key);
    if (!NT_SUCCESS(status)) {
        os_trace_err("%s: unable to open '%S' key: %!STATUS!", dev_ctx->name,
                     sdioemb_key_name.Buffer, status);
        goto out;
    }
    status = WdfRegistryQueryULong(sdioemb_key, &quirks_value_name, quirks);
    if (!NT_SUCCESS(status)) {
        os_trace_err("%s: unable to query '%S' value: %!STATUS!", dev_ctx->name,
                     quirks_value_name.Buffer, status);
        goto out;
    }
  out:
    if (sdioemb_key) {
        WdfRegistryClose(sdioemb_key);
    }
    if (hardware_key) {
        WdfRegistryClose(hardware_key);
    }
    return status;
}
