/*
 * Sdioemb Windows KMDF SHC driver -- slot driver entry points/callbacks.
 *
 * Copyright (C) 2009-2010 Cambridge Silicon Radio Ltd.
 *
 * Refer to LICENSE.txt included with this source code for details on
 * the license terms.
 */
#include "shc_internal.h"

#include "slot.tmh"

static void shc_enable_int(struct sdioemb_slot *slot, uint32_t ints)
{
    struct shc_data *shc = slot->drv_data;
    struct shc_device_context *dev_ctx = shc_get_device_context(shc->device);
    uint32_t stat_en;

    WdfInterruptAcquireLock(dev_ctx->interrupt);

    stat_en = os_io_read32(shc->shc.addr, SHC_INT_STATUS_EN);
    os_io_write32(shc->shc.addr, SHC_INT_STATUS_EN, stat_en | ints);

    (void)os_io_read32(shc->shc.addr, SHC_INT_STATUS_EN); /* flush posted writes */
    WdfInterruptReleaseLock(dev_ctx->interrupt);
}

static void shc_disable_int(struct sdioemb_slot *slot, uint32_t ints)
{
    struct shc_data *shc = slot->drv_data;
    struct shc_device_context *dev_ctx = shc_get_device_context(shc->device);
    uint32_t stat_en;

    WdfInterruptAcquireLock(dev_ctx->interrupt);

    stat_en = os_io_read32(shc->shc.addr, SHC_INT_STATUS_EN);
    os_io_write32(shc->shc.addr, SHC_INT_STATUS_EN, stat_en & ~ints);

    (void)os_io_read32(shc->shc.addr, SHC_INT_STATUS_EN); /* flush posted writes */
    WdfInterruptReleaseLock(dev_ctx->interrupt);
}

static void shc_cmd_complete(struct sdioemb_slot *slot, struct sdioemb_cmd *cmd)
{
    struct shc_data *shc = slot->drv_data;

    if (shc->use_dma && (cmd->flags & SDIOEMB_CMD_FLAG_READ)) {
        memcpy(cmd->data, shc->dma_buf_virt, cmd->len);
    }

    sdioemb_cmd_complete(slot, cmd);
}

static int shc_set_bus_freq(struct sdioemb_slot *slot, int clk)
{
    struct shc_data *shc = slot->drv_data;

    return sdioemb_shc_set_bus_freq(&shc->shc, clk);
}

static int shc_set_bus_width(struct sdioemb_slot *slot, int bus_width)
{
    struct shc_data *shc = slot->drv_data;

    return sdioemb_shc_set_bus_width(&shc->shc, bus_width);
}

static int shc_start_cmd(struct sdioemb_slot *slot, struct sdioemb_cmd *cmd)
{
    struct shc_data *shc = slot->drv_data;

    if (cmd->data) {
        shc->use_dma = true;
        if (cmd->len > SHC_DMA_BUFFER_LEN) {
            os_trace_warn("%s: DMA buffer is too small (%d > %d)",
                          slot->name, cmd->len, SHC_DMA_BUFFER_LEN);
            shc->use_dma = false;
        }
        if ((shc->shc.quirks & SHC_QUIRK_DMA_NEEDS_DWORD_ALIGNED_LEN)
            && ((unsigned long)cmd->len & 0x3)) {
            shc->use_dma = false;
        }
    } else {
        shc->use_dma = false;
    }

    if (shc->use_dma) {
        if (!(cmd->flags & SDIOEMB_CMD_FLAG_READ)) {
            memcpy(shc->dma_buf_virt, cmd->data, cmd->len);
        }
    }

    sdioemb_shc_start_cmd(&shc->shc, cmd, shc->use_dma, shc->dma_buf_phys);
    return 0;
}

static int shc_card_present(struct sdioemb_slot *slot)
{
    struct shc_data *shc = slot->drv_data;

    return sdioemb_shc_card_present(&shc->shc);
}

static int shc_card_power(struct sdioemb_slot *slot, enum sdioemb_power power)
{
    struct shc_data *shc = slot->drv_data;
    int ret;

    ret = sdioemb_shc_card_power(&shc->shc, power);
    if (ret < 0) {
        return ret;
    }

    /* JMicron 385 chip needs config space write to enable the power switch */
    if (shc->shc.quirks & SHC_QUIRK_JMICRON_JMB38X) {
        uint8_t config_byte;

        shc_pci_config_read8(shc->device, 0xAE, &config_byte);

        if (power != SDIOEMB_POWER_OFF) {
            config_byte |= 0x47;
        } else {
            config_byte &= ~(0x47);
        }
        shc_pci_config_write8(shc->device, 0xAE, config_byte);
    }

    return 0;
}

static void shc_enable_card_int(struct sdioemb_slot *slot)
{
    struct shc_data *shc = slot->drv_data;

    sdioemb_shc_enable_card_int(&shc->shc);
}

static void shc_disable_card_int(struct sdioemb_slot *slot)
{
    struct shc_data *shc = slot->drv_data;

    sdioemb_shc_disable_card_int(&shc->shc);
}

static int shc_hard_reset(struct sdioemb_slot *slot)
{
    struct shc_data *shc = slot->drv_data;

    return sdioemb_shc_hard_reset(&shc->shc);
}

static NTSTATUS shc_create_dma_buffer(WDFDEVICE device, struct sdioemb_slot *slot)
{
    struct shc_device_context *dev_ctx = shc_get_device_context(device);
    struct shc_data *shc = slot->drv_data;
    NTSTATUS status;

    status = WdfCommonBufferCreate(dev_ctx->dma_enabler, SHC_DMA_BUFFER_LEN,
                                   WDF_NO_OBJECT_ATTRIBUTES, &shc->dma_buf);
    if (!NT_SUCCESS(status)) {
        return status;
    }

    shc->dma_buf_virt = WdfCommonBufferGetAlignedVirtualAddress(shc->dma_buf);
    shc->dma_buf_phys = WdfCommonBufferGetAlignedLogicalAddress(shc->dma_buf).QuadPart;

    return STATUS_SUCCESS;
}

static void shc_destroy_dma_buffer(WDFDEVICE device, struct sdioemb_slot *slot)
{
    struct shc_data *shc = slot->drv_data;

    if (shc->dma_buf) {
        WdfObjectDelete(shc->dma_buf);
    }
}

NTSTATUS shc_create_slot(WDFDEVICE device, struct sdioemb_slot **new_slot, int slot_idx)
{
    struct shc_device_context *dev_ctx = shc_get_device_context(device);
    struct sdioemb_slot *slot;
    struct shc_data *shc;
    NTSTATUS status;

    slot = sdioemb_slot_alloc(sizeof(*shc));
    if (slot == NULL) {
        return STATUS_INSUFFICIENT_RESOURCES;
    }
    shc = slot->drv_data;

    snprintf(slot->name, sizeof(slot->name), "%s slot %d", dev_ctx->name, slot_idx);
    slot->set_bus_freq     = shc_set_bus_freq;
    slot->set_bus_width    = shc_set_bus_width;
    slot->start_cmd        = shc_start_cmd;
    slot->card_present     = shc_card_present;
    slot->card_power       = shc_card_power;
    slot->enable_card_int  = shc_enable_card_int;
    slot->disable_card_int = shc_disable_card_int;
    slot->hard_reset       = shc_hard_reset;

    sdioemb_shc_init(&shc->shc);
    shc->shc.slot         = slot;
    shc->shc.enable_int   = shc_enable_int;
    shc->shc.disable_int  = shc_disable_int;
    shc->shc.cmd_complete = shc_cmd_complete;
    shc->shc.quirks       = dev_ctx->quirks;

    sdioemb_shc_show_quirks(&shc->shc);

    shc->device = device;

    status = sdioemb_wdf_create_child_list(device, slot);
    if (!NT_SUCCESS(status)) {
        return status;
    }

    status = shc_create_dma_buffer(device, slot);
    if (!NT_SUCCESS(status)) {
        return status;
    }    

    *new_slot = slot;
    return STATUS_SUCCESS;
}

void shc_destroy_slot(WDFDEVICE device, struct sdioemb_slot *slot)
{
    struct shc_data *shc;

    if (slot == NULL) {
        return;
    }

    shc = slot->drv_data;

    shc_destroy_dma_buffer(device, slot);
    sdioemb_wdf_destroy_child_list(slot);
    sdioemb_shc_clean_up(&shc->shc);
    sdioemb_slot_free(slot);
}
