//
// Sdioemb Windows KMDF USHC driver.
//
// Copyright (C) 2010 Cambridge Silicon Radio Ltd.
//
// Refer to LICENSE.txt included with this source code for details on
// the license terms.
//
#define WPP_CONTROL_GUIDS                                               \
    WPP_DEFINE_CONTROL_GUID(UshcTraceGuid,                              \
                            (085936a8,e620,4606,8b51,f2306bb2ac0d),     \
                            WPP_DEFINE_BIT(OS_TRACE_DFLT)               \
        )
#include <initguid.h>
#include <wdmguid.h>
#include <sdioemb/windows/guid.h>

#include "ushc_internal.h"

#include "driver.tmh"

//
// DriverEntry
//
// Arguments
//  DriverObject - Pointer to a DRIVER_OBJECT structure. This is the driver's driver object
//  RegistryPath - Pointer to a counted Unicode string specifying the path to 
//                 the driver's registry key
// Returns
//  NTSTATUS - STATUS_SUCCESS on success otherwise a NTSTATUS error code
//
NTSTATUS DriverEntry(__inout PDRIVER_OBJECT DriverObject, __in PUNICODE_STRING RegistryPath)
{
    WDF_DRIVER_CONFIG DriverConfig;
    WDF_OBJECT_ATTRIBUTES DriverAttributes;
    NTSTATUS Status;

    WPP_INIT_TRACING(DriverObject, RegistryPath);

    WDF_DRIVER_CONFIG_INIT(&DriverConfig, UshcEvtDeviceAdd);
    WDF_OBJECT_ATTRIBUTES_INIT(&DriverAttributes);
    DriverAttributes.EvtCleanupCallback = UshcDriverCleanup;

    Status = WdfDriverCreate(DriverObject, RegistryPath, &DriverAttributes, &DriverConfig, WDF_NO_HANDLE);
    if(STATUS_SUCCESS != Status) 
    {
        WPP_CLEANUP(DriverObject);
    }

    return Status;
}

//
// UshcDriverCleanup
//
// Arguments
//  Driver - Handle to a framework driver object that represents the driver. 
//
// Returns
//  VOID
//
VOID UshcDriverCleanup(__in WDFDRIVER Driver)
{
    WPP_CLEANUP(WdfDriverWdmGetDriverObject(Driver));
}

//
// UshcEvtDeviceAdd
//
// Arguments
//  Driver     - Handle to a framework driver object that represents the driver. 
//  DeviceInit - Pointer to a framework-allocated WDFDEVICE_INIT structure. 
// 
// Returns
//  NTSTATUS - STATUS_SUCCESS
//
NTSTATUS UshcEvtDeviceAdd(__in WDFDRIVER Driver, __in PWDFDEVICE_INIT DeviceInit)
{
    WDF_PNPPOWER_EVENT_CALLBACKS PnpPowerEventCallbacks;
    WDF_OBJECT_ATTRIBUTES DriverAttributes;
    WDFDEVICE Device;
    PUSHC_DEVICE_CONTEXT Dc;
    UINT8 NumSlots=1;
    INT i;
    NTSTATUS Status = STATUS_UNSUCCESSFUL;
    
    WdfDeviceInitSetIoType(DeviceInit, WdfDeviceIoBuffered);

    WDF_PNPPOWER_EVENT_CALLBACKS_INIT(&PnpPowerEventCallbacks);
    PnpPowerEventCallbacks.EvtDevicePrepareHardware = UshcEvtDevicePrepareHardware;
    PnpPowerEventCallbacks.EvtDeviceReleaseHardware = UshcEvtDeviceReleaseHardware;
    PnpPowerEventCallbacks.EvtDeviceD0Entry = UshcEvtDeviceD0Entry;
    PnpPowerEventCallbacks.EvtDeviceD0Exit = UshcEvtDeviceD0Exit;
    WdfDeviceInitSetPnpPowerEventCallbacks(DeviceInit, &PnpPowerEventCallbacks);

    WDF_OBJECT_ATTRIBUTES_INIT(&DriverAttributes);
    WDF_OBJECT_ATTRIBUTES_INIT_CONTEXT_TYPE(&DriverAttributes, USHC_DEVICE_CONTEXT);
    DriverAttributes.EvtCleanupCallback = UshcEvtDeviceCleanup;

    sdioemb_wdf_event_log_init(DeviceInit);

    __try
    {
        Status = WdfDeviceCreate(&DeviceInit, &DriverAttributes, &Device);
        if(STATUS_SUCCESS != Status)
        {
            os_trace_err("Failed to create WDF device");
            __leave;
        }

        Status = sdioemb_wdf_event_log_create(Device);
        if (!NT_SUCCESS(Status))
        {
            os_trace_err("Failed to create event log");
            __leave;
        }

        Dc = UshcGetDeviceContext(Device);

        Dc->Slots = os_alloc(NumSlots * sizeof(*Dc->Slots));
        if(Dc->Slots == NULL)
        {
            os_trace_err("Failed to alloc %d bytes", sizeof(*Dc->Slots));
            Status = STATUS_INSUFFICIENT_RESOURCES;
            __leave;
        }

        Dc->NumSlots = NumSlots;
        for(i=0; i<Dc->NumSlots; i++)
        {
            Status = UshcCreateSlot(Device, &Dc->Slots[i], i);
            if(Dc->Slots[i] == NULL)
            {
                os_trace_err("Failed to create Slot[%d]",i);
                __leave;
            }
        }
    }
    __finally
    {
        if(STATUS_SUCCESS != Status)
        {
            os_trace_err("Failed UshcEvtDeviceAdd with Status = 0x%x", Status);
        }
    }

    return Status;
}

//
// UshcEvtDeviceCleanup
//
// Arguments
//  Device - Framework device handle
//
// Returns
//  VOID
//
VOID UshcEvtDeviceCleanup(__in WDFOBJECT Object)
{
    WDFDEVICE Device = Object;
    PUSHC_DEVICE_CONTEXT Dc = UshcGetDeviceContext(Device);
    INT i;

    for(i = 0; i < Dc->NumSlots; i++)
    {
        UshcDestroySlot(Object, Dc->Slots[i]);
    }
    os_free(Dc->Slots);
}

//
// UshcEvtDevicePrepareHardware
//
// Arguments
//  Device                 - Framework device handle
//  ResourceList           - Unussed
//  ResourceListTranslated - Unussed
//
// Returns
//  NTSTATUS - STATUS_SUCCESS on success otherwise a NTSTATUS error code
//
NTSTATUS UshcEvtDevicePrepareHardware(__in WDFDEVICE Device, 
                                      __in WDFCMRESLIST ResourceList, 
                                      __in WDFCMRESLIST ResourceListTranslated)
{
    PUSB_CONFIGURATION_DESCRIPTOR ConfigurationDescriptor = NULL;
    WDF_USB_CONTINUOUS_READER_CONFIG ReaderConfig;
    PUSHC_DEVICE_CONTEXT Dc = UshcGetDeviceContext(Device);
    UINT16 Size = 0;
    WDFMEMORY MemoryObject;
    WDF_OBJECT_ATTRIBUTES MemoryObjectAttributes;
    NTSTATUS Status = STATUS_UNSUCCESSFUL;

    UNREFERENCED_PARAMETER(ResourceList);
    UNREFERENCED_PARAMETER(ResourceListTranslated);

    __try
    {
        if(NULL == Dc->WdfUsbTargetDevice) 
        {
            Status = WdfUsbTargetDeviceCreate(Device, WDF_NO_OBJECT_ATTRIBUTES, &Dc->WdfUsbTargetDevice);
            if(STATUS_SUCCESS != Status) 
            {
                os_trace_err("Failed to create target device");
                __leave;
            }
        }

        WdfUsbTargetDeviceGetDeviceDescriptor(Dc->WdfUsbTargetDevice, &Dc->UsbDeviceDescriptor);
        if(Dc->UsbDeviceDescriptor.bNumConfigurations < 1)
        {
            os_trace_err("Failed to get usb target device descriptor");
            Status = STATUS_DEVICE_CONFIGURATION_ERROR;
            __leave;
        }

        //
        // Read the first configuration descriptor
        // This requires two steps:
        // 1. Ask the WDFUSBDEVICE how big it is
        // 2. Allocate it and get it from the WDFUSBDEVICE
        //
        Status = WdfUsbTargetDeviceRetrieveConfigDescriptor(Dc->WdfUsbTargetDevice, NULL, &Size);
        if((STATUS_BUFFER_TOO_SMALL != Status) || (Size == 0)) 
        {
            os_trace_err("Failed to retrieve config descriptor size");
            __leave;
        }

        //
        // Create a memory object and specify usbdevice as the parent so that
        // it will be freed automatically.
        //
        WDF_OBJECT_ATTRIBUTES_INIT(&MemoryObjectAttributes);
        MemoryObjectAttributes.ParentObject = Dc->WdfUsbTargetDevice;
        Status = WdfMemoryCreate(&MemoryObjectAttributes,
                                 NonPagedPool,
                                 'chsU',
                                 Size,
                                 &MemoryObject,
                                 &ConfigurationDescriptor);
        if(STATUS_SUCCESS != Status)
        {
            os_trace_err("Failed to create memory");
            __leave;
        }

        Status = WdfUsbTargetDeviceRetrieveConfigDescriptor(Dc->WdfUsbTargetDevice,
                                                            ConfigurationDescriptor,
                                                            &Size);
        if(STATUS_SUCCESS != Status)
        {
            os_trace_err("Failed to retrieve config descriptor");
            __leave;
        }

        Dc->UsbConfigurationDescriptor = ConfigurationDescriptor;

        Status = SelectInterface(Device);
        if(STATUS_SUCCESS != Status)
        {
            __leave;
        }

        WDF_USB_CONTINUOUS_READER_CONFIG_INIT(&ReaderConfig, InterruptPipeReadComplete, Dc, 64);
        ReaderConfig.EvtUsbTargetPipeReadersFailed = InterruptPipeReadFailed;
        //
        // Reader requests are not posted to the target automatically.
        // We call WdfIoTargetStart in UshcEvtDeviceD0Entry() to start the reader. 
        //
        Status = WdfUsbTargetPipeConfigContinuousReader(Dc->InterruptPipe, &ReaderConfig);
        if(STATUS_SUCCESS != Status)
        {
            os_trace_err("Failed to configure continous reader");
            __leave;
        }
    }
    __finally
    {
        if(STATUS_SUCCESS != Status)
        {
            os_trace_err("Failed UshcEvtDevicePrepareHardware with Status = 0x%x", Status);
        }
    }

    return Status;
}

//
// SelectInterface
//
// Arguments
//  Device - Framework device handle
//
// Returns
//  NTSTATUS - STATUS_SUCCESS on success otherwise a NTSTATUS error code
//
NTSTATUS SelectInterface(__in WDFDEVICE Device)
{
    PUSHC_DEVICE_CONTEXT Dc = UshcGetDeviceContext(Device);
    WDF_USB_DEVICE_SELECT_CONFIG_PARAMS ConfigParams;
    WDFUSBPIPE Pipe;
    WDF_USB_PIPE_INFORMATION PipeInfo;
    UCHAR NumberConfiguredPipes; 
    UCHAR p;
    UCHAR f = 0;
    NTSTATUS Status;

    //
    // The device has only 1 interface.
    //
    WDF_USB_DEVICE_SELECT_CONFIG_PARAMS_INIT_SINGLE_INTERFACE(&ConfigParams);
    Status = WdfUsbTargetDeviceSelectConfig(Dc->WdfUsbTargetDevice, WDF_NO_OBJECT_ATTRIBUTES, &ConfigParams);
    if(STATUS_SUCCESS == Status)
    {
        if(WdfUsbTargetDeviceGetNumInterfaces(Dc->WdfUsbTargetDevice) > 0)
        {
            Dc->UsbInterface = ConfigParams.Types.SingleInterface.ConfiguredUsbInterface;
            NumberConfiguredPipes = ConfigParams.Types.SingleInterface.NumberConfiguredPipes;

            //
            // Get pipe handles
            //
            for(p=0; p < NumberConfiguredPipes; p++) 
            {
                WDF_USB_PIPE_INFORMATION_INIT(&PipeInfo);
        
                Pipe = WdfUsbInterfaceGetConfiguredPipe(Dc->UsbInterface, p, &PipeInfo);

                //
                // Tell the framework that it's okay to read less than
                // MaximumPacketSize
                //
                WdfUsbTargetPipeSetNoMaximumPacketSizeCheck(Pipe);
        
                switch(PipeInfo.PipeType)
                {
                    case WdfUsbPipeTypeControl:
                    break;
        
                    case WdfUsbPipeTypeIsochronous:
                    break;
        
                    case WdfUsbPipeTypeBulk:
                        if(WdfUsbTargetPipeIsInEndpoint(Pipe))
                        {
                            os_trace_dbg("Bulk IN Pipe");
                            os_trace_dbg(" Maximum Packet Size 0x%x",PipeInfo.MaximumPacketSize);
                            os_trace_dbg(" Endpoint Address 0x%x",PipeInfo.EndpointAddress);
                            Dc->DataInPipe = Pipe;
                            ++f;
                        }
                        else
                        {
                            if(WdfUsbTargetPipeIsOutEndpoint(Pipe))
                            {
                                os_trace_dbg("Bulk OUT Pipe");
                                os_trace_dbg(" Maximum Packet Size 0x%x", PipeInfo.MaximumPacketSize);
                                os_trace_dbg(" Endpoint Address 0x%x", PipeInfo.EndpointAddress);
                                Dc->DataOutPipe = Pipe;
                                ++f;
                            }
                        }
                    break;
        
                    case WdfUsbPipeTypeInterrupt:
                        os_trace_dbg("Interrupt Pipe");
                        os_trace_dbg(" Maximum Packet Size 0x%x", PipeInfo.MaximumPacketSize);
                        os_trace_dbg(" Endpoint Address 0x%x", PipeInfo.EndpointAddress);
                        os_trace_dbg(" Interval 0x%x", PipeInfo.Interval);
                        Dc->InterruptPipe = Pipe;
                        ++f;
                    break;

                    default:
                        os_trace_err("Unknown PipeType 0x%x", PipeInfo.PipeType);
                    break;
                }
            }
        }
        else
        {
            os_trace_err("Single USB configuration has no interfaces");
        }
    }
    else
    {
        os_trace_err("Failed to slect single USB configuration");
    }

    //
    // We should have found 3 pipes (IN OUT and Interrupt)
    //
    if(3 != f)
    {
        os_trace_err("USB Interface only has %d of the requiered 3 pipes", f);
        Status = STATUS_INVALID_DEVICE_STATE;
    }

    return Status;
}

//
// InterruptPipeReadFailed
//
// Arguments
//  Pipe       - A handle to a framework pipe object
//  Status     - The NTSTATUS value that the pipe's I/O target returned
//  UsbdStatus - The USBD_STATUS-typed status value that the pipe's I/O target returned 
//
// Returns
//  BOOLEAN - TRUE, causes the framework to reset the USB pipe and then restart the continuous reader.
//            FALSE, the framework does not reset the device or restart the continuous reader. 
//            
//
BOOLEAN InterruptPipeReadFailed(__in WDFUSBPIPE Pipe, __in NTSTATUS Status, __in USBD_STATUS UsbdStatus)
{
    UNREFERENCED_PARAMETER(Pipe);

    os_trace_dbg("Read on INT pipe fails with Status 0x%x USBD Status 0x%x", Status, UsbdStatus);

    return TRUE;
}


//
// UshcEvtDeviceReleaseHardware
//
// Arguments
//  Device                 - Framework device handle
//  ResourceListTranslated - UNREFERENCED
//
// Returns
//  NTSTATUS - STATUS_SUCCESS
//
NTSTATUS UshcEvtDeviceReleaseHardware(__in WDFDEVICE Device, __in WDFCMRESLIST ResourceListTranslated)
{
    PUSHC_DEVICE_CONTEXT Dc = UshcGetDeviceContext(Device);
    INT i;

    UNREFERENCED_PARAMETER(ResourceListTranslated);
    for(i = 0; i < Dc->NumSlots; i++)
    {
        struct sdioemb_slot *slot;
        USHC_DATA *Ushc;

        slot = Dc->Slots[i];
        Ushc = slot->drv_data;
    }

    return STATUS_SUCCESS;
}

//
// UshcEvtDeviceD0Entry
//
// Arguments
//  Device        - Framework device handle
//  PreviousState - Previous device power state
//
// Returns
//  NTSTATUS - STATUS_SUCCESS on success otherwise a NTSTATUS error code
//
NTSTATUS UshcEvtDeviceD0Entry(__in WDFDEVICE Device, __in WDF_POWER_DEVICE_STATE PreviousState)
{
    PUSHC_DEVICE_CONTEXT Dc = UshcGetDeviceContext(Device);
    INT i;
    NTSTATUS Status;

    for(i=0; i<Dc->NumSlots; i++)
    {
        PSDIOEMB_SLOT Slot;
        USHC_DATA *Ushc;

        Slot = Dc->Slots[i];
        Ushc = Slot->drv_data;

        sdioemb_ushc_start(&Ushc->Ushc);
        if(!Dc->Registered)
        {
            sdioemb_wdf_slot_register(Slot);
        }
    }
    Dc->Registered = true;
 
    Status = WdfIoTargetStart(WdfUsbTargetPipeGetIoTarget(Dc->InterruptPipe));

    return Status;
}

//
// UshcEvtDeviceD0Exit
//
// Arguments
//  Device      - Framework device handle
//  TargetState - Device target power state
//
// Returns
//  NTSTATUS - STATUS_SUCCESS
//
NTSTATUS UshcEvtDeviceD0Exit(__in WDFDEVICE Device, __in WDF_POWER_DEVICE_STATE TargetState)
{
    PUSHC_DEVICE_CONTEXT Dc = UshcGetDeviceContext(Device);
    INT i;

    WdfIoTargetStop(WdfUsbTargetPipeGetIoTarget(Dc->InterruptPipe), WdfIoTargetCancelSentIo);

    for(i=0; i<Dc->NumSlots; i++)
    {
        PSDIOEMB_SLOT Slot;
        PUSHC_DATA Ushc; 
        Slot = Dc->Slots[i];
        Ushc = Slot->drv_data;

        switch(TargetState)
        {
            case WdfPowerDeviceD3Final:
               sdioemb_wdf_slot_unregister(Slot);
               //  |
               //  | Falling now
               //  V
            default:
                sdioemb_ushc_stop(&Ushc->Ushc);
            break;
        }
    }

    return STATUS_SUCCESS;
}
