//
// Sdioemb Windows KMDF USHC driver.
//
// Copyright (C) 2010 Cambridge Silicon Radio Ltd.
//
// Refer to LICENSE.txt included with this source code for details on
// the license terms.
//
#include "ushc_internal.h"

#include "slot.tmh"

#define USHC_CBW_REQUEST_TIMEOUT_MS 2000


//
// UshcSubmitCbwRequest
//
// Arguments
//  Slot      - Pointer to a SDIOEMB_SLOT structure
//  CmdIndex  - Sdio command index
//  BlockSize - Block size for data transfer
//  CmdArg    - sdio command argument
//
// Returns
//  INT - 0 on success otherwise a POSIX error code
//
INT UshcSubmitCbwRequest(__in PSDIOEMB_SLOT Slot, __in UCHAR CmdIndex, __in UINT16 BlockSize, __in UINT32 CmdArg)
{
    PUSHC_DATA Ushc = Slot->drv_data;
    WDF_OBJECT_ATTRIBUTES RequestAttributes;
    WDF_REQUEST_REUSE_PARAMS  Rp;
    WDF_REQUEST_SEND_OPTIONS SendOptions;
    PUSHC_DEVICE_CONTEXT Dc = UshcGetDeviceContext(Ushc->Device);
    LONG c;
    NTSTATUS Status = STATUS_SUCCESS;

    Ushc->Cbw.CmdIndex = CmdIndex;
    Ushc->Cbw.BlockSize = BlockSize;
    Ushc->Cbw.Arg = CmdArg;

    WDF_REQUEST_REUSE_PARAMS_INIT(&Rp, WDF_REQUEST_REUSE_NO_FLAGS, STATUS_SUCCESS);
    WdfRequestReuse(Ushc->CbwRequest, &Rp);

    Status = WdfUsbTargetPipeFormatRequestForWrite(Dc->DataOutPipe, Ushc->CbwRequest, Ushc->CbwMemory, NULL);
    if(STATUS_SUCCESS == Status)
    {
        WdfRequestSetCompletionRoutine(Ushc->CbwRequest, CbwRequestCompletionRoutine, Ushc);

        WDF_REQUEST_SEND_OPTIONS_INIT(&SendOptions, WDF_REQUEST_SEND_OPTION_TIMEOUT);
        WDF_REQUEST_SEND_OPTIONS_SET_TIMEOUT(&SendOptions, WDF_REL_TIMEOUT_IN_MS(USHC_CBW_REQUEST_TIMEOUT_MS));

        c = InterlockedExchange(&Ushc->CommandIoCount, 2);
        if(0 != c)
        {
            os_trace_err("IoCount (at CbwSubmit)) = %d", c);
        }

        if(FALSE == WdfRequestSend(Ushc->CbwRequest, WdfUsbTargetPipeGetIoTarget(Dc->DataOutPipe), &SendOptions))
        {
            InterlockedExchange(&Ushc->CommandIoCount, 0);
            Status = WdfRequestGetStatus(Ushc->CbwRequest);
            os_trace_err("Failed to sbumit Cbw request Status = 0x%x", Status);
        }
    }
    else
    {
        os_trace_err("Failed to format Cbw request Status = 0x%x", Status);
    }
    
    return NTStatus2Posix(Status);
}

//
// CbwRequestCompletionRoutine
//
// Arguments
//  Request - Request handle
//  Target  - Device handle
//  Params  - Request completion params
//  Context - Driver supplied context
//
// Returns
//  VOID
//
VOID CbwRequestCompletionRoutine(__in WDFREQUEST Request, 
                                 __in WDFIOTARGET Target,
                                 __in PWDF_REQUEST_COMPLETION_PARAMS CompletionParams,
                                 __in WDFCONTEXT Context)
{
    PUSHC_DATA Ushc = (PUSHC_DATA)Context;
    PUSHC_DEVICE_CONTEXT Dc = UshcGetDeviceContext(Ushc->Device);
    NTSTATUS Status = CompletionParams->IoStatus.Status;
    USBD_STATUS UsbStatus = CompletionParams->Parameters.Usb.Completion->UsbdStatus;

    if(STATUS_SUCCESS != Status)
    {
        os_trace_err("Cbw complete fails Status = 0x%x UrbStatus = 0x%x Reseting DataOutPipe",
                     Status, UsbStatus);

        QPipeResetWorkItem(Ushc, Dc->DataOutPipe);
    }

    if(0 == InterlockedDecrement(&Ushc->CommandIoCount))
    {
        sdioemb_ushc_command_complete(&Ushc->Ushc, Ushc->Csw.Status, Ushc->Csw.Response);
    }
}

#define USHC_DATA_REQUEST_TIMEOUT_MS 2000
//
// UshcSubmitDataRequest
//
// Arguments
//  Slot                 - Pointer to a SDIOEMB_SLOT structure
//  Direction            - USHC_HOST_TO_DEVICE or USHC_DEVICE_TO_HOST
//  RequestBuffer        - Pointer to a buffer for the transfer 
//  RequestBufferLength  - Length, in bytes, of the buffer specified in RequestBuffer
//
// Returns
//  INT - 0 on success otherwise a POSIX error code
//
INT UshcSubmitDataRequest(__in PSDIOEMB_SLOT Slot,
                          __in USHC_REQUEST_DIRECTION Direction,
                          __in_bcount(RequestBufferLength) PVOID RequestBuffer,
                          __in UINT32 RequestBufferLength)
{
    PUSHC_DATA Ushc = Slot->drv_data;
    PUSHC_DEVICE_CONTEXT Dc = UshcGetDeviceContext(Ushc->Device);
    WDFREQUEST DataRequest;
    WDF_REQUEST_SEND_OPTIONS SendOptions;
    WDF_OBJECT_ATTRIBUTES RequestAttributes;
    WDFMEMORY DataMemory;
    WDF_OBJECT_ATTRIBUTES MemoryAttributes;
    WDFIOTARGET IoTarget;
    NTSTATUS Status = STATUS_UNSUCCESSFUL;

    __try
    {
        WDF_OBJECT_ATTRIBUTES_INIT(&RequestAttributes);
        Status = WdfRequestCreate(&RequestAttributes, WdfUsbTargetDeviceGetIoTarget(Dc->WdfUsbTargetDevice), &DataRequest);
        if(STATUS_SUCCESS != Status)
        {
            os_trace_err("RequestCreate fails (%x)", Status);
            __leave;
        }

        WDF_OBJECT_ATTRIBUTES_INIT(&MemoryAttributes);
        MemoryAttributes.ParentObject = DataRequest;
        Status = WdfMemoryCreatePreallocated(&MemoryAttributes, RequestBuffer, RequestBufferLength, &DataMemory);
        if(STATUS_SUCCESS != Status)
        {
            os_trace_err("MemoryCreate fails (%x)", Status);
            __leave;
        }

        if(USHC_DEVICE_TO_HOST == Direction)
        {
            Status = WdfUsbTargetPipeFormatRequestForRead(Dc->DataInPipe, DataRequest, DataMemory, NULL);
            if(STATUS_SUCCESS != Status)
            {
                os_trace_err("TargetPipeFormatRequestForRead fails (%x)", Status);
                __leave;
            }
        }
        else
        {
            if(USHC_HOST_TO_DEVICE == Direction)
            {
                Status = WdfUsbTargetPipeFormatRequestForWrite(Dc->DataOutPipe, DataRequest, DataMemory, NULL);
                if(STATUS_SUCCESS != Status)
                {
                    os_trace_err("TargetPipeFormatRequestForWrite fails (%x)", Status);
                    __leave;
                }
            }
            else
            {
                Status = STATUS_INVALID_PARAMETER;
                os_trace_err("Unknown Direction for data request");
                __leave;
            }
        }

        WdfRequestSetCompletionRoutine(DataRequest, DataRequestCompletionRoutine, Ushc);

        WDF_REQUEST_SEND_OPTIONS_INIT(&SendOptions, WDF_REQUEST_SEND_OPTION_TIMEOUT);
        WDF_REQUEST_SEND_OPTIONS_SET_TIMEOUT(&SendOptions, WDF_REL_TIMEOUT_IN_MS(USHC_DATA_REQUEST_TIMEOUT_MS));

        InterlockedIncrement(&Ushc->CommandIoCount);

        if(FALSE == WdfRequestSend(DataRequest, WdfUsbTargetPipeGetIoTarget(Dc->DataOutPipe), &SendOptions))
        {
            Status = WdfRequestGetStatus(DataRequest);
            os_trace_err("RequestSend fails (%x)", Status);
            __leave;
        }
    }
    __finally
    {
        if(STATUS_SUCCESS != Status)
        {
            InterlockedExchange(&Ushc->CommandIoCount, 0);
            os_trace_err("Failed to sbumit Data request Status = 0x%x", Status);
        }
    }
    
    return NTStatus2Posix(Status);
}

// 
// DataRequestCompletionRoutine
//
// Arguments
//  Request - Request handle
//  Target  - Device handle
//  Params  - Request completion params
//  Context - Driver supplied context
//
// Returns
//  VOID
//
VOID DataRequestCompletionRoutine(__in WDFREQUEST Request, 
                                  __in WDFIOTARGET Target,
                                  __in PWDF_REQUEST_COMPLETION_PARAMS CompletionParams,
                                  __in WDFCONTEXT Context)
{
    PUSHC_DATA Ushc = (PUSHC_DATA)Context;

    if(0 == InterlockedDecrement(&Ushc->CommandIoCount))
    {
        sdioemb_ushc_command_complete(&Ushc->Ushc, Ushc->Csw.Status, Ushc->Csw.Response);
    }

    WdfObjectDelete(Request);
}

#define USHC_CSW_REQUEST_TIMEOUT_MS 2000

//
// UshcSubmitCswRequest
//
// Arguments
//  Slot      - Pointer to a SDIOEMB_SLOT structure
//
// Returns
//  INT - 0 on success otherwise a POSIX error code
//
INT UshcSubmitCswRequest(__in PSDIOEMB_SLOT Slot)
{
    PUSHC_DATA Ushc = Slot->drv_data;
    PUSHC_DEVICE_CONTEXT Dc = UshcGetDeviceContext(Ushc->Device);
    WDF_REQUEST_REUSE_PARAMS  Rp;
    WDF_REQUEST_SEND_OPTIONS SendOptions;
    NTSTATUS Status = STATUS_SUCCESS;

    WDF_REQUEST_REUSE_PARAMS_INIT(&Rp, WDF_REQUEST_REUSE_NO_FLAGS, STATUS_SUCCESS);
    WdfRequestReuse(Ushc->CswRequest, &Rp);

    Status = WdfUsbTargetPipeFormatRequestForRead(Dc->DataInPipe, Ushc->CswRequest, Ushc->CswMemory, NULL);
    if(STATUS_SUCCESS == Status)
    {
        WdfRequestSetCompletionRoutine(Ushc->CswRequest, CswRequestCompletionRoutine, Ushc);

        WDF_REQUEST_SEND_OPTIONS_INIT(&SendOptions, WDF_REQUEST_SEND_OPTION_TIMEOUT);
        WDF_REQUEST_SEND_OPTIONS_SET_TIMEOUT(&SendOptions, WDF_REL_TIMEOUT_IN_MS(USHC_CSW_REQUEST_TIMEOUT_MS));

        if(FALSE == WdfRequestSend(Ushc->CswRequest, WdfUsbTargetPipeGetIoTarget(Dc->DataInPipe), &SendOptions))
        {
            InterlockedExchange(&Ushc->CommandIoCount, 0);
            Status = WdfRequestGetStatus(Ushc->CswRequest);
            os_trace_err("Failed to sbumit Csw request Status = 0x%x", Status);
        }
    }
    else
    {
        os_trace_err("Failed to format Csw request Status = 0x%x", Status);
    }

    return NTStatus2Posix(Status);
}

//
// CswRequestCompletionRoutine
//
// Arguments
//  Request - Request handle
//  Target  - Device handle
//  Params  - Request completion params
//  Context - Driver supplied context
//
// Returns
//  VOID
//
VOID CswRequestCompletionRoutine(__in WDFREQUEST Request, 
                                 __in WDFIOTARGET Target,
                                 __in PWDF_REQUEST_COMPLETION_PARAMS CompletionParams,
                                 __in WDFCONTEXT Context)
{
    PUSHC_DATA Ushc = (PUSHC_DATA)Context;
    PUSHC_DEVICE_CONTEXT Dc = UshcGetDeviceContext(Ushc->Device);
    NTSTATUS Status = CompletionParams->IoStatus.Status;
    USBD_STATUS UsbStatus = CompletionParams->Parameters.Usb.Completion->UsbdStatus;

    if(STATUS_SUCCESS != Status)
    {
        os_trace_err("Csw complete fails. Status = 0x%x UrbStatus = 0x%x Reseting DataInPipe",
                     Status, UsbStatus);
        Ushc->Csw.Status = SDIOEMB_CMD_ERR_CMD_OTHER;
        QPipeResetWorkItem(Ushc, Dc->DataInPipe);
    }

    if(0 == InterlockedDecrement(&Ushc->CommandIoCount))
    {
        sdioemb_ushc_command_complete(&Ushc->Ushc, Ushc->Csw.Status, Ushc->Csw.Response);
    }
}

//
// InterruptPipeReadComplete
//
// Arguments
//  Pipe      - Handle to a framework pipe object
//  Buffer    - Handle to a memory object that represents a buffer that contains data from the device
//  BytesRead - Nurmber of bytes in Buffer
//  Context   - Context information specified in WDF_USB_CONTINUOUS_READER_CONFIG
//
// Returns
//  VOID
//
VOID InterruptPipeReadComplete(__in WDFUSBPIPE Pipe, __in WDFMEMORY Buffer, __in size_t BytesRead, __in WDFCONTEXT Context)
{
    PUSHC_DEVICE_CONTEXT Dc = Context;
    PSDIOEMB_SLOT Slot;
    PUSHC_DATA Ushc;
    INT32 s;
    PUSHC_INT_DATA IntData;

    UNREFERENCED_PARAMETER(Pipe);

    if(sizeof(USHC_INT_DATA) == BytesRead)
    {
        IntData = WdfMemoryGetBuffer(Buffer, NULL);

        for(s=0; s<Dc->NumSlots; s++) 
        {
            Slot = Dc->Slots[s];
            Ushc = Slot->drv_data;
            sdioemb_ushc_isr(&Ushc->Ushc, IntData->Status);
        }
    }
    else
    {
        os_trace_err("Int pipe read only %d bytes, expected %d", BytesRead, sizeof(USHC_INT_DATA));
    }
}

//
// QPipeResetWorkItem
//
// Arguments
//  Ushc - Pointer to a USHC_DATA structure
//  Pipe - Pipe object handle
//
// Returns
//
//  NTSTATUS - STATUS_SUCCESS on success otherwise a NTSTATUS error code
//
NTSTATUS QPipeResetWorkItem(__in PUSHC_DATA Ushc, __in WDFUSBPIPE Pipe)
{
    WDF_OBJECT_ATTRIBUTES WorkItemAttributes;
    WDF_WORKITEM_CONFIG WorkitemConfig;
    WDFWORKITEM WorkItem;
    PWORKER_CONTEXT WorkerContext;
    NTSTATUS Status;

    WDF_OBJECT_ATTRIBUTES_INIT(&WorkItemAttributes);
    WDF_OBJECT_ATTRIBUTES_SET_CONTEXT_TYPE(&WorkItemAttributes, WORKER_CONTEXT);
    WorkItemAttributes.ParentObject = Ushc->Device;

    WDF_WORKITEM_CONFIG_INIT(&WorkitemConfig, Worker);

    WDF_WORKITEM_CONFIG_INIT(&WorkitemConfig, Worker);
    Status = WdfWorkItemCreate(&WorkitemConfig, &WorkItemAttributes, &WorkItem);
    if(STATUS_SUCCESS == Status)
    {
        InterlockedIncrement(&Ushc->CommandIoCount);
        WorkerContext = GetWorkerContext(WorkItem);
        WorkerContext->Ushc = Ushc;
        WorkerContext->Pipe = Pipe;
        WdfWorkItemEnqueue(WorkItem);
    }

    return Status;
}

//
// Worker
//
// Arguments
//  WorkItem - Handle to a work-item object
//
// Returns
//  VOID
//
VOID Worker(__in WDFWORKITEM WorkItem)
{
    PWORKER_CONTEXT  Context = GetWorkerContext(WorkItem);
    PUSHC_DEVICE_CONTEXT Dc = UshcGetDeviceContext(Context->Ushc->Device);

    SubmitVendorRequest(Dc->WdfUsbTargetDevice, USHC_RESET, BmRequestHostToDevice, BmRequestToDevice, 0, 0, NULL, 0);
    ResetPipe(Context->Ushc, Context->Pipe);

    WdfObjectDelete(WorkItem);
}

//  
//  ResetPipe
//
// Arguments
//  Ushc - Pointer to a USHC_DATA structure
//  Pipe - Pipe object handle to reset
//  
// Returns
//  NTSTATUS - STATUS_SUCCESS on success otherwise a NTSTATUS error code
//
NTSTATUS ResetPipe(__in PUSHC_DATA Ushc, __in WDFUSBPIPE Pipe)
{
    NTSTATUS Status;

    Status = WdfUsbTargetPipeResetSynchronously(Pipe, WDF_NO_HANDLE, NULL);
    if(0 == InterlockedDecrement(&Ushc->CommandIoCount))
    {
        sdioemb_ushc_command_complete(&Ushc->Ushc, Ushc->Csw.Status, Ushc->Csw.Response);
    }

    return Status;
}

//
// UshcCmdComplete
//
// Arguments
//  Slot - Pointer to a SDIOEMB_SLOT structure
//  Cmd  - Pointer to a a SDIO_CMD structure
//
// Returns
//  VOID
//
VOID UshcCmdComplete(__in PSDIOEMB_SLOT Slot, __in PSDIO_CMD Cmd)
{
    PUSHC_DATA Ushc = Slot->drv_data;

    sdioemb_cmd_complete(Slot, Cmd);
}

//
// UshcSetBusFreq
//
// Arguments
//  Slot  - Pointer to a SDIOEMB_SLOT structure
//  Clock - clock value
//
// Returns
//  INT - Clock on success otherwise a POSIX error code
//
INT UshcSetBusFreq(__in PSDIOEMB_SLOT Slot, __in INT Clock)
{
    PUSHC_DATA Ushc = Slot->drv_data;
    
    /* 
     * Hardware can't detect interrupts while the clock is off
     */
    if (Clock == SDIOEMB_BUS_FREQ_IDLE) {
        Clock = 400000;
    }

    /*
     * UshcSetBusWidth may be called at DISPATCH IRQL but
     * sdioemb_ushc_set_bus_width() can only be called at PASSIVE so
     * queue a work item.
     *
     * A command cannot be started while a change to bus frequency is
     * pending so set a flag to defer the command to the work item.
     */
    WdfSpinLockAcquire(Ushc->Lock);
    Ushc->SetBusFreqPending = TRUE;
    Ushc->PendingBusFreq = Clock;
    WdfSpinLockRelease(Ushc->Lock);

    WdfWorkItemEnqueue(Ushc->SetBusFreqWork);

    return Clock;
}

//
// UshcSetBusWidth
//
// Arguments
//  Slot     - Pointer to a SDIOEMB_SLOT structure
//  BusWidth - 1 or 4
//
// Returns
//  INT - 0 on success otherwise a POSIX error code
//
INT UshcSetBusWidth(__in PSDIOEMB_SLOT Slot, __in INT BusWidth)
{
    PUSHC_DATA Ushc = Slot->drv_data;

    return sdioemb_ushc_set_bus_width(&Ushc->Ushc, BusWidth);
}

//
// UshcStartCmd
//
// Arguments
//  Slot - Pointer to a SDIOEMB_SLOT structure
//  Cmd  - Pointer to a a SDIO_CMD structure
//
// Returns
//  INT
//
INT UshcStartCmd(__in PSDIOEMB_SLOT Slot, __in PSDIO_CMD Cmd)
{
    PUSHC_DATA Ushc = Slot->drv_data;
    bool deferred = false;

    WdfSpinLockAcquire(Ushc->Lock);
    if (Ushc->SetBusFreqPending) {
        Ushc->PendingCmd = Cmd;
        deferred = true;
    }
    WdfSpinLockRelease(Ushc->Lock);

    if (!deferred) {
        return sdioemb_ushc_start_cmd(&Ushc->Ushc, Cmd);
    }
    return 0;
}

//
// UshcCardPresent
//
// Arguments
//  Slot  - Pointer to a SDIOEMB_SLOT structure
//
// Returns
//  INT
//
INT UshcCardPresent(__in PSDIOEMB_SLOT Slot)
{
    PUSHC_DATA Ushc = Slot->drv_data;

    return sdioemb_ushc_card_present(&Ushc->Ushc);
}

//
// UshcCardPower
//
// Arguments
//  Slot  - Pointer to a SDIOEMB_SLOT structure
//  Power - 
//
// Returns
//  INT
//
INT UshcCardPower(__in PSDIOEMB_SLOT Slot, __in SDIOEMB_POWER Power)
{
    PUSHC_DATA Ushc = Slot->drv_data;

    sdioemb_ushc_card_power(&Ushc->Ushc, Power);
    return 0;
}

//
// UshcEnableCardInt
//
// Arguments
//  Slot  - Pointer to a SDIOEMB_SLOT structure
//
// Returns
//  VOID
//
VOID UshcEnableCardInt(__in PSDIOEMB_SLOT Slot)
{
    PUSHC_DATA Ushc = Slot->drv_data;

    sdioemb_ushc_enable_card_int(&Ushc->Ushc);
}

//
// UshcDisableCardInt
// 
// Arguments
//  Slot  - Pointer to a SDIOEMB_SLOT structure
//
// Returns
//  VOID
//
VOID UshcDisableCardInt(__in PSDIOEMB_SLOT Slot)
{
    PUSHC_DATA Ushc = Slot->drv_data;

    sdioemb_ushc_disable_card_int(&Ushc->Ushc);
}

//
// UshcHardReset
//
// Arguments
//  Slot  - Pointer to a SDIOEMB_SLOT structure
//
// Returns
//  INT - 0
//
INT UshcHardReset(__in PSDIOEMB_SLOT Slot)
{
    PUSHC_DATA Ushc = Slot->drv_data;

    return 0;
}

static VOID UshcSetBusFreqWork(__in WDFWORKITEM work)
{
    PWORKER_CONTEXT WorkerContext = GetWorkerContext(work);
    PUSHC_DATA Ushc = WorkerContext->Ushc;
    struct sdioemb_slot *Slot = Ushc->Ushc.slot;
    struct sdioemb_cmd *Cmd;
    int ret;

    ret = sdioemb_ushc_set_bus_freq(&Ushc->Ushc, Ushc->PendingBusFreq);

    WdfSpinLockAcquire(Ushc->Lock);
    Ushc->SetBusFreqPending = FALSE;
    Cmd = Ushc->PendingCmd;
    Ushc->PendingCmd = NULL;
    WdfSpinLockRelease(Ushc->Lock);

    /* Now start the command that was wait for the frequency
       change. */
    if (Cmd) {
        if (ret >= 0) {
            ret = UshcStartCmd(Slot, Cmd);
        }
        if (ret < 0) {
            Cmd->status = SDIOEMB_CMD_ERR_CMD_OTHER;
            sdioemb_cmd_complete(Slot, Cmd);
        }
    }
}

//
// UshcCreateSlot
//
// Arguments
//  Device    -
//  NewSlot   - Pointer to a SDIOEMB_SLOT structure
//  SlotIndex -
//
// Returns
//  NTSTATUS - STATUS_SUCCESS on success otherwise a NTSTATUS error code
//
NTSTATUS UshcCreateSlot(__in WDFDEVICE Device, __out PSDIOEMB_SLOT* NewSlot, __in INT SlotIndex)
{
    PUSHC_DEVICE_CONTEXT Dc = UshcGetDeviceContext(Device);
    WDF_OBJECT_ATTRIBUTES Attributes;
    WDF_WORKITEM_CONFIG WorkItemConfig;
    PWORKER_CONTEXT WorkerContext;
    PSDIOEMB_SLOT Slot = NULL;
    PUSHC_DATA Ushc;
    NTSTATUS Status = STATUS_UNSUCCESSFUL;

    __try
    {
        Slot = sdioemb_slot_alloc(sizeof(*Ushc));
        if(Slot == NULL) 
        {
            Status = STATUS_INSUFFICIENT_RESOURCES;
            __leave;
        }
        Ushc = Slot->drv_data;

        Slot->set_bus_freq     = UshcSetBusFreq;
        Slot->set_bus_width    = UshcSetBusWidth;
        Slot->start_cmd        = UshcStartCmd;
        Slot->card_present     = UshcCardPresent;
        Slot->card_power       = UshcCardPower;
        Slot->enable_card_int  = UshcEnableCardInt;
        Slot->disable_card_int = UshcDisableCardInt;
        Slot->hard_reset       = UshcHardReset;

        sdioemb_ushc_init(&Ushc->Ushc);
        Ushc->Ushc.slot                  = Slot;
        Ushc->Ushc.cmd_complete          = UshcCmdComplete;
        Ushc->Ushc.set_host_ctrl         = UshcSetHostControl;
        Ushc->Ushc.submit_vendor_request = UshcSubmitVendorRequest;
        Ushc->Ushc.submit_cbw_request    = UshcSubmitCbwRequest;
        Ushc->Ushc.submit_data_request   = UshcSubmitDataRequest;
        Ushc->Ushc.submit_csw_request    = UshcSubmitCswRequest;

        Ushc->Device = Device;

        Status = sdioemb_wdf_create_child_list(Device, Slot);
        if(STATUS_SUCCESS != Status) 
        {
            __leave;
        }

        WDF_OBJECT_ATTRIBUTES_INIT(&Attributes);
        Attributes.ParentObject = Device;
        Status = WdfSpinLockCreate(&Attributes, &Ushc->Lock);
        if(STATUS_SUCCESS != Status)
        {
            __leave;
        }

        Ushc->Cbw.Signature = USHC_SDIO_CBW_SIGNATURE;
        WDF_OBJECT_ATTRIBUTES_INIT(&Attributes);
        Attributes.ParentObject = Dc->WdfUsbTargetDevice;
        Status = WdfRequestCreate(&Attributes, 
                                  WdfUsbTargetDeviceGetIoTarget(Dc->WdfUsbTargetDevice), 
                                  &Ushc->CbwRequest);
        if(STATUS_SUCCESS != Status)
        {
            __leave;
        }

        WDF_OBJECT_ATTRIBUTES_INIT(&Attributes);
        Attributes.ParentObject = Ushc->CbwRequest;
        Status = WdfMemoryCreatePreallocated(&Attributes, &Ushc->Cbw, sizeof(CBW), &Ushc->CbwMemory);
        if(STATUS_SUCCESS != Status)
        {
            __leave;
        }

        WDF_OBJECT_ATTRIBUTES_INIT(&Attributes);
        Attributes.ParentObject = Dc->WdfUsbTargetDevice;
        Status = WdfRequestCreate(&Attributes, 
                                  WdfUsbTargetDeviceGetIoTarget(Dc->WdfUsbTargetDevice), 
                                  &Ushc->CswRequest);
        if(STATUS_SUCCESS != Status)
        {
            __leave;
        }

        WDF_OBJECT_ATTRIBUTES_INIT(&Attributes);
        Attributes.ParentObject = Ushc->CswRequest;
        Status = WdfMemoryCreatePreallocated(&Attributes, &Ushc->Csw, sizeof(CSW), &Ushc->CswMemory);
        if(STATUS_SUCCESS != Status)
        {
            __leave;
        }

        WDF_WORKITEM_CONFIG_INIT(&WorkItemConfig, UshcSetBusFreqWork);
        WDF_OBJECT_ATTRIBUTES_INIT_CONTEXT_TYPE(&Attributes, WORKER_CONTEXT);
        Attributes.ParentObject = Device;
        Status = WdfWorkItemCreate(&WorkItemConfig, &Attributes, &Ushc->SetBusFreqWork);
        if(STATUS_SUCCESS != Status)
        {
             __leave;
        }
        WorkerContext = GetWorkerContext(Ushc->SetBusFreqWork);
        WorkerContext->Ushc = Ushc;

        *NewSlot = Slot;
    }
    __finally
    {
        if(STATUS_SUCCESS != Status)
        {
            os_trace_err("Failed UshcCreateSlot with Status = 0x%x", Status);
        }
    }

    return Status;
}

//
// UshcDestroySlot
//
// Arguments
//  Device -
//  Slot   - Pointer to a SDIOEMB_SLOT structure
//
// Returns
//  VOID
//
VOID UshcDestroySlot(__in WDFDEVICE Device, __in PSDIOEMB_SLOT Slot)
{
    PUSHC_DATA Ushc;

    if(NULL == Slot) 
    {
        return;
    }

    Ushc = Slot->drv_data;

    sdioemb_wdf_destroy_child_list(Slot);
    sdioemb_ushc_clean_up(&Ushc->Ushc);
    sdioemb_slot_free(Slot);
}

//
// UshcSubmitVendorRequest
//
// Arguments
//  Slot           - Pointer to the SDIOEMB_SLOT.
//  Request        - Vendor-defined request code
//  Direction      - USHC_HOST_TO_DEVICE or USHC_DEVICE_TO_HOST
//  Recipient      - One of: USHC_RECIPIENT_DEVICE, USHC_RECIPIENT_INTERFACE, USHC_RECIPIENT_ENDPOINT, USHC_RECIPIENT_OTHER
//  Value          - Value, specific to Request (can be 0)
//  Index          - Index, specific to Request (can be 0)
//  IoBuffer       - Pointer to a buffer for the transfer (can be NULL)
//  IoBufferLength - Length, in bytes, of the buffer specified in RequestBuffer (can be 0) 
//
// Returns
//  INT - 0 on success otherwise a POSIX error code
//
INT UshcSubmitVendorRequest(__in PSDIOEMB_SLOT Slot, 
                            __in USHC_REQUEST Request,
                            __in USHC_REQUEST_DIRECTION Direction, 
                            __in USHC_REQUEST_RECIPIENT Recipient,
                            __in UINT16 Value, 
                            __in UINT16 Index, 
                            __in PVOID IoBuffer, 
                            __in UINT32 IoBufferLength)
{
    PUSHC_DEVICE_CONTEXT Dc = UshcGetDeviceContext(((PUSHC_DATA)(Slot->drv_data))->Device);
    WDF_USB_BMREQUEST_DIRECTION _Direction; 
    WDF_USB_BMREQUEST_RECIPIENT _Recipient;
    NTSTATUS Status = STATUS_SUCCESS;

    switch(Direction)
    {
        case USHC_HOST_TO_DEVICE:
            _Direction = BmRequestHostToDevice;
        break;

        case USHC_DEVICE_TO_HOST:
            _Direction = BmRequestDeviceToHost;
        break;

        default:
            Status = STATUS_INVALID_PARAMETER;
        break;
    }

    switch(Recipient)
    {
        case USHC_RECIPIENT_DEVICE:
            _Recipient = BmRequestToDevice;
        break;

        case USHC_RECIPIENT_INTERFACE:
            _Recipient = BmRequestToInterface; 
        break;

        case USHC_RECIPIENT_ENDPOINT:
            _Recipient = BmRequestToEndpoint;
        break;

        case USHC_RECIPIENT_OTHER:
            _Recipient = BmRequestToOther;
        break;

        default:
            Status = STATUS_INVALID_PARAMETER;
        break;
    }

    if(STATUS_SUCCESS == Status)
    {
        Status = SubmitVendorRequest(Dc->WdfUsbTargetDevice, 
                                     Request, 
                                     _Direction,
                                     _Recipient,
                                     Value,
                                     Index,
                                     IoBuffer,
                                     IoBufferLength);
    }

    return NTStatus2Posix(Status);
}

//
// UshcSetHostControl
//
// Arguments
//  Dc             - Pointer to a SDIOEMB_SLOT structure
//  ControlerState -
//
// Returns
//  INT - 0 on success otherwise a POSIX error code
//
INT UshcSetHostControl(__in PSDIOEMB_SLOT Slot, __in UINT16 ControlerState)
{
    PUSHC_DEVICE_CONTEXT Dc = UshcGetDeviceContext(((PUSHC_DATA)(Slot->drv_data))->Device);
    NTSTATUS Status;

    Status = SetControler(Dc, ControlerState);

    return NTStatus2Posix(Status);
}

//
// SetControler
//
// Arguments
//  Dc             - Pointer to a USHC_DEVICE_CONTEXT structure
//  ControlerState -
//
// Returns
//  NTSTATUS - STATUS_STATUS on success otherwise a NTSTATUS error code
//
NTSTATUS SetControler(__in PUSHC_DEVICE_CONTEXT Dc, __in UINT16 ControlerState)
{
    NTSTATUS Status;

    Status = SubmitVendorRequest(Dc->WdfUsbTargetDevice,
                                 USHC_HOST_CTRL,
                                 BmRequestHostToDevice,
                                 BmRequestToDevice,
                                 ControlerState,
                                 0,
                                 NULL,
                                 0);

    return Status;
}

#define USHC_VENDOR_REQUEST_TIMEOUT_MS 2000
//
// SubmitVendorRequest
//
// Arguments
//  WdfUsbTargetDevice   - Pointer to the WDFUSBDEVICE.
//  RequestCode          - Vendor-defined request code
//  RequestDirection     - BmRequestHostToDevice or BmRequestDeviceToHost
//  RequestRecipient     - One of: BmRequestToDevice, BmRequestToInterface, BmRequestToEndpoint, BmRequestToOther
//  Value                - Value, specific to Request (can be 0)
//  Index                - Index, specific to Request (can be 0)
//  RequestBuffer        - Pointer to a buffer for the transfer (can be NULL)
//  RequestBufferLength  - Length, in bytes, of the buffer specified in RequestBuffer (can be 0) 
//
// Returns
//  NTSTATUS - STATUS_STATUS on success otherwise a NTSTATUS error code
//
NTSTATUS SubmitVendorRequest(__in WDFUSBDEVICE WdfUsbTargetDevice, 
                             __in UCHAR RequestCode, 
                             __in WDF_USB_BMREQUEST_DIRECTION RequestDirection, 
                             __in WDF_USB_BMREQUEST_RECIPIENT RequestRecipient,
                             __in UINT16 Value,
                             __in UINT16 Index,
                             __in_bcount_opt(RequestBufferLength) PVOID RequestBuffer,
                             __in ULONG RequestBufferLength)
{
    WDF_USB_CONTROL_SETUP_PACKET ControlSetupPacket;
    WDF_REQUEST_SEND_OPTIONS SendOptions;
    WDF_MEMORY_DESCRIPTOR MemoryDescriptor;
    ULONG BytesTransferred;
    NTSTATUS Status;

    WDF_REQUEST_SEND_OPTIONS_INIT(&SendOptions, WDF_REQUEST_SEND_OPTION_TIMEOUT);
    WDF_REQUEST_SEND_OPTIONS_SET_TIMEOUT(&SendOptions, WDF_REL_TIMEOUT_IN_MS(USHC_VENDOR_REQUEST_TIMEOUT_MS));
    WDF_USB_CONTROL_SETUP_PACKET_INIT_VENDOR(&ControlSetupPacket,
                                             RequestDirection,
                                             RequestRecipient,
                                             RequestCode,
                                             Value,
                                             Index);

    WDF_MEMORY_DESCRIPTOR_INIT_BUFFER(&MemoryDescriptor, RequestBuffer, RequestBufferLength);

    Status = WdfUsbTargetDeviceSendControlTransferSynchronously(WdfUsbTargetDevice,
                                                                WDF_NO_HANDLE,
                                                                &SendOptions,
                                                                &ControlSetupPacket,
                                                                &MemoryDescriptor,
                                                                &BytesTransferred);

    return Status;
}


//
// NTStatus2Posix
//
// Arguments
//  Status - NTSTATUS value
//
// Returns
//  INT - appropriate Posix error code or EINVAL if nothing appropriate
//
INT NTStatus2Posix(__in NTSTATUS Status)
{
    switch(Status)
    {
        case STATUS_SUCCESS:
            return 0;
        break;

        case STATUS_TIMEOUT:
        case STATUS_IO_TIMEOUT:
            return -ETIMEDOUT;
        break;

        case STATUS_CRC_ERROR:
            return -EIO;
        break;

        case STATUS_IO_DEVICE_ERROR:
            return -ENODEV;
        break;

        case STATUS_NOT_SUPPORTED:
            return -ENOTSUPP;
        break;

        case STATUS_BUFFER_TOO_SMALL:
        case STATUS_INVALID_PARAMETER:
        case STATUS_ACCESS_DENIED:
        case STATUS_INVALID_DEVICE_STATE:
        default:
            return -EINVAL;
        break;
    }
}
