/*
 * Sdioemb Userspace Interface driver for Windows.
 *
 * Copyright (C) Cambridge Silicon Radio Ltd.
 *
 * Refer to LICENSE.txt included with this source code for details on
 * the license terms.
 */
#define WPP_CONTROL_GUIDS                                               \
    WPP_DEFINE_CONTROL_GUID(UifTraceGuid,                               \
                            (92a4e2fd,1d9b,4625,b62e,2fe26ebbd300),     \
                            WPP_DEFINE_BIT(OS_TRACE_DFLT)               \
        )
#include <initguid.h>
#include <sdioemb/windows/guid.h>
#include "uif_internal.h"
#include <sdioemb/slot_api.h> /* for sdioemb_add_function() and sdioemb_del_function() */

#include "driver.tmh"

DRIVER_INITIALIZE DriverEntry;

static EVT_WDF_OBJECT_CONTEXT_CLEANUP uif_driver_cleanup;
static EVT_WDF_DRIVER_DEVICE_ADD uif_evt_driver_device_add;
static EVT_WDF_OBJECT_CONTEXT_CLEANUP uif_device_evt_cleanup;
static EVT_WDF_DEVICE_D0_ENTRY uif_evt_device_d0_entry;
static EVT_WDF_DEVICE_D0_EXIT uif_evt_device_d0_exit;

NTSTATUS DriverEntry(
    IN OUT PDRIVER_OBJECT driver_object,
    IN PUNICODE_STRING registry_path)
{
    WDF_DRIVER_CONFIG config;
    WDF_OBJECT_ATTRIBUTES attribs;
    NTSTATUS status;
    int ret;

    WPP_INIT_TRACING(driver_object, registry_path);

    WDF_DRIVER_CONFIG_INIT(&config, uif_evt_driver_device_add);
    WDF_OBJECT_ATTRIBUTES_INIT(&attribs);
    attribs.EvtCleanupCallback = uif_driver_cleanup;

    status = WdfDriverCreate(driver_object, registry_path, &attribs, &config, WDF_NO_HANDLE);
    if (!NT_SUCCESS(status)) {
        WPP_CLEANUP(driver_object);
    }

    ret = uif_register();
    if (ret < 0) {
        return os_err_to_ntstatus(ret);
    }

    return STATUS_SUCCESS;
}

static void uif_driver_cleanup(
    IN WDFDRIVER driver)
{
    uif_unregister();
    WPP_CLEANUP(WdfDriverWdmGetDriverObject(driver));
}

NTSTATUS uif_evt_driver_device_add(
    IN WDFDRIVER driver,
    IN PWDFDEVICE_INIT device_init)
{
    WDF_PNPPOWER_EVENT_CALLBACKS pnp_power_callbacks;
    WDF_OBJECT_ATTRIBUTES attribs;
    WDFDEVICE device;
    struct uif_device_context *dev_ctx;
    struct sdioemb_wdf_interface interface;
    NTSTATUS status;

    WdfDeviceInitSetIoType(device_init, WdfDeviceIoDirect);

    WDF_PNPPOWER_EVENT_CALLBACKS_INIT(&pnp_power_callbacks);
    pnp_power_callbacks.EvtDeviceD0Entry = uif_evt_device_d0_entry;
    pnp_power_callbacks.EvtDeviceD0Exit = uif_evt_device_d0_exit;
    WdfDeviceInitSetPnpPowerEventCallbacks(device_init, &pnp_power_callbacks);

    WDF_OBJECT_ATTRIBUTES_INIT_CONTEXT_TYPE(&attribs, UIF_DEVICE_CONTEXT);
    attribs.EvtCleanupCallback = uif_device_evt_cleanup;

    uif_init_file_config(device_init);

    status = WdfDeviceCreate(&device_init, &attribs, &device);
    if (!NT_SUCCESS(status)) {
        return status;
    }

    dev_ctx = uif_get_device_context(device);
    
    os_event_init(&dev_ctx->request_event);
    os_spinlock_init(&dev_ctx->interrupt_lock);

    /*
     * Get the sdioemb function driver interface.  This inteface
     * provides the pointer to sdioemb's function device (fdev).
     *
     * This references both the driver interface and the fdev.
     */
    status = WdfFdoQueryForInterface(device, &GUID_DRVINTERFACE_SDIOEMB_FUNC,
                                     &dev_ctx->interface.header, sizeof(interface), 1, NULL);
    if (!NT_SUCCESS(status)) {
        os_trace_err("can't get function interface: %!STATUS!", status);
        return status;
    }

    /*
     * Use the function's FDO as the OS device in the fdev.
     */
    dev_ctx->interface.fdev->os_device = device;

    status = uif_create_interface(device);
    if (!NT_SUCCESS(status)) {
        return status;
    }

    return STATUS_SUCCESS;
}

VOID uif_device_evt_cleanup(
    IN WDFOBJECT object)
{
    WDFDEVICE device = object;
    struct uif_device_context *dev_ctx = uif_get_device_context(device);

    /*
     * Dereference the driver interface and the fdev.
     */
    dev_ctx->interface.header.InterfaceDereference(dev_ctx->interface.header.Context);

    os_spinlock_destroy(&dev_ctx->interrupt_lock);
    os_event_destroy(&dev_ctx->request_event);
}

NTSTATUS uif_evt_device_d0_entry(
    IN WDFDEVICE device,
    IN WDF_POWER_DEVICE_STATE previous_state)
{
    struct uif_device_context *dev_ctx = uif_get_device_context(device);
    struct sdioemb_dev *fdev = dev_ctx->interface.fdev;

    os_trace_dbg("uif%d: D0 entry", fdev->slot_id);

    /*
     * Add the function the first time the D0 state is entered.
     *
     * On subsequent calls the function will have been suspended so
     * resume it.
     */
    if (!dev_ctx->added) {
        sdioemb_add_function(fdev);
        dev_ctx->added = true;
    } else {
        sdioemb_resume_function(fdev);
    }

    return STATUS_SUCCESS;
}

NTSTATUS uif_evt_device_d0_exit(
    IN WDFDEVICE device,
    IN WDF_POWER_DEVICE_STATE target_state)
{
    struct uif_device_context *dev_ctx = uif_get_device_context(device);
    struct sdioemb_dev *fdev = dev_ctx->interface.fdev;

    os_trace_dbg("uif%d: D0 exit", fdev->slot_id);

    /*
     * Suspend the function unless the D3Final state is being entered.
     */
    switch (target_state) {
    case WdfPowerDeviceD3Final:
        sdioemb_del_function(fdev);
        dev_ctx->added = false; 
        break;
    default:
        sdioemb_suspend_function(fdev);
        break;
    }

    return STATUS_SUCCESS;
}
