/*
 * Sdioemb Userspace Interface driver for Windows.
 *
 * Copyright (C) Cambridge Silicon Radio Ltd.
 *
 * Refer to LICENSE.txt included with this source code for details on
 * the license terms.
 *
 * When the device interface is first opened the function is
 * 'activated'.  This allows the uif driver to always be loaded
 * without affecting other function drivers.
 *
 * The interface is not deactivated when the device interface is
 * closed as this would require a call to sdioemb_power_off() which
 * would result in the card repeatedly being powered on and off.
 */
#include "uif_internal.h"
#include <ntstrsafe.h>

static EVT_WDF_DEVICE_FILE_CREATE uif_evt_device_file_create;
static EVT_WDF_FILE_CLEANUP uif_evt_file_cleanup;
static EVT_WDF_IO_IN_CALLER_CONTEXT  uif_evt_io_in_caller_context;
static EVT_WDF_IO_QUEUE_STATE uif_evt_io_queue_ready;

void uif_init_file_config(PWDFDEVICE_INIT device_init)
{
    WDF_FILEOBJECT_CONFIG file_config;
    WDF_OBJECT_ATTRIBUTES attribs;

    WDF_OBJECT_ATTRIBUTES_INIT(&attribs);
    attribs.SynchronizationScope = WdfSynchronizationScopeNone;
    WDF_FILEOBJECT_CONFIG_INIT(&file_config,
                               uif_evt_device_file_create,
                               NULL,
                               uif_evt_file_cleanup);
    WdfDeviceInitSetFileObjectConfig(device_init, &file_config, &attribs);    

    WdfDeviceInitSetIoInCallerContextCallback(device_init, uif_evt_io_in_caller_context);
}

NTSTATUS uif_create_interface(WDFDEVICE device)
{
    struct uif_device_context *dev_ctx = uif_get_device_context(device);
    WDF_IO_QUEUE_CONFIG queue_config;
    DECLARE_UNICODE_STRING_SIZE(refstr, 32);
    NTSTATUS status;

    WDF_IO_QUEUE_CONFIG_INIT(&queue_config, WdfIoQueueDispatchManual);
    status = WdfIoQueueCreate(device, &queue_config, WDF_NO_OBJECT_ATTRIBUTES, &dev_ctx->queue);
    if (!NT_SUCCESS(status)) {
        return status;
    }
    status = WdfIoQueueReadyNotify(dev_ctx->queue, uif_evt_io_queue_ready, NULL);
    if (!NT_SUCCESS(status)) {
        return status;
    }
    status = WdfDeviceConfigureRequestDispatching(device, dev_ctx->queue,
                                                  WdfRequestTypeDeviceControl);
    if (!NT_SUCCESS(status)) {
        return status;
    }

    RtlUnicodeStringPrintf(&refstr, L"sdio_uif%d", dev_ctx->interface.fdev->slot_id);
    status = WdfDeviceCreateDeviceInterface(device, &GUID_DEVINTERFACE_SDIOEMB_UIF, &refstr);
    if (!NT_SUCCESS(status)) {
        return status;
    }

    return STATUS_SUCCESS;
}

static VOID uif_evt_device_file_create(
    IN WDFDEVICE device,
    IN WDFREQUEST request,
    IN WDFFILEOBJECT file)
{
    struct uif_device_context *dev_ctx = uif_get_device_context(device);

    if (!dev_ctx->active) {
        sdioemb_power_on(dev_ctx->interface.fdev);
        dev_ctx->active = true;
    }

    sdioemb_interrupt_enable(dev_ctx->interface.fdev);

    WdfRequestComplete(request, STATUS_SUCCESS);
}

static VOID uif_evt_file_cleanup(
    IN WDFFILEOBJECT file)
{
    WDFDEVICE device = WdfFileObjectGetDevice(file);
    struct uif_device_context *dev_ctx = uif_get_device_context(device);

    sdioemb_interrupt_disable(dev_ctx->interface.fdev);
    sdioemb_interrupt_acknowledge(dev_ctx->interface.fdev);
}

static VOID uif_evt_io_in_caller_context(WDFDEVICE device, WDFREQUEST request)
{
    struct uif_request_context *req_ctx;
    WDF_REQUEST_PARAMETERS params;
    WDF_OBJECT_ATTRIBUTES attribs;
    struct sdioemb_uif_input *input;
    struct sdioemb_uif_cmd *cmd;
    NTSTATUS status;

    WDF_REQUEST_PARAMETERS_INIT(&params);
    WdfRequestGetParameters(request, &params);

    /*
     * Probe and lock an SDIO command's data buffer.
     */
    if (params.Type == WdfRequestTypeDeviceControl
        && params.Parameters.DeviceIoControl.IoControlCode == SDIOEMB_UIF_IOC_CMD) {

        WDF_OBJECT_ATTRIBUTES_INIT_CONTEXT_TYPE(&attribs, UIF_REQUEST_CONTEXT);
        status = WdfObjectAllocateContext(request, &attribs, &req_ctx);
        if (!NT_SUCCESS(status)) {
            goto error;
        }

        status = WdfRequestRetrieveInputBuffer(request, sizeof(*input), &input, NULL);
        if (!NT_SUCCESS(status)) {
            goto error;
        }

        cmd = &input->cmd;

        if (cmd->data == NULL || cmd->len < 1) {
            status = STATUS_INVALID_PARAMETER;
            goto error;
        }

        switch (cmd->type) {
        case SDIOEMB_UIF_CMD52_READ:
        case SDIOEMB_UIF_CMD53_READ:
            status = WdfRequestProbeAndLockUserBufferForWrite(request, cmd->data, cmd->len,
                                                              &req_ctx->data);
            break;
        case SDIOEMB_UIF_CMD52_WRITE:
        case SDIOEMB_UIF_CMD53_WRITE:
            status = WdfRequestProbeAndLockUserBufferForRead(request, cmd->data, cmd->len,
                                                             &req_ctx->data);
            break;
        default:
            status = STATUS_INVALID_PARAMETER;
            break;
        }
        if (!NT_SUCCESS(status)) {
            goto error;
        }
    }

    status = WdfDeviceEnqueueRequest(device, request);
    if (!NT_SUCCESS(status)) {
        goto error;
    }
    return;

  error:
    WdfRequestComplete(request, status);
}

static VOID uif_evt_io_queue_ready(WDFQUEUE queue, WDFCONTEXT context)
{
    WDFDEVICE device = WdfIoQueueGetDevice(queue);
    struct uif_device_context *dev_ctx = uif_get_device_context(device);

    os_event_raise(&dev_ctx->request_event, 1);
}
