/*
 * Sdioemb Userspace Interface driver for Windows.
 *
 * Copyright (C) Cambridge Silicon Radio Ltd.
 *
 * Refer to LICENSE.txt included with this source code for details on
 * the license terms.
 */
#include "uif_internal.h"
#include "../../../core/sdio_internal.h"

static EVT_WDF_REQUEST_CANCEL uif_evt_request_cancel_wait_for_int;

static void uif_ioctl_request_complete(WDFREQUEST request, NTSTATUS status)
{
    size_t output_len = sizeof(struct sdioemb_uif_output);

    if (!NT_SUCCESS(status)) {
        output_len = 0;
    }
    WdfRequestCompleteWithInformation(request, status, output_len);
}

static void uif_evt_request_cancel_wait_for_int(WDFREQUEST request)
{
    WDFFILEOBJECT file = WdfRequestGetFileObject(request);
    WDFDEVICE device = WdfFileObjectGetDevice(file);
    struct uif_device_context *dev_ctx = uif_get_device_context(device);
    os_int_status_t istate;

    os_spinlock_lock_intsave(&dev_ctx->interrupt_lock, &istate);

    if (dev_ctx->wait_for_int_request == request) {
        dev_ctx->wait_for_int_request = NULL;
    }
    uif_ioctl_request_complete(request, STATUS_CANCELLED);

    os_spinlock_unlock_intrestore(&dev_ctx->interrupt_lock, &istate);
}

static NTSTATUS uif_get_card_info(WDFDEVICE device, struct sdioemb_dev *fdev,
                                  struct sdioemb_uif_output *output)
{
    output->card_info.num_functions = (uint8_t)fdev->priv->slot->priv->num_functions;

    return STATUS_SUCCESS;
}

static NTSTATUS uif_get_func_info(WDFDEVICE device, struct sdioemb_dev *fdev,
                                  struct sdioemb_uif_input *input,
                                  struct sdioemb_uif_output *output)
{
    struct sdioemb_slot_priv *slotp = fdev->priv->slot->priv;
    uint8_t func;
    struct sdioemb_uif_func_info func_info;

    func = input->get_func_info.function;
    if (func > fdev->priv->slot->priv->num_functions) {
        return STATUS_INVALID_PARAMETER;
    }

    output->func_info.manf_id        = fdev->vendor_id;
    output->func_info.card_id        = fdev->device_id;
    output->func_info.std_if         = slotp->functions[func]->interface;
    output->func_info.max_block_size = slotp->functions[func]->max_blocksize;
    output->func_info.block_size     = slotp->functions[func]->blocksize;

    return STATUS_SUCCESS;
}

static NTSTATUS uif_cmd(WDFDEVICE device, WDFREQUEST request, struct sdioemb_dev *fdev, 
                        struct sdioemb_uif_input *input,
                        struct sdioemb_uif_output *output)
{
    struct uif_request_context *req_ctx = uif_get_request_context(request);
    struct sdioemb_slot_priv *slotp = fdev->priv->slot->priv;
    struct sdioemb_uif_cmd *cmd;
    uint8_t *data;
    int ret;

    cmd  = &input->cmd;
    data = WdfMemoryGetBuffer(req_ctx->data, NULL);

    switch (cmd->type) {
    case SDIOEMB_UIF_CMD52_READ:
        ret = slotp->io_ops.read8(fdev, cmd->function, cmd->address, data);
        break;
    case SDIOEMB_UIF_CMD52_WRITE:
        ret = slotp->io_ops.write8(fdev, cmd->function, cmd->address, *data);
        break;
    case SDIOEMB_UIF_CMD53_READ:
        fdev->blocksize = cmd->block_size;
        ret = slotp->io_ops.read(fdev, cmd->function, cmd->address, data, cmd->len);
        break;
    case SDIOEMB_UIF_CMD53_WRITE:
        fdev->blocksize = cmd->block_size;
        ret = slotp->io_ops.write(fdev, cmd->function, cmd->address, data, cmd->len);
        break;
    default:
        ret = -EINVAL;
    }
    return os_err_to_ntstatus(ret);
}

static NTSTATUS uif_wait_for_int(WDFDEVICE device, WDFREQUEST request, struct sdioemb_dev *fdev)
{
    struct uif_device_context *dev_ctx = uif_get_device_context(device);
    os_int_status_t istate;

    sdioemb_interrupt_acknowledge(fdev);

    os_spinlock_lock_intsave(&dev_ctx->interrupt_lock, &istate);

    if (dev_ctx->interrupt_pending) {
        uif_ioctl_request_complete(request, STATUS_SUCCESS);
        dev_ctx->interrupt_pending = false;
    } else {
        dev_ctx->wait_for_int_request = request;
        WdfRequestMarkCancelable(request, uif_evt_request_cancel_wait_for_int);
    }
    os_spinlock_unlock_intrestore(&dev_ctx->interrupt_lock, &istate);

    return STATUS_SUCCESS;
}

static NTSTATUS uif_reinsert(WDFDEVICE device, struct sdioemb_dev *fdev)
{
    struct sdioemb_slot *slot = fdev->priv->slot;

    sdioemb_card_removed(slot);
    os_sleep_ms(100);
    sdioemb_card_inserted(slot);

    return STATUS_SUCCESS;
}

static NTSTATUS uif_set_bus_width(WDFDEVICE device, struct sdioemb_dev *fdev,
                                  struct sdioemb_uif_input *input)
{
    struct sdioemb_slot_priv *slotp = fdev->priv->slot->priv;
    uint8_t func;
    uint8_t bus_width;
    int ret;

    func      = input->set_bus_width.function;
    bus_width = input->set_bus_width.bus_width;
    if (func > fdev->priv->slot->priv->num_functions) {
        return STATUS_INVALID_PARAMETER;
    }

    ret = sdioemb_set_block_size(slotp->functions[func], bus_width);
    return os_err_to_ntstatus(ret);
}

static int uif_set_bus_freq(WDFDEVICE device, struct sdioemb_dev *fdev,
                            struct sdioemb_uif_input *input)
{
    struct sdioemb_slot *slot = fdev->priv->slot;
    int32_t bus_freq;
    NTSTATUS status;
    int ret;

    bus_freq = input->set_bus_freq.bus_freq;

    if (bus_freq < 0 && bus_freq != SDIOEMB_BUS_FREQ_DEFAULT) {
        return STATUS_INVALID_PARAMETER;
    }

    if (bus_freq != SDIOEMB_BUS_FREQ_OFF) {
        sdioemb_set_max_bus_freq(fdev, bus_freq);
        bus_freq = slot->clock_freq;
    }

    /* Set clock rate and start it if it isn't already running at the
       correct speed. */
    slot_set_bus_freq(slot, bus_freq);

    return 0;
}

static NTSTATUS uif_set_block_size(WDFDEVICE device, struct sdioemb_dev *fdev,
                                   struct sdioemb_uif_input *input)
{
    struct sdioemb_slot_priv *slotp = fdev->priv->slot->priv;
    uint8_t func;
    uint16_t blksz;
    int ret;

    func  = input->set_block_size.function;
    blksz = input->set_block_size.block_size;
    if (func > fdev->priv->slot->priv->num_functions) {
        return STATUS_INVALID_PARAMETER;
    }

    ret = sdioemb_set_block_size(slotp->functions[func], blksz);

    return os_err_to_ntstatus(ret);
}

static NTSTATUS do_one_ioctl_request(WDFDEVICE device, WDFREQUEST request, ULONG io_control_code)
{
    struct uif_device_context *dev_ctx = uif_get_device_context(device);
    struct sdioemb_dev *fdev;
    struct sdioemb_uif_input *input;
    struct sdioemb_uif_output *output;
    NTSTATUS status;

    fdev = dev_ctx->interface.fdev;

    status = WdfRequestRetrieveInputBuffer(request, sizeof(*input), &input, NULL);
    if (!NT_SUCCESS(status)) {
        return status;
    }
    status = WdfRequestRetrieveOutputBuffer(request, sizeof(*output), &output, NULL);
    if (!NT_SUCCESS(status)) {
        return status;
    }

    switch (io_control_code) {
    case SDIOEMB_UIF_IOC_GET_CARD_INFO:
        return uif_get_card_info(device, fdev, output);
    case SDIOEMB_UIF_IOC_GET_FUNC_INFO:
        return uif_get_func_info(device, fdev, input, output);
    case SDIOEMB_UIF_IOC_CMD:
        return uif_cmd(device, request, fdev, input, output);
    case SDIOEMB_UIF_IOC_WAITFORINT:
        return uif_wait_for_int(device, request, fdev);
    case SDIOEMB_UIF_IOC_REINSERT:
        return uif_reinsert(device, fdev);
    case SDIOEMB_UIF_IOC_SET_BUS_FREQ:
        return uif_set_bus_freq(device, fdev, input);
    case SDIOEMB_UIF_IOC_SET_BLOCK_SIZE:
        return uif_set_block_size(device, fdev, input);
    default:
        break;
    }
    return STATUS_INVALID_PARAMETER;
}

void request_thread_func(void *arg)
{
    WDFDEVICE device = arg;
    struct uif_device_context *dev_ctx = uif_get_device_context(device);

    while (!os_thread_should_stop(&dev_ctx->request_thread)) {
        WDFREQUEST request;
        WDF_REQUEST_PARAMETERS params;
        ULONG io_control_code;
        NTSTATUS status;

        status = WdfIoQueueRetrieveNextRequest(dev_ctx->queue, &request);
        if (!NT_SUCCESS(status)) {
            os_event_wait(&dev_ctx->request_event);
            continue;
        }

        WDF_REQUEST_PARAMETERS_INIT(&params);
        WdfRequestGetParameters(request, &params);
        io_control_code = params.Parameters.DeviceIoControl.IoControlCode;

        status = do_one_ioctl_request(device, request, io_control_code);

        if (io_control_code != SDIOEMB_UIF_IOC_WAITFORINT) {
            uif_ioctl_request_complete(request, status);
        }
    }
}

static int uif_probe(struct sdioemb_dev *fdev)
{
    WDFDEVICE device = fdev->os_device;
    struct uif_device_context *dev_ctx = uif_get_device_context(device);
    NTSTATUS status;
    char name[32];
    int ret;

    snprintf(name, sizeof(name), "sdio_uif%d", dev_ctx->interface.fdev->slot_id);
    ret = os_thread_create(&dev_ctx->request_thread, name, request_thread_func, device);
    if (ret < 0) {
        return ret;
    }

    return 0;
}

static void uif_remove(struct sdioemb_dev *fdev)
{
    WDFDEVICE device = fdev->os_device;
    struct uif_device_context *dev_ctx = uif_get_device_context(device);

    os_thread_stop(&dev_ctx->request_thread, &dev_ctx->request_event);
}

static void uif_interrupt(struct sdioemb_dev *fdev)
{
    WDFDEVICE device = fdev->os_device;
    struct uif_device_context *dev_ctx = uif_get_device_context(device);
    os_int_status_t istate;

    os_spinlock_lock_intsave(&dev_ctx->interrupt_lock, &istate);

    if (dev_ctx->wait_for_int_request) {
        WdfRequestUnmarkCancelable(dev_ctx->wait_for_int_request);
        uif_ioctl_request_complete(dev_ctx->wait_for_int_request, STATUS_SUCCESS);
        dev_ctx->wait_for_int_request = NULL;
    } else {
        dev_ctx->interrupt_pending = true;
    }

    os_spinlock_unlock_intrestore(&dev_ctx->interrupt_lock, &istate);
}

static struct sdioemb_id_table uif_id_table[] = {
    { SDIOEMB_ANY_ID, SDIOEMB_ANY_ID, 0, SDIOEMB_ANY_IFACE },
    { 0 },
};

static struct sdioemb_func_driver uif_driver = {
    "uif",
    uif_id_table,
    uif_probe,
    uif_remove,
    uif_interrupt,
    NULL,
    NULL,
};

int uif_register(void)
{
    return sdioemb_driver_register(&uif_driver);
}

void uif_unregister(void)
{
    sdioemb_driver_unregister(&uif_driver);
}
